
/*
 * Computer Algebra Kit (c) 1993,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

#ifndef __CAOBJECT_HEADER__
#define __CAOBJECT_HEADER__

#include "stdio.h"

#ifdef CAKIT_NEXT
#include <objc/Object.h>
#include <mach/cthreads.h>
#endif
#ifdef __STPSTN__
#include <Object.h>
#endif
#ifdef __GNUOBJC__
#include <objc/Object.h>
#endif
#ifdef __PORTABLE_OBJC__
#include "Object.h"
#endif

@interface CAObject : Object
{
#ifdef CAKIT_MACH
    struct mutex		lock;
#endif /* CAKIT_MACH */
    int				refCount;
    STR				_strValue;
}

+ initialize;
+ (STR) cakitRevision;

+ new;
+ newStrValue:(STR)aString;
+ str:(STR)aString;
+ int:(int)intValue;
- copy;
- deepCopy;
- cheapCopy;
- clear;
- free;

- (BOOL) sameClass:a;
- (BOOL) differentClass:a;
- (BOOL) isKindOfSequence;

- (STR) str;
- (STR) strValue;
- (const char *) stringValue;
- strValue:(STR)aString;
- (int) intValue;
- intValue:(int)i;
- asModp:(unsigned short)p;
- (BOOL) isFloatingPoint;
- asNumerical;
- (float) floatValue;
- floatValue:(float)f;
- asTotalFraction;
- asScalar;
- asSymbol;

- (BOOL) isEqual:a;
- (BOOL) notEqual:a;

- (BOOL) inOrderedSet;
- (int) compare:b;
- (int) sign;
- (BOOL) isLess:a;
- (BOOL) isGreater:a;
- (BOOL) isLessEqual:a;
- (BOOL) isGreaterEqual:a;
- absValue;

- (BOOL) inAdditiveSemiGroup;
- (BOOL) inAdditiveMonoid;
- (BOOL) inAdditiveGroup;
- (BOOL) isZero;
- (BOOL) notZero;
- (BOOL) isOpposite:b;
- (BOOL) notOpposite:b;
- zero;
- negate;
- add:b;
- subtract:b;
- increment;
- decrement;
- multiplyIntValue:(int)b;
- double;
- quadruple;
- divideIntValue:(int)b;
- half;
- quarter;

- (BOOL) commutes;
- (BOOL) commutesWith:b;
- (BOOL) inSemiGroup;
- (BOOL) inMonoid;
- (BOOL) inGroup;
- one;
- minusOne;
- (BOOL) isOne;
- (BOOL) notOne;
- (BOOL) isMinusOne;
- (BOOL) notMinusOne;
- square;
- multiply:b;
- power:(int)n;
- inverse;
- (BOOL) isUnit;
- (BOOL) notUnit;
- divide:b;

- remainder:b quotient:(id *)q;
- remainder:b;
- quotient:b;

- (BOOL) inEuclideanDomain;
- (BOOL) isCoprime:b;
- (BOOL) notCoprime:b;
- (BOOL) isGcd:a:b;
- (BOOL) isLcm:a:b;
- gcd:b;
- bezout:b gcd:(id *)gcd;
- lcm:b;

- modulo:m;
- multiply:b modulo:m;
- squareModulo:m;
- power:(int)n modulo:m;
- inverseModulo:m;

- random;

- (int) characteristic;
- (BOOL) isCharacteristicZero;
- (BOOL) notCharacteristicZero;
- (BOOL) isCharacteristicTwo;
- (BOOL) notCharacteristicTwo;
- frobenius;
- frobeniusInverse;
- (int) dimensionOverPrimeField;

- (BOOL) inRing;
- (BOOL) inIntegralDomain;
- (BOOL) inField;
- (BOOL) inFieldOfFractions;

- scalarZero;
- scalarContent;
- divideScalarContent;
- multiplyScalar:s;
- divideScalar:s;
- addScalar:s;
- subtractScalar:s;

- asExpression;
- expression:anExpression;

- (BOOL) printsLeadingSign;
- (BOOL) printsSum;
- (BOOL) printsProduct;
- printToFile:(FILE *)aFile;

#ifdef CAKIT_NEXT
- printToStream:(NXStream *)aStream;
- (void) printForDebugger:(NXStream *)aStream;
#endif /* CAKIT_NEXT */

#ifdef __STPSTN__
- printOn:(IOD)anIOD;
#endif /* __STPSTN__ */
- fileOutOn:aFiler;
- fileInFrom:aFiler;
- addReference; /* Obsoleted, use -cheapCopy */

@end

#ifdef CAKIT_NEXT
static inline id idReplace(id o,id n) {return (n)?([o free],n):nil;}
#else
id idReplace(id o,id n);
#endif /* CAKIT_NEXT */

/* global variables */

extern id CACollectionClass;
extern id CAComplexClass;
extern id CAExpressionClass;
extern id CAFloatClass;
extern id CAFractionClass;
extern id CAIntegerClass;
extern id CAIntegerModpClass;
extern id CAMatrixClass;
extern id CAMonomialClass;
extern id CAPolynomialClass;
extern id CASequenceClass;
extern id CASymbolClass;
extern id CATermClass;
extern id CAVectorClass;

#endif /* __CAOBJECT_HEADER__ */

