
/*
 * Computer Algebra Kit (c) 1995,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"
#define headEq(x)  (strEq(head,(x)))

@implementation Expression
static BOOL numericalAtomHead(STR head)
{
    if (strEq(head,"Integer"))		return YES;
    if (strEq(head,"Float"))		return YES;
    return NO;
}

static BOOL atomHead(STR head)
{
    if (numericalAtomHead(head))	return YES;
    if (strEq(head,"Symbol"))		return YES;
    return NO;
}

- (BOOL)isErrorMsg
{
    return headEq("Error");
}

static BOOL headAdditive(STR head)
{
    if (strEq(head,"Negate")) return YES;
    if (strEq(head,"Add")) return YES;
    if (strEq(head,"Subtract")) return YES;
    return NO;
}

static BOOL headMultiplicative(STR head)
{
    if (strEq(head,"Multiply")) return YES;
    if (strEq(head,"Divide")) return YES;
    return NO;
}

static id binary(STR head,id a,id b)
{
    return [CAExpressionClass newBinaryHead:head:a:b];
}

static id errMsg(STR msg)
{
    return [CAExpressionClass newErrorMsg:msg];
}


#ifdef CAKIT_MACH

#endif

- newStrValueNoEval:(STR)strValue
{
    return [self notImplemented:_cmd];
}

+ newStrValueNoEval:(STR)strValue
{
    id res;
#ifdef CAKIT_MACH
    static struct mutex parseLock = MUTEX_INITIALIZER;
#endif /* CAKIT_MACH */
    extern id  ca_yyres;
    extern int ca_yyparse();
    extern int ca_yysetstr(STR str);

    assert(strValue != NULL);
    	
#ifdef CAKIT_MACH
    mutex_lock(&parseLock);
#endif /* CAKIT_MACH */

    ca_yysetstr(strValue);
    res = (ca_yyparse())?[self newErrorMsg:"Syntax Error"]:ca_yyres;
    
#ifdef CAKIT_MACH
    mutex_unlock(&parseLock);
#endif /* CAKIT_MACH */

    return res;
}

/******************************************************************************
 *
 * Input/Output Tables
 *
 *****************************************************************************/

static int ioCount   = 0;
static id inputCltn  = nil;
static id outputCltn = nil;

static void setUpIOTables()
{
    if (inputCltn==nil) inputCltn  = [CACollection new];
    if (outputCltn==nil) outputCltn = [CACollection new];
    assert([inputCltn numMembers] == ioCount);
    assert([outputCltn numMembers] == ioCount);
}

static void insertIOCopies(id i,id o)
{
    setUpIOTables();
    assert([i isKindOf:CAExpressionClass]);
    assert([o isKindOf:CAExpressionClass]);
    [inputCltn insert:[i cheapCopy]];
    [outputCltn insert:[o cheapCopy]];
    ioCount++;
}

static id copyOfOutputAt(int anIndex)
{
    int i = (anIndex >= 0)?anIndex:anIndex+ioCount;
    
    if (0 <= i && i < ioCount) {
	return [[outputCltn memberAt:i] cheapCopy];
    } else {
	char aBuffer[512];
	sprintf(aBuffer,"Out(%i) does not exist",i);
	return errMsg(aBuffer);
    }
}

static id copyOfInputAt(int anIndex)
{
    int i = (anIndex >= 0)?anIndex:anIndex+ioCount;
    
    if (0 <= i && i < ioCount) {
	return [[inputCltn memberAt:i] cheapCopy];
    } else {
	char aBuffer[512];
	sprintf(aBuffer,"In(%i) does not exist",i);
	return errMsg(aBuffer);
    }
}

+ newStrValue:(STR)strValue
{
    id output,input;
    if (strValue == NULL) [self error:"newStrValue: NULL Argument"];

    input  = [self newStrValueNoEval:strValue];assert(input);
    output = [input simplify];
    
    if (NO == [output isKindOf:CAExpressionClass]) {
	[self error:"-simplify should return an expression"];
    }

    insertIOCopies(input,output);
        
    input = [input free];return output;
}

- newIntValue:(int)intValue
{
    return [self notImplemented:_cmd];	
}

+ newIntValue:(int)intValue
{
    id anInteger = [CAIntegerClass newIntValue:intValue];
    return [self newHead:"Integer" object:anInteger];
}

- newErrorMsg:(STR)aMsg
{
    return [self notImplemented:_cmd];
}

+ newErrorMsg:(STR)aMsg
{
    id symbol;
    symbol = [CASymbolClass newChars:aMsg count:strlen(aMsg)];
    symbol = [CAExpressionClass newHead:"Symbol" object:symbol];
    return [self newUnaryHead:"Error":symbol];
}

- _setUpHead:(STR)strValue expressions:aCollection
{
    assert([aCollection numMembers] >= 0);
    head = (char *)OC_Malloc(strlen(strValue) + 1);strcpy(head,strValue);
    expressions = aCollection;object = nil;return self;
}
- _setUpHead:(STR)strValue object:anObject
{
    head = (char *)OC_Malloc(strlen(strValue) + 1);strcpy(head,strValue);
    object = anObject;expressions = nil;return self;
}

- newHead:(STR)strValue expressions:aCollection
{
    return [self notImplemented:_cmd];
}

+ newHead:(STR)strValue expressions:aCollection
{
    return [[super new] _setUpHead:strValue expressions:aCollection];
}

- newHead:(STR)strValue object:anObject
{
    return [self notImplemented:_cmd];
}

+ newHead:(STR)strValue object:anObject
{
    return [[super new] _setUpHead:strValue object:anObject];
}

- newNullaryHead:(STR)strValue
{
    return [self notImplemented:_cmd];
}

+ newNullaryHead:(STR)strValue
{
    return [self newHead:strValue expressions:[CACollectionClass new]];
}

- newUnaryHead:(STR)strValue:a
{
    return [self notImplemented:_cmd];
}

+ newUnaryHead:(STR)strValue:a
{
    id cltn;assert(a);
    cltn = [[CACollectionClass new] insert:a];
    return [self newHead:strValue expressions:cltn];
}

- newBinaryHead:(STR)strValue:a:b
{
    return [self notImplemented:_cmd];
}

+ newBinaryHead:(STR)strValue:a:b
{
    id cltn;assert(a && b);
    cltn = [[[CACollectionClass new] insert:a] insert:b];
    return [self newHead:strValue expressions:cltn];
}

- copy
{
    STR copyHead = (STR)OC_Malloc(strlen(head) + 1);strcpy(copyHead,head);
    
    self = [super copy];head = copyHead;
    if (expressions) expressions = [expressions cheapCopy];
    if (object) object = [object cheapCopy];
    return self;
}

- deepCopy
{
    STR copyHead = (STR)OC_Malloc(strlen(head) + 1);strcpy(copyHead,head);

    self = [super copy];head = copyHead;
    if (expressions) expressions = [expressions deepCopy];
    if (object) object = [object deepCopy];
    return self;
}

- clear
{
    OC_Free(head);
    if (expressions) expressions = [expressions free];
    if (object) object = [object free];
    return [super clear];
}


- (STR) head
{
    return head;
}

- expressions
{
    return expressions;
}

- object
{
    return object;
}

- firstExpression
{
    if ([expressions numMembers]) {
	return [expressions memberAt:0];
    } else {
	[self error:"firstExpression: no such expression"];return 0;
    }
}

- lastExpression
{
    if ([expressions numMembers]) {
	return [expressions lastMember];
    } else {
	[self error:"lastExpression: no such expression"];return 0;
    }
}

- (int) numExpressions
{
    if (expressions) {
	return [expressions numMembers];
    } else {
        [self error:"numExpressions : no subexpressions"];return 0;
    }
}

- (int) intValue
{
    assert(strEq(head,"Integer"));return [object intValue];
}


static BOOL argsContainHead(id expressions,STR aHead)
{
    int i;
    for(i=0;i<[expressions numMembers];i++) {
	if ([[expressions memberAt:i] containsHead:aHead]) return YES;
    }
    return NO;
}

- (BOOL) containsHead:(STR)aHead
{
    if (strEq(head,aHead)) {
	return YES;
    } else {
	if (expressions) {
	    return argsContainHead(expressions,aHead);
	} else {
	    return NO;
	}
    }
}

- (BOOL) isEqual:anExpression
{
    if (self == anExpression) {
	return YES;
    } else {
	if (headEq([anExpression head])) {
	    if (expressions) {
		return [expressions isEqual:[anExpression expressions]];
	    } else {
		return [object isEqual:[anExpression object]];
	    }
	} else {
	    return NO;
	}
    }
}

- (int) compare:anExpression
{
    if ([self isEqual:anExpression]) return 0;
    
    if (headEq("Symbol") && strEq([anExpression head],"Symbol")) {
	return [[self object] compare:[anExpression object]];
    } else {
	[self error:"Can't compare expressions"];return 0;
    }
}


static BOOL doPatternMatch(id aPattern,id anExpression,id kTable,id vTable)
{
    int i = [kTable findMember:aPattern];
    
    if (i == -1) {
	[kTable insert:[aPattern cheapCopy]];
	[vTable insert:[anExpression cheapCopy]];
	return YES;
    } else {
	return [anExpression isEqual:[vTable memberAt:i]];
    }
}

static BOOL doCltnMatch(id aCltn,id bCltn,id kTable,id vTable)
{
    if ([aCltn numMembers] != [bCltn numMembers]) {
	return NO;
    } else {
	int i,n = [aCltn numMembers];
	for(i=0;i<n;i++) {
	    id aMember = [aCltn memberAt:i];
	    id bMember = [bCltn memberAt:i];
	    if (NO == [aMember isMatch:bMember keyTable:kTable valueTable:vTable]) return NO;
	}
	return YES;
    }
}

- (BOOL) isMatch:anExpression keyTable:kTable valueTable:vTable
{
    if (self == anExpression) {
	return YES;
    } else {
    
	if (headEq("Blank")) return YES; /* matches anything */
	if (headEq("Pattern")) {
	    return doPatternMatch([self firstExpression],anExpression,kTable,vTable);
	}
	
	if (headEq([anExpression head])) {
	    if (expressions) {
		return doCltnMatch(expressions,[anExpression expressions],kTable,vTable);
	    } else {
		return [object isEqual:[anExpression object]];
	    }
	} else {
	    return NO;
	}
    }
}

/*
 * Key - Value Defining - Undefining.
 */

static id keyTable = (id)1;	/* nil is a valid receiver for [] */
static id valueTable = (id)1;

static void setUpKeyValueTables()
{
    if (keyTable == (id)1) keyTable = [CACollectionClass new];
    if (valueTable == (id)1) valueTable = [CACollectionClass new];
    assert([keyTable numMembers] == [keyTable numMembers]);
}

/*
 * Check whether suitable as a key
 */

static BOOL checkKey(id anExpression)
{
    STR head = [anExpression head];
    if (headEq("Integer")) return NO;
    if (headEq("Float")) return NO;
    if (headEq("Part")) return NO;
    return YES;
}

/*
 * Search key equal in the sense of |isEqual:|.
 * Return index of the key or -1 if not found.
 * Work from the end toward the beginning s.t. defs added to the
 * end of the table take precedence over those at the beginning
 */

static int findKeyEqual(id anExpression)
{
    setUpKeyValueTables();return [keyTable findMember:anExpression];
}

/*
 * Add a copy of "aKey" as a key and a copy of "aValue" as value.
 * Because the pair is added to the end of the table, it will effectively
 * replace a previous definition of "aKey".
 */

static void defineKeyValue(id aKey,id aValue)
{
    setUpKeyValueTables();
    [keyTable insert:[aKey cheapCopy]];
    [valueTable insert:[aValue cheapCopy]];
}

/*
 * Remove and free the last "aKey" in the keyTable and valueTable.
 * "aKey" shouldn't be a key itself.  Returns zero if nothing removed.
 */

static int undefineLastKey(id aKey)
{
    int i;
    
    setUpKeyValueTables();
    
    i = findKeyEqual(aKey);
    if (i == -1) {
	return 0;
    } else {
	[[keyTable removeMemberAt:i] free];
	[[valueTable removeMemberAt:i] free];
	return 1;
    }
}

/*
 * Remove and free *all* occurences of "aKey" in the keyTable and valueTable.
 * "aKey" shouldn't be any of the keys itself.
 */

static void undefineKey(id aKey)
{
    int notDone; do { notDone = undefineLastKey(aKey); } while (notDone);
}

/*
 * Rule Defining
 */


/*
 * Check whether an expression is a well formed rule or a collection of rules
 */

static BOOL checkRule(id aRule)
{
    if (strEq([aRule head],"Rule") == NO) return NO;
    if ([aRule numExpressions] != 2) return NO;
    if (checkKey([aRule firstExpression]) == NO) return NO;
    return YES;
}

static BOOL checkRuleCltn(id aCltn)
{
    int i,n = [aCltn numMembers];
    for(i=0;i<n;i++) if (NO == checkRule([aCltn memberAt:i])) return NO;
    return YES;
}

static BOOL checkRules(id anExpression)
{
    STR head = [anExpression head];
    if (headEq("Rule")) return checkRule(anExpression);
    if (headEq("Collection")) return checkRuleCltn([anExpression expressions]);
    return NO;
}

/*
 * Given a Rule, add a copy of the left side as key and a copy of the right
 * side as value.
 * Define a Rule or a Collection of Rules
 */

static void defineRule(id aRule)
{
    assert(checkRule(aRule));
    defineKeyValue([aRule firstExpression],[aRule lastExpression]);
}

static void defineCltnRules(id aCltn)
{
    int i,n = [aCltn numMembers];
    for(i=0;i<n;i++) defineRule([aCltn memberAt:i]);
}

static void defineRules(id anExpression)
{
    assert(checkRules(anExpression));
    if (strEq([anExpression head],"Rule")) {
	defineRule(anExpression);
    } else {
	defineCltnRules([anExpression expressions]);
    }
}

/*
 * Undefine given a Rule or a Collection of Rules
 */

static void undefineRule(id aRule)
{
    assert(checkRule(aRule));
    undefineLastKey([aRule firstExpression]);
}

static void undefineCltnRules(id aCltn)
{
    int i,n = [aCltn numMembers];
    for(i=0;i<n;i++) undefineRule([aCltn memberAt:i]);
}

static void undefineRules(id anExpression)
{
    assert(checkRules(anExpression));
    if (strEq([anExpression head],"Rule")) {
	undefineRule(anExpression);
    } else {
	undefineCltnRules([anExpression expressions]);
    }
}

/*
 * Performing a "one pass" substitution given a rule or a set of rules.
 *
 * Invokes |substitute| and not |simplify| since that would cause an infinite
 * loop in cases such as X /. X->X
 *
 * Works by temporarily adding items to (the end of) the symbol
 * tables, and then removing them again.  Works if findKeyEqual works from the
 * end towards the beginning of the table.
 */

static id subsRules(id lh,id rh)
{
    id res;
    
    if (checkRules(rh)) {
	defineRules(rh);
	res = [lh substitute];
	undefineRules(rh);
    } else {
	res = errMsg("Usage: anExpression /. aRule or {rules...}");
    }
    
    return res;
}

/*
 * Given a cltn of keys and values, perform a "one pass" substitution
 * Defined in terms of "subsRules" (maybe we could do it the other way?)
 */

static id subsKeyValues(id anExpression,id kTable,id vTable)
{
    id res;
    int i,n = [kTable numMembers];
    id rules = [CACollectionClass new];
    for(i=0;i<n;i++) {
	id aKey   = [[kTable memberAt:i] cheapCopy];
	id aValue = [[vTable memberAt:i] cheapCopy];
	[rules insert:[CAExpressionClass newBinaryHead:"Rule":aKey:aValue]];
    }
    rules = [CAExpressionClass newHead:"Collection" expressions:rules];
    res   = subsRules(anExpression,rules);
    [rules free];return res;
}

/*
 * Substitute anExpression by a match.
 * An exact match (|isEqual:|) has precedence over any pattern matching
 * Then try to match anExpression by |isMatch:keyTable:valueTable:|.
 *
 * If that fails returns *nil*.
 */

static id subsExactMatch(id anExpression)
{
    int n = findKeyEqual(anExpression);
    if (n != -1) return [[valueTable memberAt:n] cheapCopy];
    
    return nil;
}

static id subsMatch(id anExpression)
{
    int n;
    id exact = subsExactMatch(anExpression);if (exact) return exact;
    
    n = [keyTable numMembers];
    while (n--) {
	id aKey;
	BOOL match;
	id kTable = [CACollectionClass new];
	id vTable = [CACollectionClass new];
	aKey  = [keyTable memberAt:n];
	match = [aKey isMatch:anExpression keyTable:kTable valueTable:vTable];
	if (match) {
	    id res = subsKeyValues([valueTable memberAt:n],kTable,vTable);
	    [kTable free];[vTable free];
	    return res;
	} else {
	    [kTable free];[vTable free];
	}
    }
    
    return nil;
}


static id subsSymbol(id self)
{
    int i = findKeyEqual(self);
    if (i == -1) {
	return [self cheapCopy];
    } else {
	return [[valueTable memberAt:i] cheapCopy];
    }
}

static id subsReplaceAll(id lh,id rh)
{
    return subsRules(lh,rh);
}

/*
 *
 * Default Substituter, Examples :
 *
 * (1) Cos(x) /. Cos(x) -> 1 gives a subsDefault(Cos(x)) after defining
 * "Cos(x)" as "1".   Here we need subsExactMatch to get "1".
 *
 * (2) f(n_) := f(n-1) + f(n-2) defines f(n_).  f(2) gives a subsDefault
 * of f(n-1) + f(n-2) after defining "n" as "2".  Here we need subsDescend to
 * get f(2-1) + f(2-2)
 *
 * There was a nasty bug in (2) when matching (using subsMatch) instead of
 * subsExactMatch.  f(2-1) was matched towards f(1-1-1) and so on.
 *
 * Note: another solution might me in implementing something else than
 * subsMatch that does only one match ie. no recursion ? Like that we might
 * be able to allow patterns...
 */

static id subsDefaultDescend(STR head,id aCltn)
{
    int i,n = [aCltn numMembers];
    id newCltn = [CACollectionClass new];
    for(i=0;i<n;i++) {
	id e = [[aCltn memberAt:i] substitute];
	assert([e isErrorMsg] == NO);
	[newCltn insert:e];
    }
    return [CAExpressionClass newHead:head expressions:newCltn];
}

static id subsDefault(id self)
{
    id match = subsExactMatch(self);
    
    if (match) {
	return match;
    } else {
	return subsDefaultDescend([self head],[self expressions]);
    }
}

- substitute
{
    if (headEq("Symbol")) return subsSymbol(self);
    if (headEq("ReplaceAll")) {
    return subsReplaceAll([self firstExpression],[self lastExpression]);
    }
    return (expressions)?subsDefault(self):[self cheapCopy];
}


- addAsVariableTo:aCltn
{
    if (numericalAtomHead(head)) return aCltn;
    [aCltn insertIfAbsent:[self cheapCopy]];
    return aCltn;
}

static id AddAsVariablesTo(id aCltn,id expressions)
{
    int i;
    for(i=0;i<[expressions numMembers];i++) {
	[[expressions memberAt:i] addAsVariableTo:aCltn];
    }
    return aCltn;
}

/* Additive Variables : in 1/a + a there are two vars {1/a,a}
 * (x+y)^2 also counts as a single variable
 */

static id AddAdditiveVariablesTo(id aCltn,id expressions)
{
    int i;
    for(i=0;i<[expressions numMembers];i++) {
	[[expressions memberAt:i] addAdditiveVariablesTo:aCltn];
    }
    return aCltn;
}

- addAdditiveVariablesTo:aCltn
{
    BOOL descend = NO;
    
    descend = (descend || strEq(head,"Collection"));
    descend = (descend || headAdditive(head));
    
    if (descend) {
	return AddAdditiveVariablesTo(aCltn,expressions);
    } else {
	if (strEq(head,"Multiply")) {
	    return AddAsVariablesTo(aCltn,expressions);
	} else {
	    return [self addAsVariableTo:aCltn];
	}
    }
}

- additiveVariables
{
    return [self addAdditiveVariablesTo:[CACollectionClass new]];
}

/* Multipicative Variables : in 1/a * a there's one var {a}
 * (x+y)^2 contains the variable {x+y}
 */

static id AddMultiplicativeVariablesTo(id aCltn,id expressions)
{
    int i;
    for(i=0;i<[expressions numMembers];i++) {
	[[expressions memberAt:i] addMultiplicativeVariablesTo:aCltn];
    }
    return aCltn;
}

- addMultiplicativeVariablesTo:aCltn
{
    BOOL descend = NO;
    
    descend = (descend || strEq(head,"Collection"));
    descend = (descend || headMultiplicative(head));
    
    if (descend) {
	return AddMultiplicativeVariablesTo(aCltn,expressions);
    } else {
	if (strEq(head,"Power")) {
	    return AddAsVariablesTo(aCltn,expressions);
	} else {
	    return [self addAsVariableTo:aCltn];
	}
    }
}

- multiplicativeVariables
{
    return [self addMultiplicativeVariablesTo:[CACollectionClass new]];
}

/* Expand Variables : basically any symbol or non-elementary operator */

static id AddExpandVariablesTo(id aCltn,id expressions)
{
    int i;
    for(i=0;i<[expressions numMembers];i++) {
	[[expressions memberAt:i] addExpandVariablesTo:aCltn];
    }
    return aCltn;
}

- addExpandVariablesTo:aCltn
{
    BOOL descend = NO;
    
    descend = (descend || strEq(head,"Collection"));
    descend = (descend || headAdditive(head));
    descend = (descend || headMultiplicative(head));
    descend = (descend || strEq(head,"Power"));
    
    if (descend) {
	return AddExpandVariablesTo(aCltn,expressions);
    } else {
	return [self addAsVariableTo:aCltn];
    }
}

- expandVariables
{
    return [self addExpandVariablesTo:[CACollectionClass new]];
}


static id simpSymbol(id self)
{
    int i = findKeyEqual(self);
    if (i == -1) {
	return [self cheapCopy];
    } else {
	return [[valueTable memberAt:i] simplify];
    }
}

static id simpSetDelayed(id lh,id rh)
{
    if (checkKey(lh)) {
	defineKeyValue(lh,rh);
	/* first we did a return [rh simplify]
	 * but I think it's better not to since a) it doesn't reflect
	 * that the assignment is unevaluated and b) it doesnt work with
	 * the Fibonacci example Fib(n_) := Fib(n-1) + Fib(n-2)
         */
	return [rh cheapCopy];
    } else {
	return errMsg("Left side cannot be redefined");
    }
}

static id simpSet(id lh,id rh)
{
    id r,t = [rh simplify];
    r = simpSetDelayed(lh,t); /* evaluated right side */
    [t free];return r;
}

static id simpUnset(id lh)
{
    if (checkKey(lh)) {
	undefineKey(lh);return [lh cheapCopy];
    } else {
	return errMsg("Left side cannot be undefined");
    }
}

static id simpReplaceAll(id lh,id rh)
{
    id res = [lh simplify]; /* f(x) /. x->y is not equivalent to f(y) */
    res = idReplace(res,subsRules(res,rh));
    return idReplace(res,[res simplify]);
}

static id simpDefaultDescend(STR head,id aCltn)
{
    int i,n = [aCltn numMembers];
    id newCltn = [CACollectionClass new];
    for(i=0;i<n;i++) {
	id e = [[aCltn memberAt:i] simplify];
	if ([e isErrorMsg]) {
	    [newCltn free];return e;
	} else {
	    [newCltn insert:e];
	}
    }
    return [CAExpressionClass newHead:head expressions:newCltn];
}

static id simpDefault(STR head,id aCltn)
{
    int i;
    id descend,match;
    
    descend = simpDefaultDescend(head,aCltn);
    
    if ((match = subsMatch(descend))) {
	[descend free];return idReplace(match,[match simplify]);
    } else {
	return descend;
    }
}


static id simpFactorialInt(int n)
{
    id c,d;
    
    if (n < 0) return errMsg("Argument of Factorial is a negative integer");
    if (n > 10000) return errMsg("Argument of Factorial is unreasonable large");
    
    c = [CAIntegerClass newFactorial:n];
    d = [c asExpression];[c free];return d;
}

static id simpFactorial(id a)
{
    id b = [a simplify];
    
    if (strEq([b head],"Integer")) {
	id c = simpFactorialInt([b intValue]);
	[b free];return c;
    }
    
    return [CAExpressionClass newUnaryHead:"Factorial":b];
}


static id simpPartInt(id a,int n)
{
    id aCltn = [a expressions];assert(aCltn != nil);
    if (0 <= n && n < [aCltn numMembers]) {
	return [[aCltn memberAt:n] cheapCopy];
    } else {
	return errMsg("Index out of bounds");
    }
}

static id simpPart(id a,id b)
{
    id aa = [a simplify];
    if (numericalAtomHead([aa head])) {
	[aa free];return errMsg("Can't index numerical value");
    } else {
	id bb = [b simplify];
	if (strEq([aa head],"Collection") && strEq([bb head],"Integer")) {
	    id c = simpPartInt(aa,[bb intValue]);
	    [aa free];[bb free];return c;
	} else {
	    return [CAExpressionClass newBinaryHead:"Part":aa:bb];
	}
    }
}

static id simpNumericalPowerInt(id a,int n)
{
    id c,d;
    if ([a isZero] && n == 0) return errMsg("Indeterminate 0^0");
    c = [a power:+n];
    d = [c asExpression];[c free];return d;
}

static id simpPowerInt(id a,int n)
{
    id e = [a simplify];
    
    if (numericalAtomHead([e head]) && n >= 0) {
	id c = simpNumericalPowerInt([e object],n);
	[e free];return c;
    } else {
	id m;
	if (n == 1) return e;
	if (n == 0) { [e free];return [CAExpressionClass newIntValue:1]; }
	m = [CAExpressionClass newIntValue:n];
	return [CAExpressionClass newBinaryHead:"Power":e:m];
    }
}

static id simpPower(id a,id b)
{
    id n = [b simplify];
    
    if (strEq([n head],"Integer")) {
	id c = simpPowerInt(a,[n intValue]);
	[n free];return c;
    } else {
	return [CAExpressionClass newBinaryHead:"Power":[a simplify]:n];
    }
}

static id simpAlgInteger(id e)
{
    id c;static id dummy;
    if (!dummy) dummy = [CAIntegerClass new];
    
    if ((c = [dummy expression:e])) {
	id d = [c asExpression];[c free];return d;
    } else {
	return nil;
    }
}

static id simpAlgFloat(id e)
{
    /* we don't want 1/2 to evaluate to .5 */
    if ([e containsHead:"Float"]) {
    id c;static id dummy;
    if (!dummy) dummy = [CAFloatClass new];
    
    if ((c = [dummy expression:e])) {
	id d = [c asExpression];[c free];return d;
    } else {
	return nil;
    }
    } else {
	return nil;
    }
}


static id simpAlgRational(id e)
{
    id c;static id dummy;
    if (!dummy) {
    dummy = [CAFractionClass newTotalFraction:[CAIntegerClass new]];
    }
    
    if ((c = [dummy expression:e])) {
	id d = [c asExpression];[c free];return d;
    } else {
	return nil;
    }
}

static id simpAlgIntegerPolynomial(id e,id vars)
{
    id c,dummy;
    dummy = [CAPolynomialClass newScalar:[CAIntegerClass new]];
    dummy = idReplace(dummy,[dummy collect:vars]); /* vars remains in scope */
    dummy = idReplace(dummy,[dummy makeExpanded]);

    if ((c = [dummy expression:e])) {
	id d = [c asExpression];[dummy free];[c free];return d;
    } else {
	[dummy free];return nil;
    }
}

static id simpAlgFloatPolynomial(id e,id vars)
{
    if ([e containsHead:"Float"]) {
    id c,dummy;
    dummy = [CAPolynomialClass newScalar:[CAFloatClass new]];
    dummy = idReplace(dummy,[dummy collect:vars]); /* vars remains in scope */
    dummy = idReplace(dummy,[dummy makeExpanded]);

    if ((c = [dummy expression:e])) {
	id d = [c asExpression];[dummy free];[c free];return d;
    } else {
	[dummy free];return nil;
    }
    } else {
	return nil;
    }
}

static id simpAlg(id ee,id vars)
{
    id s;
    
    if ([vars numMembers] == 0) {
	s = simpAlgInteger(ee);if (s) return s;
	s = simpAlgFloat(ee);if (s) return s;
	s = simpAlgRational(ee);if (s) return s;
    } else {
	s = simpAlgIntegerPolynomial(ee,vars);if (s) return s;
	s = simpAlgFloatPolynomial(ee,vars);if (s) return s;
    }
    
    return nil;
}

- simplifyIfNotAdditive
{    
    if (headAdditive(head)) {
	int i,n = [expressions numMembers];
	id newCltn = [CACollectionClass new];
	for(i=0;i<n;i++) {
	    id e = [[expressions memberAt:i] simplifyIfNotAdditive];
	    if ([e isErrorMsg]) {
		[newCltn free];return e;
	    } else {
		[newCltn insert:e];
	    }
	}
	return [CAExpressionClass newHead:head expressions:newCltn];
    } else {
	return [self simplify];
    }
}

static id simpAdditive(id e)
{
    id res,vars,ee;
      
    /* descend and simplify selectively */
    ee  = [e simplifyIfNotAdditive];

    /* get vars */
    vars = [ee additiveVariables];
    
    /* try to recognize */
    res = simpAlg(ee,vars);

    [vars free];if (res) {
	[ee free];return res;
    } else {
	return ee;
    }
}

- simplifyIfNotMultiplicative
{    
    if (headMultiplicative(head)) {
	int i,n = [expressions numMembers];
	id newCltn = [CACollectionClass new];
	for(i=0;i<n;i++) {
	    id e = [[expressions memberAt:i] simplifyIfNotMultiplicative];
	    if ([e isErrorMsg]) {
		[newCltn free];return e;
	    } else {
		[newCltn insert:e];
	    }
	}
	return [CAExpressionClass newHead:head expressions:newCltn];
    } else {
	return [self simplify];
    }
}

static id simpMultiplicative(id e)
{
    id res,vars,ee;
      
    /* descend and simplify selectively */
    ee  = [e simplifyIfNotMultiplicative];

    /* get vars */
    vars = [ee multiplicativeVariables];
    
    /* try to recognize */
    res = simpAlg(ee,vars);

    [vars free];if (res) {
	[ee free];return res;
    } else {
	return ee;
    }
}

- simplifyIfNotElementary
{    
    if (headAdditive(head) || headMultiplicative(head) || strEq(head,"Power")) {
	int i,n = [expressions numMembers];
	id newCltn = [CACollectionClass new];
	for(i=0;i<n;i++) {
	    id e = [[expressions memberAt:i] simplifyIfNotElementary];
	    if ([e isErrorMsg]) {
		[newCltn free];return e;
	    } else {
		[newCltn insert:e];
	    }
	}
	return [CAExpressionClass newHead:head expressions:newCltn];
    } else {
	return [self simplify];
    }
}

static id simpExpand(id e)
{
    id res,vars,ee;
      
    /* descend and simplify selectively */
    ee  = [e simplifyIfNotElementary];

    /* get expanding vars */
    vars = [ee expandVariables];
    
    /* try to recognize */
    res = simpAlg(ee,vars);

    [vars free];if (res) {
	[ee free];return res;
    } else {
	return ee;
    }
}

static id simpAdditiveVariables(id e)
{
    id res,aCltn;
    e = [e simplify];
    aCltn = [e additiveVariables];
    res = [aCltn asExpression];
    [aCltn free];[e free];
    return res;
}

static id simpMultiplicativeVariables(id e)
{
    id res,aCltn;
    e = [e simplify];
    aCltn = [e multiplicativeVariables];
    res = [aCltn asExpression];
    [aCltn free];[e free];
    return res;
}

static id simpExpandVariables(id e)
{
    id res,aCltn;
    e = [e simplify];
    aCltn = [e expandVariables];
    res = [aCltn asExpression];
    [aCltn free];[e free];
    return res;
}

static id simpVariablesFor(id e,id s)
{
    if (strEq([s head],"Symbol")) {
	STR operator = [[s object] strValue];
	if (strEq(operator,"Expand")) return simpExpandVariables(e);
	if (strEq(operator,"Add")) return simpAdditiveVariables(e);
	if (strEq(operator,"Multiply")) return simpMultiplicativeVariables(e);
    }
    return errMsg("Usage: Variables(anExpression)");
}

static id simpVariables(id self)
{
    if ([self numExpressions] == 1) return simpExpandVariables([self firstExpression]);
    if ([self numExpressions] == 2) return simpVariablesFor([self firstExpression],[self lastExpression]);
    return errMsg("Usage: Variables(anExpression)");
}

static id simpOut(id e)
{
    id ee = [e simplify];
    if (strEq([ee head],"Integer")) {
	int i = [ee intValue];
	ee = idReplace(ee,copyOfOutputAt(i));
	return idReplace(ee,[ee simplify]);
    } else {
	return [CAExpressionClass newUnaryHead:"Out":ee];
    }
}

static id simpIn(id e)
{
    id ee = [e simplify];
    if (strEq([ee head],"Integer")) {
	int i = [ee intValue];
	ee = idReplace(ee,copyOfInputAt(i));
	/* evaluating the input doesnt make much sense to me as
	 * then you'd more likely use Out(n)
         */
	return ee;
    } else {
	return [CAExpressionClass newUnaryHead:"In":ee];
    }
}


static id getExpression(STR fileName)
{
    char aBuffer[24 * 1024];/* MPWCPPC has a max of 32KB */
    FILE *aFile = fopen(fileName,"r");
    if (aFile) {
	int i = 0;id res; /* get everything, not just the first line */
	while (!feof(aFile)) aBuffer[i++] = fgetc(aFile);
	fclose(aFile);
	res = [CAExpressionClass newStrValueNoEval:aBuffer];
	 /* don't use +newStrValue: because it changes history */
	return idReplace(res,[res simplify]);
    } else {
	return errMsg("Can't read from file");
    }
}

static id putExpression(id anExpression,STR fileName)
{
    FILE *aFile = fopen(fileName,"w");
    if (aFile) {
	[anExpression printToFile:aFile];fprintf(aFile,"\n");
	fclose(aFile);return [anExpression cheapCopy];
    } else {
	return errMsg("Can't write to file");
    }
}

static id simpGet(id aFilename)
{
    if (strEq([aFilename head],"Symbol")) {
	return getExpression([[aFilename object] strValue]);
    } else {
	return errMsg("Usage : << fileName");
    }
}

static id simpPut(id anExpression,id aFilename)
{
    if (strEq([aFilename head],"Symbol")) {
	id r,s = [anExpression simplify];
	r = putExpression(s,[[aFilename object] strValue]);
	[s free];return r;
    } else {
	return errMsg("Usage : anExpression >> fileName");
    }
}

static id simpCount(id self)
{
    if ([self numExpressions] == 1) {
	int n;
	id a = [[self firstExpression] simplify];
	n = ([a object])?0:[[a expressions] numMembers];
	return [CAExpressionClass newIntValue:n];
    } else {
	return errMsg("Usage: Count(anExpression)");
    }
}

static id simpInsertAt(id cltn,id member,id anIndex)
{
    cltn   = [cltn simplify];
    anIndex  = [anIndex simplify];
    member = [member simplify];
    
    if (strEq([cltn head],"Collection") && strEq([anIndex head],"Integer")) {
	id res;int i = [anIndex intValue];

	/* |insert:at:| is equivalent to |insert:| if i == numMembers */
	
	if (0 <= i && i <= [cltn numExpressions]) {
	    id new = [[[cltn expressions] copy] insert:member at:i];
	    res = [CAExpressionClass newHead:"Collection" expressions:new];
	} else {
	    [member free];res = errMsg("Index out of bounds");
	}
	
	[cltn free];[anIndex free];return res;
    } else {
	id new = [CACollectionClass new];
	[new insert:cltn];[new insert:member];[new insert:anIndex];
	return [CAExpressionClass newHead:"Insert" expressions:new];
    }
}

static id simpInsert(id self)
{
    if ([self numExpressions] == 3) {
	id cltn,member,anIndex;
	id expressions = [self expressions];
	cltn   = [expressions memberAt:0];
	member = [expressions memberAt:1];
	anIndex  = [expressions memberAt:2];
	return simpInsertAt(cltn,member,anIndex);
    } else {
	return errMsg("Usage: Insert(aCollection,anExpression,index)");
    }
}

static id simpAppendAux(id cltn,id member)
{
    cltn   = [cltn simplify];
    member = [member simplify];
    
    if (strEq([cltn head],"Collection")) {
	id new = [[[cltn expressions] copy] insert:member];
	new = [CAExpressionClass newHead:"Collection" expressions:new];
	[cltn free];return new;
    } else {
	id new = [CACollectionClass new];
	[new insert:cltn];[new insert:member];
	return [CAExpressionClass newHead:"Append" expressions:new];
    }
}

static id simpAppend(id self)
{
    if ([self numExpressions] == 2) {
	id cltn,member;
	cltn   = [self firstExpression];
	member = [self lastExpression];
	return simpAppendAux(cltn,member);
    } else {
	return errMsg("Usage: Append(aCollection,anExpression)");
    }
}

static id indexRule(id anIndex,int i)
{
    id aKey   = [anIndex cheapCopy];
    id aValue = [CAExpressionClass newIntValue:i];
    return [CAExpressionClass newBinaryHead:"Rule":aKey:aValue];
}

static id newTableRow(id anExpression,id anIndex,int imin,int imax)
{
    int i;
    id aCltn = [CACollectionClass new];
    
    for(i=imin;i<imax;i++) {
	id aRule = indexRule(anIndex,i);
	[aCltn insert:subsRules(anExpression,aRule)];
	aRule = [aRule free];
    }
    
    return [CAExpressionClass newHead:"Collection" expressions:aCltn];
}

static id newTable(id anExpression,id indexi,int imin,int imax,id indexj,int jmin,int jmax)
{
    int i;
    id aCltn = [CACollectionClass new];
    
    for(i=imin;i<imax;i++) {
	id e,aRule = indexRule(indexi,i);
	e = subsRules(anExpression,aRule);
	[aCltn insert:newTableRow(e,indexj,jmin,jmax)];
	e = [e free];aRule = [aRule free];
    }
    
    return [CAExpressionClass newHead:"Collection" expressions:aCltn];
}

static BOOL checkTuple(id anExpression)
{
    if (strEq([anExpression head],"Collection")) {
	return ([anExpression numExpressions] == 2);
    } else {
	return NO;
    }
}

/*
 * Returns *nil* for symbolic max values
 */

static id simpTableOne(id e,id i)
{
    if (checkTuple(i)) {
	id  anIndex = [i firstExpression];
	id  imax = [[i lastExpression] simplify];
	if (strEq([imax head],"Integer")) {
	    /* negative value will give empty collection */
	    id res = newTableRow(e,anIndex,0,[imax intValue]);
	    [imax free];return idReplace(res,[res simplify]);
	} else {
	    [imax free];return nil;
	}
    } else {
	return errMsg("Usage: Table(expr,{i,m}) or Table(expr,{i,m},{j,n})");
    }
}

/*
 * Maybe to generalize to more dimensions, we could use recursion on expression
 * ie. set expression = Table() and invoke simpTableOne()
 *
 * Returns *nil* for symbolic max values
 */

static id simpTableTwo(id e,id i,id j)
{
    if (checkTuple(i)) {
	id imax,jmax;
	id indexi,indexj;
	indexi = [i firstExpression];
	indexj = [j firstExpression];
	imax   = [[i lastExpression] simplify];
	jmax   = [[j lastExpression] simplify];
	if (strEq([imax head],"Integer") && strEq([jmax head],"Integer")) {
	    int m = [imax intValue];
	    int n = [jmax intValue];
	    /* negative value will give empty collection */
	    id res = newTable(e,indexi,0,m,indexj,0,n);
	    [imax free];[jmax free];return idReplace(res,[res simplify]);
	} else {
	    [imax free];[jmax free];return nil;
	}
    } else {
	return errMsg("Usage: Table(expr,{i,m}) or Table(expr,{i,m},{j,n})");
    }
}

static id simpTable(id self)
{
    id res;
    id expressions = [self expressions];
    
    if ([self numExpressions] == 2) {
	id e = [expressions memberAt:0];
	id i = [expressions memberAt:1];
	res = simpTableOne(e,i);return (res)?res:[self cheapCopy];
    }
    if ([self numExpressions] == 3) {
	id e = [expressions memberAt:0];
	id i = [expressions memberAt:1];
	id j = [expressions memberAt:2];
	res = simpTableTwo(e,i,j);return (res)?res:[self cheapCopy];
    }
    return errMsg("Usage: Table(expr,{i,m}) or Table(expr,{i,m},{j,n})");
}

static id simpJoinCltn(id a,id b)
{
    int i;
    id new = [CACollectionClass new];
    for(i=0;i<[a numMembers];i++) [new insert:[[a memberAt:i] cheapCopy]];
    for(i=0;i<[b numMembers];i++) [new insert:[[b memberAt:i] cheapCopy]];
    return [CAExpressionClass newHead:"Collection" expressions:new];
}

static id simpJoin(id self)
{
    if ([self numExpressions] == 2) {
	id a = [[self firstExpression] simplify];
	id b = [[self lastExpression] simplify];
	if (strEq([a head],"Collection") && strEq([b head],"Collection")) {
	    id res = simpJoinCltn([a expressions],[b expressions]);
	    [a free];[b free];return res;
	} else {
	    return [CAExpressionClass newBinaryHead:"Join":a:b];
	}
    } else {
	return errMsg("Usage: Join(aCltn,bCltn)");
    }
}

static BOOL isMatrixExpression(id anExpression)
{
    if (strEq([anExpression head],"Collection")) {
	int i,n;
	id aCltn = [anExpression expressions];
	n = [aCltn numMembers];
	for(i=0;i<n;i++) {
	    id row = [aCltn memberAt:i];
	    if (!strEq([row head],"Collection")) return NO;
	    if ([row numExpressions] != n) return NO;
	}
	return YES;
    } else {
	return NO;
    }
}

/*
 * returns a CAMatrix or nil if not possible
 */

static id simpIntegerMatrix(id e)
{
    static id dummy;
    
    if (!dummy) {
	dummy = [CAIntegerClass new];
	dummy = [CAMatrixClass newScalar:dummy numRows:0 numColumns:0];
    }
    
    return [dummy expression:e];
}

static id simpRationalMatrix(id e)
{
    static id dummy;
    
    if (!dummy) {
	dummy = [CAFractionClass newTotalFraction:[CAIntegerClass new]];
	dummy = [CAMatrixClass newScalar:dummy numRows:0 numColumns:0];
    }
    
    return [dummy expression:e];
}


static id simpIntegerPolynomialMatrixInVars(id e,id vars)
{
    id m,dummy = [CAPolynomialClass newScalar:[CAIntegerClass new]];
    dummy = idReplace(dummy,[dummy collect:vars]); /* vars remains in scope */
    dummy = idReplace(dummy,[dummy makeExpanded]);
    dummy = [CAMatrixClass newScalar:dummy numRows:0 numColumns:0];
    
    m = [dummy expression:e];[dummy free];return m;
}

static id simpMatrixInVars(id e,id vars)
{
    id m;
    if ([vars numMembers] == 0) {
	m = simpIntegerMatrix(e);if (m) return m;
	m = simpRationalMatrix(e);if (m) return m;
	return nil;
    } else {
	m = simpIntegerPolynomialMatrixInVars(e,vars);if (m) return m;
	return nil;
    }
}

static id simpMatrix(id e)
{
    if (isMatrixExpression(e)) {
	id m,vars = [e expandVariables];
	m = simpMatrixInVars(e,vars);
	[vars free];return m;
    } else {
	return nil;
    }
}

/*
 * Simplify Det Expression, or if fails for whatever reason, leave unevaluated
 */

static id simpDetMatrix(id self)
{
#ifdef __BEOS__
    id m;
    id s;
    
    s    = [self simplify];
    m    = simpMatrix(s);
    
    if (m) {
	id det = [m determinant];[m free];
	if (det) {
	    [s free];return idReplace(det,[det asExpression]);
	} else {
	    return [CAExpressionClass newUnaryHead:"Det":s];
	}
    } else {
	return [CAExpressionClass newUnaryHead:"Det":s];
    }
#else
    id m;
    
    self = [self simplify];
    m    = simpMatrix(self);
    
    if (m) {
	id det = [m determinant];[m free];
	if (det) {
	    [self free];return idReplace(det,[det asExpression]);
	} else {
	    return [CAExpressionClass newUnaryHead:"Det":self];
	}
    } else {
	return [CAExpressionClass newUnaryHead:"Det":self];
    }
#endif
}

static id simpDet(id self)
{
    if ([self numExpressions] == 1) {
	return simpDetMatrix([self firstExpression]);
    } else {
	return errMsg("Usage: Det(aMatrix)");
    }
}

- simplify
{
    if (headEq("Hold")) return [self cheapCopy];
    if (headEq("Symbol")) return simpSymbol(self);
    if (headEq("Integer")) return [self cheapCopy];
    if (headEq("Float")) return [self cheapCopy];
    if (headAdditive(head)) return simpAdditive(self);
    if (headMultiplicative(head)) return simpMultiplicative(self);
    if (headEq("Power")) return simpPower([self firstExpression],[self lastExpression]);
    if (headEq("Factorial")) return simpFactorial([self firstExpression]);
    if (headEq("Set")) return simpSet([self firstExpression],[self lastExpression]);
    if (headEq("SetDelayed")) return simpSetDelayed([self firstExpression],[self lastExpression]);
    if (headEq("Unset")) return simpUnset([self firstExpression]);
    if (headEq("ReplaceAll")) return simpReplaceAll([self firstExpression],[self lastExpression]);
    if (headEq("Out")) return simpOut([self firstExpression]);
    if (headEq("In")) return simpIn([self firstExpression]);
    if (headEq("Variables")) return simpVariables(self);
    if (headEq("Part")) return simpPart([self firstExpression],[self lastExpression]);
    if (headEq("Expand")) return simpExpand([self firstExpression]);
    if (headEq("Det")) return simpDet(self);
    if (headEq("Get")) return simpGet([self firstExpression]);
    if (headEq("Put")) return simpPut([self firstExpression],[self lastExpression]);
    if (headEq("Count")) return simpCount(self);
    if (headEq("Insert")) return simpInsert(self);
    if (headEq("Append")) return simpAppend(self);
    if (headEq("Table")) return simpTable(self);
    if (headEq("Join")) return simpJoin(self);
    return (expressions)?simpDefault(head,expressions):[self cheapCopy];
}


- asExpression
{
    return [self cheapCopy];
}

- expression:anExpression
{
    return [anExpression cheapCopy];
}

/******************************************************************************
 *
 * Printing
 *
 *****************************************************************************/


static void fullFormToFileDelimited(FILE *aFile,STR head,id expressions)
{
    int i,n;
    fprintf(aFile,head);fprintf(aFile,"(");n = [expressions numMembers];
    if (n) [[expressions memberAt:0] fullFormToFile:aFile];
    for(i=1;i<n;i++) {
	fprintf(aFile,",");[[expressions memberAt:i] fullFormToFile:aFile];
    }
    fprintf(aFile,")");
}

- fullFormToFile:(FILE *)aFile
{
    if (atomHead(head)) { [object printToFile:aFile]; return self; }
    fullFormToFileDelimited(aFile,head,expressions);return self;
}

static void printToFileDelimited(FILE *aFile,STR head,id expressions)
{
    int i,n;
    fprintf(aFile,head);fprintf(aFile,"(");n = [expressions numMembers];
    if (n) [[expressions memberAt:0] printToFile:aFile];
    for(i=1;i<n;i++) {
	fprintf(aFile,",");[[expressions memberAt:i] printToFile:aFile];
    }
    fprintf(aFile,")");
}

static void printToFileNegate(FILE *aFile,id a)
{
    BOOL aParen = [a printsSum];
    
    fprintf(aFile,"-");
    if (aParen) fprintf(aFile,"(");
    [a printToFile:aFile];
    if (aParen) fprintf(aFile,")");
}

static void printToFileFactorial(FILE *aFile,id a)
{
    BOOL aParen = [a printsSum] || [a printsProduct];
    
    if (aParen) fprintf(aFile,"(");
    [a printToFile:aFile];
    if (aParen) fprintf(aFile,")");
    
    fprintf(aFile,"!");
}

static void printToFileAdd(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," + ");
    [b printToFile:aFile];
}

static void printToFileSubtract(FILE *aFile,id a,id b)
{
    BOOL bParen = [b printsSum];
    
    [a printToFile:aFile];
    fprintf(aFile," - ");
    if (bParen) fprintf(aFile,"(");
    [b printToFile:aFile];
    if (bParen) fprintf(aFile,")");
}

static void printToFileMultiply(FILE *aFile,id a,id b)
{
    BOOL aParen = [a printsSum];
    BOOL bParen = [b printsSum];

    if (aParen) fprintf(aFile,"(");
    [a printToFile:aFile];
    if (aParen) fprintf(aFile,")");
    fprintf(aFile," ");
    if (bParen) fprintf(aFile,"(");
    [b printToFile:aFile];
    if (bParen) fprintf(aFile,")");
}

static void printToFileDivide(FILE *aFile,id a,id b)
{
    BOOL aParen = [a printsSum];
    BOOL bParen = [b printsSum] || [b printsProduct];

    if (aParen) fprintf(aFile,"(");
    [a printToFile:aFile];
    if (aParen) fprintf(aFile,")");
    fprintf(aFile,"/");
    if (bParen) fprintf(aFile,"(");
    [b printToFile:aFile];
    if (bParen) fprintf(aFile,")");
}

static void printToFilePower(FILE *aFile,id a,id b)
{
    BOOL aParen = [a printsSum] || [a printsProduct];
    BOOL bParen = [b printsSum] || [b printsProduct];

    if (aParen) fprintf(aFile,"(");
    [a printToFile:aFile];
    if (aParen) fprintf(aFile,")");
    fprintf(aFile,"^");
    if (bParen) fprintf(aFile,"(");
    [b printToFile:aFile];
    if (bParen) fprintf(aFile,")");
}

static void printToFileRule(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," -> ");
    [b printToFile:aFile];
}

static void printToFileEqual(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," == ");
    [b printToFile:aFile];
}

static void printToFileNotEqual(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," != ");
    [b printToFile:aFile];
}

static void printToFileLess(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," < ");
    [b printToFile:aFile];
}

static void printToFileGreater(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," > ");
    [b printToFile:aFile];
}

static void printToFileLessEqual(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," <= ");
    [b printToFile:aFile];
}

static void printToFileGreaterEqual(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile," >= ");
    [b printToFile:aFile];
}

static void printToFilePart(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    fprintf(aFile,"[");
    [b printToFile:aFile];
    fprintf(aFile,"]");
}

static void printToFilePattern(FILE *aFile,id a,id b)
{
    [a printToFile:aFile];
    [b printToFile:aFile];
}

static void printToFileFullForm(FILE *aFile,id expressions)
{
    assert([expressions numMembers] == 1);[[expressions memberAt:0] fullFormToFile:aFile];
}

static void printToFileError(FILE *aFile,id expressions)
{
   fprintf(aFile,"\"");[[expressions memberAt:0] printToFile:aFile];fprintf(aFile,"\"");
}

- (BOOL)printsLeadingSign
{
    if (atomHead(head)) return [object printsLeadingSign];
    return strEq(head,"Negate") || strEq(head,"Subtract");
}

- (BOOL)printsSum
{
    if (atomHead(head)) return [object printsSum];
    return strEq(head,"Add") || strEq(head,"Subtract");
}

- (BOOL)printsProduct
{
    if (atomHead(head)) return [object printsProduct];
    return strEq(head,"Divide") || strEq(head,"Multiply");
}

- printToFile:(FILE *)aFile
{
    /* atom expressions */
    if (atomHead(head)) { [object printToFile:aFile]; return self; }

    if (headEq("Collection")) { [expressions printToFile:aFile]; return self; }
    if (headEq("Null")) { return self; }
    if (headEq("Blank")) { fprintf(aFile,"_"); return self; }
    if (headEq("FullForm")) { printToFileFullForm(aFile, expressions);return self; }
    if (headEq("Error")) { printToFileError(aFile, expressions);return self; }
    if (headEq("CompoundExpression")) {
	[[expressions lastMember] printToFile:aFile];return self;
    }
    
    /* unary and binary printing */
    if (headEq("Negate")) {
	printToFileNegate(aFile,[expressions memberAt:0]);return self;
    }
    if (headEq("Factorial")) {
	printToFileFactorial(aFile,[expressions memberAt:0]);return self;
    }
    if (headEq("Add")) {
	printToFileAdd(aFile,[expressions memberAt:0],[expressions memberAt:1]);return self;
    }
    if (headEq("Subtract")) {
	printToFileSubtract(aFile,[expressions memberAt:0],[expressions memberAt:1]);return self;
    }
    if (headEq("Multiply")) {
	printToFileMultiply(aFile,[expressions memberAt:0],[expressions memberAt:1]);return self;
    }
    if (headEq("Divide")) {
	printToFileDivide(aFile,[expressions memberAt:0],[expressions memberAt:1]);return self;
    }
    if (headEq("Power")) {
    	printToFilePower(aFile,[expressions memberAt:0],[expressions memberAt:1]);return self;
    }
    if (headEq("Equal")) {
	printToFileEqual(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("NotEqual")) {
	printToFileNotEqual(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("Less")) {
	printToFileLess(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("LessEqual")) {
	printToFileLessEqual(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("Greater")) {
	printToFileGreater(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("GreaterEqual")) {
	printToFileGreaterEqual(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("Rule")) {
	printToFileRule(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("Part")) {
	printToFilePart(aFile,[self firstExpression],[self lastExpression]);return self;
    }
    if (headEq("Pattern")) {
	printToFilePattern(aFile,[self firstExpression],[self lastExpression]);return self;
    }

    /* default printing */
    printToFileDelimited(aFile,head, expressions);return self;
}

@end

