
/*
 * Computer Algebra Kit (c) 1993,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"

static ca_float_args ca_float_getargs(id self)
{
    ca_float_args args;args.unused = 0;return args;
}

static int ca_float_size(ca_float_args args)
{
    return 1;				/* pointer increment */
}

static int ca_float_bsize(ca_float_args args)
{
    return sizeof(ca_float_c);		/* bytes */
}

static ca_float_t ca_float_id2t(id object)
{
    return [object ca_float_value];	/* read-only */
}

static ca_float_t ca_float_id2ref(id object,ca_float_c *ref)
{
    return [object ca_float_value];	/* read-only, scratch space "ref" */
}

static ca_float_c *ca_float_id2c(id object)
{
    return [object ca_float_reference];	/* write for inplace arithmetic */
} 

static id ca_float_t2id(id proto,ca_float_t f)
{
    return [proto ca_float_value:f];	/* new object containing copy of "f" */
}

static id ca_float_c2id(id proto,ca_float_c *f)
{
    return [proto ca_float_reference:f];/* new object containing "f" */
}

static int ca_float_check(ca_float_t self,ca_float_args args)
{
    return 1;
}

static ca_float_t ca_float_zero(ca_float_args args)
{
    return 0.0;
}

static ca_float_t ca_float_one(ca_float_args args)
{
    return 1.0;
}

static unsigned ca_float_hash(ca_float_t a,ca_float_args args)
{
    return (unsigned)a;
}

static int ca_float_iszero(ca_float_t a,ca_float_args args)
{
    return a == 0.0;
}

static int ca_float_isone(ca_float_t a,ca_float_args args)
{
    return a == +1.0;
}

static int ca_float_isminusone(ca_float_t a,ca_float_args args)
{
    return a == -1.0;
}

static int ca_float_sgn(ca_float_t a,ca_float_args args)
{
    return (a == 0.0)?0:((a<0.0)?-1:+1);
}

static int ca_float_cmp(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return (a == b)?0:((a<b)?-1:+1);
}

static int ca_float_iseq(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a == b;
}

static int ca_float_isop(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a + b == 0.0;
}

static ca_float_t ca_float_neg(ca_float_t a,ca_float_args args)
{
    return -a;
}

static ca_float_t ca_float_dbl(ca_float_t a,ca_float_args args)
{
    return a+a;
}

static ca_float_t ca_float_add(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a + b;
}

static ca_float_t ca_float_sub(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a - b;
}

static ca_float_t ca_float_sqr(ca_float_t a,ca_float_args args)
{
    return a*a;
}

static ca_float_t ca_float_mul(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a*b;
}

static ca_float_t ca_float_inv(ca_float_t a,ca_float_args args)
{
    return 1.0/a;
}

static ca_float_t ca_float_div(ca_float_t a,ca_float_t b,ca_float_args args)
{
    return a/b;
}

static ca_float_t ca_float_pow(ca_float_t a,int n,ca_float_args args)
{
    /* on some systems, pow() needs libm.a
     * return pow(a,n);
     */

    ca_float_t o;
    
    if (n==0) return 1.0;
    if (n==1) return a;
    if (n==2) return a * a;
    
    o = 1.0;
    while (n>1) {
	if (n&1) o = o * a;
	a = a * a;
	n >>= 1;
    }
    
    return o * a;
}

#if !ca_float_isvalue
#error "ca_float_copy_val.fab should not be included"
#endif

static void ca_float_clear(ca_float_c *c,ca_float_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_float_t)0xcafebabe;
    #endif
}

static void ca_float_copy(ca_float_c *c,ca_float_t a,ca_float_args args)
{
    *c = a;
}

static void ca_float_deepcopy(ca_float_c *c,ca_float_t a,ca_float_args args)
{
    *c = a;
}

static void ca_float_move(ca_float_c *c,ca_float_c *a,ca_float_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_float_t)0xcafebabe;
    #endif
}

#if !ca_float_isvalue
#error "ca_float_add_val.fab should not be included"
#endif

static void ca_float_oozero(ca_float_c *c,ca_float_args args)
{
    *c = ca_float_zero(args);
}

static void ca_float_inzero(ca_float_c *a,ca_float_args args)
{
    *a = ca_float_zero(args);
}

static void ca_float_inneg(ca_float_c *a,ca_float_args args)
{
    *a = ca_float_neg(*a,args);
}

static void ca_float_ooneg(ca_float_c *c,ca_float_t a,ca_float_args args)
{
    *c = ca_float_neg(a,args);
}

static ca_float_t ca_float_dblv(ca_float_t a,int v,ca_float_args args)
{
    ca_float_t c = ca_float_dbl(a,args);return (v==+1)?c:ca_float_neg(c,args);
}

static void ca_float_indbl(ca_float_c *a,int v,ca_float_args args)
{
    *a = ca_float_dblv(*a,v,args);
}

static void ca_float_oodbl(ca_float_c *c,ca_float_t a,int v,ca_float_args args)
{
    *c = ca_float_dblv(a,v,args);
}

static ca_float_t ca_float_addvw(ca_float_t a,int v,ca_float_t b,int w,ca_float_args args)
{
    ca_float_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_float_add(a,b,args);
	} else {
	    c = ca_float_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_float_sub(b,a,args);
	} else {
	    c = ca_float_neg(ca_float_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_float_inadd(ca_float_c *a,int v,ca_float_t b,int w,ca_float_args args)
{
    *a = ca_float_addvw(*a,v,b,w,args);
}

static void ca_float_ooadd(ca_float_c *c,ca_float_t a,int v,ca_float_t b,int w,ca_float_args args)
{
    *c = ca_float_addvw(a,v,b,w,args);
}
#if !ca_float_isvalue
#error "ca_float_mul_val.fab should not be included"
#endif

static void ca_float_insqr(ca_float_c *a,ca_float_args args)
{
    *a = ca_float_sqr(*a,args);
}

static void ca_float_oosqr(ca_float_c *c,ca_float_t a,ca_float_args args)
{
    *c = ca_float_sqr(a,args);
}

static void ca_float_inmul(ca_float_c *a,ca_float_t b,ca_float_args args)
{
    *a = ca_float_mul(*a,b,args);
}

static void ca_float_oomul(ca_float_c *c,ca_float_t a,ca_float_t b,ca_float_args args)
{
    *c = ca_float_mul(a,b,args);
}
#if !ca_float_isvalue
#error "ca_float_inv_val.fab should not be included"
#endif

static void ca_float_ininv(ca_float_c *a,ca_float_args args)
{
    *a = ca_float_inv(*a,args);
}

static void ca_float_ooinv(ca_float_c *c,ca_float_t a,ca_float_args args)
{
    *c = ca_float_inv(a,args);
}

#if !ca_float_isvalue
#error "ca_float_div_val.fab should not be included"
#endif

static void ca_float_indiv(ca_float_c *a,ca_float_t b,ca_float_args args)
{
    *a = ca_float_div(*a,b,args);
}

static BOOL ca_float_oodiv(ca_float_c *c,ca_float_t a,ca_float_t b,ca_float_args args)
{
    *c = ca_float_div(a,b,args);return SUCCESS;
}
#if !ca_float_isvalue
#error "ca_float_add_val.fab should not be included"
#endif

static ca_float_t ca_float_mulv(ca_float_t a,ca_float_t b,int v,ca_float_args args)
{
    ca_float_t c = ca_float_mul(a,b,args);return (v==+1)?c:ca_float_neg(c,args);
}

static void ca_float_inmulv(ca_float_c *a,ca_float_t b,int v,ca_float_args args)
{
    *a = ca_float_mulv(*a,b,v,args);
}

static void ca_float_oomulv(ca_float_c *c,ca_float_t a,ca_float_t b,int v,ca_float_args args)
{
    *c = ca_float_mulv(a,b,v,args);
}

static ca_float_t ca_float_addmul(ca_float_t a,int v,ca_float_t b,ca_float_t s,int w,ca_float_args args)
{
    return ca_float_addvw(a,v,ca_float_mul(b,s,args),w,args);
}

static ca_float_t ca_float_addsqr(ca_float_t a,int v,ca_float_t b,int w,ca_float_args args)
{
    return ca_float_addvw(a,v,ca_float_sqr(b,args),w,args);
}

static ca_float_t ca_float_muladd(ca_float_t a,ca_float_t s,int v,ca_float_t b,int w,ca_float_args args)
{
    return ca_float_addvw(ca_float_mul(a,s,args),v,b,w,args);
}

static ca_float_t ca_float_muladdmul(ca_float_t a,ca_float_t s,int v,ca_float_t b,ca_float_t t,int w,ca_float_args args)
{
    return ca_float_addvw(ca_float_mul(a,s,args),v,ca_float_mul(b,t,args),w,args);
}

static void ca_float_inaddmul(ca_float_c *a,int v,ca_float_t b,ca_float_t t,int w,ca_float_args args)
{
    *a = ca_float_addmul(*a,v,b,t,w,args);
}

static void ca_float_ooaddmul(ca_float_c *c,ca_float_t a,int v,ca_float_t b,ca_float_t t,int w,ca_float_args args)
{
    *c = ca_float_addmul(a,v,b,t,w,args);
}

static void ca_float_inaddsqr(ca_float_c *a,int v,ca_float_t b,int w,ca_float_args args)
{
    *a = ca_float_addsqr(*a,v,b,w,args);
}

static void ca_float_ooaddsqr(ca_float_c *c,ca_float_t a,int v,ca_float_t b,int w,ca_float_args args)
{
    *c = ca_float_addsqr(a,v,b,w,args);
}

static void ca_float_inmuladd(ca_float_c *a,ca_float_t s,int v,ca_float_t b,int w,ca_float_args args)
{
    *a = ca_float_muladd(*a,s,v,b,w,args);
}

static void ca_float_oomuladd(ca_float_c *c,ca_float_t a,ca_float_t s,int v,ca_float_t b,int w,ca_float_args args)
{
    *c = ca_float_muladd(a,s,v,b,w,args);
}

static void ca_float_inmuladdmul(ca_float_c *a,ca_float_t s,int v,ca_float_t b,ca_float_t t,int w,ca_float_args args)
{
    *a = ca_float_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_float_oomuladdmul(ca_float_c *c,ca_float_t a,ca_float_t s,int v,ca_float_t b,ca_float_t t,int w,ca_float_args args)
{
    *c = ca_float_muladdmul(a,s,v,b,t,w,args);
}
@implementation Float

+ new
{
    return [self newFloatValue:0];
}

- _setUpFloatValue:(float)floatValue
{
    value = floatValue;return self;
}

- newFloatValue:(float)floatValue
{
    return [self notImplemented:_cmd];
}

+ newFloatValue:(float)floatValue
{
    return [[super new] _setUpFloatValue:floatValue];
}

+ newStrValue:(STR)aString
{
    static id dummy = nil;
    if (!dummy) dummy = [self new]; return [dummy strValue:aString];
}

- empty
{
    self = [self clone];value = 0;return self;
}

- (ca_float_t) ca_float_value
{
    return value;
}

- ca_float_value:(ca_float_t)f
{
    self = [self clone];value = f;return self;
}

- (ca_float_t *) ca_float_reference
{
    return &value;
}

- ca_float_reference:(ca_float_t *)f
{
    self = [self clone];value = *f;return self;
}


- (BOOL) isFloatingPoint
{
    return YES;
}

- (BOOL)inAdditiveSemiGroup		{ return YES; }
- (BOOL)inAdditiveMonoid		{ return YES; }
- (BOOL)inAdditiveGroup			{ return YES; }

- (BOOL)inSemiGroup			{ return YES; }
- (BOOL)inMonoid			{ return YES; }
- (BOOL)inGroup				{ return NO;  }

- (BOOL)inEuclideanDomain		{ return NO; }
- (BOOL)inIntegralDomain		{ return NO; }
- (BOOL)inField				{ return NO; }

- (BOOL)inOrderedSet			{ return YES; }
- (int)characteristic			{ return 0;   }


- (unsigned) hash
{
    return ca_float_hash(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_float_iseq(ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_float_iseq(ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
    }
}

- (float) floatValue
{
    return value;
}

- floatValue:(float)f
{
    return [self ca_float_value:f];
}


- (int) sign
{
    return ca_float_sgn(ca_float_id2t(self),ca_float_getargs(self));
}

- (int) compare:b
{
    [self checkSameClass:b];

    if (self==b) {
    return 0;
    } else {
    return ca_float_cmp(ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
    }
}

- zero
{
    static id zero = nil;
    return (zero)?[zero cheapCopy]:[zero=[self floatValue:0] cheapCopy];
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_float_iszero(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return NO == ca_float_iszero(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) isOpposite:b
{
    [self checkSameClass:b];
    
    if (self == b) {
	return [self isZero] || [self isCharacteristicTwo];
    } else {
	return ca_float_isop(ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
    }
}

- (BOOL) notOpposite:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self notZero] && [self notCharacteristicTwo];
    } else {
	return NO == ca_float_isop(ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
    }
}

- negate
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self cheapCopy];
    } else {
	ca_float_c c;assert([self check]);
	ca_float_ooneg(&c,ca_float_id2t(self),ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return self;
    } else {
	assert([self check]);
	ca_float_inneg(ca_float_id2c(self),ca_float_getargs(self));
	return [self invalidate];
    }
}

- _double:(int)v
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_float_c c;assert([self check]);
	ca_float_oodbl(&c,ca_float_id2t(self),v,ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_float_indbl(ca_float_id2c(self),v,ca_float_getargs(self));
	return [self invalidate];
    }
}

- double
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_float_c c;assert([self check]);
	ca_float_oodbl(&c,ca_float_id2t(self),+1,ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_float_indbl(ca_float_id2c(self),+1,ca_float_getargs(self));
	return [self invalidate];
    }
}

- _add:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);

    if (self == b) {
	return (v==w)?[self _double:v]:[self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b _negate:w];
	    if ([b isZero]) return [self _negate:v];
	}
	if (1) {
	    ca_float_c c;
	    ca_float_ooadd(&c,ca_float_id2t(self),v,ca_float_id2t(b),w,ca_float_getargs(self));
	    return ca_float_c2id(self,&c);
	}
    }

    return nil; /* to quiet some compilers, e.g. WATCOM */
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);

    if (self == b) {
	return (v==w)?[self _doubleSelf:v]:idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return [self _negateSelf:v];
	    if ([self isZero]) return idReplace(self,[b _negate:w]);
	}
	ca_float_inadd(ca_float_id2c(self),v,ca_float_id2t(b),w,ca_float_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- add:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self double];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b cheapCopy];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_float_c c;
	    ca_float_ooadd(&c,ca_float_id2t(self),+1,ca_float_id2t(b),+1,ca_float_getargs(self));
	    return ca_float_c2id(self,&c);
	}
    }

    return nil;
}

- addSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self add:b]);

    if (self == b) {
	return [self doubleSelf];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b cheapCopy]);
	}
	ca_float_inadd(ca_float_id2c(self),+1,ca_float_id2t(b),+1,ca_float_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- subtract:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b negate];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_float_c c;
	    ca_float_ooadd(&c,ca_float_id2t(self),+1,ca_float_id2t(b),-1,ca_float_getargs(self));
	    return ca_float_c2id(self,&c);
	}
    }

    return nil;
}

- subtractSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self subtract:b]);

    if (self == b) {
	return idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b negate]);
	}
	ca_float_inadd(ca_float_id2c(self),+1,ca_float_id2t(b),-1,ca_float_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- one
{
    static id one = nil;
    return (one)?[one cheapCopy]:[one = [self floatValue:1] cheapCopy];
}

- minusOne
{
    static id minusOne = nil;
    return (minusOne)?[minusOne cheapCopy]:[minusOne = [self floatValue:-1] cheapCopy];
}

- (BOOL) isOne
{
    assert([self check]);
    return ca_float_isone(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) notOne
{
    assert([self check]);
    return NO == ca_float_isone(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) isMinusOne
{
    assert([self check]);
    return ca_float_isminusone(ca_float_id2t(self),ca_float_getargs(self));
}

- (BOOL) notMinusOne
{
    assert([self check]);
    return NO == ca_float_isminusone(ca_float_id2t(self),ca_float_getargs(self));
}

- square
{
    if ([self inMonoid] && [self isOne]) return [self cheapCopy];
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return [self one];
	if ([self isZero])     return [self cheapCopy];
    }
    if (1) {
	ca_float_c c;assert([self check]);
	ca_float_oosqr(&c,ca_float_id2t(self),ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }

    return nil;
}

- squareSelf
{
    if (refCount>1) return idReplace(self,[self square]);

    if ([self inMonoid] && [self isOne]) return self;
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return idReplace(self,[self one]);
	if ([self isZero])     return self;
    }
    
    assert([self check]);
    ca_float_insqr(ca_float_id2c(self),ca_float_getargs(self));
    return [self invalidate];
}

- multiply:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self square];
    } else {
	if ([self inMonoid]) {
	    if ([self isOne]) return [b cheapCopy];
	    if ([b isOne]) return [self cheapCopy];
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]||[b isZero]) return [self zero];
	    if ([self isMinusOne]) return [b negate];
	    if ([b isMinusOne]) return [self negate];
	}
	if (1) {
	    ca_float_c c;
	    ca_float_oomul(&c,ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
	    return ca_float_c2id(self,&c);
	}
    }

    return nil;
}

- multiplySelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self multiply:b]);

    if (self == b) {
	return [self squareSelf];
    } else {
	if ([self inMonoid]) {
	    if ([b isOne]) return self;
	    if ([self isOne]) return idReplace(self,[b cheapCopy]);
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return self;
	    if ([b isZero]) return idReplace(self,[b cheapCopy]);
	    if ([b isMinusOne]) return [self negateSelf];
	    if ([self isMinusOne]) return idReplace(self,[b negate]);
	}
	ca_float_inmul(ca_float_id2c(self),ca_float_id2t(b),ca_float_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- inverse
{
    if ([self inMonoid] && [self isOne]) return [self cheapCopy];
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return [self cheapCopy];
	if ([self isZero])     return nil;
    }
    if (1) {
	ca_float_c c;assert([self check]);
	ca_float_ooinv(&c,ca_float_id2t(self),ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }

    return nil;
}

- inverseSelf
{
    if (refCount>1) return idReplace(self,[self inverse]);

    if ([self inMonoid] && [self isOne]) return self;
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return self;
	if ([self isZero])     return nil;
    }
    
    assert([self check]);
    ca_float_ininv(ca_float_id2c(self),ca_float_getargs(self));
    return [self invalidate];
}

- divide:b
{
    [self checkSameClass:b];

    if ([self inAdditiveMonoid]) {
	if ([b isZero]) return nil;
	if ([self isZero]) return [self cheapCopy];
    }

    if (self == b) {
	return [self one];
    } else {
	if ([self inMonoid]) {
	    if ([b isOne]) return [self cheapCopy];
	    if ([self inAdditiveMonoid]&&[b isMinusOne]) return [self negate];
	}
	if (1) {
	    ca_float_c c;
	    BOOL failed = ca_float_oodiv(&c,ca_float_id2t(self),ca_float_id2t(b),ca_float_getargs(self));
	    return (failed)?nil:ca_float_c2id(self,&c);
	}
    }

    return nil;
}

- divideSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self divide:b]);

    if ([self inAdditiveMonoid]) {
	if ([b isZero]) return nil;
	if ([self isZero]) return self;
    }

    if (self == b) {
	return idReplace(self,[self one]);
    } else {
	if ([self inMonoid]) {
	 if ([b isOne]) return self;
	 if ([self inAdditiveMonoid]&&[b isMinusOne]) return [self negateSelf];
	}
	ca_float_indiv(ca_float_id2c(self),ca_float_id2t(b),ca_float_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- _add:(int)v:B multiply:b:(int)w
{
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _negate:v];
    if ([b isZero]) return [self _negate:v];

    if ([b isOne]) return [self _add:v:B:+w];
    if ([B isOne]) return [self _add:v:b:+w];
    if ([b isMinusOne]) return [self _add:v:B:-w];
    if ([B isMinusOne]) return [self _add:v:b:-w];

    if (1) {
	ca_float_c c;
	ca_float_ooaddmul(&c,ca_float_id2t(self),v,ca_float_id2t(B),ca_float_id2t(b),w,ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }

    return nil;
}

- _addSelf:(int)v:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _add:v:B multiply:b:w]);
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _negateSelf:v];
    if ([b isZero]) return [self _negateSelf:v];

    if ([b isOne]) return [self _addSelf:v:B:+w];
    if ([B isOne]) return [self _addSelf:v:b:+w];
    if ([b isMinusOne]) return [self _addSelf:v:B:-w];
    if ([B isMinusOne]) return [self _addSelf:v:b:-w];

    ca_float_inaddmul(ca_float_id2c(self),v,ca_float_id2t(B),ca_float_id2t(b),w,ca_float_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _negate:+w];
    if ([a isZero]) return [B _negate:+w];
    if ([B isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B:+w];
    if ([a isOne]) return [self _add:+v:B:+w];
    if ([self isMinusOne]) return [a _add:-v:B:+w];
    if ([a isMinusOne]) return [self _add:-v:B:+w];

    if (1) {
	ca_float_c c;
	ca_float_oomuladd(&c,ca_float_id2t(self),ca_float_id2t(a),v,ca_float_id2t(B),w,ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B:w]);
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _negate:+w]);
    if ([a isZero]) return idReplace(self,[B _negate:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B:+w]);
    if ([a isOne]) return [self _addSelf:+v:B:+w];
    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B:+w];

    ca_float_inmuladd(ca_float_id2c(self),ca_float_id2t(a),v,ca_float_id2t(B),w,ca_float_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B multiply:b:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([a isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _multiply:a:+v];
    if ([b isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B multiply:b:+w];
    if ([a isOne]) return [self _add:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiply:a:+v add:B:+w];
    if ([B isOne]) return [self _multiply:a:+v add:b:+w];

    if ([self isMinusOne]) return [a _add:-v:B multiply:b:+w];
    if ([a isMinusOne]) return [self _add:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiply:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiply:a:+v add:b:-w];

    if (1) {
	ca_float_c c;
	ca_float_oomuladdmul(&c,ca_float_id2t(self),ca_float_id2t(a),v,ca_float_id2t(B),ca_float_id2t(b),w,ca_float_getargs(self));
	return ca_float_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B multiply:b:w]);
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([a isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];
    if ([b isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B multiply:b:+w]);
    if ([a isOne]) return [self _addSelf:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiplySelf:a:+v add:B:+w];
    if ([B isOne]) return [self _multiplySelf:a:+v add:b:+w];

    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B multiply:b:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiplySelf:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiplySelf:a:+v add:b:-w];

    ca_float_inmuladdmul(ca_float_id2c(self),ca_float_id2t(a),v,ca_float_id2t(B),ca_float_id2t(b),w,ca_float_getargs(self));
    return [self invalidate];
}


- strValue:(STR)aString
{
    return [self floatValue:atof(aString)];
}


- asExpression
{
    return [CAExpressionClass newHead:"Float" object:[self cheapCopy]];
}

- expression:anExpression
{
    if ([anExpression containsHead:"Symbol"]) return nil;
    
    if (strEq([anExpression head],"Integer")) {
	id anInteger = [anExpression object];
        return [CAFloatClass newFloatValue:[anInteger floatValue]];
    }
    
    if (strEq([anExpression head],"Float")) {
	id aFloat = [anExpression object];
	assert([self sameClass:aFloat]);return [aFloat cheapCopy];
    }
    
    return [self stdExpression:anExpression];
}


- (BOOL) printsLeadingSign
{
    return [self sign] < 0;
}

- printToFile:(FILE *)aFile
{
    fprintf(aFile,"%g",[self floatValue]);return self;
}

@end

