
/*
 * Computer Algebra Kit (c) 1993,96 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"

@implementation CASequence

- _setUpCarrier:aSequence
{
    carrier = aSequence;return self;
}

+ newCarrier:aSequence
{
    return [[[super new] _setUpCarrier:aSequence] toFirstMember];
}

- clear
{
    [carrier free];return [super clear];
}


- nextMember
{
    id m;
    if (m = [carrier nextMember]) {
	return m;
    } else {
	assert([carrier isEmpty] || [carrier previousMember] != nil);
	assert([carrier isEmpty] || m == [carrier nextMember]);
	return m;
    }
}

- previousMember
{
    id m;
    if (m = [carrier previousMember]) {
	return m;
    } else {
	assert([carrier isEmpty] || [carrier nextMember] != nil);
	assert([carrier isEmpty] || m == [carrier previousMember]);
	return m;
    }
}


- (int) numMembers
{
    return [carrier numMembers];
}

- (BOOL) isEmpty
{
    return [carrier isEmpty];
}

- (BOOL) isKindOfSequence
{
    return YES;
}


- toFirstMember
{
    [carrier toFirstMember];return self;
}

- toLastMember
{
    [carrier toLastMember];return self;
}

- toMemberAt:(int)i
{
    [carrier toMemberAt:i];return self;
}


- lastMember
{
    id res = [[self toLastMember] previousMember];
    assert([self isEmpty] || [res check] != nil);
    return res;
}

- firstMember
{
    id res = [[self toFirstMember] nextMember];
    assert([self isEmpty] || [res check] != nil);
    return res;
}

- memberAt:(int)i
{
    id res = [[self toMemberAt:i] nextMember];
    #ifndef NDEBUG
    if (0 <= i && i < [self numMembers]) {
	assert([res check]);
    } else {
	assert(res == nil);
    }
    #endif
    return res;
}

- commonDenominator
{
    id d,m = [self firstMember];
    if (m) {
	d = [[m denominator] cheapCopy];
	while (m = [self nextMember]) d = [d lcmSelf:[m denominator]];
	return d;
    } else {
	return nil;
    }
}

- content
{
    id d,m = [self firstMember];
    if (m) {
	d = [m cheapCopy];
	while ([d notZero] && [d notOne] && (m = [self nextMember])) {
	    d = [d gcdSelf:m];
	}
	return d;
    } else {
	return nil;
    }
}


- printToFile:(FILE *)aFile
{
    id aMember;
    
    fprintf(aFile,"{");
    if (aMember = [self nextMember]) [aMember printToFile:aFile];
    while (aMember = [self nextMember]) {
	fprintf(aFile,",");[aMember printToFile:aFile];
    }
    fprintf(aFile,"}");
    
    return self;
}

@end

