#include "cakit.h"

#define VAL_MIN -32500
#define VAL_MAX +32500

static int ca_intobj_check(ca_intobj_t self,ca_intobj_args args)
{
    assert(args.proto);[args.proto check];
    
    assert(VAL_MIN == - VAL_MAX); /* for opposite tests */
    assert(VAL_MAX * VAL_MAX + VAL_MAX * VAL_MAX < INT_MAX);
    assert(VAL_MIN * VAL_MAX - VAL_MIN * VAL_MAX > INT_MIN);

    if (self->hit) {
	assert(VAL_MIN <= self->val && self->val <= VAL_MAX);
	if (self->obj) {
	    id obj = [args.proto intValue:self->val];
	    assert([obj isEqual:self->obj]);
	    [obj free];
	}
    } else {
	int val;id obj = self->obj;
	assert([obj check]);val = [obj intValue];
	assert(val < VAL_MIN || VAL_MAX < val);
    }
    
    return 1;
}

/*
 * return new reference to object or create it if necessary
 */

static id ca_intobj_obj(ca_intobj_t self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return [self->obj cheapCopy];
    } else {
	id obj;assert(self->hit);
	obj = [args.proto intValue:self->val];
	assert([obj intValue] == self->val);
	return obj;
    }
}

/*
 * works inplace on "self"; creates object if it doesn't exists; returns it.
 */

static id ca_intobj_makeobj(ca_intobj_c *self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return self->obj;
    } else {
	assert(self->hit);return self->obj = [args.proto intValue:self->val];
    }
}

/*
 * initializing "c" by object
 */

static void ca_intobj_ooobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by object result of inplace operation
 * used like :
 *
 *		ca_intobj_makeobj(c,args);
 *		ca_intobj_inobj(c,[c->obj negateSelf],args)
 */

static void ca_intobj_inobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    /* c->obj is invalid reference now... replace it by result inplace op
     * we cannot check "c" here (as the object is invalid...)
     */
    assert(c->obj != nil);
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * initializing "c" by int value
 */

static void ca_intobj_ooval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = nil;
    } else {
	c->hit = 0;c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by int value
 */

static void ca_intobj_inval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;
	if (c->obj) c->obj = [c->obj free];
    } else {
	c->hit = 0;
	if (c->obj) c->obj = [c->obj free];
	c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_oozero(ca_intobj_c *c,ca_intobj_args args)
{
    c->hit = 1;c->val = 0;c->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_copy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj cheapCopy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_deepcopy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj copy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_move(ca_intobj_c *c,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    c->hit = a->hit;
    c->val = a->val;a->val = SHRT_MAX;
    c->obj = a->obj;a->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_clear(ca_intobj_c *c,ca_intobj_args args)
{
    c->val = SHRT_MAX;if (c->obj) c->obj = [c->obj free];
}

static ca_intobj_args ca_intobj_getargs(id self)
{
    ca_intobj_args args;args.proto = self;return args;
}

static int ca_intobj_size(ca_intobj_args args)
{
    return 1;				/* pointer increment */
}

static int ca_intobj_bsize(ca_intobj_args args)
{
    return sizeof(ca_intobj_c);		/* byte size carrier */
}

#if 0
static ca_intobj_t  ca_intobj_id2t(id object) { /* not needed */ }
static ca_intobj_c *ca_intobj_id2t(id object) { /* not needed */ }
#endif

/*
 * initialize scratch space "ref" with a read-only value 
 */

static ca_intobj_t ca_intobj_id2ref(id object,ca_intobj_c *ref)
{
    ca_intobj_args args = ca_intobj_getargs(object);	/* use object as proto */
    
    ca_intobj_ooobj(ref,object,args);return ref;
}

/*
 * return a new object, containing a copy of "t"
 */

static id ca_intobj_t2id(id proto,ca_intobj_t t)
{
    ca_intobj_args args = ca_intobj_getargs(proto);
    return ca_intobj_obj(t,args);
}

/*
 * return a new object, containing "t" itself (moved into object)
 */

static id ca_intobj_c2id(id proto,ca_intobj_c *c)
{
    id res;
    ca_intobj_args args = ca_intobj_getargs(proto);
    
    res = ca_intobj_obj(c,args);ca_intobj_clear(c,args);
     
    return res;
}

static unsigned ca_intobj_hash(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit)?a->val:[a->obj hash];
}

static int ca_intobj_iseq(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return YES;
    } else {
	if (a->hit) {
	    return (b->hit)?a->val==b->val:NO;
	} else {
	    return (b->hit)?NO:[a->obj isEqual:b->obj];
	}
    }
}

static int ca_intobj_sgn(ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	short int sgn = a->val;return (sgn==0)?0:((sgn<0)?-1:+1);
    } else {
	return [a->obj sign];
    }
}

static int ca_intobj_cmp(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    int cmp;
    
    if (a==b) {
	cmp = 0;
    } else {
	if (a->hit && b->hit) {
	    cmp = a->val - b->val;
	    cmp = (cmp==0)?0:((cmp<0)?-1:+1);
	} else {
	    id A = ca_intobj_obj(a,args);
	    id B = ca_intobj_obj(b,args);
	    cmp = [A compare:B];[A free];[B free];
	}
    }
    
    return cmp;
}

static int ca_intobj_iszero(ca_intobj_t a,ca_intobj_args args)
{
    int iszero = (a->hit && a->val == 0);
    assert(iszero || (a->hit&&a->val!=0) || (a->hit==0&&[a->obj notZero]));
    return iszero;
}

static int ca_intobj_isop(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return NO;
    } else {
	if (a->hit) {
	    assert(VAL_MAX + VAL_MIN == 0);
	    assert(VAL_MIN <= a->val && a->val <= VAL_MAX);
	    assert(VAL_MIN <= b->val && b->val <= VAL_MAX);
	    return (b->hit)?a->val+b->val==0:NO;
	} else {
	    return (b->hit)?NO:[a->obj isOpposite:b->obj];
	}
    }
}

static void ca_intobj_ooneg(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = - a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj negate],args);
    }
    
    assert(ca_intobj_isop(c,a,args));
}

static void ca_intobj_inneg(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = - c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj negateSelf],args);
    }
}

static void ca_intobj_oodbl(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_args args)
{
    if (a->hit) {
	int C = 2 * a->val;C = (v==+1)?+C:-C;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj _double:v],args);
    }
}

static void ca_intobj_indbl(ca_intobj_c *c,int v,ca_intobj_args args)
{
    if (c->hit) {
	int C = 2 * c->val;C = (v==+1)?+C:-C;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj _doubleSelf:v],args);
    }
}

static void ca_intobj_ooadd(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _add:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inadd(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B:w],args);
	[B free];
    }
}

static int ca_intobj_isone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == +1);
}

static int ca_intobj_isminusone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == -1);
}

static void ca_intobj_oosqr(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = a->val * a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj square],args);
    }
}

static void ca_intobj_insqr(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = c->val * c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj squareSelf],args);
    }
}

static void ca_intobj_oomul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A multiply:B];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmul(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj multiplySelf:B],args);
	[B free];
    }
}

static void ca_intobj_oomulv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val * v;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _multiply:B:v];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmulv(ca_intobj_c *c,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val * v;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:B:v],args);
	[B free];
    }
}

static void ca_intobj_ooaddmul(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && b->hit && t->hit) {
	int C = a->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	T = ca_intobj_obj(t,args);
	C = [A _add:v:B multiply:T:w];[A free];[B free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddmul(ca_intobj_c *c,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && b->hit && t->hit) {
	int C = c->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B multiply:T:w],args);
	[B free];[T free];
    }
}

static void ca_intobj_oomuladd(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit) {
	int C = a->val * s->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	C = [A _multiply:S:v add:B:w];
	[A free];[B free];[S free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladd(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit) {
	int C = c->val * s->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B:w],args);
	[B free];[S free];
    }
}

static void ca_intobj_oomuladdmul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit && t->hit) {
	int C = a->val * s->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	T = ca_intobj_obj(t,args);
	C = [A _multiply:S:v add:B multiply:T:w];
	[A free];[B free];[S free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladdmul(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit && t->hit) {
	int C = c->val * s->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B multiply:T:w],args);
	[B free];[S free];[T free];
    }
}

static void ca_intobj_ooaddsqr(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _addSquare:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddsqr(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelfSquare:v:B:w],args);
	[B free];
    }
}

static BOOL ca_intobj_oodiv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C;
	if (b->val == 0 || a->val % b->val != 0) {
	    return FAILURE;
	} else {
	    C = a->val / b->val;
	}
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A divide:B];[A free];[B free];
	if (C==nil) return FAILURE;
	ca_intobj_ooobj(c,C,args);
    }
    
    return SUCCESS;
}

static void ca_intobj_indiv(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C;
	if (b->val == 0 || c->val % b->val != 0) {
	    fprintf(stderr,"-exact division failed\n");abort();
	} else {
	    C = c->val / b->val;
	}
	ca_intobj_inval(c,C,args);
    } else {
	id C,B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	C = [c->obj divideSelf:B];
	if (C==nil) {fprintf(stderr,"-exact division failed\n");abort();}
	ca_intobj_inobj(c,C,args);
	[B free];
    }
}

static void ca_intobj_args_fileOut(id aFile,ca_intobj_args *args) {}
static void ca_intobj_args_fileIn(id aFile,ca_intobj_args *args) {}

static void ca_intobj_fileOut(id aFiler,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(a,args));
    [aFiler fileOut:&a->hit type:'S']; 
    if (a->hit) {
	[aFiler fileOut:&a->val type:'s']; /* signed ! */
    } else {
	[aFiler fileOut:&a->obj type:'@'];
    }
}

static void ca_intobj_fileIn(id aFiler,ca_intobj_c *c,ca_intobj_args args)
{
    [aFiler fileIn:&c->hit type:'S'];
    if (c->hit) {
	[aFiler fileIn:&c->val type:'s'];
        c->obj = nil;
    } else {
	[aFiler fileIn:&c->obj type:'@'];
    }
}


/*
 * Computer Algebra Kit (c) 1993,95 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_unimonom_size(ca_unimonom_args args)
{
    return 1;
}

static int ca_unimonom_bsize(ca_unimonom_args args)
{
    return sizeof(ca_unimonom_t);
}

static int ca_unimonom_check(ca_unimonom_t self,ca_unimonom_args args)
{
    assert(self >= 0);return 1;
}

static void ca_unimonom_clear(ca_unimonom_c *c,ca_unimonom_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_copy(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_args args)
{
    *c = a;
}

static void ca_unimonom_move(ca_unimonom_c *c,ca_unimonom_c *a,ca_unimonom_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_args_fileOut(id aFiler,ca_unimonom_args *args) {}
static void ca_unimonom_args_fileIn(id aFiler,ca_unimonom_args *args) {}

static void ca_unimonom_fileOut(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileOut:c type:'i'];
}

static void ca_unimonom_fileIn(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileIn:c type:'i'];
}

static unsigned ca_unimonom_hash(ca_unimonom_t self,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_deg(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_isone(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self == 0;
}

static int ca_unimonom_iseq(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == b;
}

static int ca_unimonom_iseqmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == x+b;
}

static int ca_unimonom_cmp(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-b;
}

static int ca_unimonom_cmpmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-x-b;
}

static void ca_unimonom_oomul(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    *c = a + b;
}

static void ca_unimonom_inmul(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    *c += b;
}

static BOOL ca_unimonom_oodiv(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    if (a < b) {
	return FAILURE;
    } else {
	*c = a - b;
	return SUCCESS;
    }
}

static void ca_unimonom_indiv(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    if (*c < b) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
    *c -= b;
}

#if ca_intobj_pervalue
static ca_intobj_t ca_intobjseq_tval(ca_intobjseq_t self)
{
    return *self;
}
static ca_intobj_c *ca_intobjseq_cval(ca_intobjseq_t self)
{
    return self;
}
#else
static ca_intobj_t ca_intobjseq_tval(ca_intobjseq_t self)
{
    return self;
}
static ca_intobj_c *ca_intobjseq_cval(ca_intobjseq_t self)
{
    return self;
}
#endif

static int ca_intobjseq_check(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_check(ca_intobjseq_tval(self),args);
	self += ca_intobj_size(args);
    }
    return 1;
}

static ca_intobjseq_t ca_intobjseq_alloc(int n,ca_intobj_args args)
{
    return (ca_intobjseq_t)OC_Malloc(n * ca_intobj_bsize(args));
}

static ca_intobjseq_t ca_intobjseq_realloc(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    return (ca_intobjseq_t)OC_Realloc(self,n * ca_intobj_bsize(args));
}

static ca_intobjseq_t ca_intobjseq_free(ca_intobjseq_t self,ca_intobj_args args)
{
    OC_Free(self);return NULL;
}

static void ca_intobjseq_oozero(ca_intobjseq_t c,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_oozero(ca_intobjseq_cval(c),args);
	c += ca_intobj_size(args);
    }
}

static void ca_intobjseq_copy(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_copy(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_deepcopy(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_deepcopy(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

#if ca_intobj_isvalue
static void ca_intobjseq_clear(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    /* nothing to clear */
}
#else
static void ca_intobjseq_clear(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_clear(self,args);
	self+=ca_intobj_size(args);
    }
}
#endif /* ca_intobj_isvalue */

static int ca_intobjseq_iseq(ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_intobj_iseq(ca_intobjseq_tval(a),ca_intobjseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_intobj_size(args);b+=ca_intobj_size(args);
	}
    }
    return 1;
}

static unsigned ca_intobjseq_hash(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    return (n)?ca_intobj_hash(ca_intobjseq_tval(a),args):0;
}

static void ca_intobjseq_place(ca_intobjseq_t self,ca_intobj_t a,int i,ca_intobj_args args)
{
    ca_intobj_c *c = ca_intobjseq_cval(self + i * ca_intobj_size(args));
    ca_intobj_clear(c,args);ca_intobj_copy(c,a,args);
}

static int ca_intobjseq_insertlast(ca_intobjseq_t self,ca_intobj_t a,int n,ca_intobj_args args)
{
    ca_intobj_c *c;
    assert(n >= 0);
    c = ca_intobjseq_cval(self + n * ca_intobj_size(args));
    ca_intobj_copy(c,a,args);return ++n;
}

static int ca_intobjseq_insertfirst(ca_intobjseq_t self,ca_intobj_t a,int n,ca_intobj_args args)
{
    int j;
    assert(n > 0);
    self += n * ca_intobj_size(args);
    j=n;while(j--) {
	ca_intobjseq_t t;
	t = self - ca_intobj_size(args);
	ca_intobj_move(ca_intobjseq_cval(self),ca_intobjseq_cval(t),args);
	self = t;
    }
    ca_intobj_copy(ca_intobjseq_cval(self),a,args);
    return ++n;
}

static int ca_intobjseq_insertat(ca_intobjseq_t self,ca_intobj_t a,int i,int n,ca_intobj_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_intobj_size(args);
    return i + ca_intobjseq_insertfirst(self,a,n-i,args);
}

static int ca_intobjseq_freelast(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    ca_intobj_c *c;assert(0 < n);
    --n;c = ca_intobjseq_cval(self + n * ca_intobj_size(args));
    ca_intobj_clear(c,args);return n;
}

static int ca_intobjseq_freefirst(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    int j;
    assert(0 < n);
    ca_intobj_clear(ca_intobjseq_cval(self),args);
    --n;j=n;while(j--) {
	ca_intobjseq_t t;
	t = self+ca_intobj_size(args);
	ca_intobj_move(ca_intobjseq_cval(self),ca_intobjseq_cval(t),args);
	self = t;
    }
    return n;
}

static int ca_intobjseq_freeat(ca_intobjseq_t self,int i,int n,ca_intobj_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_intobj_size(args);
    return i + ca_intobjseq_freefirst(self,n-i,args);
}

static ca_intobj_t ca_intobjseq_at(ca_intobjseq_t self,int i,ca_intobj_args args)
{
    ca_intobj_t res;assert(0 <= i);
    res = ca_intobjseq_tval(self + i * ca_intobj_size(args));
    assert(ca_intobj_check(res,args));
    return res;
}

static ca_intobj_t ca_intobjseq_last(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    return ca_intobjseq_at(self,n-1,args);
}

static ca_intobj_c *ca_intobjseq_cvalat(ca_intobjseq_t self,int i,ca_intobj_args args)
{
    assert(0 <= i);return ca_intobjseq_cval(self + i * ca_intobj_size(args));
}

static int ca_intobjseq_iszero(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	if (ca_intobj_iszero(ca_intobjseq_tval(a),args)) {
	    a+=ca_intobj_size(args);
	} else {
	    return 0;
	}
    }
    return 1;
}

static int ca_intobjseq_isop(ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_intobj_isop(ca_intobjseq_tval(a),ca_intobjseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_intobj_size(args);b+=ca_intobj_size(args);
	}
    }
    return 1;
}

static void ca_intobjseq_ooneg(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_ooneg(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inneg(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_inneg(ca_intobjseq_cval(a),args);
	a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_oonegv(ca_intobjseq_t c,ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(v*v == 1);if (v == -1) ca_intobjseq_ooneg(c,a,n,args);
    else ca_intobjseq_copy(c,a,n,args);
}

static void ca_intobjseq_innegv(ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(v*v == 1);if (v == -1) ca_intobjseq_inneg(a,n,args);
}

static void ca_intobjseq_oodbl(ca_intobjseq_t c,ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_oodbl(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_indbl(ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_indbl(ca_intobjseq_cval(a),v,args);
	a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_ooadd(ca_intobjseq_t c,ca_intobjseq_t a,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_ooadd(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,ca_intobjseq_tval(b),w,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inadd(ca_intobjseq_t a,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inadd(ca_intobjseq_cval(a),v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_oomul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulsc(ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_inmul(ca_intobjseq_cval(a),s,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscv(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_oomulv(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscv(ca_intobjseq_t a,ca_intobj_t s,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_inmulv(ca_intobjseq_cval(a),s,v,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_ooaddmulsc(ca_intobjseq_t c,ca_intobjseq_t a,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_ooaddmul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscadd(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_oomuladd(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscaddmulsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_oomuladdmul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inaddmulsc(ca_intobjseq_t a,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inaddmul(ca_intobjseq_cval(a),v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscadd(ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inmuladd(ca_intobjseq_cval(a),s,v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscaddmulsc(ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inmuladdmul(ca_intobjseq_cval(a),s,v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static BOOL ca_intobjseq_oodivsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    int i;
    BOOL failed;
    ca_intobjseq_t orgc = c;
    assert(n >= 0);
   
    for(i=0;i<n;i++) {
    
	failed = ca_intobj_oodiv(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,args);
	
	if (failed) {
	    ca_intobjseq_clear(orgc,i - 1,args);
	    return FAILURE;
	} else {
	    c+=ca_intobj_size(args);a+=ca_intobj_size(args);
	}
    }
    
    return SUCCESS;
}

static void ca_intobjseq_indivsc(ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_indiv(ca_intobjseq_cval(a),s,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oodotsqr(ca_intobj_c *c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    ca_intobj_oozero(c,args);
    while (n--) {
	ca_intobj_inaddsqr(c,+1,ca_intobjseq_tval(a),+1,args);a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_oodotmul(ca_intobj_c *c,ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0);
    ca_intobj_oozero(c,args);
    while (n--) {
	ca_intobj_inaddmul(c,+1,ca_intobjseq_tval(a),ca_intobjseq_tval(b),+1,args);
	a += ca_intobj_size(args);b += ca_intobj_size(args);
    }
}

static void ca_intobjseq_args_fileOut(id aFiler,ca_intobj_args *args)
{
    ca_intobj_args_fileOut(aFiler,args);
}
static void ca_intobjseq_args_fileIn(id aFiler,ca_intobj_args *args)
{
    ca_intobj_args_fileIn(aFiler,args);
}

static void ca_intobjseq_fileOut(id aFiler,ca_intobjseq_t self,int n,ca_intobj_args args)
{
    assert(n>=0);
    while (n--) {
	ca_intobj_fileOut(aFiler,ca_intobjseq_cval(self),args);
	self += ca_intobj_size(args);
    }
}

static void ca_intobjseq_fileIn(id aFiler,ca_intobjseq_t self,int n,ca_intobj_args args)
{
    assert(n>=0);
    while (n--) {
	ca_intobj_fileIn(aFiler,ca_intobjseq_cval(self),args);
	self += ca_intobj_size(args);
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_check(ca_intobj_vardnsrecdegdnspolseq_t self,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    if (n) {
	ca_intobj_t lcf = ca_intobjseq_last(self,n,args.sargs);
	assert(!ca_intobj_iszero(lcf,args.sargs));
    }
    
    return ca_intobjseq_check(self,n,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_copy(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_copy(c,a,n,args.sargs);return n;
}

static int ca_intobj_vardnsrecdegdnspolseq_deepcopy(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_deepcopy(c,a,n,args.sargs);return n;
}

static int ca_intobj_vardnsrecdegdnspolseq_count(ca_intobj_vardnsrecdegdnspolseq_t self,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    int count = 0;
    while (n--) {
	if (!ca_intobj_iszero(ca_intobjseq_tval(self),args.sargs)) count++;
	self += ca_intobj_size(args.sargs);
    }
    return count;
}

static int ca_intobj_vardnsrecdegdnspolseq_ord(ca_intobj_vardnsrecdegdnspolseq_t self,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    int ord = -1;
    while (n--) {
	if (ca_intobj_iszero(ca_intobjseq_tval(self),args.sargs)) {
	    ord++;self += ca_intobj_size(args.sargs);
	} else {
	    break;
	}
    }
    return ord;
}

static int ca_intobj_vardnsrecdegdnspolseq_iseq(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_vardnsrecdegdnspol_args args)
{
    return (an==bn)?ca_intobjseq_iseq(a,b,an,args.sargs):0;
}

static int ca_intobj_vardnsrecdegdnspolseq_isop(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_vardnsrecdegdnspol_args args)
{
    return (an==bn)?ca_intobjseq_isop(a,b,an,args.sargs):0;
}

static int ca_intobj_vardnsrecdegdnspolseq_hash(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobjseq_hash(a,an,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_isone(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_vardnsrecdegdnspol_args args)
{
    return an == 1 && ca_intobj_isone(ca_intobjseq_tval(a),args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_isminusone(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_vardnsrecdegdnspol_args args)
{
    return an == 1 && ca_intobj_isminusone(ca_intobjseq_tval(a),args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_removezeroes(ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    a += n * ca_intobj_size(args.sargs);
    while (n--) {
	a -= ca_intobj_size(args.sargs);
	if (ca_intobj_iszero(ca_intobjseq_tval(a),args.sargs)) {
	    ca_intobj_clear(ca_intobjseq_cval(a),args.sargs);
	} else {
	    break;
	}
    }
    return n+1;
}

static int ca_intobj_vardnsrecdegdnspolseq_delete(ca_intobj_vardnsrecdegdnspolseq_t self,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    n = n - 1;assert(n >= 0);
    ca_intobj_clear(ca_intobjseq_cvalat(self,n,args.sargs),args.sargs);
    return ca_intobj_vardnsrecdegdnspolseq_removezeroes(self,n,args);
}

static int ca_intobj_vardnsrecdegdnspolseq_insert(ca_intobj_vardnsrecdegdnspolseq_t self,int n,int e,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_iszero(s,args.sargs) == 0);
    if (e < n) {
	ca_intobj_inadd(ca_intobjseq_cvalat(self,e,args.sargs),+1,s,+1,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(self,n,args);
    } else {
	ca_intobjseq_oozero(self + n * ca_intobj_size(args.sargs),e + 1 - n,args.sargs);
	ca_intobj_inadd(ca_intobjseq_cvalat(self,e,args.sargs),+1,s,+1,args.sargs);
	return e + 1;
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_ooneg(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_ooneg(c,a,n,args.sargs);return n;
}

static int ca_intobj_vardnsrecdegdnspolseq_inneg(ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_inneg(a,n,args.sargs);return n;
}

static int ca_intobj_vardnsrecdegdnspolseq_oodbl(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,int v,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_oodbl(c,a,n,v,args.sargs);return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,n,args);
}

static int ca_intobj_vardnsrecdegdnspolseq_indbl(ca_intobj_vardnsrecdegdnspolseq_t a,int n,int v,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_indbl(a,n,v,args.sargs);return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,n,args);
}

static int ca_intobj_vardnsrecdegdnspolseq_ooadd(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_intobjseq_ooadd(c,a,v,b,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_ooadd(c,a,v,b,w,an,args.sargs);
	    c += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(c,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_ooadd(c,a,v,b,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(c,a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inadd(ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_intobjseq_inadd(a,v,b,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_inadd(a,v,b,w,an,args.sargs);
	    a += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(a,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_inadd(a,v,b,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_innegv(a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_oomulsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_oomulsc(c,a,s,n,args.sargs);return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,n,args);
}

static int ca_intobj_vardnsrecdegdnspolseq_inmulsc(ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_inmulsc(a,s,n,args.sargs);return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,n,args);
}

static int ca_intobj_vardnsrecdegdnspolseq_ooaddmulsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,an,args.sargs);
	    c += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(c,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(c,a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inaddmulsc(ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_intobjseq_inaddmulsc(a,v,b,t,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_inaddmulsc(a,v,b,t,w,an,args.sargs);
	    a += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(a,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_innegv(a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_oomulscadd(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_intobjseq_oomulscadd(c,a,s,v,b,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_oomulscadd(c,a,s,v,b,w,an,args.sargs);
	    c += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(c,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_oomulscadd(c,a,s,v,b,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(c,a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inmulscadd(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_intobjseq_inmulscadd(a,s,v,b,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_inmulscadd(a,s,v,b,w,an,args.sargs);
	    a += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oonegv(a,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_inmulscadd(a,s,v,b,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_inmulscv(a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_oomulscaddmulsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,an,args.sargs);
	    c += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(c,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(c,a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inmulscaddmulsc(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,an,args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,an,args.sargs);
	    a += an*ca_intobj_size(args.sargs);
	    b += an*ca_intobj_size(args.sargs);
	    ca_intobjseq_oomulscv(a,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    ca_intobjseq_inmulscv(a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static BOOL ca_intobj_vardnsrecdegdnspolseq_oodivsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobjseq_oodivsc(c,a,s,n,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_indivsc(ca_intobj_vardnsrecdegdnspolseq_t a,int n,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_indivsc(a,s,n,args.sargs);return n;
}

static int ca_intobj_vardnsrecdegdnspolseq_oomulmonomsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,int m,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_oozero(c,m,args.sargs);
    c += m * ca_intobj_size(args.sargs);
    ca_intobjseq_oomulscv(c,a,s,v,n,args.sargs);
    return n + m;
}

static BOOL ca_intobj_vardnsrecdegdnspolseq_oodivmonomsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int n,int m,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(n >= m);
    a += m * ca_intobj_size(args.sargs);
    return ca_intobjseq_oodivsc(c,a,s,n - m,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspolseq_ooaddmulmonomsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_intobjseq_oonegv(c,a,v,m,args.sargs);
	c += m*ca_intobj_size(args.sargs);
	a += m*ca_intobj_size(args.sargs);
	assert(an >= bn);ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	c -= m*ca_intobj_size(args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_intobjseq_oonegv(c,a,v,an,args.sargs);
		c += an*ca_intobj_size(args.sargs);
		a += an*ca_intobj_size(args.sargs);
		ca_intobjseq_oozero(c,m-an,args.sargs);
		c += m*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(c,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_intobjseq_oonegv(c,a,v,m,args.sargs);
		c += m*ca_intobj_size(args.sargs);
		a += m*ca_intobj_size(args.sargs);
		assert(bn > an-m);ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,an-m,args.sargs);
		c += (an-m)*ca_intobj_size(args.sargs);
		a += (an-m)*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(c,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_intobjseq_oonegv(c,a,v,m,args.sargs);
	    c += m*ca_intobj_size(args.sargs);
	    a += m*ca_intobj_size(args.sargs);
	    assert(an >= bn);ca_intobjseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    assert(an > bn+m);ca_intobjseq_oonegv(c,a,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inaddmulmonomsc(ca_intobj_vardnsrecdegdnspolseq_t a,int an,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_intobjseq_innegv(a,v,m,args.sargs);
	a += m*ca_intobj_size(args.sargs);
	assert(an >= bn);ca_intobjseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	a -= m*ca_intobj_size(args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_intobjseq_innegv(a,v,an,args.sargs);
		a += an*ca_intobj_size(args.sargs);
		ca_intobjseq_oozero(a,m-an,args.sargs);
		a += m*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(a,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_intobjseq_innegv(a,v,m,args.sargs);
		a += m*ca_intobj_size(args.sargs);
		assert(bn > an-m);ca_intobjseq_inaddmulsc(a,v,b,t,w,an-m,args.sargs);
		a += (an-m)*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(a,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_intobjseq_innegv(a,v,m,args.sargs);
	    a += m*ca_intobj_size(args.sargs);
	    assert(an >= bn);ca_intobjseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    assert(an > bn+m);ca_intobjseq_innegv(a,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_oomulscaddmulmonomsc(ca_intobj_vardnsrecdegdnspolseq_t c,ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_intobjseq_oomulscv(c,a,s,v,m,args.sargs);
	c += m*ca_intobj_size(args.sargs);
	a += m*ca_intobj_size(args.sargs);
	assert(an >= bn);ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	c -= m*ca_intobj_size(args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_intobjseq_oomulscv(c,a,s,v,an,args.sargs);
		c += an*ca_intobj_size(args.sargs);
		a += an*ca_intobj_size(args.sargs);
		ca_intobjseq_oozero(c,m-an,args.sargs);
		c += m*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(c,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_intobjseq_oomulscv(c,a,s,v,m,args.sargs);
		c += m*ca_intobj_size(args.sargs);
		a += m*ca_intobj_size(args.sargs);
		assert(bn > an-m);ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,an-m,args.sargs);
		c += (an-m)*ca_intobj_size(args.sargs);
		a += (an-m)*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(c,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_intobjseq_oomulscv(c,a,s,v,m,args.sargs);
	    c += m*ca_intobj_size(args.sargs);
	    a += m*ca_intobj_size(args.sargs);
	    assert(an >= bn);ca_intobjseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	    c += bn*ca_intobj_size(args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    assert(an > bn+m);ca_intobjseq_oomulscv(c,a,s,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_intobj_vardnsrecdegdnspolseq_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegdnspolseq_t a,int an,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspolseq_t b,int bn,int m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_intobjseq_inmulscv(a,s,v,m,args.sargs);
	a += m*ca_intobj_size(args.sargs);
	assert(an >= bn);ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	a -= m*ca_intobj_size(args.sargs);
	return ca_intobj_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_intobjseq_inmulscv(a,s,v,an,args.sargs);
		a += an*ca_intobj_size(args.sargs);
		ca_intobjseq_oozero(a,m-an,args.sargs);
		a += m*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(a,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_intobjseq_inmulscv(a,s,v,m,args.sargs);
		a += m*ca_intobj_size(args.sargs);
		assert(bn > an-m);ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,an-m,args.sargs);
		a += (an-m)*ca_intobj_size(args.sargs);
		ca_intobjseq_oomulscv(a,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_intobjseq_inmulscv(a,s,v,m,args.sargs);
	    a += m*ca_intobj_size(args.sargs);
	    assert(an >= bn);ca_intobjseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	    a += bn*ca_intobj_size(args.sargs);
	    assert(an > bn+m);ca_intobjseq_inmulscv(a,s,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}
static void ca_intobj_vardnsrecdegdnspol_ix_init(ca_intobj_vardnsrecdegdnspol_ixt ix,id self)
{
    ca_intobj_vardnsrecdegdnspol_t v = [self ca_intobj_vardnsrecdegdnspol_value];
    ix->i = -1;ix->n = v->n;
}

static void ca_intobj_vardnsrecdegdnspol_ix_clear(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_intobj_vardnsrecdegdnspol_ix_count(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    return ix->n; /* total number, zero or nonzero */
}

static int ca_intobj_vardnsrecdegdnspol_ix_isempty(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    return ix->n == 0; 
}

static int ca_intobj_vardnsrecdegdnspol_ix_ismatch(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    assert(-1 <= ix->i && ix->i <= ix->n);
    return (ix->i != -1 && ix->i != ix->n);
}

static void ca_intobj_vardnsrecdegdnspol_ix_tofirst(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    ix->i = -1;	/* next = first */
}

static void ca_intobj_vardnsrecdegdnspol_ix_tolast(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    ix->i = ix->n;	/* prev = last */
}

static void ca_intobj_vardnsrecdegdnspol_ix_to(ca_intobj_vardnsrecdegdnspol_ixt ix,int i)
{
    /* conversion to "n-i-1" done in elting routine... */
    if (i < 0) ix->i = -1;else if (i > ix->n) ix->i = ix->n;
    else ix->i = i-1; /* next -> at i */
}

static void ca_intobj_vardnsrecdegdnspol_ix_next(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    if (ix->i == ix->n) {
	/* remain on position just after last term */
    } else {
	++(ix->i);
    }
}

static void ca_intobj_vardnsrecdegdnspol_ix_prev(ca_intobj_vardnsrecdegdnspol_ixt ix)
{
    if (ix->i == -1) {
	/* remain on position just before first term */
    } else {
	--(ix->i);
    }
}
static int ca_intobj_vardnsrecdegdnspol_check(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(self->n <= self->c);
    return ca_intobj_vardnsrecdegdnspolseq_check(self->ptr,self->n,args);
}

static ca_intobj_vardnsrecdegdnspol_args ca_intobj_vardnsrecdegdnspol_getargs(id self)
{
    ca_intobj_vardnsrecdegdnspol_args res;
    res.sargs = ca_intobj_getargs([[self termZero] coefficient]);
    return res; 
}

static ca_intobj_vardnsrecdegdnspol_t ca_intobj_vardnsrecdegdnspol_id2t(id object)
{
    return [object ca_intobj_vardnsrecdegdnspol_value];
}

static ca_intobj_vardnsrecdegdnspol_c *ca_intobj_vardnsrecdegdnspol_id2c(id object)
{
    return [object ca_intobj_vardnsrecdegdnspol_reference];
}

static id ca_intobj_vardnsrecdegdnspol_t2id(id proto,ca_intobj_vardnsrecdegdnspol_t f)
{
    return [proto ca_intobj_vardnsrecdegdnspol_value:f];
}

static id ca_intobj_vardnsrecdegdnspol_c2id(id proto,ca_intobj_vardnsrecdegdnspol_c *r)
{
    return [proto ca_intobj_vardnsrecdegdnspol_reference:r];
}

static int ca_intobj_vardnsrecdegdnspol_isempty(ca_intobj_vardnsrecdegdnspol_t self)
{
    return self->n == 0;
}

static void ca_intobj_vardnsrecdegdnspol_init(ca_intobj_vardnsrecdegdnspol_t self,int c,ca_intobj_vardnsrecdegdnspol_args args)
{
    self->n   = 0;
    self->c   = c;
    self->ptr = ca_intobjseq_alloc(c,args.sargs);
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args));
}

static void ca_intobj_vardnsrecdegdnspol_copy(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_copy(c->ptr,a->ptr,a->n,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_deepcopy(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_deepcopy(c->ptr,a->ptr,a->n,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_move(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    c->ptr = a->ptr;c->n = a->n;c->c = a->c;
    a->ptr = NULL;assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_destroy(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    /* needed e.g. after a failed oodiv operation */
    self->ptr = ca_intobjseq_free(self->ptr,args.sargs);
}

static void ca_intobj_vardnsrecdegdnspol_clear(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobjseq_clear(self->ptr,self->n,args.sargs);
    ca_intobj_vardnsrecdegdnspol_destroy(self,args);
}

static int ca_intobj_vardnsrecdegdnspol_count(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_count(self->ptr,self->n,args); /* num non-zero coefs */
}

static unsigned ca_intobj_vardnsrecdegdnspol_hash(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_hash(a->ptr,a->n,args);
}

static int ca_intobj_vardnsrecdegdnspol_iseq(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_vardnsrecdegdnspol_args args)
{
    return (a==b)?YES:ca_intobj_vardnsrecdegdnspolseq_iseq(a->ptr,a->n,b->ptr,b->n,args);
}

static int ca_intobj_vardnsrecdegdnspol_isone(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_isone(self->ptr,self->n,args);
}

static int ca_intobj_vardnsrecdegdnspol_isminusone(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_isminusone(self->ptr,self->n,args);
}

static int ca_intobj_vardnsrecdegdnspol_deg(ca_intobj_vardnsrecdegdnspol_t self ,ca_intobj_vardnsrecdegdnspol_args args)
{
    return self->n - 1;
}

static int ca_intobj_vardnsrecdegdnspol_ord(ca_intobj_vardnsrecdegdnspol_t self ,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_ord(self->ptr,self->n,args);
}

static int ca_intobj_vardnsrecdegdnspol_maxdeg(ca_intobj_vardnsrecdegdnspol_t self ,ca_intobj_vardnsrecdegdnspol_args args)
{
    return self->n - 1;
}

static int ca_intobj_vardnsrecdegdnspol_mindeg(ca_intobj_vardnsrecdegdnspol_t self ,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspolseq_ord(self->ptr,self->n,args);
}

static ca_intobj_t ca_intobj_vardnsrecdegdnspol_scalat(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_ixt ix,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_ix_ismatch(ix) && ix->n == self->n);
    return ca_intobjseq_at(self->ptr,ix->n - ix->i - 1,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspol_expat(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_ixt ix,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_ix_ismatch(ix) && ix->n == self->n);
    return ix->n - ix->i - 1;
}

static ca_intobj_t ca_intobj_vardnsrecdegdnspol_scalfirst(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_isempty(self) == NO);
    return ca_intobjseq_last(self->ptr,self->n,args.sargs);
}

static int ca_intobj_vardnsrecdegdnspol_expfirst(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_isempty(self) == NO);
    return self->n - 1;
}

static void ca_intobj_vardnsrecdegdnspol_expand(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    self->c   = self->n + self->c + 1;
    self->ptr = ca_intobjseq_realloc(self->ptr,self->c,args.sargs);
    assert(self->n < self->c);
}

static void ca_intobj_vardnsrecdegdnspol_remove(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args) && ca_intobj_vardnsrecdegdnspol_isempty(self) == NO);
    self->n = ca_intobj_vardnsrecdegdnspolseq_delete(self->ptr,self->n,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args));
}

static void ca_intobj_vardnsrecdegdnspol_insert(ca_intobj_vardnsrecdegdnspol_t self,int e,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args));
    while (self->c < e + 1) ca_intobj_vardnsrecdegdnspol_expand(self,args);
    self->n = ca_intobj_vardnsrecdegdnspolseq_insert(self->ptr,self->n,e,s,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args));
}

static int ca_intobj_vardnsrecdegdnspol_iszero(ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    return ca_intobj_vardnsrecdegdnspol_isempty(self);
}

static int ca_intobj_vardnsrecdegdnspol_isop(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(a != b);return ca_intobj_vardnsrecdegdnspolseq_isop(a->ptr,a->n,b->ptr,b->n,args);
}

static void ca_intobj_vardnsrecdegdnspol_ooneg(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_ooneg(c->ptr,a->ptr,a->n,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inneg(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_vardnsrecdegdnspol_args args)
{
    a->n = ca_intobj_vardnsrecdegdnspolseq_inneg(a->ptr,a->n,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oodbl(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oodbl(c->ptr,a->ptr,a->n,v,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_indbl(ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspolseq_indbl(a->ptr,a->n,v,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_ooadd(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(a!=b);
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_ooadd(c->ptr,a->ptr,a->n,v,b->ptr,b->n,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inadd(ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    assert(a!=b);
    while (a->c < b->n) ca_intobj_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_intobj_vardnsrecdegdnspolseq_inadd(a->ptr,a->n,v,b->ptr,b->n,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oomulsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oomulsc(c->ptr,a->ptr,a->n,s,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inmulsc(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    a->n = ca_intobj_vardnsrecdegdnspolseq_inmulsc(a->ptr,a->n,s,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_ooaddmulsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n=ca_intobj_vardnsrecdegdnspolseq_ooaddmulsc(c->ptr,a->ptr,a->n,v,b->ptr,b->n,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inaddmulsc(ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_intobj_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_intobj_vardnsrecdegdnspolseq_inaddmulsc(a->ptr,a->n,v,b->ptr,b->n,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oomulscadd(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,int w, ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oomulscadd(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inmulscadd(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_intobj_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_intobj_vardnsrecdegdnspolseq_inmulscadd(a->ptr,a->n,s,v,b->ptr,b->n,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oomulscaddmulsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oomulscaddmulsc(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inmulscaddmulsc(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_intobj_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_intobj_vardnsrecdegdnspolseq_inmulscaddmulsc(a->ptr,a->n,s,v,b->ptr,b->n,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static BOOL ca_intobj_vardnsrecdegdnspol_oodivsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    BOOL failed;
    ca_intobj_vardnsrecdegdnspol_init(c,a->n,args);
    failed = ca_intobj_vardnsrecdegdnspolseq_oodivsc(c->ptr,a->ptr,a->n,s,args);
    if (failed) {
	/* the actual scalars are already freed */
	ca_intobj_vardnsrecdegdnspol_destroy(c,args);
    } else {
	c->n = a->n;
	assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
    }
    return failed;
}

static void ca_intobj_vardnsrecdegdnspol_indivsc(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegdnspol_args args)
{
    a->n = ca_intobj_vardnsrecdegdnspolseq_indivsc(a->ptr,a->n,s,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oomulmonomsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,b->n + m,args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oomulmonomsc(c->ptr,b->ptr,b->n,m,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inmulmonomsc(ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_c tmp;ca_intobj_vardnsrecdegdnspol_oomulmonomsc(&tmp,b,m,t,w,args);
    ca_intobj_vardnsrecdegdnspol_clear(b,args);ca_intobj_vardnsrecdegdnspol_move(b,&tmp,args);
}

static void ca_intobj_vardnsrecdegdnspol_ooaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n + m),args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_ooaddmulmonomsc(c->ptr,a->ptr,a->n,v,b->ptr,b->n,m,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_t a,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n + m) ca_intobj_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_intobj_vardnsrecdegdnspolseq_inaddmulmonomsc(a->ptr,a->n,v,b->ptr,b->n,m,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static void ca_intobj_vardnsrecdegdnspol_oomulscaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_init(c,MAX(a->n,b->n + m),args);
    c->n = ca_intobj_vardnsrecdegdnspolseq_oomulscaddmulmonomsc(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,m,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
}

static void ca_intobj_vardnsrecdegdnspol_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegdnspol_args args)
{
    a->n = ca_intobj_vardnsrecdegdnspolseq_inmulscaddmulmonomsc(a->ptr,a->n,s,v,b->ptr,b->n,m,t,w,args);
    assert(ca_intobj_vardnsrecdegdnspol_check(a,args));
}

static BOOL ca_intobj_vardnsrecdegdnspol_oodivmonomsc(ca_intobj_vardnsrecdegdnspol_t c,ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t, ca_intobj_vardnsrecdegdnspol_args args)
{
    if (ca_intobj_vardnsrecdegdnspol_ord(b,args) < m) {
	return FAILURE;
    } else {
	BOOL failed;
	ca_intobj_vardnsrecdegdnspol_init(c,b->n - m,args);
	failed = ca_intobj_vardnsrecdegdnspolseq_oodivmonomsc(c->ptr,b->ptr,b->n,m,t,args);
	if (failed) {
	    /* the actual scalars are already freed */
	    ca_intobj_vardnsrecdegdnspol_destroy(c,args);
	} else {
	    c->n = b->n - m;
	    assert(ca_intobj_vardnsrecdegdnspol_check(c,args));
	}
	return failed;
    }
}

static void ca_intobj_vardnsrecdegdnspol_indivmonomsc(ca_intobj_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_intobj_t t,ca_intobj_vardnsrecdegdnspol_args args)
{
    ca_intobj_vardnsrecdegdnspol_c tmp;
    BOOL failed = ca_intobj_vardnsrecdegdnspol_oodivmonomsc(&tmp,b,m,t,args);
    if (failed) { fprintf(stderr,"Exact Division Failed");abort(); }
    ca_intobj_vardnsrecdegdnspol_clear(b,args);ca_intobj_vardnsrecdegdnspol_move(b,&tmp,args);
}

static void ca_intobj_vardnsrecdegdnspol_args_fileOut(id aFiler,ca_intobj_vardnsrecdegdnspol_args *args)
{
    ca_intobjseq_args_fileOut(aFiler,&args->sargs);
}
static void ca_intobj_vardnsrecdegdnspol_args_fileIn(id aFiler,ca_intobj_vardnsrecdegdnspol_args *args)
{
    ca_intobjseq_args_fileIn(aFiler,&args->sargs);
}

static void ca_intobj_vardnsrecdegdnspol_fileOut(id aFiler,ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    int n = self->n;
    [aFiler fileOut:&n type:'i'];
    assert(ca_intobj_vardnsrecdegdnspol_check(self,args));
    ca_intobjseq_fileOut(aFiler,self->ptr,n,args.sargs);
}

static void ca_intobj_vardnsrecdegdnspol_fileIn(id aFiler,ca_intobj_vardnsrecdegdnspol_t self,ca_intobj_vardnsrecdegdnspol_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    ca_intobj_vardnsrecdegdnspol_init(self,n,args);
    ca_intobjseq_fileIn(aFiler,self->ptr,n,args.sargs);self->n = n;
    assert(self->n == n && ca_intobj_vardnsrecdegdnspol_check(self,args));
}

@implementation ca_intobj_vardnsrecdegdns_polynomial : ca_vardnsrecdegdns_polynomial

- _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    id aSymbol;
    
    assert([aCoef isZero] && [aScalar isZero]);
    assert([aCoef class_vardnsrecdegdns_polynomial] == self);
     
    aScalar  = [aScalar cheapCopy];
    aSymbol  = [[aCltn lastMember] cheapCopy];
    termZero = [CATermClass newScalarZero:aScalar coefficient:aCoef symbol:aSymbol exponent:1];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [CAMonomialClass newScalar:aScalar symbols:aCltn];
    scalarZero   = aScalar;
    symbols      = aCltn;
    
    ca_intobj_vardnsrecdegdnspol_init(&value,2 /* capacity */,ca_intobj_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

+ newScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn];
}

- empty
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_init(&c,2 /* capacity */ ,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- (int) numTerms
{
    /* doesn't take getargs in degsps case */
    assert([self check]);return ca_intobj_vardnsrecdegdnspol_count(&value ,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (ca_intobj_vardnsrecdegdnspol_t)ca_intobj_vardnsrecdegdnspol_value
{
    return &value;
}

- ca_intobj_vardnsrecdegdnspol_value:(ca_intobj_vardnsrecdegdnspol_t)aValue
{
    self = [self clone];
    ca_intobj_vardnsrecdegdnspol_copy(&value,aValue,ca_intobj_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

- (ca_intobj_vardnsrecdegdnspol_t)ca_intobj_vardnsrecdegdnspol_reference
{
    return &value;
}

- ca_intobj_vardnsrecdegdnspol_reference:(ca_intobj_vardnsrecdegdnspol_c *)aReference
{
    self = [self clone];
    ca_intobj_vardnsrecdegdnspol_move(&value,aReference,ca_intobj_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_copy(&c,&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- deepCopy
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_deepcopy(&c,&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_intobj_vardnsrecdegdnspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_intobj_vardnsrecdegdnspol_isone(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_intobj_vardnsrecdegdnspol_isminusone(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (unsigned) hash
{
    return ca_intobj_vardnsrecdegdnspol_hash(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_intobj_vardnsrecdegdnspol_iseq(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_id2t(b),ca_intobj_vardnsrecdegdnspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_intobj_vardnsrecdegdnspol_iseq(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_id2t(b),ca_intobj_vardnsrecdegdnspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_intobj_vardnsrecdegdnspol_iszero(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_intobj_vardnsrecdegdnspol_iszero(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_intobj_vardnsrecdegdnspol_isop(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_id2t(b),ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_intobj_vardnsrecdegdnspol_isop(ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_id2t(b),ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- negate
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_ooneg(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_intobj_vardnsrecdegdnspol_inneg(ca_intobj_vardnsrecdegdnspol_id2c(self),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_oodbl(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),v,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_intobj_vardnsrecdegdnspol_indbl(ca_intobj_vardnsrecdegdnspol_id2c(self),v,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_intobj_vardnsrecdegdnspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegdnspol_oodbl(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_intobj_vardnsrecdegdnspol_indbl(ca_intobj_vardnsrecdegdnspol_id2c(self),+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_intobj_vardnsrecdegdnspol_ooadd(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),v,ca_intobj_vardnsrecdegdnspol_id2t(b),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_intobj_vardnsrecdegdnspol_inadd(ca_intobj_vardnsrecdegdnspol_id2c(self),v,ca_intobj_vardnsrecdegdnspol_id2t(b),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_intobj_vardnsrecdegdnspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsrecdegdnspol_ooadd(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),+1,ca_intobj_vardnsrecdegdnspol_id2t(b),+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_intobj_vardnsrecdegdnspol_inadd(ca_intobj_vardnsrecdegdnspol_id2c(self),+1,ca_intobj_vardnsrecdegdnspol_id2t(b),+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_intobj_vardnsrecdegdnspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsrecdegdnspol_ooadd(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),+1,ca_intobj_vardnsrecdegdnspol_id2t(b),-1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_intobj_vardnsrecdegdnspol_inadd(ca_intobj_vardnsrecdegdnspol_id2c(self),+1,ca_intobj_vardnsrecdegdnspol_id2t(b),-1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_intobj_vardnsrecdegdnspol_deg(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (int) lastDegree
{
    return ca_intobj_vardnsrecdegdnspol_ord(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (int) maxDegree
{
    return ca_intobj_vardnsrecdegdnspol_maxdeg(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- (int) minDegree
{
    return ca_intobj_vardnsrecdegdnspol_mindeg(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
}

- asScalar
{
    if ([self numMembers] == 1) {
	id asScalar,lastTerm,aSequence = [self eachMember];
	asScalar = (lastTerm = [aSequence lastMember])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self numMembers] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachMember];
	asSymbol = (lastTerm = [aSequence lastMember])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols numMembers] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);
    assert([termZero isZero]&&[termZero exponent]==1&&[termZero symbol]!=nil);
    assert([[symbols lastMember] isEqual:[termZero symbol]]);

    ca_intobj_vardnsrecdegdnspol_check(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return self;
}

- clear
{
    ca_intobj_vardnsrecdegdnspol_clear(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- eachTerm
{
    id aSequence = [ca_intobj_vardnsrecdegdnspol_sequence newContent:[self cheapCopy]];
    return [CASequenceClass newCarrier:aSequence];
}

- elt_ca_intobj_vardnsrecdegdnspol_ix:(ca_intobj_vardnsrecdegdnspol_ixt)ix
{
    if (ca_intobj_vardnsrecdegdnspol_ix_ismatch(ix)) {
	id c,s;int e;ca_intobj_t t;
	t = ca_intobj_vardnsrecdegdnspol_scalat(&value,ix,ca_intobj_vardnsrecdegdnspol_getargs(self));
	e = ca_intobj_vardnsrecdegdnspol_expat(&value,ix,ca_intobj_vardnsrecdegdnspol_getargs(self));
	c = ca_intobj_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	return [termZero coefficient:c symbol:s exponent:e];
    } else {
	return nil;
    }
}
- removeTerm
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id c,s;int e;ca_intobj_t t;
	[self checkRefCountOne];assert([self check]);
	
	t = ca_intobj_vardnsrecdegdnspol_scalfirst(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
	e = ca_intobj_vardnsrecdegdnspol_expfirst(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
	c = ca_intobj_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	lt = [termZero coefficient:c symbol:s exponent:e];
	
	ca_intobj_vardnsrecdegdnspol_remove(&value,ca_intobj_vardnsrecdegdnspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertTerm:aTerm
{
    if ([aTerm isZero])  {
	[aTerm free];return self;
    } else {
	ca_intobj_c c;ca_intobj_t r;int e;
	assert([aTerm check]);[self checkRefCountOne];assert([self check]);
	r = ca_intobj_id2ref([aTerm coefficient],&c);e = [aTerm exponent];
	ca_intobj_vardnsrecdegdnspol_insert(&value,e,r,ca_intobj_vardnsrecdegdnspol_getargs(self));
	assert([aTerm check]);[aTerm free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyCoefficient:s
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c sc;
    ca_intobj_vardnsrecdegdnspol_oomulsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- multiplySelfCoefficient:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyCoefficient:s]);
    ca_intobj_vardnsrecdegdnspol_inmulsc(ca_intobj_vardnsrecdegdnspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- divideCoefficient:s
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c sc;
    BOOL failed = ca_intobj_vardnsrecdegdnspol_oodivsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- divideSelfCoefficient:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideCoefficient:s]);
    ca_intobj_vardnsrecdegdnspol_indivsc(ca_intobj_vardnsrecdegdnspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c bc;
    ca_intobj_vardnsrecdegdnspol_ooaddmulsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),v,ca_intobj_vardnsrecdegdnspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_intobj_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyCoefficient:b:w]);
    ca_intobj_vardnsrecdegdnspol_inaddmulsc(ca_intobj_vardnsrecdegdnspol_id2c(self),v,ca_intobj_vardnsrecdegdnspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c ac;
    ca_intobj_vardnsrecdegdnspol_oomulscadd(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegdnspol_id2t(B),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B:(int)w
{
    ca_intobj_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B:w]);
    ca_intobj_vardnsrecdegdnspol_inmulscadd(ca_intobj_vardnsrecdegdnspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegdnspol_id2t(B),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c ac,bc;
    ca_intobj_vardnsrecdegdnspol_oomulscaddmulsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegdnspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_intobj_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyCoefficient:b:w]);
    ca_intobj_vardnsrecdegdnspol_inmulscaddmulsc(ca_intobj_vardnsrecdegdnspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegdnspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}
- multiplyTerm:aTerm
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegdnspol_oomulmonomsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),[aTerm exponent],s,+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- multiplySelfTerm:aTerm
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyTerm:aTerm]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegdnspol_inmulmonomsc(ca_intobj_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,+1,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyTerm:aTerm:(int)v
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegdnspol_oomulmonomsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),[aTerm exponent],s,v,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfTerm:aTerm:(int)v
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self _multiplyTerm:aTerm:v]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegdnspol_inmulmonomsc(ca_intobj_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,v,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- divideTerm:aTerm
{
    int e;
    ca_intobj_vardnsrecdegdnspol_c c;
    BOOL failed;
    ca_intobj_t s;
    ca_intobj_c sc;
    
    e = [aTerm exponent];
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    failed=ca_intobj_vardnsrecdegdnspol_oodivmonomsc(&c,ca_intobj_vardnsrecdegdnspol_id2t(self),e,s,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- divideSelfTerm:aTerm
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideTerm:aTerm]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegdnspol_indivmonomsc(ca_intobj_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyTerm:b:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c bc;
    ca_intobj_t bs;
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegdnspol_ooaddmulmonomsc(&c,ca_intobj_vardnsrecdegdnspol_id2c(self),v,ca_intobj_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyTerm:b:(int)w
{
    ca_intobj_c bc;
    ca_intobj_t bs;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyTerm:b:w]);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegdnspol_inaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_id2c(self),v,ca_intobj_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_intobj_vardnsrecdegdnspol_c c;
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    as = ca_intobj_id2ref(a,&ac);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegdnspol_oomulscaddmulmonomsc(&c,ca_intobj_vardnsrecdegdnspol_id2c(self),as,v,ca_intobj_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return ca_intobj_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyTerm:b:w]);
    as = ca_intobj_id2ref(a,&ac);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegdnspol_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegdnspol_id2c(self),as,v,ca_intobj_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_intobj_vardnsrecdegdnspol_args args;
    [super fileOutOn:aFiler];
    args = ca_intobj_vardnsrecdegdnspol_getargs(self); /* can't do this in fileIn case */
    ca_intobj_vardnsrecdegdnspol_args_fileOut(aFiler,&args);
    ca_intobj_vardnsrecdegdnspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_intobj_vardnsrecdegdnspol_args args;
    [super fileInFrom:aFiler];
    ca_intobj_vardnsrecdegdnspol_args_fileIn(aFiler,&args);
    ca_intobj_vardnsrecdegdnspol_fileIn(aFiler,&value,args);
    return self;
}

@end

