#include "cakit.h"

#define VAL_MIN -32500
#define VAL_MAX +32500

static int ca_intobj_check(ca_intobj_t self,ca_intobj_args args)
{
    assert(args.proto);[args.proto check];
    
    assert(VAL_MIN == - VAL_MAX); /* for opposite tests */
    assert(VAL_MAX * VAL_MAX + VAL_MAX * VAL_MAX < INT_MAX);
    assert(VAL_MIN * VAL_MAX - VAL_MIN * VAL_MAX > INT_MIN);

    if (self->hit) {
	assert(VAL_MIN <= self->val && self->val <= VAL_MAX);
	if (self->obj) {
	    id obj = [args.proto intValue:self->val];
	    assert([obj isEqual:self->obj]);
	    [obj free];
	}
    } else {
	int val;id obj = self->obj;
	assert([obj check]);val = [obj intValue];
	assert(val < VAL_MIN || VAL_MAX < val);
    }
    
    return 1;
}

/*
 * return new reference to object or create it if necessary
 */

static id ca_intobj_obj(ca_intobj_t self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return [self->obj cheapCopy];
    } else {
	id obj;assert(self->hit);
	obj = [args.proto intValue:self->val];
	assert([obj intValue] == self->val);
	return obj;
    }
}

/*
 * works inplace on "self"; creates object if it doesn't exists; returns it.
 */

static id ca_intobj_makeobj(ca_intobj_c *self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return self->obj;
    } else {
	assert(self->hit);return self->obj = [args.proto intValue:self->val];
    }
}

/*
 * initializing "c" by object
 */

static void ca_intobj_ooobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by object result of inplace operation
 * used like :
 *
 *		ca_intobj_makeobj(c,args);
 *		ca_intobj_inobj(c,[c->obj negateSelf],args)
 */

static void ca_intobj_inobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    /* c->obj is invalid reference now... replace it by result inplace op
     * we cannot check "c" here (as the object is invalid...)
     */
    assert(c->obj != nil);
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * initializing "c" by int value
 */

static void ca_intobj_ooval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = nil;
    } else {
	c->hit = 0;c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by int value
 */

static void ca_intobj_inval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;
	if (c->obj) c->obj = [c->obj free];
    } else {
	c->hit = 0;
	if (c->obj) c->obj = [c->obj free];
	c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_oozero(ca_intobj_c *c,ca_intobj_args args)
{
    c->hit = 1;c->val = 0;c->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_copy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj cheapCopy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_deepcopy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj copy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_move(ca_intobj_c *c,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    c->hit = a->hit;
    c->val = a->val;a->val = SHRT_MAX;
    c->obj = a->obj;a->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_clear(ca_intobj_c *c,ca_intobj_args args)
{
    c->val = SHRT_MAX;if (c->obj) c->obj = [c->obj free];
}

static ca_intobj_args ca_intobj_getargs(id self)
{
    ca_intobj_args args;args.proto = self;return args;
}

static int ca_intobj_size(ca_intobj_args args)
{
    return 1;				/* pointer increment */
}

static int ca_intobj_bsize(ca_intobj_args args)
{
    return sizeof(ca_intobj_c);		/* byte size carrier */
}

#if 0
static ca_intobj_t  ca_intobj_id2t(id object) { /* not needed */ }
static ca_intobj_c *ca_intobj_id2t(id object) { /* not needed */ }
#endif

/*
 * initialize scratch space "ref" with a read-only value 
 */

static ca_intobj_t ca_intobj_id2ref(id object,ca_intobj_c *ref)
{
    ca_intobj_args args = ca_intobj_getargs(object);	/* use object as proto */
    
    ca_intobj_ooobj(ref,object,args);return ref;
}

/*
 * return a new object, containing a copy of "t"
 */

static id ca_intobj_t2id(id proto,ca_intobj_t t)
{
    ca_intobj_args args = ca_intobj_getargs(proto);
    return ca_intobj_obj(t,args);
}

/*
 * return a new object, containing "t" itself (moved into object)
 */

static id ca_intobj_c2id(id proto,ca_intobj_c *c)
{
    id res;
    ca_intobj_args args = ca_intobj_getargs(proto);
    
    res = ca_intobj_obj(c,args);ca_intobj_clear(c,args);
     
    return res;
}

static unsigned ca_intobj_hash(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit)?a->val:[a->obj hash];
}

static int ca_intobj_iseq(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return YES;
    } else {
	if (a->hit) {
	    return (b->hit)?a->val==b->val:NO;
	} else {
	    return (b->hit)?NO:[a->obj isEqual:b->obj];
	}
    }
}

static int ca_intobj_sgn(ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	short int sgn = a->val;return (sgn==0)?0:((sgn<0)?-1:+1);
    } else {
	return [a->obj sign];
    }
}

static int ca_intobj_cmp(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    int cmp;
    
    if (a==b) {
	cmp = 0;
    } else {
	if (a->hit && b->hit) {
	    cmp = a->val - b->val;
	    cmp = (cmp==0)?0:((cmp<0)?-1:+1);
	} else {
	    id A = ca_intobj_obj(a,args);
	    id B = ca_intobj_obj(b,args);
	    cmp = [A compare:B];[A free];[B free];
	}
    }
    
    return cmp;
}

static int ca_intobj_iszero(ca_intobj_t a,ca_intobj_args args)
{
    int iszero = (a->hit && a->val == 0);
    assert(iszero || (a->hit&&a->val!=0) || (a->hit==0&&[a->obj notZero]));
    return iszero;
}

static int ca_intobj_isop(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return NO;
    } else {
	if (a->hit) {
	    assert(VAL_MAX + VAL_MIN == 0);
	    assert(VAL_MIN <= a->val && a->val <= VAL_MAX);
	    assert(VAL_MIN <= b->val && b->val <= VAL_MAX);
	    return (b->hit)?a->val+b->val==0:NO;
	} else {
	    return (b->hit)?NO:[a->obj isOpposite:b->obj];
	}
    }
}

static void ca_intobj_ooneg(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = - a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj negate],args);
    }
    
    assert(ca_intobj_isop(c,a,args));
}

static void ca_intobj_inneg(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = - c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj negateSelf],args);
    }
}

static void ca_intobj_oodbl(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_args args)
{
    if (a->hit) {
	int C = 2 * a->val;C = (v==+1)?+C:-C;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj _double:v],args);
    }
}

static void ca_intobj_indbl(ca_intobj_c *c,int v,ca_intobj_args args)
{
    if (c->hit) {
	int C = 2 * c->val;C = (v==+1)?+C:-C;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj _doubleSelf:v],args);
    }
}

static void ca_intobj_ooadd(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _add:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inadd(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B:w],args);
	[B free];
    }
}

static int ca_intobj_isone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == +1);
}

static int ca_intobj_isminusone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == -1);
}

static void ca_intobj_oosqr(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = a->val * a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj square],args);
    }
}

static void ca_intobj_insqr(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = c->val * c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj squareSelf],args);
    }
}

static void ca_intobj_oomul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A multiply:B];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmul(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj multiplySelf:B],args);
	[B free];
    }
}

static void ca_intobj_oomulv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val * v;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _multiply:B:v];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmulv(ca_intobj_c *c,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val * v;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:B:v],args);
	[B free];
    }
}

static void ca_intobj_ooaddmul(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && b->hit && t->hit) {
	int C = a->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	T = ca_intobj_obj(t,args);
	C = [A _add:v:B multiply:T:w];[A free];[B free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddmul(ca_intobj_c *c,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && b->hit && t->hit) {
	int C = c->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B multiply:T:w],args);
	[B free];[T free];
    }
}

static void ca_intobj_oomuladd(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit) {
	int C = a->val * s->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	C = [A _multiply:S:v add:B:w];
	[A free];[B free];[S free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladd(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit) {
	int C = c->val * s->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B:w],args);
	[B free];[S free];
    }
}

static void ca_intobj_oomuladdmul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit && t->hit) {
	int C = a->val * s->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	T = ca_intobj_obj(t,args);
	C = [A _multiply:S:v add:B multiply:T:w];
	[A free];[B free];[S free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladdmul(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit && t->hit) {
	int C = c->val * s->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B multiply:T:w],args);
	[B free];[S free];[T free];
    }
}

static void ca_intobj_ooaddsqr(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _addSquare:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddsqr(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelfSquare:v:B:w],args);
	[B free];
    }
}

static BOOL ca_intobj_oodiv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C;
	if (b->val == 0 || a->val % b->val != 0) {
	    return FAILURE;
	} else {
	    C = a->val / b->val;
	}
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A divide:B];[A free];[B free];
	if (C==nil) return FAILURE;
	ca_intobj_ooobj(c,C,args);
    }
    
    return SUCCESS;
}

static void ca_intobj_indiv(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C;
	if (b->val == 0 || c->val % b->val != 0) {
	    fprintf(stderr,"-exact division failed\n");abort();
	} else {
	    C = c->val / b->val;
	}
	ca_intobj_inval(c,C,args);
    } else {
	id C,B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	C = [c->obj divideSelf:B];
	if (C==nil) {fprintf(stderr,"-exact division failed\n");abort();}
	ca_intobj_inobj(c,C,args);
	[B free];
    }
}

static void ca_intobj_args_fileOut(id aFile,ca_intobj_args *args) {}
static void ca_intobj_args_fileIn(id aFile,ca_intobj_args *args) {}

static void ca_intobj_fileOut(id aFiler,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(a,args));
    [aFiler fileOut:&a->hit type:'S']; 
    if (a->hit) {
	[aFiler fileOut:&a->val type:'s']; /* signed ! */
    } else {
	[aFiler fileOut:&a->obj type:'@'];
    }
}

static void ca_intobj_fileIn(id aFiler,ca_intobj_c *c,ca_intobj_args args)
{
    [aFiler fileIn:&c->hit type:'S'];
    if (c->hit) {
	[aFiler fileIn:&c->val type:'s'];
        c->obj = nil;
    } else {
	[aFiler fileIn:&c->obj type:'@'];
    }
}


/*
 * Computer Algebra Kit (c) 1993,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_unimonom_size(ca_unimonom_args args)
{
    return 1;
}

static int ca_unimonom_bsize(ca_unimonom_args args)
{
    return sizeof(ca_unimonom_t);
}

static int ca_unimonom_check(ca_unimonom_t self,ca_unimonom_args args)
{
    assert(self >= 0);return 1;
}

static void ca_unimonom_clear(ca_unimonom_c *c,ca_unimonom_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_copy(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_args args)
{
    *c = a;
}

static void ca_unimonom_move(ca_unimonom_c *c,ca_unimonom_c *a,ca_unimonom_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_args_fileOut(id aFiler,ca_unimonom_args *args) {}
static void ca_unimonom_args_fileIn(id aFiler,ca_unimonom_args *args) {}

static void ca_unimonom_fileOut(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileOut:c type:'i'];
}

static void ca_unimonom_fileIn(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileIn:c type:'i'];
}

static unsigned ca_unimonom_hash(ca_unimonom_t self,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_deg(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_isone(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self == 0;
}

static int ca_unimonom_iseq(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == b;
}

static int ca_unimonom_iseqmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == x+b;
}

static int ca_unimonom_cmp(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-b;
}

static int ca_unimonom_cmpmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-x-b;
}

static void ca_unimonom_oomul(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    *c = a + b;
}

static void ca_unimonom_inmul(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    *c += b;
}

static BOOL ca_unimonom_oodiv(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    if (a < b) {
	return FAILURE;
    } else {
	*c = a - b;
	return SUCCESS;
    }
}

static void ca_unimonom_indiv(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    if (*c < b) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
    *c -= b;
}

/*
 * scalar is accessed by the C "->" operator since having variable sized
 * scalars _and_ monomials gets difficult on machines like the DEC alpha
 * (where we'd have to manually pad to get pointer aligned structures).
 * 
 * the monomial (exponent) is however not accessed by the C "->" operator
 * but rather by functions that can take offsets as arguments
 * this is important the monomials in variable dense representation
 * (fixarray of exp)
 *
 * stes - mar31, 1997 - CHANGES: for DEC alpha coefficient needs to be aligned
 * properly, put it before the monomial and drop the idea of variable sized
 * coefficients.
 */

static ca_unimonom_c *ca_intobj_vardnsrecdegspsterm_monom_cval(ca_intobj_vardnsrecdegspsterm_t self)
{
    return &self->data;
}

static ca_unimonom_t ca_intobj_vardnsrecdegspsterm_monom(ca_intobj_vardnsrecdegspsterm_t self)
{
#if ca_unimonom_pervalue
    return self->data;
#else
    return  ca_intobj_vardnsrecdegspsterm_monom_cval(self);
#endif
}

/* stes - mar31,1997
 * change for DEC alpha to exclude varsized scalars
 */

static ca_intobj_c *ca_intobj_vardnsrecdegspsterm_scal_cval(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
	return &self->scal;
}

static ca_intobj_t ca_intobj_vardnsrecdegspsterm_scal(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
#if ca_intobj_pervalue
	return self->scal;
#else
    return  ca_intobj_vardnsrecdegspsterm_scal_cval(self,args);
#endif
}

static int ca_intobj_vardnsrecdegspsterm_check(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_check(ca_intobj_vardnsrecdegspsterm_scal(self,args),args.sargs));
    assert(!ca_intobj_iszero(ca_intobj_vardnsrecdegspsterm_scal(self,args),args.sargs));
    assert(ca_unimonom_check(ca_intobj_vardnsrecdegspsterm_monom(self),args.margs));
    return 1;
}

static int ca_intobj_vardnsrecdegspsterm_bsize(ca_intobj_vardnsrecdegspsterm_args args)
{
	/* this is sizeof(ca_unimonom_c) too much but why bother ? */
    return sizeof(struct ca_intobj_vardnsrecdegspsterm) + ca_unimonom_bsize(args.margs);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_alloc(ca_intobj_vardnsrecdegspsterm_args args)
{
    return (ca_intobj_vardnsrecdegspsterm_t)OC_Malloc(ca_intobj_vardnsrecdegspsterm_bsize(args));
}

static void ca_intobj_vardnsrecdegspsterm_clear(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_unimonom_clear(ca_intobj_vardnsrecdegspsterm_monom_cval(self),args.margs);
    ca_intobj_clear(ca_intobj_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
}

static void ca_intobj_vardnsrecdegspsterm_free(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    /* invalidating links not always helps : buggy code might work because
     * links were cleared...        self->nlink = NULL;self->plink = NULL;
     */
    OC_Free(self);
}

static int ca_intobj_vardnsrecdegspsterm_cmp(ca_intobj_vardnsrecdegspsterm_t a,ca_intobj_vardnsrecdegspsterm_t b,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(a,args) && ca_intobj_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_cmp(ca_intobj_vardnsrecdegspsterm_monom(a),ca_intobj_vardnsrecdegspsterm_monom(b),args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_cmpmonom(ca_intobj_vardnsrecdegspsterm_t a,ca_unimonom_t b,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(a,args) && ca_unimonom_check(b,args.margs));
    return ca_unimonom_cmp(ca_intobj_vardnsrecdegspsterm_monom(a),b,args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_cmpmul(ca_intobj_vardnsrecdegspsterm_t a,ca_intobj_vardnsrecdegspsterm_t b,ca_unimonom_t m ,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(a,args) && ca_intobj_vardnsrecdegspsterm_check(b,args));
    assert(ca_unimonom_check(m,args.margs));
    return ca_unimonom_cmpmul(ca_intobj_vardnsrecdegspsterm_monom(a),ca_intobj_vardnsrecdegspsterm_monom(b),m,args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_deg(ca_intobj_vardnsrecdegspsterm_t a,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_unimonom_deg(ca_intobj_vardnsrecdegspsterm_monom(a),args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_iszero(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_iszero(ca_intobj_vardnsrecdegspsterm_scal(self,args),args.sargs);
}

static int ca_intobj_vardnsrecdegspsterm_isone(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_isone(ca_intobj_vardnsrecdegspsterm_scal(self,args),args.sargs) && ca_unimonom_isone(ca_intobj_vardnsrecdegspsterm_monom(self),args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_isminusone(ca_intobj_vardnsrecdegspsterm_t self,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_isminusone(ca_intobj_vardnsrecdegspsterm_scal(self,args),args.sargs) && ca_unimonom_isone(ca_intobj_vardnsrecdegspsterm_monom(self),args.margs);
}

static int ca_intobj_vardnsrecdegspsterm_iseq(ca_intobj_vardnsrecdegspsterm_t a,ca_intobj_vardnsrecdegspsterm_t b,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(a,args) && ca_intobj_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_iseq(ca_intobj_vardnsrecdegspsterm_monom(a),ca_intobj_vardnsrecdegspsterm_monom(b),args.margs) && ca_intobj_iseq(ca_intobj_vardnsrecdegspsterm_scal(a,args),ca_intobj_vardnsrecdegspsterm_scal(b,args),args.sargs);
}

static int ca_intobj_vardnsrecdegspsterm_isop(ca_intobj_vardnsrecdegspsterm_t a,ca_intobj_vardnsrecdegspsterm_t b,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(a,args) && ca_intobj_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_iseq(ca_intobj_vardnsrecdegspsterm_monom(a),ca_intobj_vardnsrecdegspsterm_monom(b),args.margs) && ca_intobj_isop(ca_intobj_vardnsrecdegspsterm_scal(a,args),ca_intobj_vardnsrecdegspsterm_scal(b,args),args.sargs);
}

static void ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A)
{
    C->nlink = A;A->plink = C;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A)
{
    ca_intobj_vardnsrecdegspsterm_glue(C,A);return A;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_delete(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t C = A->nlink;
    ca_intobj_vardnsrecdegspsterm_glue(A->plink,C);
    ca_intobj_vardnsrecdegspsterm_clear(A,args);
    ca_intobj_vardnsrecdegspsterm_free(A,args);
    return C;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_condcns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t new,ca_intobj_vardnsrecdegspsterm_args args)
{
    if (ca_intobj_vardnsrecdegspsterm_iszero(new,args)) {
	ca_intobj_vardnsrecdegspsterm_clear(new,args);ca_intobj_vardnsrecdegspsterm_free(new,args);return C;
    } else {
	assert(ca_intobj_vardnsrecdegspsterm_check(new,args));return ca_intobj_vardnsrecdegspsterm_cns(C,new);
    }
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_conddelete(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    if (ca_intobj_vardnsrecdegspsterm_iszero(A,args)) {
	return ca_intobj_vardnsrecdegspsterm_delete(A,args);
    } else {
	return A->nlink;
    }
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_copymonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    
    ca_intobj_copy(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_intobj_vardnsrecdegspsterm_check(new,args));return ca_intobj_vardnsrecdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsrecdegspsterm_copymonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_copymonomsc_cns(C,m,s,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_deepcopymonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    
    ca_intobj_deepcopy(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_intobj_vardnsrecdegspsterm_check(new,args));return ca_intobj_vardnsrecdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsrecdegspsterm_deepcopymonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_deepcopymonomsc_cns(C,m,s,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_copy_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_vardnsrecdegspsterm_copymonomsc_cns(C,ca_intobj_vardnsrecdegspsterm_monom(A),ca_intobj_vardnsrecdegspsterm_scal(A,args),args);
}

static void ca_intobj_vardnsrecdegspsterm_copy(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_copy_cns(C,A,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_deepcopy_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_vardnsrecdegspsterm_deepcopymonomsc_cns(C,ca_intobj_vardnsrecdegspsterm_monom(A),ca_intobj_vardnsrecdegspsterm_scal(A,args),args);
}

static void ca_intobj_vardnsrecdegspsterm_deepcopy(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_deepcopy_cns(C,A,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooneg_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO);

    ca_intobj_ooneg(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsrecdegspsterm_ooneg(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_ooneg_cns(C,A,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inneg(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO);
    ca_intobj_inneg(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),args.sargs);
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO);
    return A->nlink;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oonegv_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    return (v == -1)?ca_intobj_vardnsrecdegspsterm_ooneg_cns(C,A,args):ca_intobj_vardnsrecdegspsterm_copy_cns(C,A,args);
}

static void ca_intobj_vardnsrecdegspsterm_oonegv(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,A,v,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_innegv(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    return (v == -1)?ca_intobj_vardnsrecdegspsterm_inneg(A,args):A->nlink;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oodbl_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO);

    ca_intobj_oodbl(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),v,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsrecdegspsterm_oodbl(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oodbl_cns(C,A,v,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_indbl(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO);
    ca_intobj_indbl(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),v,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooaddsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    ca_intobj_ooadd(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),v,B,w,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inaddsc(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inadd(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),v,B,w,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooadd_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == ca_intobj_vardnsrecdegspsterm_cmp(A,B,args));
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO && ca_intobj_vardnsrecdegspsterm_iszero(B,args) == NO);
    return ca_intobj_vardnsrecdegspsterm_ooaddsc_cns(C,A,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),w,args);
}

static void ca_intobj_vardnsrecdegspsterm_ooadd(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_ooadd_cns(C,A,v,B,w,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inadd(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == ca_intobj_vardnsrecdegspsterm_cmp(A,B,args));
    assert(ca_intobj_vardnsrecdegspsterm_iszero(A,args) == NO && ca_intobj_vardnsrecdegspsterm_iszero(B,args) == NO);
    return ca_intobj_vardnsrecdegspsterm_inaddsc(A,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    ca_intobj_oomulv(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,v,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsrecdegspsterm_oomulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A, ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulsc(ca_intobj_vardnsrecdegspsterm_t A, ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inmulv(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,v,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    ca_intobj_oomulv(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,v,args.sargs);
    ca_unimonom_oomul(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),m,args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsrecdegspsterm_oomulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inmulv(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,v,args.sargs);
    ca_unimonom_inmul(ca_intobj_vardnsrecdegspsterm_monom_cval(A),m,args.margs);

    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooaddscmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    ca_intobj_ooaddmul(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),v,B,t,w,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inaddscmulsc(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inaddmul(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),v,B,t,w,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooaddmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_ooaddscmulsc_cns(C,A,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_ooaddmulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
}

static void ca_intobj_vardnsrecdegspsterm_ooaddmulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspsterm_ooaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inaddmulsc(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_inaddscmulsc(A,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsrecdegspsterm_inaddscmulsc(A,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulscaddsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    ca_intobj_oomuladd(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,v,B,w,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulscaddsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inmuladd(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,v,B,w,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulscadd_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_oomulscaddsc_cns(C,A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),w,args);
}

static void ca_intobj_vardnsrecdegspsterm_oomulscadd(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulscadd(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_inmulscaddsc(A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulscaddscmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);
    ca_intobj_oomuladdmul(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,v,B,t,w,args.sargs);
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulscaddscmulsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_inmuladdmul(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,v,B,t,w,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulscaddmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsrecdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static void ca_intobj_vardnsrecdegspsterm_oomulscaddmulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspsterm_oomulscaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulscaddmulsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsrecdegspsterm_inmulscaddscmulsc(A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsrecdegspsterm_inmulscaddscmulsc(A,s,v,ca_intobj_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oodivsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    BOOL failed;
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    failed = ca_intobj_oodiv(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,args.sargs);
    if (failed) {
	ca_intobj_vardnsrecdegspsterm_free(new,args);return NULL;
    }
    
    ca_unimonom_copy(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_intobj_vardnsrecdegspsterm_oodivsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A, ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    if ((C = ca_intobj_vardnsrecdegspsterm_oodivsc_cns(C,A,s,args))) {
	ca_intobj_vardnsrecdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_indivsc(ca_intobj_vardnsrecdegspsterm_t A, ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_indiv(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,args.sargs);
    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_oodivmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    BOOL sfailed,mfailed;
    ca_intobj_vardnsrecdegspsterm_t new = ca_intobj_vardnsrecdegspsterm_alloc(args);

    sfailed = ca_intobj_oodiv(ca_intobj_vardnsrecdegspsterm_scal_cval(new,args),ca_intobj_vardnsrecdegspsterm_scal(A,args),s,args.sargs);
    mfailed = ca_unimonom_oodiv(ca_intobj_vardnsrecdegspsterm_monom_cval(new),ca_intobj_vardnsrecdegspsterm_monom(A),m,args.margs);
    
    if (sfailed || mfailed) {
	/* both failures are possible in the multivariate case */
	ca_intobj_vardnsrecdegspsterm_free(new,args);return NULL;
    }
    
    return ca_intobj_vardnsrecdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_intobj_vardnsrecdegspsterm_oodivmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    if ((C = ca_intobj_vardnsrecdegspsterm_oodivmonomsc_cns(C,A,m,s,args))) {
	ca_intobj_vardnsrecdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_indivmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_indiv(ca_intobj_vardnsrecdegspsterm_scal_cval(A,args),s,args.sargs);
    ca_unimonom_indiv(ca_intobj_vardnsrecdegspsterm_monom_cval(A),m,args.margs);

    return ca_intobj_vardnsrecdegspsterm_conddelete(A,args);
}

static void ca_intobj_vardnsrecdegspsterm_args_fileOut(id aFiler,ca_intobj_vardnsrecdegspsterm_args *args)
{
    ca_intobj_args_fileOut(aFiler,&args->sargs);
    ca_unimonom_args_fileOut(aFiler,&args->margs);
}
static void ca_intobj_vardnsrecdegspsterm_args_fileIn(id aFiler,ca_intobj_vardnsrecdegspsterm_args *args)
{
    ca_intobj_args_fileIn(aFiler,&args->sargs);
    ca_unimonom_args_fileIn(aFiler,&args->margs);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_fileOut(id aFiler,ca_intobj_vardnsrecdegspsterm_t self ,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_vardnsrecdegspsterm_check(self,args));
    
    ca_intobj_fileOut(aFiler,ca_intobj_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
    ca_unimonom_fileOut(aFiler,ca_intobj_vardnsrecdegspsterm_monom_cval(self),args.margs);
    
    return self->nlink;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspsterm_fileIn_cns(id aFiler,ca_intobj_vardnsrecdegspsterm_t C ,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t self = ca_intobj_vardnsrecdegspsterm_alloc(args);
    
    ca_intobj_fileIn(aFiler,ca_intobj_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
    ca_unimonom_fileIn(aFiler,ca_intobj_vardnsrecdegspsterm_monom_cval(self),args.margs);
    
    return ca_intobj_vardnsrecdegspsterm_cns(C,self);
}

static int ca_intobj_vardnsrecdegspslist_check(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t prev;
    
    for(prev = NULL;A!=AT;prev = A,A = A->nlink) {
	if (ca_intobj_vardnsrecdegspsterm_check(A,args) == 0) return 0;
	if (prev && ca_intobj_vardnsrecdegspsterm_cmp(prev,A,args) <= 0) return 0;
    }
    
    return 1;
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_copy_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) { C = ca_intobj_vardnsrecdegspsterm_copy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_copy(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_copy_cns(C,A,AT,args),CT);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_deepcopy_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) { C = ca_intobj_vardnsrecdegspsterm_deepcopy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_deepcopy(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_deepcopy_cns(C,A,AT,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_clear(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) { ca_intobj_vardnsrecdegspsterm_clear(A,args);A = A->nlink; }
}

static void ca_intobj_vardnsrecdegspslist_free(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_t t;for(;A!=AT;A=t) { t = A->nlink;ca_intobj_vardnsrecdegspsterm_free(A,args); }
}

static int ca_intobj_vardnsrecdegspslist_count(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static unsigned ca_intobj_vardnsrecdegspslist_hash(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    return ca_intobj_vardnsrecdegspslist_count(A,AT);
}

static int ca_intobj_vardnsrecdegspslist_iseq(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_t B, ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_vardnsrecdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_intobj_vardnsrecdegspsterm_iseq(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_intobj_vardnsrecdegspslist_isop(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_vardnsrecdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_intobj_vardnsrecdegspsterm_isop(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_intobj_vardnsrecdegspslist_deg(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    return (A==AT)?-1:ca_intobj_vardnsrecdegspsterm_deg(A,args);
}

static int ca_intobj_vardnsrecdegspslist_maxdeg(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    int max = (A==AT)?-1:ca_intobj_vardnsrecdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_intobj_vardnsrecdegspsterm_deg(A,args);if (max < d) max = d;
    }
    return max;
}

static int ca_intobj_vardnsrecdegspslist_ord(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    return (A==AT)?-1:ca_intobj_vardnsrecdegspsterm_deg(AT->plink,args);
}

static int ca_intobj_vardnsrecdegspslist_mindeg(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    int min = (A==AT)?-1:ca_intobj_vardnsrecdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_intobj_vardnsrecdegspsterm_deg(A,args);if (min > d) min = d;
    }
    return min;
}

static void ca_intobj_vardnsrecdegspslist_insert(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(ca_intobj_iszero(s,args.sargs) == NO);
    
    /* adding to zero */
    if (A == AT) {
	ca_intobj_vardnsrecdegspsterm_copymonomsc(A,m,s,A,args);return;
    }
    
    /* tail append */
    if (ca_intobj_vardnsrecdegspsterm_cmpmonom(AT->plink,m,args) > 0) {
	ca_intobj_vardnsrecdegspsterm_copymonomsc(AT->plink,m,s,AT,args);return;
    }
    
    /* insertion */
    while (A != AT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmpmonom(A,m,args);
	if (cmp == 0) {
	    ca_intobj_vardnsrecdegspsterm_inaddsc(A,+1,s,+1,args);return;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_copymonomsc(A->plink,m,s,A,args);return;
	}
	A = A->nlink;
    }
 
    fprintf(stderr,"Shouldn't happen\n");
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_ooneg_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsrecdegspsterm_ooneg_cns(C,A,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_ooneg(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_ooneg_cns(C,A,AT,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inneg(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_inneg(A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oonegv_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    if (v == -1) {
	return ca_intobj_vardnsrecdegspslist_ooneg_cns(C,A,AT,args);
    } else {
	return ca_intobj_vardnsrecdegspslist_copy_cns(C,A,AT,args);
    }
}

static void ca_intobj_vardnsrecdegspslist_oonegv(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_innegv(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    if (v == -1) ca_intobj_vardnsrecdegspslist_inneg(A,AT,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oodbl_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsrecdegspsterm_oodbl_cns(C,A,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oodbl(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oodbl_cns(C,A,AT,v,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_indbl(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_indbl(A,v,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_ooadd_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_ooadd_cns(C,A,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_ooadd(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_ooadd_cns(C,A,AT,v,B,BT,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inadd(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inadd(A,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oomulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oomulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inmulsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_inmulsc(A,s,v,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oomulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsrecdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oomulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oomulmonomsc_cns(C,A,AT,m,s,v,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_inmulmonomsc(A,m,s,v,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_ooaddmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_ooaddmulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_ooaddmulsc_cns(C,A,AT,v,B,BT,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inaddmulsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inaddmulsc(A,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_ooaddmulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_ooaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_ooaddmulmonomsc_cns(C,A,AT,v,B,BT,m,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inaddmulmonomsc(A,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oomulscadd_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oomulscadd(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT, int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oomulscadd_cns(C,A,AT,s,v,B,BT,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inmulscadd(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inmulscadd(A,s,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oomulscaddmulsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oomulscaddmulsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oomulscaddmulsc_cns(C,A,AT,s,v,B,BT,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inmulscaddmulsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inmulscaddmulsc(A,s,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_oomulscaddmulmonomsc_cns(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsrecdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsrecdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_oomulscaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_oomulscaddmulmonomsc_cns(C,A,AT,s,v,B,BT,m,t,w,args),CT);
}

static void ca_intobj_vardnsrecdegspslist_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspsterm_t B,ca_intobj_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsrecdegspsterm_inmulscaddmulmonomsc(A,s,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsrecdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsrecdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static BOOL ca_intobj_vardnsrecdegspslist_oodivsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_intobj_vardnsrecdegspsterm_t nextC = ca_intobj_vardnsrecdegspsterm_oodivsc_cns(C,A,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_intobj_vardnsrecdegspsterm_glue(C,CT);return failed;
}

static void ca_intobj_vardnsrecdegspslist_indivsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_indivsc(A,s,args);
}

static BOOL ca_intobj_vardnsrecdegspslist_oodivmonomsc(ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m, ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_intobj_vardnsrecdegspsterm_t nextC = ca_intobj_vardnsrecdegspsterm_oodivmonomsc_cns(C,A,m,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_intobj_vardnsrecdegspsterm_glue(C,CT);return failed;
}

static void ca_intobj_vardnsrecdegspslist_indivmonomsc(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_indivmonomsc(A,m,s,args);
}

static void ca_intobj_vardnsrecdegspslist_fileOut(id aFiler,ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT,ca_intobj_vardnsrecdegspsterm_args args)
{
    int n;
    n = ca_intobj_vardnsrecdegspslist_count(A,AT);
    [aFiler fileOut:&n type:'i'];
    while (A != AT) A = ca_intobj_vardnsrecdegspsterm_fileOut(aFiler,A,args);
}

static ca_intobj_vardnsrecdegspsterm_t ca_intobj_vardnsrecdegspslist_fileIn_cns(id aFiler,ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    while (n--) C = ca_intobj_vardnsrecdegspsterm_fileIn_cns(aFiler,C,args);
    return C;
}

static void ca_intobj_vardnsrecdegspslist_fileIn(id aFiler,ca_intobj_vardnsrecdegspsterm_t C,ca_intobj_vardnsrecdegspsterm_t CT,ca_intobj_vardnsrecdegspsterm_args args)
{
    ca_intobj_vardnsrecdegspsterm_glue(ca_intobj_vardnsrecdegspslist_fileIn_cns(aFiler,C,args),CT);
}

static void ca_intobj_vardnsrecdegspspol_ix_init(ca_intobj_vardnsrecdegspspol_ixt ix,id self)
{
    ca_intobj_vardnsrecdegspspol_t v = [self ca_intobj_vardnsrecdegspspol_value];
    ix->head = v;ix->ix = v;
}

static void ca_intobj_vardnsrecdegspspol_ix_clear(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_intobj_vardnsrecdegspspol_ix_count_aux_aux(ca_intobj_vardnsrecdegspsterm_t A,ca_intobj_vardnsrecdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static int ca_intobj_vardnsrecdegspspol_ix_count_aux(ca_intobj_vardnsrecdegspspol_t self)
{
    return ca_intobj_vardnsrecdegspspol_ix_count_aux_aux(self->nlink,self);
}

static int ca_intobj_vardnsrecdegspspol_ix_count(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    /* merde ! can't include entire genlist for ca_intobj_vardnsrecdegspslist_count */
    return ca_intobj_vardnsrecdegspspol_ix_count_aux(ix->head); 
}

static int ca_intobj_vardnsrecdegspspol_ix_isempty_aux(ca_intobj_vardnsrecdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static int ca_intobj_vardnsrecdegspspol_ix_isempty(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    /* can't include entire genpol for ca_intobj_vardnsrecdegspspol_isempty */
    return ca_intobj_vardnsrecdegspspol_ix_isempty_aux(ix->head); 
}

static int ca_intobj_vardnsrecdegspspol_ix_ismatch(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    return (ix->ix != NULL && ix->ix != ix->head);
}

static void ca_intobj_vardnsrecdegspspol_ix_tofirst(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    ix->ix = ix->head;
}

static void ca_intobj_vardnsrecdegspspol_ix_tolast(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    ix->ix = NULL;
}

static void ca_intobj_vardnsrecdegspspol_ix_to(ca_intobj_vardnsrecdegspspol_ixt ix,int i)
{
    fprintf(stderr,"memberAt: notImplemented\n");
}

static void ca_intobj_vardnsrecdegspspol_ix_next(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    if (ix->ix == NULL) {
	/* remain on position just after last term */
    } else {
	ix->ix = ix->ix->nlink; /* next term or head = position after last */
	if (ix->ix == ix->head) ix->ix = NULL;
    }
}

static void ca_intobj_vardnsrecdegspspol_ix_prev(ca_intobj_vardnsrecdegspspol_ixt ix)
{
    if (ix->ix == ix->head) {
	/* remain on position just before first term */
    } else {
	if (ix->ix == NULL) {
	    ix->ix = ix->head->plink; /* last term or head = position before first */
	} else {
	    ix->ix = ix->ix->plink; /* previous term or head = position before first */
	}
    }
}
static int ca_intobj_vardnsrecdegspspol_check(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    assert (self->data == 0xcafebabe);
    return ca_intobj_vardnsrecdegspslist_check(self->nlink,self,args);
}

static ca_intobj_vardnsrecdegspspol_t ca_intobj_vardnsrecdegspspol_id2t(id object)
{
    return [object ca_intobj_vardnsrecdegspspol_value];
}

static ca_intobj_vardnsrecdegspspol_c *ca_intobj_vardnsrecdegspspol_id2c(id object)
{
    return [object ca_intobj_vardnsrecdegspspol_reference];
}

static id ca_intobj_vardnsrecdegspspol_t2id(id proto,ca_intobj_vardnsrecdegspspol_t f)
{
    return [proto ca_intobj_vardnsrecdegspspol_value:f];
}

static id ca_intobj_vardnsrecdegspspol_c2id(id proto,ca_intobj_vardnsrecdegspspol_c *r)
{
    return [proto ca_intobj_vardnsrecdegspspol_reference:r];
}

static int ca_intobj_vardnsrecdegspspol_isempty(ca_intobj_vardnsrecdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static void ca_intobj_vardnsrecdegspspol_init(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    self->data = 0xcafebabe;self->nlink = self;self->plink = self;
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
}

static void ca_intobj_vardnsrecdegspspol_copy(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_copy(c,a->nlink,a,c,args);
}

static void ca_intobj_vardnsrecdegspspol_deepcopy(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_deepcopy(c,a->nlink,a,c,args);
}

static void ca_intobj_vardnsrecdegspspol_move(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    if (ca_intobj_vardnsrecdegspspol_isempty(a)) {
	ca_intobj_vardnsrecdegspspol_init(c,args);
    } else {
	ca_intobj_vardnsrecdegspspol_init(c,args);
	ca_intobj_vardnsrecdegspsterm_glue(c,a->nlink);ca_intobj_vardnsrecdegspsterm_glue(a->plink,c);
    }
    
    a->data = 0;assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_clear(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_clear(self->nlink,self,args);
    self->data = 0;ca_intobj_vardnsrecdegspslist_free(self->nlink,self,args);
}

static int ca_intobj_vardnsrecdegspspol_count(ca_intobj_vardnsrecdegspspol_t self)
{
    return ca_intobj_vardnsrecdegspslist_count(self->nlink,self);
}

static unsigned ca_intobj_vardnsrecdegspspol_hash(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspslist_hash(a->nlink,a,args);
}

static int ca_intobj_vardnsrecdegspspol_iseq(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_vardnsrecdegspspol_args args)
{
    return (a==b)?YES:ca_intobj_vardnsrecdegspslist_iseq(a->nlink,a,b->nlink,b,args);
}

static int ca_intobj_vardnsrecdegspspol_isone(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    /* if the leading term is one, then it is one */
    return (ca_intobj_vardnsrecdegspspol_isempty(a))?0:ca_intobj_vardnsrecdegspsterm_isone(a->nlink,args);
}

static int ca_intobj_vardnsrecdegspspol_isminusone(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    /* if the leading term is minus one, then it is minus one */
    return (ca_intobj_vardnsrecdegspspol_isempty(a))?0:ca_intobj_vardnsrecdegspsterm_isminusone(a->nlink,args);
}

static int ca_intobj_vardnsrecdegspspol_deg(ca_intobj_vardnsrecdegspspol_t self ,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspslist_deg(self->nlink,self,args);
}

static int ca_intobj_vardnsrecdegspspol_ord(ca_intobj_vardnsrecdegspspol_t self ,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspslist_ord(self->nlink,self,args);
}

static int ca_intobj_vardnsrecdegspspol_maxdeg(ca_intobj_vardnsrecdegspspol_t self ,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspslist_maxdeg(self->nlink,self,args);
}

static int ca_intobj_vardnsrecdegspspol_mindeg(ca_intobj_vardnsrecdegspspol_t self ,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspslist_mindeg(self->nlink,self,args);
}

static ca_intobj_t ca_intobj_vardnsrecdegspspol_scalat(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_ixt ix,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_intobj_vardnsrecdegspsterm_scal(ix->ix,args);
}

static ca_unimonom_t ca_intobj_vardnsrecdegspspol_monomat(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_ixt ix,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_intobj_vardnsrecdegspsterm_monom(ix->ix);
}

static ca_intobj_t ca_intobj_vardnsrecdegspspol_scalfirst(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_isempty(self) == NO);
    return ca_intobj_vardnsrecdegspsterm_scal(self->nlink,args);
}

static ca_unimonom_t ca_intobj_vardnsrecdegspspol_monomfirst(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_isempty(self) == NO);
    return ca_intobj_vardnsrecdegspsterm_monom(self->nlink);
}

static void ca_intobj_vardnsrecdegspspol_remove(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_check(self,args) && ca_intobj_vardnsrecdegspspol_isempty(self) == NO);
    ca_intobj_vardnsrecdegspsterm_delete(self->nlink,args);
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
}

static void ca_intobj_vardnsrecdegspspol_insert(ca_intobj_vardnsrecdegspspol_t self,ca_unimonom_t m,ca_intobj_t s,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
    ca_intobj_vardnsrecdegspslist_insert(self->nlink,self,m,s,args);
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
}

static int ca_intobj_vardnsrecdegspspol_iszero(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspspol_isempty(self);
}

static int ca_intobj_vardnsrecdegspspol_isop(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(a != b);return ca_intobj_vardnsrecdegspslist_isop(a->nlink,a,b->nlink,b,args);
}

static void ca_intobj_vardnsrecdegspspol_ooneg(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_ooneg(c,a->nlink,a,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inneg(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inneg(a->nlink,a,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oodbl(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oodbl(c,a->nlink,a,v,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_indbl(ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_indbl(a->nlink,a,v,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_ooadd(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(a!=b);
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_ooadd(c,a->nlink,a,v,b->nlink,b,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inadd(ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(a!=b);
    ca_intobj_vardnsrecdegspslist_inadd(a->nlink,a,v,b->nlink,b,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oomulsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oomulsc(c,a->nlink,a,s,+1,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inmulsc(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inmulsc(a->nlink,a,s,+1,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_ooaddmulsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_ooaddmulsc(c,a->nlink,a,v,b->nlink,b,t,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inaddmulsc(ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inaddmulsc(a->nlink,a,v,b->nlink,b,t,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oomulscadd(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,int w, ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oomulscadd(c,a->nlink,a,s,v,b->nlink,b,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inmulscadd(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inmulscadd(a->nlink,a,s,v,b->nlink,b,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oomulscaddmulsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oomulscaddmulsc(c,a->nlink,a,s,v,b->nlink,b,t,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inmulscaddmulsc(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inmulscaddmulsc(a->nlink,a,s,v,b->nlink,b,t,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static BOOL ca_intobj_vardnsrecdegspspol_oodivsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegspspol_args args)
{
    BOOL failed;
    ca_intobj_vardnsrecdegspspol_init(c,args);
    failed = ca_intobj_vardnsrecdegspslist_oodivsc(c,a->nlink,a,s,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
    if (failed) ca_intobj_vardnsrecdegspspol_clear(c,args);
    return failed;
}

static void ca_intobj_vardnsrecdegspspol_indivsc(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_indivsc(a->nlink,a,s,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oomulmonomsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oomulmonomsc(c,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inmulmonomsc(ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inmulmonomsc(b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(b,args));
}

static void ca_intobj_vardnsrecdegspspol_ooaddmulmonomsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_ooaddmulmonomsc(c,a->nlink,a,v,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inaddmulmonomsc(ca_intobj_vardnsrecdegspspol_t a,int v,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inaddmulmonomsc(a->nlink,a,v,b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static void ca_intobj_vardnsrecdegspspol_oomulscaddmulmonomsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(c,args);ca_intobj_vardnsrecdegspslist_oomulscaddmulmonomsc(c,a->nlink,a,s,v,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
}

static void ca_intobj_vardnsrecdegspspol_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_inmulscaddmulmonomsc(a->nlink,a,s,v,b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsrecdegspspol_check(a,args));
}

static BOOL ca_intobj_vardnsrecdegspspol_oodivmonomsc(ca_intobj_vardnsrecdegspspol_t c,ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t, ca_intobj_vardnsrecdegspspol_args args)
{
    BOOL failed;
    ca_intobj_vardnsrecdegspspol_init(c,args);
    failed = ca_intobj_vardnsrecdegspslist_oodivmonomsc(c,b->nlink,b,m,t,c,args);
    assert(ca_intobj_vardnsrecdegspspol_check(c,args));
    if (failed) ca_intobj_vardnsrecdegspspol_clear(c,args);
    return failed;
}

static void ca_intobj_vardnsrecdegspspol_indivmonomsc(ca_intobj_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_intobj_t t,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspslist_indivmonomsc(b->nlink,b,m,t,args);
    assert(ca_intobj_vardnsrecdegspspol_check(b,args));
}

static void ca_intobj_vardnsrecdegspspol_args_fileOut(id aFiler,ca_intobj_vardnsrecdegspspol_args *args)
{
    ca_intobj_vardnsrecdegspsterm_args_fileOut(aFiler,args);
}
static void ca_intobj_vardnsrecdegspspol_args_fileIn(id aFiler,ca_intobj_vardnsrecdegspspol_args *args)
{
    ca_intobj_vardnsrecdegspsterm_args_fileIn(aFiler,args);
}

static void ca_intobj_vardnsrecdegspspol_fileOut(id aFiler,ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
    ca_intobj_vardnsrecdegspslist_fileOut(aFiler,self->nlink,self,args);
}

static void ca_intobj_vardnsrecdegspspol_fileIn(id aFiler,ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    ca_intobj_vardnsrecdegspspol_init(self,args);ca_intobj_vardnsrecdegspslist_fileIn(aFiler,self,self,args);
    assert(ca_intobj_vardnsrecdegspspol_check(self,args));
}

static ca_intobj_vardnsrecdegspspol_args ca_intobj_vardnsrecdegspspol_getargs(id self)
{
    ca_intobj_vardnsrecdegspspol_args res;
    res.sargs = ca_intobj_getargs([[self termZero] coefficient]);
    res.margs = 0; /* unused */
    return res; 
}

static int ca_intobj_vardnsrecdegspspol_expat(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_ixt ix,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspspol_monomat(self,ix,args);
}

static int ca_intobj_vardnsrecdegspspol_expfirst(ca_intobj_vardnsrecdegspspol_t self,ca_intobj_vardnsrecdegspspol_args args)
{
    return ca_intobj_vardnsrecdegspspol_monomfirst(self,args);
}
@implementation ca_intobj_vardnsrecdegsps_polynomial : ca_vardnsrecdegsps_polynomial

- _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    id aSymbol;
    assert([aCoef isZero] && [aScalar isZero]);
    assert([aCoef class_vardnsrecdegsps_polynomial] == [self class]);
    aScalar  = [aScalar cheapCopy];
    aSymbol  = [[aCltn lastMember] cheapCopy];
    termZero = [CATermClass newScalarZero:aScalar coefficient:aCoef symbol:aSymbol exponent:1];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [CAMonomialClass newScalar:aScalar symbols:aCltn];
    scalarZero   = aScalar;
    symbols      = aCltn;
    
    /* in vardns case, init takes a capacity */
    ca_intobj_vardnsrecdegspspol_init(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

+ newScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn];
}

- empty
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    /* in vardns case, init takes a capacity */
    ca_intobj_vardnsrecdegspspol_init(&c,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- (int) numTerms
{
    /* seems to take getargs in degdns case */
    assert([self check]);return ca_intobj_vardnsrecdegspspol_count(&value);
}

- (ca_intobj_vardnsrecdegspspol_t)ca_intobj_vardnsrecdegspspol_value
{
    return &value;
}

- ca_intobj_vardnsrecdegspspol_value:(ca_intobj_vardnsrecdegspspol_t)aValue
{
    self = [self clone];
    ca_intobj_vardnsrecdegspspol_copy(&value,aValue,ca_intobj_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

- (ca_intobj_vardnsrecdegspspol_t)ca_intobj_vardnsrecdegspspol_reference
{
    return &value;
}

- ca_intobj_vardnsrecdegspspol_reference:(ca_intobj_vardnsrecdegspspol_c *)aReference
{
    self = [self clone];
    ca_intobj_vardnsrecdegspspol_move(&value,aReference,ca_intobj_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegspspol_copy(&c,&value,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- deepCopy
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegspspol_deepcopy(&c,&value,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_intobj_vardnsrecdegspspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_intobj_vardnsrecdegspspol_isone(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_intobj_vardnsrecdegspspol_isminusone(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (unsigned) hash
{
    return ca_intobj_vardnsrecdegspspol_hash(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_intobj_vardnsrecdegspspol_iseq(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_id2t(b),ca_intobj_vardnsrecdegspspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_intobj_vardnsrecdegspspol_iseq(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_id2t(b),ca_intobj_vardnsrecdegspspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_intobj_vardnsrecdegspspol_iszero(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_intobj_vardnsrecdegspspol_iszero(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_intobj_vardnsrecdegspspol_isop(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_id2t(b),ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_intobj_vardnsrecdegspspol_isop(ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_id2t(b),ca_intobj_vardnsrecdegspspol_getargs(self));
}

- negate
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegspspol_ooneg(&c,ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_intobj_vardnsrecdegspspol_inneg(ca_intobj_vardnsrecdegspspol_id2c(self),ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegspspol_oodbl(&c,ca_intobj_vardnsrecdegspspol_id2t(self),v,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_intobj_vardnsrecdegspspol_indbl(ca_intobj_vardnsrecdegspspol_id2c(self),v,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_intobj_vardnsrecdegspspol_c c;assert([self check]);
    ca_intobj_vardnsrecdegspspol_oodbl(&c,ca_intobj_vardnsrecdegspspol_id2t(self),+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_intobj_vardnsrecdegspspol_indbl(ca_intobj_vardnsrecdegspspol_id2c(self),+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_intobj_vardnsrecdegspspol_ooadd(&c,ca_intobj_vardnsrecdegspspol_id2t(self),v,ca_intobj_vardnsrecdegspspol_id2t(b),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_intobj_vardnsrecdegspspol_inadd(ca_intobj_vardnsrecdegspspol_id2c(self),v,ca_intobj_vardnsrecdegspspol_id2t(b),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_intobj_vardnsrecdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsrecdegspspol_ooadd(&c,ca_intobj_vardnsrecdegspspol_id2t(self),+1,ca_intobj_vardnsrecdegspspol_id2t(b),+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_intobj_vardnsrecdegspspol_inadd(ca_intobj_vardnsrecdegspspol_id2c(self),+1,ca_intobj_vardnsrecdegspspol_id2t(b),+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_intobj_vardnsrecdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsrecdegspspol_ooadd(&c,ca_intobj_vardnsrecdegspspol_id2t(self),+1,ca_intobj_vardnsrecdegspspol_id2t(b),-1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_intobj_vardnsrecdegspspol_inadd(ca_intobj_vardnsrecdegspspol_id2c(self),+1,ca_intobj_vardnsrecdegspspol_id2t(b),-1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_intobj_vardnsrecdegspspol_deg(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (int) lastDegree
{
    return ca_intobj_vardnsrecdegspspol_ord(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (int) maxDegree
{
    return ca_intobj_vardnsrecdegspspol_maxdeg(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- (int) minDegree
{
    return ca_intobj_vardnsrecdegspspol_mindeg(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
}

- asScalar
{
    if ([self numMembers] == 1) {
	id asScalar,lastTerm,aSequence = [self eachMember];
	asScalar = (lastTerm = [aSequence lastMember])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self numMembers] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachMember];
	asSymbol = (lastTerm = [aSequence lastMember])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols numMembers] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);
    assert([termZero isZero]&&[termZero exponent]==1&&[termZero symbol]!=nil);
    assert([[symbols lastMember] isEqual:[termZero symbol]]);

    ca_intobj_vardnsrecdegspspol_check(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
    return self;
}

- clear
{
    ca_intobj_vardnsrecdegspspol_clear(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- eachTerm
{
    id aSequence = [ca_intobj_vardnsrecdegspspol_sequence newContent:[self cheapCopy]];
    return [CASequenceClass newCarrier:aSequence];
}

- elt_ca_intobj_vardnsrecdegspspol_ix:(ca_intobj_vardnsrecdegspspol_ixt)ix
{
    if (ca_intobj_vardnsrecdegspspol_ix_ismatch(ix)) {
	id c,s;int e;ca_intobj_t t;
	t = ca_intobj_vardnsrecdegspspol_scalat(&value,ix,ca_intobj_vardnsrecdegspspol_getargs(self));
	e = ca_intobj_vardnsrecdegspspol_expat(&value,ix,ca_intobj_vardnsrecdegspspol_getargs(self));
	c = ca_intobj_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	return [termZero coefficient:c symbol:s exponent:e];
    } else {
	return nil;
    }
}
- removeTerm
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id c,s;int e;ca_intobj_t t;
	[self checkRefCountOne];assert([self check]);
	
	t = ca_intobj_vardnsrecdegspspol_scalfirst(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
	e = ca_intobj_vardnsrecdegspspol_expfirst(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
	c = ca_intobj_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	lt = [termZero coefficient:c symbol:s exponent:e];
	
	ca_intobj_vardnsrecdegspspol_remove(&value,ca_intobj_vardnsrecdegspspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertTerm:aTerm
{
    if ([aTerm isZero])  {
	[aTerm free];return self;
    } else {
	ca_intobj_c c;ca_intobj_t r;int e;
	assert([aTerm check]);[self checkRefCountOne];assert([self check]);
	r = ca_intobj_id2ref([aTerm coefficient],&c);e = [aTerm exponent];
	ca_intobj_vardnsrecdegspspol_insert(&value,e,r,ca_intobj_vardnsrecdegspspol_getargs(self));
	assert([aTerm check]);[aTerm free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyCoefficient:s
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c sc;
    ca_intobj_vardnsrecdegspspol_oomulsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- multiplySelfCoefficient:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyCoefficient:s]);
    ca_intobj_vardnsrecdegspspol_inmulsc(ca_intobj_vardnsrecdegspspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- divideCoefficient:s
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c sc;
    BOOL failed = ca_intobj_vardnsrecdegspspol_oodivsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegspspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- divideSelfCoefficient:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideCoefficient:s]);
    ca_intobj_vardnsrecdegspspol_indivsc(ca_intobj_vardnsrecdegspspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c bc;
    ca_intobj_vardnsrecdegspspol_ooaddmulsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),v,ca_intobj_vardnsrecdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_intobj_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyCoefficient:b:w]);
    ca_intobj_vardnsrecdegspspol_inaddmulsc(ca_intobj_vardnsrecdegspspol_id2c(self),v,ca_intobj_vardnsrecdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c ac;
    ca_intobj_vardnsrecdegspspol_oomulscadd(&c,ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegspspol_id2t(B),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B:(int)w
{
    ca_intobj_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B:w]);
    ca_intobj_vardnsrecdegspspol_inmulscadd(ca_intobj_vardnsrecdegspspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegspspol_id2t(B),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c ac,bc;
    ca_intobj_vardnsrecdegspspol_oomulscaddmulsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_intobj_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyCoefficient:b:w]);
    ca_intobj_vardnsrecdegspspol_inmulscaddmulsc(ca_intobj_vardnsrecdegspspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsrecdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}
- multiplyTerm:aTerm
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegspspol_oomulmonomsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),[aTerm exponent],s,+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- multiplySelfTerm:aTerm
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyTerm:aTerm]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegspspol_inmulmonomsc(ca_intobj_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,+1,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyTerm:aTerm:(int)v
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegspspol_oomulmonomsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),[aTerm exponent],s,v,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfTerm:aTerm:(int)v
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self _multiplyTerm:aTerm:v]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegspspol_inmulmonomsc(ca_intobj_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,v,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- divideTerm:aTerm
{
    int e;
    ca_intobj_vardnsrecdegspspol_c c;
    BOOL failed;
    ca_intobj_t s;
    ca_intobj_c sc;
    
    e = [aTerm exponent];
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    failed=ca_intobj_vardnsrecdegspspol_oodivmonomsc(&c,ca_intobj_vardnsrecdegspspol_id2t(self),e,s,ca_intobj_vardnsrecdegspspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- divideSelfTerm:aTerm
{
    ca_intobj_t s;
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideTerm:aTerm]);
    s = ca_intobj_id2ref([aTerm coefficient],&sc);
    ca_intobj_vardnsrecdegspspol_indivmonomsc(ca_intobj_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyTerm:b:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c bc;
    ca_intobj_t bs;
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegspspol_ooaddmulmonomsc(&c,ca_intobj_vardnsrecdegspspol_id2c(self),v,ca_intobj_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyTerm:b:(int)w
{
    ca_intobj_c bc;
    ca_intobj_t bs;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyTerm:b:w]);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegspspol_inaddmulmonomsc(ca_intobj_vardnsrecdegspspol_id2c(self),v,ca_intobj_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_intobj_vardnsrecdegspspol_c c;
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    as = ca_intobj_id2ref(a,&ac);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegspspol_oomulscaddmulmonomsc(&c,ca_intobj_vardnsrecdegspspol_id2c(self),as,v,ca_intobj_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return ca_intobj_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyTerm:b:w]);
    as = ca_intobj_id2ref(a,&ac);
    bs = ca_intobj_id2ref([b coefficient],&bc);
    ca_intobj_vardnsrecdegspspol_inmulscaddmulmonomsc(ca_intobj_vardnsrecdegspspol_id2c(self),as,v,ca_intobj_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_intobj_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_intobj_vardnsrecdegspspol_args args;
    [super fileOutOn:aFiler];
    args = ca_intobj_vardnsrecdegspspol_getargs(self); /* can't do this in fileIn case */
    ca_intobj_vardnsrecdegspspol_args_fileOut(aFiler,&args);
    ca_intobj_vardnsrecdegspspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_intobj_vardnsrecdegspspol_args args;
    [super fileInFrom:aFiler];
    ca_intobj_vardnsrecdegspspol_args_fileIn(aFiler,&args);
    ca_intobj_vardnsrecdegspspol_fileIn(aFiler,&value,args);
    return self;
}

@end

