
/*
 * Computer Algebra Kit (c) 1992,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"
#include <ctype.h>	/* isdigit() */

static int ca_intwrd_create(ca_intdig_t self,DIGIT *digits,int numdigits)
{
    int n;
    assert(sizeof(DIGIT) == sizeof(ca_intwrd_t));
    n = numdigits;while (n--) *self++ = *digits++;
    return numdigits;
}

static void ca_intwrd_extract(ca_intdig_t self,int count,DIGIT *digits,int numdigits)
{
    assert(numdigits == count);while (count--) *digits++ = *self++;
}

static int ca_intwrd_numdigits(ca_intdig_t a,int n)
{
    assert(n && sizeof(ca_intwrd_t) == sizeof(DIGIT));
    return n;
}

static DIGIT ca_intwrd_digitat(ca_intdig_t a,int i,int n)
{
    assert(n && 0 <= i && i < n && sizeof(ca_intwrd_t) == sizeof(DIGIT));
    return a[n-i-1];
}

static int ca_intwrd_insert(ca_intdig_t a,int n,DIGIT d)
{
    assert(n && sizeof(ca_intwrd_t) == sizeof(DIGIT));
    a[n] = d;return n+1;
}

static int ca_intwrd_remove(ca_intdig_t a,int n)
{
    assert(n && sizeof(ca_intwrd_t) == sizeof(DIGIT));
    return n-1;
}

static ca_intwrd_t ca_intwrd_oodbl(ca_intdig_t hic,ca_intdig_t hia,int an)
{
    unsigned c = 0;
    while (an--) {
	ca_intwrd_t w = *--hia;
	c = c + w + w;
	*--hic = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_indbl(ca_intdig_t hia,int an)
{
    unsigned c = 0;
    while (an--) {
	ca_intwrd_t w = *--hia;
	c = c + w + w;
	*hia = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_ooadd(ca_intdig_t hic,ca_intdig_t hia,ca_intdig_t hib,int n,int m)
{
    unsigned c = 0;
    while (n--) {
	c = c + (*--hia) + (*--hib);
	*--hic = c;c >>= 16;
    }
    while (m--) {			/* don't stop if c == 0 ! */
	c += (*--hia);
	*--hic = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_inadd(ca_intdig_t hia,ca_intdig_t hib,int n,int m)
{
    unsigned c = 0;
    while (n--) {
	c = c + (*--hia) + (*--hib);
	*hia = c;c >>= 16;
    }
    while (c && m--) {	/* for inplace, don't need to continue if c == 0 */
	c += (*--hia);
	*hia = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_oosub(ca_intdig_t hic,ca_intdig_t hia,ca_intdig_t hib,int n,int m)
{
    unsigned c = 0;
    ca_intwrd_t cc = 0;	/* alternatively, we could use a negative "carry"
    			 * by using c = 0xffff0000 + (c >> 16) and adding
                         */

    while (n--) {
	c = (*--hia) - (*--hib + cc);
	*--hic = c;cc = (c >> 16);cc = -cc;
    }
    while (m--) {			/* don't stop if cc == 0 ! */
	c = (*--hia) - cc;
	*--hic = c;cc = (c >> 16);cc = -cc;
    }

    return cc;
}

static ca_intwrd_t ca_intwrd_insub(ca_intdig_t hia,ca_intdig_t hib,int n,int m)
{
    unsigned c = 0;
    ca_intwrd_t cc = 0;

    while (n--) {
	c = (*--hia) - (*--hib + cc);
	*hia = c;cc = (c >> 16);cc = -cc;
    }
    while (cc && m--) {	/* for inplace, don't need to continue if cc == 0 */
	c = (*--hia) - cc;
	*hia = c;cc = (c >> 16);cc = -cc;
    }

    return cc;
}

static ca_intwrd_t ca_intwrd_oomul(ca_intdig_t hic,ca_intwrd_t s,ca_intdig_t hib,int n)
{
    unsigned c = 0;
    while (n--) {
	c = s * (*--hib) + c;
	*--hic = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_inaddmul(ca_intdig_t hia,ca_intwrd_t s,ca_intdig_t hib,int n)
{
    unsigned c = 0;
    while (n--) {
	c = c + (*--hia) + s * (*--hib);
	*hia = c;c >>= 16;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_insubmul(ca_intdig_t hia,ca_intwrd_t s,ca_intdig_t hib,int n)
{
    unsigned c = 0;
    ca_intwrd_t cc = 0;
    while (n--) {
	c = (*--hia) - (s * (*--hib) + cc);
	*hia = c;cc = (c >> 16);cc = -cc; /* single precision negate */
    }
    return cc;
}

static ca_intwrd_t ca_intwrd_ooquorem(ca_intdig_t q,ca_intdig_t a,int n,ca_intwrd_t b)
{
    unsigned c = 0;
    while (n--) {
	c = (c << 16) + (*a++);
	*q++ = (c / b);c %= b;
    }
    return c;
}

static ca_intwrd_t ca_intwrd_rem(ca_intdig_t a,int n,ca_intwrd_t b)
{
    unsigned c = 0;
    while (n--) { c = (c << 16) + (*a++);c %= b; }
    return c;
}

static int ca_intwrd_ffo(ca_intwrd_t a)
{
    int c;
    for(c = 0;c<16;c++) if (a & 0x0008000) return c; else a <<= 1;
    return c;
}

static void ca_intwrd_bsl(int cnt,ca_intdig_t dst,int len)
{
    unsigned c;
    
    assert(0 <= cnt && cnt < 16 && len >= 1);
    
    while (--len) {
	c = (*dst << 16) + (*(dst+1));
	c <<= cnt;
	c >>= 16;
	*dst++ = c;
    }
    
    c  = 0;
    c  = *dst;
    c <<= cnt;
    *dst  = c;
}

static void ca_intwrd_bsr(int cnt,ca_intdig_t dst,int len)
{
    unsigned c;
    
    assert(0 <= cnt && cnt < 16 && len >= 1);
    
    while (--len) {
	dst--;
	c = ((*(dst-1)) << 16) + *dst;
	c >>= cnt;
	*dst = c;
    }
    
    c = 0;
    c = *--dst;
    c >>= cnt;
    *dst = c;
}

static ca_intwrd_t ca_intwrd_qguess(ca_intdig_t dvdd,ca_intdig_t dvsr)
{
    unsigned r,s;
    ca_intwrd_t a,b,q;
    
    if ((a=*dvdd) == (b=*dvsr)) {
	q = -1;
	r = dvdd[1] + b;
	if (r > 0x0000ffff) return q;
    } else {
	r = (a << 16) + dvdd[1];
	q = (r / b);
	r = (r % b);
    }

    s = dvsr[1] * q;
    while(s > ((r << 16)+dvdd[2])) {
	q--;r += b; if (r > 0x0000ffff) return q;
	s -= dvsr[1];
    }
    
    return q;
}
static int ca_intdig_check(ca_intdig_t loptr,int n)
{
    if (n) assert(*loptr != 0);return 1;
}

static int ca_intdig_hicheck(ca_intdig_t hiptr,int n)
{
    if (n) assert(*(hiptr - n) != 0);return 1;
}

static int ca_intdig_create(ca_intdig_t self,DIGIT *digits,int numdigits)
{
    return (numdigits)?ca_intwrd_create(self,digits,numdigits):0;
}

static void ca_intdig_extract(ca_intdig_t self,int count,DIGIT *digits,int numdigits)
{
    if (numdigits==0) {
	assert(count==0);
    } else {
	ca_intwrd_extract(self,count,digits,numdigits);
    }
}

static ca_intdig_t ca_intdig_alloc(int n)
{
    return (ca_intdig_t)OC_MallocAtomic(n * sizeof(ca_intwrd_t));
}

static ca_intdig_t ca_intdig_realloc(ca_intdig_t loptr,int n)
{
    assert(ca_intdig_check(loptr,n));
    return (ca_intdig_t)OC_Realloc(loptr,n * sizeof(ca_intwrd_t));
}

static void ca_intdig_copy(ca_intdig_t lodst,ca_intdig_t losrc,int n)
{
    assert(n == 0 || lodst != losrc);assert(ca_intdig_check(losrc,n));
    while (n--) *lodst++ = *losrc++;
}

static void ca_intdig_hicopy(ca_intdig_t hidst,ca_intdig_t hisrc,int n)
{
    assert(n == 0 || hidst != hisrc);assert(ca_intdig_check(hisrc-n,n));
    while (n--) *--hidst = *--hisrc;
}

static ca_intdig_t ca_intdig_free(ca_intdig_t loptr)
{
    OC_Free(loptr);return NULL;
}

static unsigned ca_intdig_hash(int sgn,ca_intdig_t loptr,int n)
{
    assert(ca_intdig_check(loptr,n));
    if (n) {
	unsigned hash = (loptr[0] & 0x0000ffff) + ((unsigned)n << 16);
	return (sgn == -1)?(0xf0000000 | hash):hash;
    } else {
	return 0;
    }
}

static int ca_intdig_isone(int asgn,ca_intdig_t aloptr,int an)
{
    assert(ca_intdig_check(aloptr,an));
    return (an == 1)?(asgn == +1 && aloptr[0] == 1):0;
}

static int ca_intdig_isminusone(int asgn,ca_intdig_t aloptr,int an)
{
    assert(ca_intdig_check(aloptr,an));
    return (an == 1)?(asgn == -1 && aloptr[0] == 1):0;
}

static int ca_intdig_isabseq(ca_intdig_t a,int an,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn));
    if (an == bn) {
	while (an--) { if (*a++ != *b++) return 0; }
	return 1;
    } else {
	return 0;
    }
}

static int ca_intdig_iseq(int asgn,ca_intdig_t a,int an,int bsgn,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn));
    if (an && bn) {
	return asgn == bsgn && ca_intdig_isabseq(a,an,b,bn);
    } else {
	/* don't depend on sign of zero */
	return an == 0 && bn == 0;
    }
}

static int ca_intdig_isop(int asgn,ca_intdig_t a,int an,int bsgn,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn));
    if (an && bn) {
	return asgn + bsgn == 0 && ca_intdig_isabseq(a,an,b,bn);
    } else {
	/* don't depend on sign of zero */
	return an == 0 && bn == 0;
    }
}

static int ca_intdig_abscmp(ca_intdig_t a,int an,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn));
    if (an == bn) {
	while (an--) {
	    ca_intwrd_t ad = *a++;
	    ca_intwrd_t bd = *b++;
	    if (ad == bd) continue;else return (ad > bd)?+1:-1;
	}
	return 0;
    } else {
	return (an < bn)?-1:+1;
    }
}

static int ca_intdig_cmp(int asgn,ca_intdig_t a,int an,int bsgn,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn));

    /* don't depend on sign of zero */
    if (an == 0) return (bn)?-bsgn:0;
    if (bn == 0) return (an)?+asgn:0;

    return (asgn!=bsgn)?asgn:asgn * ca_intdig_abscmp(a,an,b,bn);
}

static int ca_intdig_numdigits(ca_intdig_t a,int n)
{
    assert(ca_intdig_check(a,n));
    return (n)?ca_intwrd_numdigits(a,n):0;
}

static DIGIT ca_intdig_digitat(ca_intdig_t a,int i,int n)
{
    assert(ca_intdig_check(a,n));
    return (n)?ca_intwrd_digitat(a,i,n):0;
}

static DIGIT ca_intdig_lastdigit(ca_intdig_t a,int n)
{
    assert(ca_intdig_check(a,n));
    return (n)?ca_intwrd_digitat(a,0,n):0;
}

static DIGIT ca_intdig_firstdigit(ca_intdig_t a,int n)
{
    assert(ca_intdig_check(a,n));
    return (n)?ca_intwrd_digitat(a,ca_intwrd_numdigits(a,n) - 1,n):0;
}

static int ca_intdig_insert(ca_intdig_t a,int n,DIGIT d)
{
    assert(ca_intdig_check(a,n));
    if (n) {
	return ca_intwrd_insert(a,n,d);
    } else {
	assert(d);a[0] = d;return 1; /* regardless of wordsize */
    }
}

static int ca_intdig_remove(ca_intdig_t a,int n)
{
    assert(ca_intdig_check(a,n));
    return (n)?ca_intwrd_remove(a,n):0;
}

static int ca_intdig_adjust(ca_intdig_t loptr,int n)
{
    while (n && *loptr == 0) { n--;loptr++; }
    return n;
}

static int ca_intdig_oodbl(ca_intdig_t hic,ca_intdig_t hia,int an)
{
    ca_intwrd_t cc;
    assert(ca_intdig_hicheck(hia,an));
    cc = ca_intwrd_oodbl(hic,hia,an);if (cc) { an++;*(hic - an) = 1; }
    return an;
}

static int ca_intdig_indbl(ca_intdig_t hia,int an)
{
    ca_intwrd_t cc;
    assert(ca_intdig_hicheck(hia,an));
    cc = ca_intwrd_indbl(hia,an);if (cc) { an++;*(hia - an) = 1; }
    return an;
}

static int ca_intdig_ooadd(ca_intdig_t hic,int asgn,ca_intdig_t hia,int an,int bsgn,ca_intdig_t hib,int bn)
{
    assert(an >= bn && ca_intdig_abscmp(hia-an,an,hib-bn,bn) > 0);
    assert(ca_intdig_hicheck(hia,an) && ca_intdig_hicheck(hib,bn));

    if (asgn == bsgn) {
	ca_intwrd_t cc = ca_intwrd_ooadd(hic,hia,hib,bn,an-bn);
	if (cc) { an++;*(hic - an) = 1; }
    } else {
	#ifndef NDEBUG
	ca_intwrd_t cc = ca_intwrd_oosub(hic,hia,hib,bn,an-bn);assert(cc==0);
	#else
	ca_intwrd_oosub(hic,hia,hib,bn,an-bn);
	#endif
	an = ca_intdig_adjust(hic-an,an);
    }

    assert(ca_intdig_hicheck(hic,an));return an;
}

static int ca_intdig_inadd(int asgn,ca_intdig_t hia,int an,int bsgn,ca_intdig_t hib,int bn)
{
    assert(an >= bn && ca_intdig_abscmp(hia-an,an,hib-bn,bn) > 0);
    assert(ca_intdig_hicheck(hia,an) && ca_intdig_hicheck(hib,bn));

    if (asgn == bsgn) {
	ca_intwrd_t cc = ca_intwrd_inadd(hia,hib,bn,an-bn);
	if (cc) { an++;*(hia - an) = 1; }
    } else {
	#ifndef NDEBUG
	ca_intwrd_t cc = ca_intwrd_insub(hia,hib,bn,an-bn);assert(cc == 0);
	#else
	ca_intwrd_insub(hia,hib,bn,an-bn);
	#endif
	an = ca_intdig_adjust(hia-an,an);
    }

    assert(ca_intdig_hicheck(hia,an));return an;
}

static int ca_intdig_oomul(ca_intdig_t hic,ca_intdig_t hia,int an,ca_intdig_t hib,int bn)
{
    ca_intwrd_t cc;
    int cn = an + bn;

    assert(an && an <= bn);
    assert(ca_intdig_hicheck(hia,an) && ca_intdig_hicheck(hib,bn));
    
    cc = *--hia;
    cc = ca_intwrd_oomul(hic,cc,hib,bn);
    hic--;*(hic - bn) = cc;

    while (--an) {
	cc = *--hia;
	if (cc) { cc = ca_intwrd_inaddmul(hic,cc,hib,bn); }
	hic--;*(hic - bn) = cc;
    }

    return ca_intdig_adjust(hic-bn,cn);
}

static int ca_intdig_ooquoremword(ca_intdig_t q,ca_intdig_t loa,int n,ca_intwrd_t b,ca_intwrd_t *r)
{
    assert(ca_intdig_check(loa,n) && b != 0);
    if (r) {
	*r = ca_intwrd_ooquorem(q-n,loa,n,b);	/* lo-pointer for q ! */
    } else {
	ca_intwrd_ooquorem(q-n,loa,n,b);
    }
    return ca_intdig_adjust(q-n,n);
}

static ca_intwrd_t ca_intdig_remword(ca_intdig_t loa,int n,ca_intwrd_t b)
{
    assert(ca_intdig_check(loa,n) && b != 0);
    return ca_intwrd_rem(loa,n,b);
}

static int ca_intdig_remquo(ca_intdig_t dvd,ca_intdig_t dvs,int n,int m,ca_intdig_t quo,int *qnref)
{
    int i,f;
    ca_intwrd_t cc,q;
    
    assert(n > 1 && m >= 0);
    f = ca_intwrd_ffo(*dvs);
    if (f) ca_intwrd_bsl(f,dvs,n);

    *--dvd = 0;m++;			/* also if "f" is zero */
    if (f) ca_intwrd_bsl(f,dvd,n+m);
    
    assert(ca_intwrd_ffo(*dvs) == 0);
    if (quo) quo -= m;			/* make quo a lo-pointer */

    for(i=0;i<m;i++) {
	q = ca_intwrd_qguess(dvd,dvs);
	if (q) {
	    cc = ca_intwrd_insubmul(dvd+n+1,q,dvs+n,n);
	    if (cc > *dvd) { ca_intwrd_inadd(dvd+n+1,dvs+n,n,0);q--; }
	}
	*dvd++ = 0;if (quo) *quo++ = q;
    }
    
    if (f) ca_intwrd_bsr(f,dvd+n,n);	/* denormalize takes hi-ptr */
    if (f) ca_intwrd_bsr(f,dvs+n,n);	/* currenlty useless, but inexpensive */

    if (qnref) *qnref = ca_intdig_adjust(quo-m,m);
    return ca_intdig_adjust(dvd,n);
}

static int ca_intdig_inremquo_gt(ca_intdig_t a,int an,ca_intdig_t b,int bn ,ca_intdig_t q,int *qnref)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn) && 0<bn && bn<=an);
    if (bn == 1) {
	ca_intwrd_t rem;
	*qnref = ca_intdig_ooquoremword(q,a,an,b[0],&rem);
	if (rem) { a[an-1] = rem;return 1; } else { return 0; }
    } else {
	return ca_intdig_remquo(a,b,bn,an-bn,q,qnref);
    }
}

static int ca_intdig_inremquo(ca_intdig_t a,int an,ca_intdig_t b,int bn,ca_intdig_t q,int *qnref)
{
    int cmp;
    assert(q && ca_intdig_check(a,an) && ca_intdig_check(b,bn) && 0<bn);
    if (cmp = ca_intdig_abscmp(a,an,b,bn)) {
	if (cmp < 0) {
	    *qnref = 0;return an;
	} else {
	    return ca_intdig_inremquo_gt(a,an,b,bn,q,qnref);
	}
    } else {
	*--q = 1;*qnref = 1;return 0;
    }
}

static int ca_intdig_inrem_gt(ca_intdig_t a,int an,ca_intdig_t b,int bn)
{
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn) && 0<bn && bn<=an);
    if (bn == 1) {
	ca_intwrd_t rem = ca_intdig_remword(a,an,b[0]);
	if (rem) { a[an-1] = rem;return 1; } else { return 0; }
    } else {
	return ca_intdig_remquo(a,b,bn,an-bn,NULL,NULL);
    }
}

static int ca_intdig_inrem(ca_intdig_t a,int an,ca_intdig_t b,int bn)
{
    int cmp;
    assert(ca_intdig_check(a,an) && ca_intdig_check(b,bn) && 0<bn);
    if (cmp = ca_intdig_abscmp(a,an,b,bn)) {
	return (cmp < 0)?an:ca_intdig_inrem_gt(a,an,b,bn);
    } else {
	return 0;
    }
}
typedef struct ca_intreg {
    int			sign;
    int			count;
    ca_intdig_t		top;
    ca_intdig_t		base;	/* not used, just for malloc-debug */
    struct ca_intreg *	next;
} *ca_intreg_t;

#define REGISTER_BSIZE (128 * 1024) /* size in bytes */
#define REGISTER_SIZE (REGISTER_BSIZE / sizeof(ca_intwrd_t))

static int ca_intreg_check(ca_intreg_t self)
{
    assert(self->sign == +1 || self->sign == -1);
    assert(self->top == self->base + REGISTER_SIZE);
    assert(ca_intdig_hicheck(self->top,self->count));
    return 1;
}
    
static ca_intreg_t avail;
#ifdef CAKIT_MACH
static struct mutex availLock = MUTEX_INITIALIZER;
#endif /* CAKIT_MACH */

static ca_intreg_t ca_intreg_init(ca_intreg_t self)
{
    self->sign  = +1;
    self->count = 0;
    /* malloc(pagesize) translates to vm_allocate on NextStep
     * keep a reference to base, because malloc-debug doesn't know about hiptr
     */
    self->base  = ca_intdig_alloc(REGISTER_SIZE);
    self->top   = self->base + REGISTER_SIZE;
    self->next  = NULL;

    return self;
}

static ca_intreg_t ca_intreg_new()
{
    ca_intreg_t new;
#ifdef CAKIT_MACH
    mutex_lock(&availLock);
#endif
    if (avail) {
	new = avail;avail = avail->next;
    } else {
	new = ca_intreg_init((ca_intreg_t)OC_Malloc(sizeof(struct ca_intreg)));
    }
#ifdef CAKIT_MACH
    mutex_unlock(&availLock);
#endif

    assert(ca_intreg_check(new));return new;
}

static void ca_intreg_clear(ca_intreg_t self)
{
    assert(ca_intreg_check(self));
    self->count = 0;
    assert(ca_intreg_check(self));
}

static void ca_intreg_free(ca_intreg_t self)
{
    ca_intreg_clear(self);
#ifdef CAKIT_MACH
    mutex_lock(&availLock);
#endif
    self->next = avail;avail = self;
#ifdef CAKIT_MACH
    mutex_unlock(&availLock);
#endif
}

static void ca_intreg_load(ca_intreg_t r,ca_int_t a)
{
    int count = a->count;
    assert(ca_intreg_check(r) && r->count == 0);
    r->sign = a->sign;
    ca_intdig_hicopy(r->top,a->digits + count,count);
    r->count = count;
    assert(ca_intreg_check(r));
    assert(ca_intdig_iseq(a->sign,a->digits,a->count,r->sign,r->top-r->count,r->count));
}

static void ca_intreg_oodbl_aux(ca_intreg_t self,int asgn,ca_intdig_t a,int an)
{
    assert(self->count == 0);
    self->count = ca_intdig_oodbl(self->top,a+an,an);
    self->sign  = asgn;
    assert(self->count == an || self->count == an+1);
}

static void ca_intreg_oodbl(ca_intreg_t self,ca_int_t a,int v)
{
    assert(ca_intreg_check(self) && v*v == 1);
    ca_intreg_oodbl_aux(self,a->sign * v,a->digits,a->count);
    assert(ca_intreg_check(self));
}

static void ca_intreg_ooadd_aux(ca_intreg_t self,int asgn,ca_intdig_t a,int an,int bsgn,ca_intdig_t b,int bn)
{
    int cmp = ca_intdig_abscmp(a,an,b,bn);

    if (cmp == 0) {
	if (asgn == bsgn) {
	    ca_intreg_oodbl_aux(self,asgn,a,an);
	} else {
	    assert(self->count == 0);
	}
    } else {
	assert(self->count == 0);

	if (cmp > 0) {
	    self->count = ca_intdig_ooadd(self->top,asgn,a+an,an,bsgn,b+bn,bn);
	    self->sign  = asgn;
	} else {
	    self->count = ca_intdig_ooadd(self->top,bsgn,b+bn,bn,asgn,a+an,an);
	    self->sign  = bsgn;
	}

	assert(self->count > 0);
    }
}

static void ca_intreg_ooadd(ca_intreg_t self,ca_int_t a,int v,ca_int_t b,int w)
{
    assert(a != b && ca_intreg_check(self) && v*v == 1 && w*w == 1);
    ca_intreg_ooadd_aux(self,a->sign * v,a->digits,a->count,b->sign * w,b->digits,b->count);
    assert(ca_intreg_check(self));
}

static void ca_intreg_inadd_aux(ca_intreg_t *ar,int asgn,ca_intdig_t a,int an,ca_intreg_t *br,int bsgn,ca_intdig_t b,int bn)
{
    int cmp = ca_intdig_abscmp(a,an,b,bn);

    if (cmp == 0) {
	if (asgn == bsgn) {
	    (*ar)->count = ca_intdig_indbl(a+an,an);
	    (*ar)->sign  = asgn;
	} else {
	    (*ar)->count = 0;
	}
    } else {
	if (cmp > 0) {
	    (*ar)->count = ca_intdig_inadd(asgn,a+an,an,bsgn,b+bn,bn);
	    (*ar)->sign  = asgn;
	} else {
	    (*br)->count = ca_intdig_inadd(bsgn,b+bn,bn,asgn,a+an,an);
	    (*br)->sign  = bsgn;
	    if (1) { ca_intreg_t t = *ar;*ar = *br;*br = t; }
	}
    }
}

static void ca_intreg_inadd(ca_intreg_t *ar,int v,ca_intreg_t *br,int w)
{
    int an,bn;
    ca_intreg_t a,b;
    a = *ar;an = a->count;b = *br;bn = b->count;
    assert(a != b && ca_intreg_check(a) && ca_intreg_check(b) && v*v == 1 && w*w == 1);
    ca_intreg_inadd_aux(ar,v*a->sign,a->top-an,an,br,w*b->sign,b->top-bn,bn);
    assert(ca_intreg_check(*ar) && ca_intreg_check(*br));
}

static void ca_intreg_oomul_aux(ca_intreg_t self,int sgn,ca_intdig_t a,int an,ca_intdig_t b,int bn)
{
    assert(self->count == 0);
    if (an <= bn) {
	if (an) {
	    self->sign = sgn;
	    self->count = ca_intdig_oomul(self->top,a+an,an,b+bn,bn);
	}
    } else {
	if (bn) {
	    self->sign = sgn;
	    self->count = ca_intdig_oomul(self->top,b+bn,bn,a+an,an);
	}
    }
    assert(self->count == 0 || self->count == an + bn - 1 || self->count == an + bn);
}

static void ca_intreg_oomul(ca_intreg_t self,ca_int_t a,ca_int_t b)
{
    assert((a != b || a == b) && ca_intreg_check(self)); /* allow a==b for sqr */
    ca_intreg_oomul_aux(self,a->sign * b->sign,a->digits,a->count,b->digits,b->count);
    assert(ca_intreg_check(self));
}

static void ca_intreg_inremquo(ca_intreg_t a,ca_intreg_t b,ca_intreg_t q)
{
    assert(ca_intreg_check(q) && q->count == 0);
    assert(ca_intreg_check(a) && ca_intreg_check(b) && b->count);
    a->count = ca_intdig_inremquo(a->top-a->count,a->count,b->top-b->count,b->count ,q->top,&q->count);
    q->sign  = a->sign * b->sign;
    assert(ca_intreg_check(a) && ca_intreg_check(b) && a->count <= b->count);
    assert(ca_intreg_check(q));
}

static void ca_intreg_inrem(ca_intreg_t a,ca_intreg_t b)
{
    assert(ca_intreg_check(a) && ca_intreg_check(b) && b->count);
    a->count = ca_intdig_inrem(a->top-a->count,a->count,b->top-b->count,b->count);
    assert(ca_intreg_check(a) && ca_intreg_check(b) && b->count);
}

static void ca_intreg_ooquoremword(ca_intreg_t q,ca_int_t a,ca_intwrd_t b,ca_intwrd_t *r)
{
    assert(b != 0 && ca_intreg_check(q) && q->count == 0);
    q->count = ca_intdig_ooquoremword(q->top,a->digits,a->count,b,r);
    q->sign  = a->sign;
    assert(q->count == a->count || q->count == a->count - 1);
}
typedef struct ca_int ca_int_c;
typedef struct ca_int *ca_intseq_t;
typedef struct { unsigned unused; } ca_int_args;

#define ca_int_isvalue	0
#define ca_int_pervalue	0

static ca_int_args ca_int_getargs(id self)
{
    ca_int_args args;args.unused = 0;return args;
}

static int ca_int_size(ca_int_args args)
{
    return 1;
}

static int ca_int_bsize(ca_int_args args)
{
    return sizeof(ca_int_c);
}

static ca_int_t ca_int_id2t(id object)
{
    return [object ca_int_value];	/* read-only */
}

static ca_int_t ca_int_id2ref(id object,ca_int_c *ref)
{
    return [object ca_int_value];	/* read-only, scratch space "ref" */
}

static ca_int_c *ca_int_id2c(id object)
{
    return [object ca_int_reference];	/* write for inplace arithmetic */
}

static id ca_int_t2id(id proto,ca_int_t f)
{
    return [proto ca_int_value:f];	/* new object containing copy of "f" */
}

static id ca_int_c2id(id proto,ca_int_c* c)
{
    return [proto ca_int_reference:c];  /* new object containing "c" */
}

static int ca_int_check(ca_int_t self,ca_int_args args)
{
    assert(self->sign == -1 || self->sign == +1);
    assert(0 <= self->count && self->count <= self->capacity);
    assert(ca_intdig_check(self->digits,self->count));
    return 1;
}

static void ca_int_create(ca_int_c *c,int sign,DIGIT *digits,int numdigits,ca_int_args args)
{
    c->sign     = sign;
    c->digits   = ca_intdig_alloc(numdigits);	/* that's enough */
    c->count    = ca_intdig_create(c->digits,digits,numdigits);
    c->capacity = numdigits;assert(c->count <= numdigits);
    assert(ca_int_check(c,args));
}

static void ca_int_extract(ca_int_t a,DIGIT *digits,int numdigits,ca_int_args args)
{
    assert(ca_int_check(a,args));
    assert(numdigits == ca_intdig_numdigits(a->digits,a->count));
    ca_intdig_extract(a->digits,a->count,digits,numdigits);
    assert(ca_int_check(a,args));
}

static void ca_int_oozero(ca_int_c *c,ca_int_args args)
{
    c->sign     = 1;
    c->count    = 0;
    c->capacity = 4;
    c->digits   = ca_intdig_alloc(4);

    assert(ca_int_check(c,args));
}

static void ca_int_copy(ca_int_c *c,ca_int_t a,ca_int_args args)
{
    int count;
    assert(c != a && ca_int_check(a,args));

    if ((count = a->count)) {
	c->sign     = a->sign;
	c->count    = count;
	c->capacity = count;
	c->digits   = ca_intdig_alloc(count);
	ca_intdig_copy(c->digits,a->digits,count);
    } else {
	ca_int_oozero(c,args);
    }

    assert(ca_int_check(c,args) && ca_int_check(a,args));
}

static void ca_int_move(ca_int_c *c,ca_int_c *a,ca_int_args args)
{
    assert(c != a && ca_int_check(a,args));

    c->sign     = a->sign;
    c->count    = a->count;
    c->capacity = a->capacity;
    c->digits   = a->digits;
#ifndef NDEBUG
    a->digits = NULL;
#endif

    assert(ca_int_check(c,args));
}

static void ca_int_clear(ca_int_c *c,ca_int_args args)
{
    ca_int_check(c,args);c->digits = ca_intdig_free(c->digits);
}

static void ca_int_expand(ca_int_c *c)
{
    c->capacity += 1 + c->count;
    c->digits    = ca_intdig_realloc(c->digits,c->capacity);
}

static void ca_int_store(ca_int_c *c,ca_intreg_t r,ca_int_args args)
{
    int count;
    assert(ca_intreg_check(r));

    if ((count = r->count)) {
	c->sign     = r->sign;
	c->count    = count;
	c->capacity = count;
	c->digits   = ca_intdig_alloc(count);
	ca_intdig_hicopy(c->digits + count,r->top,count);
    } else {
	ca_int_oozero(c,args);
    }

    assert(ca_int_check(c,args) && ca_intreg_check(r));
    assert(ca_intdig_iseq(c->sign,c->digits,c->count,r->sign,r->top-r->count,r->count));
}

static void ca_int_restore(ca_int_c *c,ca_intreg_t r,ca_int_args args)
{
    int count;
    assert(ca_int_check(c,args) && ca_intreg_check(r));

    count    = r->count;
    while (count > c->capacity) ca_int_expand(c);
    c->sign  = r->sign;
    c->count = count;
    ca_intdig_hicopy(c->digits + count,r->top,count);

    assert(ca_int_check(c,args) && ca_intreg_check(r));
}

static unsigned ca_int_hash(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_hash(a->sign,a->digits,a->count);
}

static int ca_int_iseq(ca_int_t a,ca_int_t b,ca_int_args args)
{
    assert(a != b && ca_int_check(a,args) && ca_int_check(b,args));
    return ca_intdig_iseq(a->sign,a->digits,a->count,b->sign,b->digits,b->count);
}

static int ca_int_sgn(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args)); return (a->count)?a->sign:0;
}

static int ca_int_cmp(ca_int_t a,ca_int_t b,ca_int_args args)
{
    assert(a != b && ca_int_check(a,args) && ca_int_check(b,args));
    return ca_intdig_cmp(a->sign,a->digits,a->count,b->sign,b->digits,b->count);
}

static int ca_int_numdigits(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_numdigits(a->digits,a->count);
}

static DIGIT ca_int_digitat(ca_int_t a,int i,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_digitat(a->digits,i,a->count);
}

static DIGIT ca_int_lastdigit(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_lastdigit(a->digits,a->count);
}

static DIGIT ca_int_firstdigit(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_firstdigit(a->digits,a->count);
}

static void ca_int_insert(ca_int_t a,DIGIT d,ca_int_args args)
{
    assert(ca_int_check(a,args));
    if (a->count == 0 && d == 0) {
	/* don't insert - keep leading digit nonzero */
    } else {
	while (a->count + 1 > a->capacity) ca_int_expand(a);
	a->count = ca_intdig_insert(a->digits,a->count,d);
    }
    assert(ca_int_check(a,args));
}

static DIGIT ca_int_remove(ca_int_t a,ca_int_args args)
{
    DIGIT d;assert(ca_int_check(a,args));
    if (a->count) {
	d = ca_intdig_lastdigit(a->digits,a->count);
	a->count = ca_intdig_remove(a->digits,a->count);
    } else {
	d = 0;
    }
    assert(ca_int_check(a,args));return d;
}

static int ca_int_iszero(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));return (a->count == 0);
}

static int ca_int_isop(ca_int_t a,ca_int_t b,ca_int_args args)
{
    assert(a != b && ca_int_check(a,args) && ca_int_check(b,args));
    return ca_intdig_isop(a->sign,a->digits,a->count,b->sign,b->digits,b->count);
}

static void ca_int_ooneg(ca_int_c *c,ca_int_t a,ca_int_args args)
{
    ca_int_copy(c,a,args);c->sign = -c->sign;assert(ca_int_check(c,args));
}

static void ca_int_inneg(ca_int_c *c,ca_int_args args)
{
    assert(ca_int_check(c,args));c->sign = -c->sign;assert(ca_int_check(c,args));
}

static void ca_int_oodbl(ca_int_c *c,ca_int_t a,int v,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(c != a && ca_int_check(a,args));
    ca_intreg_oodbl(r0,a,v);ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args));

    ca_intreg_free(r0);
}

static void ca_int_indbl(ca_int_c *c,int v,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(ca_int_check(c,args));
    ca_intreg_oodbl(r0,c,v);ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args));

    ca_intreg_free(r0);
}

static void ca_int_ooadd(ca_int_c *c,ca_int_t a,int v,ca_int_t b,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(a != b && ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_ooadd(r0,a,v,b,w);ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args) && ca_int_check(b,args));

    ca_intreg_free(r0);
}

static void ca_int_inadd(ca_int_c *c,int v,ca_int_t b,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(c != b && ca_int_check(c,args));
    ca_intreg_ooadd(r0,c,v,b,w);ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args));

    ca_intreg_free(r0);
}

static void ca_int_ooaddword(ca_int_c *c,ca_int_t a,int v,ca_intwrd_t b,int w,ca_int_args args)
{
    struct ca_int tmp;assert(b != 0);

    tmp.sign     = +1;
    tmp.count    =  1;
    tmp.capacity =  1;
    tmp.digits   = &b;
    ca_int_ooadd(c,a,v,&tmp,w,args);
}

static void ca_int_inaddword(ca_int_c *c,int v,ca_intwrd_t b,int w,ca_int_args args)
{
    struct ca_int tmp;assert(b != 0);

    tmp.sign     = +1;
    tmp.count    =  1;
    tmp.capacity =  1;
    tmp.digits   = &b;
    ca_int_inadd(c,v,&tmp,w,args);
}

static int ca_int_isone(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_isone(a->sign,a->digits,a->count);
}

static int ca_int_isminusone(ca_int_t a,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_isminusone(a->sign,a->digits,a->count);
}

static void ca_int_oosqr(ca_int_c *c,ca_int_t a,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(c != a && ca_int_check(a,args));
    ca_intreg_oomul(r0,a,a);ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args));

    ca_intreg_free(r0);
}

static void ca_int_insqr(ca_int_c *c,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(ca_int_check(c,args));
    ca_intreg_oomul(r0,c,c);ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args));

    ca_intreg_free(r0);
}

static void ca_int_oomul(ca_int_c *c,ca_int_t a,ca_int_t b,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(a != b && ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_oomul(r0,a,b);ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args) && ca_int_check(b,args));

    ca_intreg_free(r0);
}

static void ca_int_inmul(ca_int_c *c,ca_int_t b,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(c != b && ca_int_check(c,args));
    ca_intreg_oomul(r0,c,b);ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args));

    ca_intreg_free(r0);
}

static void ca_int_oomulword(ca_int_c *c,ca_int_t a,ca_intwrd_t b,ca_int_args args)
{
    struct ca_int tmp;assert(b != 0);

    tmp.sign     = +1;
    tmp.count    =  1;
    tmp.capacity =  1;
    tmp.digits   = &b;
    ca_int_oomul(c,a,&tmp,args);
}

static void ca_int_inmulword(ca_int_c *c,ca_intwrd_t b,ca_int_args args)
{
    struct ca_int tmp;assert(b != 0);

    tmp.sign     = +1;
    tmp.count    =  1;
    tmp.capacity =  1;
    tmp.digits   = &b;
    ca_int_inmul(c,&tmp,args);
}

static void ca_int_ooaddmul(ca_int_c *c,ca_int_t a,int v,ca_int_t b,ca_int_t t,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(b,args) && ca_int_check(t,args));
    ca_intreg_load(r0,a);ca_intreg_oomul(r1,b,t);ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args) && ca_int_check(b,args) && ca_int_check(t,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_inaddmul(ca_int_c *c,int v,ca_int_t b,ca_int_t t,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(c,args) && ca_int_check(b,args) && ca_int_check(t,args));
    ca_intreg_load(r0,c);ca_intreg_oomul(r1,b,t);ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(b,args) && ca_int_check(t,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_oomuladd(ca_int_c *c,ca_int_t a,ca_int_t s,int v,ca_int_t b,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(s,args) && ca_int_check(b,args));
    ca_intreg_oomul(r0,a,s);ca_intreg_load(r1,b);ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_store(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(a,args) && ca_int_check(s,args) && ca_int_check(b,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_inmuladd(ca_int_c *c,ca_int_t s,int v,ca_int_t b,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(c,args) && ca_int_check(s,args) && ca_int_check(b,args));
    ca_intreg_oomul(r0,c,s);ca_intreg_load(r1,b);ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(s,args) && ca_int_check(b,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_oomuladdmul(ca_int_c *c,ca_int_t a,ca_int_t s,int v,ca_int_t b,ca_int_t t,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(s,args));
    assert(ca_int_check(b,args) && ca_int_check(t,args));
    ca_intreg_oomul(r0,a,s);ca_intreg_oomul(r1,b,t);
    ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_store(c,r0,args);
    assert(ca_int_check(c,args)&&ca_int_check(a,args));
    assert(ca_int_check(s,args)&&ca_int_check(b,args)&&ca_int_check(t,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_inmuladdmul(ca_int_c *c,ca_int_t s,int v,ca_int_t b,ca_int_t t,int w,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();
    ca_intreg_t r1 = ca_intreg_new();

    assert(ca_int_check(c,args) && ca_int_check(s,args));
    assert(ca_int_check(b,args) && ca_int_check(t,args));
    ca_intreg_oomul(r0,c,s);ca_intreg_oomul(r1,b,t); ca_intreg_inadd(&r0,v,&r1,w);
    ca_int_restore(c,r0,args);
    assert(ca_int_check(c,args) && ca_int_check(s,args));
    assert(ca_int_check(b,args) && ca_int_check(t,args));

    ca_intreg_free(r0);ca_intreg_free(r1);
}

static void ca_int_ooremquo(ca_int_c *r,ca_int_t a,ca_int_t b,ca_int_c *q,ca_int_args args)
{
    ca_intreg_t dvdd = ca_intreg_new();
    ca_intreg_t dvsr = ca_intreg_new();
    ca_intreg_t quot = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_load(dvdd,a);ca_intreg_load(dvsr,b);
    ca_intreg_inremquo(dvdd,dvsr,quot);
    ca_int_store(r,dvdd,args);ca_int_store(q,quot,args);
    assert(ca_int_check(r,args) && ca_int_check(q,args));
    assert(ca_int_check(a,args) && ca_int_check(b,args));
    
    ca_intreg_free(dvdd);
    ca_intreg_free(dvsr);
    ca_intreg_free(quot);
}

static void ca_int_inremquo(ca_int_c *a,ca_int_t b,ca_int_c *q,ca_int_args args)
{
    ca_intreg_t dvdd = ca_intreg_new();
    ca_intreg_t dvsr = ca_intreg_new();
    ca_intreg_t quot = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_load(dvdd,a);ca_intreg_load(dvsr,b);
    ca_intreg_inremquo(dvdd,dvsr,quot);
    ca_int_restore(a,dvdd,args);ca_int_store(q,quot,args);
    assert(ca_int_check(q,args));
    assert(ca_int_check(a,args) && ca_int_check(b,args));
    
    ca_intreg_free(dvdd);
    ca_intreg_free(dvsr);
    ca_intreg_free(quot);
}

static void ca_int_oorem(ca_int_c *r,ca_int_t a,ca_int_t b,ca_int_args args)
{
    ca_intreg_t dvdd = ca_intreg_new();
    ca_intreg_t dvsr = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_load(dvdd,a);ca_intreg_load(dvsr,b);
    ca_intreg_inrem(dvdd,dvsr);ca_int_store(r,dvdd,args);
    assert(ca_int_check(r,args));
    assert(ca_int_check(a,args) && ca_int_check(b,args));
    
    ca_intreg_free(dvdd);
    ca_intreg_free(dvsr);
}

static void ca_int_inrem(ca_int_c *a,ca_int_t b,ca_int_args args)
{
    ca_intreg_t dvdd = ca_intreg_new();
    ca_intreg_t dvsr = ca_intreg_new();

    assert(ca_int_check(a,args) && ca_int_check(b,args));
    ca_intreg_load(dvdd,a);ca_intreg_load(dvsr,b);
    ca_intreg_inrem(dvdd,dvsr);ca_int_restore(a,dvdd,args);
    assert(ca_int_check(a,args) && ca_int_check(b,args));
    
    ca_intreg_free(dvdd);
    ca_intreg_free(dvsr);
}

static void ca_int_ooquoremword(ca_int_c *q,ca_int_t a,ca_intwrd_t b,ca_intwrd_t *r,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(ca_int_check(a,args));
    ca_intreg_ooquoremword(r0,a,b,r);ca_int_store(q,r0,args);
    assert(ca_int_check(q,args) && ca_int_check(a,args));
    assert(r == NULL || *r < b);
    
    ca_intreg_free(r0);
}

static void ca_int_inquoremword(ca_int_c *q,ca_intwrd_t b,ca_intwrd_t *r,ca_int_args args)
{
    ca_intreg_t r0 = ca_intreg_new();

    assert(ca_int_check(q,args));
    ca_intreg_ooquoremword(r0,q,b,r);ca_int_restore(q,r0,args);
    assert(ca_int_check(q,args) && (r == NULL || *r < b));
    
    ca_intreg_free(r0);
}

static ca_intwrd_t ca_int_remword(ca_int_t a,ca_intwrd_t b,ca_int_args args)
{
    assert(ca_int_check(a,args));
    return ca_intdig_remword(a->digits,a->count,b);
}

static void ca_int_fileout(id aFiler,ca_int_c *a,ca_int_args args)
{
    DIGIT *digits;
    int sign,count;
    assert(ca_int_check(a,args));

    sign   = a->sign;
    count  = ca_int_numdigits(a,args);
    digits = (DIGIT *)OC_MallocAtomic(sizeof(DIGIT) * count);
    ca_int_extract(a,digits,count,args);

    [aFiler fileOut:&sign type:'i'];
    [aFiler fileOut:&count type:'i'];
    while (count--) [aFiler fileOut:digits++ type:'S'];
    OC_Free(digits);assert(ca_int_check(a,args));
}

static void ca_int_filein(id aFiler,ca_int_c *c,ca_int_args args)
{
    DIGIT *digits;
    int sign,count,i;

    [aFiler fileIn:&sign type:'i'];
    [aFiler fileIn:&count type:'i'];
    digits = (DIGIT *)OC_MallocAtomic(sizeof(DIGIT) * count);
    for(i=0;i<count;i++) [aFiler fileIn:digits+i type:'S'];
    ca_int_create(c,sign,digits,count,args);
    OC_Free(digits);
}

@implementation Integer
- check
{
    [super check];ca_int_check(&value,ca_int_getargs(self));return self;
}


- _setUpInteger
{
    ca_int_oozero(&value,ca_int_getargs(self));
    return self;
}

+ new
{
    return [[super new] _setUpInteger];
}
+ str:(STR)aString
{
    return [super str:aString];
}

+ newStrValue:(STR)aString
{
    static id dummy = nil;
    if (!dummy) dummy = [self new]; return [dummy strValue:aString];
}
- newIntValue:(int)intValue
{
    return [self notImplemented:_cmd];
}

+ newIntValue:(int)intValue
{
    static id dummy = nil;
    if (!dummy) dummy = [self new]; return [dummy intValue:intValue];
}

+ int:(int)intValue
{
    return [super int:intValue];
}

- newFactorial:(int)n
{
    return [self notImplemented:_cmd];
}

+ factorial:(int)n
{
    return [self newFactorial:n];
}

+ newFactorial:(int)n
{
    if (n < 0 || n > 10000) {
	return [self error:"Be Reasonable"];
    } else {
	DIGIT i,m;
	unsigned long j;
	id fac = [self newIntValue:1];
	
	m = (DIGIT)n;

	for(i=2;j=i*(i+1),j<USHRT_MAX && i+1<=m;i+=2) {
	    fac = [fac multiplySelfDigit:j];
	}
    
	for(;i<=m;i++) fac = [fac multiplySelfDigit:i];
    
	return fac;
    }
}

+ fibonacci:(int)n
{
    [self notImplemented:_cmd];return nil;
}

- copy
{
    id copy = [super copy];
    ca_int_copy([copy ca_int_reference],&value,ca_int_getargs(self));
    return copy;
}
- deepCopy
{
    return [self copy];
}

- clear
{
    ca_int_clear(&value,ca_int_getargs(self));return [super clear];
}

- empty
{
    self = [self clone];ca_int_oozero(&value,ca_int_getargs(self));return self;
}

- (ca_int_t) ca_int_value
{
    return &value;
}

- ca_int_value:(ca_int_t)aValue
{
    self = [self clone];ca_int_copy(&value,aValue,ca_int_getargs(self));return self;
}

- (ca_int_t) ca_int_reference
{
    return &value;
}

- ca_int_reference:(ca_int_t)aReference
{
    self = [self clone];ca_int_move(&value,aReference,ca_int_getargs(self));return self;
}

- (BOOL)sameClass:b			{ return isa == [b class]; }
- (BOOL)inAdditiveSemiGroup		{ return YES; }
- (BOOL)inAdditiveMonoid		{ return YES; }
- (BOOL)inAdditiveGroup			{ return YES; }

- (BOOL)inSemiGroup			{ return YES; }
- (BOOL)inMonoid			{ return YES; }
- (BOOL)inGroup				{ return NO;  }

- (BOOL)inEuclideanDomain		{ return YES; }
- (BOOL)inIntegralDomain		{ return YES; }
- (BOOL)inField				{ return NO; }

- (BOOL)inOrderedSet			{ return YES; }
- (int)characteristic			{ return 0;   }

- class_vector
{
    return (id)[ca_intobj_vector class];
}

- class_vardnsrecdegsps_polynomial
{
    return (id)[ca_intobj_vardnsrecdegsps_polynomial class];
}

- class_vardnsrecdegdns_polynomial
{
    return (id)[ca_intobj_vardnsrecdegdns_polynomial class];
}

- class_vardnsexpdegsps_polynomial
{
    return (id)[ca_intobj_vardnsexpdegsps_polynomial class];
}


- (unsigned) hash
{
    return ca_int_hash(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_int_iseq(ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_int_iseq(ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
    }
}

- (int) numDigits
{
    return ca_int_numdigits(&value,ca_int_getargs(self));
}
- (BOOL) isDigit
{
    return ca_int_numdigits(&value,ca_int_getargs(self)) == 1;
}

- insertDigit:(DIGIT)d
{
    if ([self isZero] && d == 0) {
	return self;
    } else {
	[self checkRefCountOne];
	ca_int_insert(&value,d,ca_int_getargs(self));
	return [self invalidate];
    }
}
- (DIGIT) removeDigit
{
    DIGIT d;
   
    if ([self isZero]) {
	d = 0;
    } else {
	[self checkRefCountOne];
	d = ca_int_remove(&value,ca_int_getargs(self));
	[self invalidate];
    }
    
    return d;
}

- (DIGIT) digitAt:(int)i
{
    return ca_int_digitat(&value,i,ca_int_getargs(self));
}
- (DIGIT) lastDigit
{
    return ca_int_lastdigit(&value,ca_int_getargs(self));
}
- (DIGIT) leadingDigit
{
    return ca_int_firstdigit(&value,ca_int_getargs(self));
}

static char *getChars(id self,int *cref,DIGIT base)
{
    id copy;
    int count,capacity;
    DIGIT r;char *stack;
    
    copy  = [self copy];assert(0 < base <= 10);
    count = 0;stack = (char *)OC_MallocAtomic(sizeof(char) * (capacity = 100));

    copy = [copy quotientSelfDigit:base remainder:&r];
    stack[count++] = r + '0';
    
    while ([copy notZero]) {
	copy = [copy quotientSelfDigit:base remainder:&r];
	stack[count++] = r + '0';
	if (count == capacity) stack = (char *)OC_Realloc(stack,capacity *= 2);
    }
    
    [copy free];*cref = count;return stack;
}

- (STR) strValue
{
    /* might use getChars here instead of going through printToFile: */
    return [super strValue];
}
static id fromDigits(id self,STR aString)
{
    int c;
    id res = [self zero];
    assert(aString != NULL && isdigit(*aString));
    
    while ((c = *aString++)) {
	if (isdigit(c)) {
	    res = [[res multiplySelfDigit:10] addSelfDigit:c - '0'];
	} else {
	    res = [res free];break;
	}
    }
    
    return res;
}

- strValue:(STR)aString
{
    if (aString) {
	int c,sign = +1;
	while ((c = *aString)) {
	    if (c == '+') { sign = +sign; aString++;continue; }
	    if (c == '-') { sign = -sign; aString++;continue; }
	    if (c == ' ' || c == '\t') { aString++;continue;  }
	    if (isdigit(c)) {
		id res = fromDigits(self,aString);
		return (sign==+1)?res:[res negateSelf];
	    } else {
		return nil;
	    }
	}
	return nil;
    } else {
	return nil;
    }
}

- asNumerical
{
    return [CAFloatClass newFloatValue:[self floatValue]];
}
- asModp:(unsigned short)p
{
    if ([self sign] < 0) {
	return [CAIntegerModpClass newIntValue:-[self remainderDigit:p] modulus:p];
    } else {
	return [CAIntegerModpClass newIntValue:+[self remainderDigit:p] modulus:p];
    }
}
- (DIGIT) digitValue
{
    return [self lastDigit];
}
- digitValue:(DIGIT)aValue
{
    self = [self empty];[self insertDigit:aValue];return self;
}
- (unsigned long) ulongValue
{
    if ([self numDigits] < 2) {
	return [self digitValue];
    } else {
	DIGIT d0 = [self digitAt:0];
	DIGIT d1 = [self digitAt:1];
	return d0 + (d1 << (CHAR_BIT * sizeof(DIGIT)));
    }
}
- ulongValue:(unsigned long)aValue
{
    id res;
    DIGIT d0 = aValue;
    DIGIT d1 = aValue >> (CHAR_BIT * sizeof(DIGIT));
    
    res = [self empty];
    if (d1) [res insertDigit:d1];
    if (d1 || d0) [res insertDigit:d0];
    return res;
}
- (int) intValue
{
    if ([self numDigits] <= 2) {
	unsigned long u = [self ulongValue];
    
	if ([self sign] < 0) {
	    return (u < INT_MAX)?-u:INT_MIN;
	} else {
	    return (u < INT_MAX)?+u:INT_MAX;
	}
    } else {
	return ([self sign] < 0)?INT_MIN:INT_MAX;
    }
}

- intValue:(int)aValue
{
    id res;

    if (aValue < 0) {
	res = [[self ulongValue:-aValue] negateSelf];
    } else {
	res = [self ulongValue:aValue];
    }

    return res;
}

- (double) doubleValue
{
    int i,numDigits;
    double doubleValue = 0.0;
    double bound = DBL_MAX / 2.0;
    
    for(numDigits = [self numDigits],i=numDigits-1;i>=0;i--) {
	DIGIT mask  = (USHRT_MAX + 1) >> 1;
	DIGIT digit = [self digitAt:i];
	while (mask) {
	    if (doubleValue >= bound) return ([self sign]<0)?DBL_MIN:DBL_MAX;
	    doubleValue *= 2.0;
	    if (mask & digit) doubleValue += 1.0;
	    mask >>= 1;
	}
    }
    
    return [self sign] * doubleValue;
}

- (float) floatValue
{
    int i,numDigits;
    float floatValue = 0.0;
    float bound = FLT_MAX / 2.0;
    
    for(numDigits = [self numDigits],i=numDigits-1;i>=0;i--) {
	DIGIT mask  = (USHRT_MAX + 1) >> 1;
	DIGIT digit = [self digitAt:i];
	while (mask) {
	    if (floatValue >= bound) return ([self sign]<0)?FLT_MIN:FLT_MAX;
	    floatValue *= 2.0;
	    if (mask & digit) floatValue += 1.0;
	    mask >>= 1;
	}
    }
    
    return [self sign] * floatValue;
}


- (int) sign
{
    return ca_int_sgn(ca_int_id2t(self),ca_int_getargs(self));
}

- (int) compare:b
{
    [self checkSameClass:b];

    if (self==b) {
    return 0;
    } else {
    return ca_int_cmp(ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
    }
}

- zero
{
    static id zero = nil;
    
#ifdef CAKIT_MACH
    static struct mutex iLock = MUTEX_INITIALIZER;

    mutex_lock(&iLock);
    if (zero == nil) zero = [self intValue:0];
    mutex_unlock(&iLock);
#else
    if (zero == nil) zero = [self intValue:0];
#endif /* CAKIT_MACH */
    
    return [zero cheapCopy];
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_int_iszero(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return NO == ca_int_iszero(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) isOpposite:b
{
    [self checkSameClass:b];
    
    if (self == b) {
	return [self isZero] || [self isCharacteristicTwo];
    } else {
	return ca_int_isop(ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
    }
}

- (BOOL) notOpposite:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self notZero] && [self notCharacteristicTwo];
    } else {
	return NO == ca_int_isop(ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
    }
}

- negate
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self cheapCopy];
    } else {
	ca_int_c c;assert([self check]);
	ca_int_ooneg(&c,ca_int_id2t(self),ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return self;
    } else {
	assert([self check]);
	ca_int_inneg(ca_int_id2c(self),ca_int_getargs(self));
	return [self invalidate];
    }
}

- _double:(int)v
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_int_c c;assert([self check]);
	ca_int_oodbl(&c,ca_int_id2t(self),v,ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_int_indbl(ca_int_id2c(self),v,ca_int_getargs(self));
	return [self invalidate];
    }
}

- double
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_int_c c;assert([self check]);
	ca_int_oodbl(&c,ca_int_id2t(self),+1,ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_int_indbl(ca_int_id2c(self),+1,ca_int_getargs(self));
	return [self invalidate];
    }
}

- _add:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);

    if (self == b) {
	return (v==w)?[self _double:v]:[self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b _negate:w];
	    if ([b isZero]) return [self _negate:v];
	}
	if (1) {
	    ca_int_c c;
	    ca_int_ooadd(&c,ca_int_id2t(self),v,ca_int_id2t(b),w,ca_int_getargs(self));
	    return ca_int_c2id(self,&c);
	}
    }

    return nil; /* to quiet some compilers, e.g. WATCOM */
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);

    if (self == b) {
	return (v==w)?[self _doubleSelf:v]:idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return [self _negateSelf:v];
	    if ([self isZero]) return idReplace(self,[b _negate:w]);
	}
	ca_int_inadd(ca_int_id2c(self),v,ca_int_id2t(b),w,ca_int_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- add:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self double];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b cheapCopy];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_int_c c;
	    ca_int_ooadd(&c,ca_int_id2t(self),+1,ca_int_id2t(b),+1,ca_int_getargs(self));
	    return ca_int_c2id(self,&c);
	}
    }

    return nil;
}

- addSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self add:b]);

    if (self == b) {
	return [self doubleSelf];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b cheapCopy]);
	}
	ca_int_inadd(ca_int_id2c(self),+1,ca_int_id2t(b),+1,ca_int_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- subtract:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b negate];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_int_c c;
	    ca_int_ooadd(&c,ca_int_id2t(self),+1,ca_int_id2t(b),-1,ca_int_getargs(self));
	    return ca_int_c2id(self,&c);
	}
    }

    return nil;
}

- subtractSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self subtract:b]);

    if (self == b) {
	return idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b negate]);
	}
	ca_int_inadd(ca_int_id2c(self),+1,ca_int_id2t(b),-1,ca_int_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- addDigit:(DIGIT)d
{
    ca_int_c c;
    if (d == 0) return [self cheapCopy];
    if ([self isZero]) return [self digitValue:d];
    ca_int_ooaddword(&c,&value,+1,d,+1,ca_int_getargs(self));
    return [self ca_int_reference:&c];
}

- addSelfDigit:(DIGIT)d
{
    if (refCount > 1) return idReplace(self,[self addDigit:d]);
    
    if (d == 0) return self;
    if ([self isZero]) return idReplace(self,[self digitValue:d]);
    ca_int_inaddword(&value,+1,d,+1,ca_int_getargs(self));
    return [self invalidate];
}

- subtractDigit:(DIGIT)d
{
    ca_int_c c;
    if (d == 0) return [self cheapCopy];
    if ([self isZero]) return [[self digitValue:d] negateSelf];
    ca_int_ooaddword(&c,&value,+1,d,-1,ca_int_getargs(self));
    return [self ca_int_reference:&c];
}

- subtractSelfDigit:(DIGIT)d
{
    if (refCount > 1) return idReplace(self,[self subtractDigit:d]);
    
    if (d == 0) return self;
    if ([self isZero]) return idReplace(self,[[self digitValue:d] negateSelf]);
    ca_int_inaddword(&value,+1,d,-1,ca_int_getargs(self));
    return [self invalidate];
}


- one
{
    static id one = nil;
    
#ifdef CAKIT_MACH
    static struct mutex iLock = MUTEX_INITIALIZER;

    mutex_lock(&iLock);
    if (one == nil) one = [self intValue:1];
    mutex_unlock(&iLock);
#else
    if (one == nil) one = [self intValue:1];
#endif /* CAKIT_MACH */
    
    return [one cheapCopy];
}

- minusOne
{
    static id minusOne = nil;
    
#ifdef CAKIT_MACH
    static struct mutex iLock = MUTEX_INITIALIZER;

    mutex_lock(&iLock);
    if (minusOne == nil) minusOne = [self intValue:-1];
    mutex_unlock(&iLock);
#else
    if (minusOne == nil) minusOne = [self intValue:-1];
#endif /* CAKIT_MACH */
    
    return [minusOne cheapCopy];
}

- _add:(int)v:B multiply:b:(int)w
{
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _negate:v];
    if ([b isZero]) return [self _negate:v];

    if ([b isOne]) return [self _add:v:B:+w];
    if ([B isOne]) return [self _add:v:b:+w];
    if ([b isMinusOne]) return [self _add:v:B:-w];
    if ([B isMinusOne]) return [self _add:v:b:-w];

    if (1) {
	ca_int_c c;
	ca_int_ooaddmul(&c,ca_int_id2t(self),v,ca_int_id2t(B),ca_int_id2t(b),w,ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }

    return nil;
}

- _addSelf:(int)v:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _add:v:B multiply:b:w]);
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _negateSelf:v];
    if ([b isZero]) return [self _negateSelf:v];

    if ([b isOne]) return [self _addSelf:v:B:+w];
    if ([B isOne]) return [self _addSelf:v:b:+w];
    if ([b isMinusOne]) return [self _addSelf:v:B:-w];
    if ([B isMinusOne]) return [self _addSelf:v:b:-w];

    ca_int_inaddmul(ca_int_id2c(self),v,ca_int_id2t(B),ca_int_id2t(b),w,ca_int_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _negate:+w];
    if ([a isZero]) return [B _negate:+w];
    if ([B isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B:+w];
    if ([a isOne]) return [self _add:+v:B:+w];
    if ([self isMinusOne]) return [a _add:-v:B:+w];
    if ([a isMinusOne]) return [self _add:-v:B:+w];

    if (1) {
	ca_int_c c;
	ca_int_oomuladd(&c,ca_int_id2t(self),ca_int_id2t(a),v,ca_int_id2t(B),w,ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B:w]);
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _negate:+w]);
    if ([a isZero]) return idReplace(self,[B _negate:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B:+w]);
    if ([a isOne]) return [self _addSelf:+v:B:+w];
    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B:+w];

    ca_int_inmuladd(ca_int_id2c(self),ca_int_id2t(a),v,ca_int_id2t(B),w,ca_int_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B multiply:b:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([a isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _multiply:a:+v];
    if ([b isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B multiply:b:+w];
    if ([a isOne]) return [self _add:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiply:a:+v add:B:+w];
    if ([B isOne]) return [self _multiply:a:+v add:b:+w];

    if ([self isMinusOne]) return [a _add:-v:B multiply:b:+w];
    if ([a isMinusOne]) return [self _add:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiply:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiply:a:+v add:b:-w];

    if (1) {
	ca_int_c c;
	ca_int_oomuladdmul(&c,ca_int_id2t(self),ca_int_id2t(a),v,ca_int_id2t(B),ca_int_id2t(b),w,ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B multiply:b:w]);
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([a isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];
    if ([b isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B multiply:b:+w]);
    if ([a isOne]) return [self _addSelf:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiplySelf:a:+v add:B:+w];
    if ([B isOne]) return [self _multiplySelf:a:+v add:b:+w];

    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B multiply:b:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiplySelf:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiplySelf:a:+v add:b:-w];

    ca_int_inmuladdmul(ca_int_id2c(self),ca_int_id2t(a),v,ca_int_id2t(B),ca_int_id2t(b),w,ca_int_getargs(self));
    return [self invalidate];
}


- (BOOL) isOne
{
    assert([self check]);
    return ca_int_isone(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) notOne
{
    assert([self check]);
    return NO == ca_int_isone(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) isMinusOne
{
    assert([self check]);
    return ca_int_isminusone(ca_int_id2t(self),ca_int_getargs(self));
}

- (BOOL) notMinusOne
{
    assert([self check]);
    return NO == ca_int_isminusone(ca_int_id2t(self),ca_int_getargs(self));
}

- square
{
    if ([self inMonoid] && [self isOne]) return [self cheapCopy];
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return [self one];
	if ([self isZero])     return [self cheapCopy];
    }
    if (1) {
	ca_int_c c;assert([self check]);
	ca_int_oosqr(&c,ca_int_id2t(self),ca_int_getargs(self));
	return ca_int_c2id(self,&c);
    }

    return nil;
}

- squareSelf
{
    if (refCount>1) return idReplace(self,[self square]);

    if ([self inMonoid] && [self isOne]) return self;
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return idReplace(self,[self one]);
	if ([self isZero])     return self;
    }
    
    assert([self check]);
    ca_int_insqr(ca_int_id2c(self),ca_int_getargs(self));
    return [self invalidate];
}

- multiply:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self square];
    } else {
	if ([self inMonoid]) {
	    if ([self isOne]) return [b cheapCopy];
	    if ([b isOne]) return [self cheapCopy];
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]||[b isZero]) return [self zero];
	    if ([self isMinusOne]) return [b negate];
	    if ([b isMinusOne]) return [self negate];
	}
	if (1) {
	    ca_int_c c;
	    ca_int_oomul(&c,ca_int_id2t(self),ca_int_id2t(b),ca_int_getargs(self));
	    return ca_int_c2id(self,&c);
	}
    }

    return nil;
}

- multiplySelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self multiply:b]);

    if (self == b) {
	return [self squareSelf];
    } else {
	if ([self inMonoid]) {
	    if ([b isOne]) return self;
	    if ([self isOne]) return idReplace(self,[b cheapCopy]);
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return self;
	    if ([b isZero]) return idReplace(self,[b cheapCopy]);
	    if ([b isMinusOne]) return [self negateSelf];
	    if ([self isMinusOne]) return idReplace(self,[b negate]);
	}
	ca_int_inmul(ca_int_id2c(self),ca_int_id2t(b),ca_int_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- multiplyDigit:(DIGIT)d
{
    ca_int_c c;

    if (d == 0) return [self zero];
    if (d == 1) return [self negate];
    if (d == 2) return [self double];

    if ([self isZero])     return [self zero];
    if ([self isOne])      return [self digitValue:d];
    if ([self isMinusOne]) return [[self digitValue:d] negateSelf];

    ca_int_oomulword(&c,&value,d,ca_int_getargs(self));
    return [self ca_int_reference:&c];
}

- multiplySelfDigit:(DIGIT)d
{
    if (refCount > 1) return idReplace(self,[self multiplyDigit:d]);
    
    if (d == 0) return idReplace(self,[self zero]);
    if (d == 1) return [self negateSelf];
    if (d == 2) return [self doubleSelf];

    if ([self isZero])     return idReplace(self,[self zero]);
    if ([self isOne])      return idReplace(self,[self digitValue:d]);
    if ([self isMinusOne]) return idReplace(self,[[self digitValue:d] negateSelf]);

    ca_int_inmulword(&value,d,ca_int_getargs(self));
    return [self invalidate];
}

- inverse
{
    return ([self isOne]||[self isMinusOne])?[self cheapCopy]:nil;
}


- divide:b
{
    if ([b isZero]) return nil;
    if (self == b) return [self one];
    if ([b isOne]) return [self cheapCopy];
    if ([b isMinusOne]) return [self negate];

    if ([b isDigit]) {
	id q = [self divideDigit:[b digitValue]];
	return ([b sign] > 0)?q:[q negateSelf];
    } else {
	id q,r;r = [self remainder:b quotient:&q];
	if ([r notZero]) {
	    [q free];[r free];return nil;
	} else {
	    [r free];return q;
	}
    }
}
- divideSelf:b
{
    if ([b isZero]) return nil;
    if (self == b) return idReplace(self,[self one]);
    if ([b isOne]) return self;
    if ([b isMinusOne]) return [self negateSelf];

    if ([b isDigit]) {
	self = [self divideSelfDigit:[b digitValue]];
	return ([b sign] > 0)?self:[self negateSelf];
    } else {
	return idReplace(self,[self divide:b]);
    }
}

- quotientSelf:b
{
    if ([b isZero]) return nil;
    if (self == b) return idReplace(self,[self one]);
    if ([b isOne]) return self;
    if ([b isMinusOne]) return [self negateSelf];

    if ([b isDigit]) {
	self = [self quotientSelfDigit:[b digitValue]];
	return ([b sign] > 0)?self:[self negateSelf];
    } else {
	return idReplace(self,[self quotient:b]);
    }
}

- divideDigit:(DIGIT)b
{
    DIGIT r;id q;
    
    if (b ==  0) return nil;
    if (b == +1) return [self cheapCopy];

    q = [self quotientDigit:b remainder:&r];
    if (r) {
	[q free];return nil;
    } else {
	return q;
    }
}

- divideSelfDigit:(DIGIT)b
{
    DIGIT r;

    if (b ==  0) return nil;
    if (b == +1) return self;

    self = [self quotientSelfDigit:b remainder:&r];
    if (r) {
	[self free];return nil;
    } else {
	return self;
    }
}

- quotientDigit:(DIGIT)b
{
    return [self quotientDigit:b remainder:NULL];
}
- quotientSelfDigit:(DIGIT)d
{
    return [self quotientSelfDigit:d remainder:NULL];
}

- remainder:b quotient:(id *)q
{
    if ([b isZero]) return nil;
    if ([b isOne]) { if (q) *q = [self cheapCopy];return [self zero]; }
    if ([b isMinusOne]) { if (q) *q = [self negate];return [self zero]; }

    if ([self isZero]) { if (q) *q = [self zero];return [self zero]; }

    if (q) {
	ca_int_c c,d;
	ca_int_ooremquo(&c,&value,[b ca_int_value],&d,ca_int_getargs(self));
	*q = [self ca_int_reference:&d];return [self ca_int_reference:&c];
    } else {
	ca_int_c c;
	ca_int_oorem(&c,&value,[b ca_int_value],ca_int_getargs(self));
	return [self ca_int_reference:&c];
    }
}

- remainderSelf:b quotient:(id *)q
{
    if (refCount > 1) return idReplace(self,[self remainder:b quotient:q]);

    if ([b isZero]) return nil;
    if ([b isOne]) { if (q) *q = [self cheapCopy];return idReplace(self,[self zero]); }
    if ([b isMinusOne]) { if (q) *q = [self negate];return idReplace(self,[self zero]); }

    if ([self isZero]) { if (q) *q = [self zero];return self; }

    if (q) {
	ca_int_c d;
	ca_int_inremquo(&value,[b ca_int_value],&d,ca_int_getargs(self));
	*q = [self ca_int_reference:&d];return [self invalidate];
    } else {
	ca_int_inrem(&value,[b ca_int_value],ca_int_getargs(self));
	return [self invalidate];
    }
}

- quotientDigit:(DIGIT)d remainder:(DIGIT *)r
{
    ca_int_c q;
   
    if (d == 0) return nil;
    if (d == 1) { if (r) *r = 0;return [self cheapCopy]; }
    if ([self isZero]) { if (r) *r = 0;return [self zero]; }

    if (r) {
	ca_intwrd_t rw;
	ca_int_ooquoremword(&q,&value,d,&rw,ca_int_getargs(self));
	assert(rw <= d);*r = rw;
    } else {
	ca_int_ooquoremword(&q,&value,d,NULL,ca_int_getargs(self));
    }

    return [self ca_int_reference:&q];
}

- quotientSelfDigit:(DIGIT)d remainder:(DIGIT *)r
{
    if (refCount > 1) return idReplace(self,[self quotientDigit:d remainder:r]);

    if (d == 0) return nil;
    if (d == 1) { if (r) *r = 0;return self; }
    if ([self isZero]) { if (r) *r = 0;return self; }

    if (r) {
	ca_intwrd_t rw;
	ca_int_inquoremword(&value,d,&rw,ca_int_getargs(self));
	assert(rw <= d);*r = rw;
    } else {
	ca_int_inquoremword(&value,d,NULL,ca_int_getargs(self));
    }

    return [self invalidate];
}

- (DIGIT) remainderDigit:(DIGIT)d
{
    if (d == 0) [self error:"Division by Zero"];
    if (d == 1) return 0;

    return ca_int_remword(&value,d,ca_int_getargs(self));
}

- (BOOL) isEven
{
    return (EVEN([self lastDigit]));
}

- (BOOL) isOdd
{
    return (ODD([self lastDigit]));
}


- gcd:b
{
    if (self == b) return [self cheapCopy];
    if ([self isZero]) return [b cheapCopy];
    if ([b isZero]) return [self cheapCopy];

    if ([b isDigit]) {
	DIGIT u,v;
	u = [self remainderDigit:v = [b digitValue]];
	while (v) { DIGIT t = u % v;u = v;v = t; }
	return [self digitValue:u];
    } else {
	self = [self cheapCopy];b = [b cheapCopy];
	while ([b notZero]) { id t = [self remainderSelf:b];self = b;b = t; }
	[b free];return self;
    }
}


- asExpression
{
    return [CAExpressionClass newHead:"Integer" object:[self cheapCopy]];
}

- expression:anExpression
{
    if ([anExpression containsHead:"Symbol"]) return nil;
    
    if (strEq([anExpression head],"Integer")) {
	id anInteger = [anExpression object];
	assert([self sameClass:anInteger]);return [anInteger cheapCopy];
    }
    
    return [self stdExpression:anExpression];
}


- (BOOL) printsLeadingSign
{
    return [self sign] < 0;
}

- printToFile:(FILE *)aFile
{
    if ([self isZero]) {
	fprintf(aFile,"0");
    } else {
	char *stack;int n;
	if ([self sign] < 0) fprintf(aFile,"-");
	stack = getChars(self,&n,10);while (n--) fprintf(aFile,"%c",stack[n]);
	OC_Free(stack);
    }
    
    return self;
}


- fileOutOn:aFiler
{
    [super fileOutOn:aFiler];
    ca_int_fileout(aFiler,&value,ca_int_getargs(self));
    return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler];
    ca_int_filein(aFiler,&value,ca_int_getargs(self));
    return self;
}

@end

