#include "cakit.h"

static ca_modp_args ca_modp_getargs(id self)
{
   ca_modp_args args;args.p = [self modulus];return args;
}

static int ca_modp_size(ca_modp_args args)
{
    return 1;				/* pointer increment */
}

static int ca_modp_bsize(ca_modp_args args)
{
    return sizeof(ca_modp_c);		/* bytes */
}

static ca_modp_t ca_modp_id2t(id object)
{
    return [object ca_modp_value];	/* read-only */
}

static ca_modp_t ca_modp_id2ref(id object,ca_modp_c *ref)
{
    return [object ca_modp_value];	/* read-only, scratch space "ref" */
}

static ca_modp_c *ca_modp_id2c(id object)
{
    return [object ca_modp_reference];	/* write for inplace arithmetic */
}

static id ca_modp_t2id(id proto,ca_modp_t f)
{
    return [proto ca_modp_value:f];	/* new object containing copy of "f" */
}

static id ca_modp_c2id(id proto,ca_modp_c* c)
{
    return [proto ca_modp_reference:c];/* new object containing "c" */
}

static ca_modp_t ca_modp_fromintvalue(int i,ca_modp_args args)
{
    ca_modp_t red;
    
    if (i==0) return 0;
    if (i <0) {
	red = -i;
	red %= args.p;	/* it's dangerous to % int & unsigned */
	red = (red)?args.p-red:0;
    } else {
	red  = i;
	red %= args.p;	/* dito */
    }
    
    return red;
}

static int ca_modp_check(ca_modp_t self,ca_modp_args args)
{
    assert(self < args.p);return 1;
}

static ca_modp_t ca_modp_zero(ca_modp_args args)
{
    return 0;
}

static ca_modp_t ca_modp_one(ca_modp_args args)
{
    return 1;
}

static unsigned ca_modp_hash(ca_modp_t a,ca_modp_args args)
{
    return a;
}

static int ca_modp_iseq(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a == b);
}

static int ca_modp_isop(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a==0)?(b==0):(a+b==args.p);
}

static int ca_modp_iszero(ca_modp_t a,ca_modp_args args)
{
    return a==0;
}

static int ca_modp_isone(ca_modp_t a,ca_modp_args args)
{
    return a==1;
}

static int ca_modp_isminusone(ca_modp_t a,ca_modp_args args)
{
    return (a+1)==args.p;
}

static ca_modp_t ca_modp_neg(ca_modp_t a,ca_modp_args args)
{
    assert(ca_modp_check(a,args));return (a)?args.p-a:0;
}

static ca_modp_t ca_modp_dbl(ca_modp_t a,ca_modp_args args)
{
    ca_modp_t c = a+a;
    assert(args.p != 2 && ca_modp_check(a,args));
    if (c) return (c<args.p)?c:c-args.p; else return 0;
}

static ca_modp_t ca_modp_add(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    ca_modp_t c = a+b;
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return ((c>=args.p)?c-args.p:c); /* compiles better than c<p?c:c-p */
}

static ca_modp_t ca_modp_sub(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a>=b)?(a-b):(args.p-(b-a));
}

static ca_modp_t ca_modp_mul(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    unsigned c = a * b;		/* limitation to 16bits to be able to do this
				 * russian peasant 50% slower
                                 */
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_sqr(ca_modp_t a,ca_modp_args args)
{
    unsigned c = a * a;
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_inv(ca_modp_t x,ca_modp_args args)
{
    unsigned q,s;
    int u1,u2,us;  /* intermediate values can be negative */
    ca_modp_t a,b,inv;
    
    assert(x != 0 && x != 1 && x+1 != args.p);
    
    a = x;b = args.p;u1 = 1;u2 = 0;
    while (b) {
	s  = a%b;q = a/b;
	a  = b;b = s;
	u1 -= q*u2;			
	us  = u1;u1 = u2;u2 = us;
    }
    
    inv = ((u1<0)?u1+args.p:u1); /* keep result positive */
    
    assert(ca_modp_check(inv,args));
    assert(ca_modp_isone(ca_modp_mul(x,inv,args),args));
    return inv;
}

static ca_modp_t ca_modp_div(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(b!=0);
    return ca_modp_mul(a,ca_modp_inv(b,args),args);
}

static ca_modp_t ca_modp_pow(ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_t o;
    
    if (n==0) return ca_modp_one(args);
    if (n==1) return a;
    if (n==2) return ca_modp_sqr(a,args);
    
    o = ca_modp_one(args);
    while (n>1) {
	if (n&1) o = ca_modp_mul(o,a,args);
	a = ca_modp_sqr(a,args);
	n >>= 1;
    }
    
    return ca_modp_mul(o,a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_copy_val.fab should not be included"
#endif

static void ca_modp_clear(ca_modp_c *c,ca_modp_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_modp_t)0xcafebabe;
    #endif
}

static void ca_modp_copy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_deepcopy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_move(ca_modp_c *c,ca_modp_c *a,ca_modp_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_modp_t)0xcafebabe;
    #endif
}

#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static void ca_modp_oozero(ca_modp_c *c,ca_modp_args args)
{
    *c = ca_modp_zero(args);
}

static void ca_modp_inzero(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_zero(args);
}

static void ca_modp_inneg(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_neg(*a,args);
}

static void ca_modp_ooneg(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_neg(a,args);
}

static ca_modp_t ca_modp_dblv(ca_modp_t a,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_dbl(a,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_indbl(ca_modp_c *a,int v,ca_modp_args args)
{
    *a = ca_modp_dblv(*a,v,args);
}

static void ca_modp_oodbl(ca_modp_c *c,ca_modp_t a,int v,ca_modp_args args)
{
    *c = ca_modp_dblv(a,v,args);
}

static ca_modp_t ca_modp_addvw(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    ca_modp_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_modp_add(a,b,args);
	} else {
	    c = ca_modp_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_modp_sub(b,a,args);
	} else {
	    c = ca_modp_neg(ca_modp_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_modp_inadd(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addvw(*a,v,b,w,args);
}

static void ca_modp_ooadd(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addvw(a,v,b,w,args);
}
#if !ca_modp_isvalue
#error "ca_modp_mul_val.fab should not be included"
#endif

static void ca_modp_insqr(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_sqr(*a,args);
}

static void ca_modp_oosqr(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_sqr(a,args);
}

static void ca_modp_inmul(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_mul(*a,b,args);
}

static void ca_modp_oomul(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_mul(a,b,args);
}
#if !ca_modp_isvalue
#error "ca_modp_inv_val.fab should not be included"
#endif

static void ca_modp_ininv(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_inv(*a,args);
}

static void ca_modp_ooinv(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_inv(a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_div_val.fab should not be included"
#endif

static void ca_modp_indiv(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_div(*a,b,args);
}

static BOOL ca_modp_oodiv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_div(a,b,args);return SUCCESS;
}
#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static ca_modp_t ca_modp_mulv(ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_mul(a,b,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_inmulv(ca_modp_c *a,ca_modp_t b,int v,ca_modp_args args)
{
    *a = ca_modp_mulv(*a,b,v,args);
}

static void ca_modp_oomulv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    *c = ca_modp_mulv(a,b,v,args);
}

static ca_modp_t ca_modp_addmul(ca_modp_t a,int v,ca_modp_t b,ca_modp_t s,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_mul(b,s,args),w,args);
}

static ca_modp_t ca_modp_addsqr(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_sqr(b,args),w,args);
}

static ca_modp_t ca_modp_muladd(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,b,w,args);
}

static ca_modp_t ca_modp_muladdmul(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,ca_modp_mul(b,t,args),w,args);
}

static void ca_modp_inaddmul(ca_modp_c *a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_addmul(*a,v,b,t,w,args);
}

static void ca_modp_ooaddmul(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_addmul(a,v,b,t,w,args);
}

static void ca_modp_inaddsqr(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addsqr(*a,v,b,w,args);
}

static void ca_modp_ooaddsqr(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addsqr(a,v,b,w,args);
}

static void ca_modp_inmuladd(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_muladd(*a,s,v,b,w,args);
}

static void ca_modp_oomuladd(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_muladd(a,s,v,b,w,args);
}

static void ca_modp_inmuladdmul(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_modp_oomuladdmul(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_muladdmul(a,s,v,b,t,w,args);
}
static void ca_modp_args_fileOut(id aFiler,ca_modp_args *args) {}
static void ca_modp_args_fileIn(id aFiler,ca_modp_args *args) {}

static void ca_modp_fileOut(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileOut:c type:'S'];
}

static void ca_modp_fileIn(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileIn:c type:'S'];
}


/*
 * Computer Algebra Kit (c) 1993,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_vardnsmonom_size(ca_vardnsmonom_args args)
{
    return args.nvars; /* size to add to int pointer */
}

static int ca_vardnsmonom_bsize(ca_vardnsmonom_args args)
{
    return args.nvars * sizeof(int); /* byte size exponent array */
}

static int ca_vardnsmonom_check(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) assert(*self++ >= 0);
    return 1;
}

static ca_vardnsmonom_t ca_vardnsmonom_id2t(id object)
{
    return [object ca_vardnsmonom_value];
}

static id ca_vardnsmonom_t2id(id proto,ca_vardnsmonom_t f)
{
    return [proto ca_vardnsmonom_value:f];
}

static ca_vardnsmonom_args ca_vardnsmonom_getargs(id self)
{
    ca_vardnsmonom_args res;
    res.nvars = [[self symbols] numMembers];
    return res; 
}

static ca_vardnsmonom_t ca_vardnsmonom_alloc(ca_vardnsmonom_args args)
{
    return (ca_vardnsmonom_t)OC_Malloc(ca_vardnsmonom_bsize(args));
}

static void ca_vardnsmonom_init(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = 0;
}

static void ca_vardnsmonom_copy(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = *a++;
}

static void ca_vardnsmonom_clear(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    /* nothing to clear... */
}

static void ca_vardnsmonom_free(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    OC_Free(c);
}

static void ca_vardnsmonom_args_fileOut(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileOut:&args->nvars type:'i'];
}

static void ca_vardnsmonom_args_fileIn(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileIn:&args->nvars type:'i'];
}

static void ca_vardnsmonom_fileOut(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileOut:c++ type:'i'];
}

static void ca_vardnsmonom_fileIn(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileIn:c++ type:'i'];
}

static unsigned ca_vardnsmonom_hash(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    assert(args.nvars>0);return self[0];
}

static int ca_vardnsmonom_deg(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int sum = 0;
    int nvars = args.nvars;
    while (nvars--) sum += *self++;
    return sum;
}

static int ca_vardnsmonom_numterms(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int numterms = 0;
    int nvars = args.nvars;
    while (nvars--) if (*self++) numterms++;
    return numterms;
}

static int ca_vardnsmonom_isone(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if (*self) return 0; self++; }
    return 1;
}

static int ca_vardnsmonom_iseq(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*b)) return 0; a++;b++; }
    return 1;
}

static int ca_vardnsmonom_iseqmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*x) + (*b)) return 0; a++;x++;b++; }
    return 1;
}

static int ca_vardnsmonom_cmp(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	if ((*a) < (*b)) return -1;
	if ((*a) > (*b)) return +1;
	a++;b++;
    }
    
    return 0;
}

static int ca_vardnsmonom_cmpmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	int sum = *x + *b;
	if ((*a) < sum) return -1;
	if ((*a) > sum) return +1;
	a++;x++;b++;
    }
    
    return 0;
}

static void ca_vardnsmonom_oomul(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) = (*a) + (*b); c++; a++; b++; }
}

static void ca_vardnsmonom_inmul(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) += (*b); c++; b++; }
}

static BOOL ca_vardnsmonom_oodiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *a - *b;
	if (q<0) return FAILURE;
	*c = q; c++; a++; b++;
    }
    
    return SUCCESS;
}

static void ca_vardnsmonom_indiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *c - *b;
	if (q<0) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
	*c = q; c++; b++;
    }
}
/*
 * scalar is accessed by the C "->" operator since having variable sized
 * scalars _and_ monomials gets difficult on machines like the DEC alpha
 * (where we'd have to manually pad to get pointer aligned structures).
 * 
 * the monomial (exponent) is however not accessed by the C "->" operator
 * but rather by functions that can take offsets as arguments
 * this is important the monomials in variable dense representation
 * (fixarray of exp)
 *
 * stes - mar31, 1997 - CHANGES: for DEC alpha coefficient needs to be aligned
 * properly, put it before the monomial and drop the idea of variable sized
 * coefficients.
 */

static ca_vardnsmonom_c *ca_modp_vardnsexpdegspsterm_monom_cval(ca_modp_vardnsexpdegspsterm_t self)
{
    return &self->data;
}

static ca_vardnsmonom_t ca_modp_vardnsexpdegspsterm_monom(ca_modp_vardnsexpdegspsterm_t self)
{
#if ca_vardnsmonom_pervalue
    return self->data;
#else
    return  ca_modp_vardnsexpdegspsterm_monom_cval(self);
#endif
}

/* stes - mar31,1997
 * change for DEC alpha to exclude varsized scalars
 */

static ca_modp_c *ca_modp_vardnsexpdegspsterm_scal_cval(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
	return &self->scal;
}

static ca_modp_t ca_modp_vardnsexpdegspsterm_scal(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
#if ca_modp_pervalue
	return self->scal;
#else
    return  ca_modp_vardnsexpdegspsterm_scal_cval(self,args);
#endif
}

static int ca_modp_vardnsexpdegspsterm_check(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_check(ca_modp_vardnsexpdegspsterm_scal(self,args),args.sargs));
    assert(!ca_modp_iszero(ca_modp_vardnsexpdegspsterm_scal(self,args),args.sargs));
    assert(ca_vardnsmonom_check(ca_modp_vardnsexpdegspsterm_monom(self),args.margs));
    return 1;
}

static int ca_modp_vardnsexpdegspsterm_bsize(ca_modp_vardnsexpdegspsterm_args args)
{
	/* this is sizeof(ca_vardnsmonom_c) too much but why bother ? */
    return sizeof(struct ca_modp_vardnsexpdegspsterm) + ca_vardnsmonom_bsize(args.margs);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_alloc(ca_modp_vardnsexpdegspsterm_args args)
{
    return (ca_modp_vardnsexpdegspsterm_t)OC_Malloc(ca_modp_vardnsexpdegspsterm_bsize(args));
}

static void ca_modp_vardnsexpdegspsterm_clear(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_vardnsmonom_clear(ca_modp_vardnsexpdegspsterm_monom_cval(self),args.margs);
    ca_modp_clear(ca_modp_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
}

static void ca_modp_vardnsexpdegspsterm_free(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    /* invalidating links not always helps : buggy code might work because
     * links were cleared...        self->nlink = NULL;self->plink = NULL;
     */
    OC_Free(self);
}

static int ca_modp_vardnsexpdegspsterm_cmp(ca_modp_vardnsexpdegspsterm_t a,ca_modp_vardnsexpdegspsterm_t b,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(a,args) && ca_modp_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_cmp(ca_modp_vardnsexpdegspsterm_monom(a),ca_modp_vardnsexpdegspsterm_monom(b),args.margs);
}

static int ca_modp_vardnsexpdegspsterm_cmpmonom(ca_modp_vardnsexpdegspsterm_t a,ca_vardnsmonom_t b,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(a,args) && ca_vardnsmonom_check(b,args.margs));
    return ca_vardnsmonom_cmp(ca_modp_vardnsexpdegspsterm_monom(a),b,args.margs);
}

static int ca_modp_vardnsexpdegspsterm_cmpmul(ca_modp_vardnsexpdegspsterm_t a,ca_modp_vardnsexpdegspsterm_t b,ca_vardnsmonom_t m ,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(a,args) && ca_modp_vardnsexpdegspsterm_check(b,args));
    assert(ca_vardnsmonom_check(m,args.margs));
    return ca_vardnsmonom_cmpmul(ca_modp_vardnsexpdegspsterm_monom(a),ca_modp_vardnsexpdegspsterm_monom(b),m,args.margs);
}

static int ca_modp_vardnsexpdegspsterm_deg(ca_modp_vardnsexpdegspsterm_t a,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_vardnsmonom_deg(ca_modp_vardnsexpdegspsterm_monom(a),args.margs);
}

static int ca_modp_vardnsexpdegspsterm_iszero(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_iszero(ca_modp_vardnsexpdegspsterm_scal(self,args),args.sargs);
}

static int ca_modp_vardnsexpdegspsterm_isone(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_isone(ca_modp_vardnsexpdegspsterm_scal(self,args),args.sargs) && ca_vardnsmonom_isone(ca_modp_vardnsexpdegspsterm_monom(self),args.margs);
}

static int ca_modp_vardnsexpdegspsterm_isminusone(ca_modp_vardnsexpdegspsterm_t self,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_isminusone(ca_modp_vardnsexpdegspsterm_scal(self,args),args.sargs) && ca_vardnsmonom_isone(ca_modp_vardnsexpdegspsterm_monom(self),args.margs);
}

static int ca_modp_vardnsexpdegspsterm_iseq(ca_modp_vardnsexpdegspsterm_t a,ca_modp_vardnsexpdegspsterm_t b,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(a,args) && ca_modp_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_iseq(ca_modp_vardnsexpdegspsterm_monom(a),ca_modp_vardnsexpdegspsterm_monom(b),args.margs) && ca_modp_iseq(ca_modp_vardnsexpdegspsterm_scal(a,args),ca_modp_vardnsexpdegspsterm_scal(b,args),args.sargs);
}

static int ca_modp_vardnsexpdegspsterm_isop(ca_modp_vardnsexpdegspsterm_t a,ca_modp_vardnsexpdegspsterm_t b,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(a,args) && ca_modp_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_iseq(ca_modp_vardnsexpdegspsterm_monom(a),ca_modp_vardnsexpdegspsterm_monom(b),args.margs) && ca_modp_isop(ca_modp_vardnsexpdegspsterm_scal(a,args),ca_modp_vardnsexpdegspsterm_scal(b,args),args.sargs);
}

static void ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A)
{
    C->nlink = A;A->plink = C;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A)
{
    ca_modp_vardnsexpdegspsterm_glue(C,A);return A;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_delete(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t C = A->nlink;
    ca_modp_vardnsexpdegspsterm_glue(A->plink,C);
    ca_modp_vardnsexpdegspsterm_clear(A,args);
    ca_modp_vardnsexpdegspsterm_free(A,args);
    return C;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_condcns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t new,ca_modp_vardnsexpdegspsterm_args args)
{
    if (ca_modp_vardnsexpdegspsterm_iszero(new,args)) {
	ca_modp_vardnsexpdegspsterm_clear(new,args);ca_modp_vardnsexpdegspsterm_free(new,args);return C;
    } else {
	assert(ca_modp_vardnsexpdegspsterm_check(new,args));return ca_modp_vardnsexpdegspsterm_cns(C,new);
    }
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_conddelete(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    if (ca_modp_vardnsexpdegspsterm_iszero(A,args)) {
	return ca_modp_vardnsexpdegspsterm_delete(A,args);
    } else {
	return A->nlink;
    }
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_copymonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    
    ca_modp_copy(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_modp_vardnsexpdegspsterm_check(new,args));return ca_modp_vardnsexpdegspsterm_cns(C,new);
}

static void ca_modp_vardnsexpdegspsterm_copymonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_copymonomsc_cns(C,m,s,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_deepcopymonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    
    ca_modp_deepcopy(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_modp_vardnsexpdegspsterm_check(new,args));return ca_modp_vardnsexpdegspsterm_cns(C,new);
}

static void ca_modp_vardnsexpdegspsterm_deepcopymonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_deepcopymonomsc_cns(C,m,s,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_copy_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_vardnsexpdegspsterm_copymonomsc_cns(C,ca_modp_vardnsexpdegspsterm_monom(A),ca_modp_vardnsexpdegspsterm_scal(A,args),args);
}

static void ca_modp_vardnsexpdegspsterm_copy(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_copy_cns(C,A,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_deepcopy_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_vardnsexpdegspsterm_deepcopymonomsc_cns(C,ca_modp_vardnsexpdegspsterm_monom(A),ca_modp_vardnsexpdegspsterm_scal(A,args),args);
}

static void ca_modp_vardnsexpdegspsterm_deepcopy(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_deepcopy_cns(C,A,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooneg_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO);

    ca_modp_ooneg(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_cns(C,new);
}

static void ca_modp_vardnsexpdegspsterm_ooneg(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_ooneg_cns(C,A,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inneg(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO);
    ca_modp_inneg(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),args.sargs);
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO);
    return A->nlink;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oonegv_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    return (v == -1)?ca_modp_vardnsexpdegspsterm_ooneg_cns(C,A,args):ca_modp_vardnsexpdegspsterm_copy_cns(C,A,args);
}

static void ca_modp_vardnsexpdegspsterm_oonegv(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oonegv_cns(C,A,v,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_innegv(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    return (v == -1)?ca_modp_vardnsexpdegspsterm_inneg(A,args):A->nlink;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oodbl_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO);

    ca_modp_oodbl(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),v,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsexpdegspsterm_oodbl(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oodbl_cns(C,A,v,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_indbl(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO);
    ca_modp_indbl(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),v,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooaddsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    ca_modp_ooadd(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),v,B,w,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inaddsc(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inadd(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),v,B,w,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooadd_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == ca_modp_vardnsexpdegspsterm_cmp(A,B,args));
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO && ca_modp_vardnsexpdegspsterm_iszero(B,args) == NO);
    return ca_modp_vardnsexpdegspsterm_ooaddsc_cns(C,A,v,ca_modp_vardnsexpdegspsterm_scal(B,args),w,args);
}

static void ca_modp_vardnsexpdegspsterm_ooadd(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_ooadd_cns(C,A,v,B,w,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inadd(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == ca_modp_vardnsexpdegspsterm_cmp(A,B,args));
    assert(ca_modp_vardnsexpdegspsterm_iszero(A,args) == NO && ca_modp_vardnsexpdegspsterm_iszero(B,args) == NO);
    return ca_modp_vardnsexpdegspsterm_inaddsc(A,v,ca_modp_vardnsexpdegspsterm_scal(B,args),w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    ca_modp_oomulv(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,v,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsexpdegspsterm_oomulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A, ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulsc(ca_modp_vardnsexpdegspsterm_t A, ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inmulv(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,v,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    ca_modp_oomulv(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,v,args.sargs);
    ca_vardnsmonom_oomul(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),m,args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsexpdegspsterm_oomulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inmulv(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,v,args.sargs);
    ca_vardnsmonom_inmul(ca_modp_vardnsexpdegspsterm_monom_cval(A),m,args.margs);

    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooaddscmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    ca_modp_ooaddmul(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),v,B,t,w,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inaddscmulsc(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inaddmul(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),v,B,t,w,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooaddmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_ooaddscmulsc_cns(C,A,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_ooaddmulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
}

static void ca_modp_vardnsexpdegspsterm_ooaddmulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspsterm_ooaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inaddmulsc(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_inaddscmulsc(A,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsexpdegspsterm_inaddscmulsc(A,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulscaddsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    ca_modp_oomuladd(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,v,B,w,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulscaddsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inmuladd(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,v,B,w,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulscadd_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_oomulscaddsc_cns(C,A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),w,args);
}

static void ca_modp_vardnsexpdegspsterm_oomulscadd(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulscadd(ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_inmulscaddsc(A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulscaddscmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);
    ca_modp_oomuladdmul(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,v,B,t,w,args.sargs);
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulscaddscmulsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_inmuladdmul(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,v,B,t,w,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulscaddmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsexpdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static void ca_modp_vardnsexpdegspsterm_oomulscaddmulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspsterm_oomulscaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulscaddmulsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsexpdegspsterm_inmulscaddscmulsc(A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_inmulscaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsexpdegspsterm_inmulscaddscmulsc(A,s,v,ca_modp_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oodivsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    BOOL failed;
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    failed = ca_modp_oodiv(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,args.sargs);
    if (failed) {
	ca_modp_vardnsexpdegspsterm_free(new,args);return NULL;
    }
    
    ca_vardnsmonom_copy(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_modp_vardnsexpdegspsterm_oodivsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A, ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    if ((C = ca_modp_vardnsexpdegspsterm_oodivsc_cns(C,A,s,args))) {
	ca_modp_vardnsexpdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_indivsc(ca_modp_vardnsexpdegspsterm_t A, ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_indiv(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,args.sargs);
    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_oodivmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    BOOL sfailed,mfailed;
    ca_modp_vardnsexpdegspsterm_t new = ca_modp_vardnsexpdegspsterm_alloc(args);

    sfailed = ca_modp_oodiv(ca_modp_vardnsexpdegspsterm_scal_cval(new,args),ca_modp_vardnsexpdegspsterm_scal(A,args),s,args.sargs);
    mfailed = ca_vardnsmonom_oodiv(ca_modp_vardnsexpdegspsterm_monom_cval(new),ca_modp_vardnsexpdegspsterm_monom(A),m,args.margs);
    
    if (sfailed || mfailed) {
	/* both failures are possible in the multivariate case */
	ca_modp_vardnsexpdegspsterm_free(new,args);return NULL;
    }
    
    return ca_modp_vardnsexpdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_modp_vardnsexpdegspsterm_oodivmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    if ((C = ca_modp_vardnsexpdegspsterm_oodivmonomsc_cns(C,A,m,s,args))) {
	ca_modp_vardnsexpdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_indivmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_indiv(ca_modp_vardnsexpdegspsterm_scal_cval(A,args),s,args.sargs);
    ca_vardnsmonom_indiv(ca_modp_vardnsexpdegspsterm_monom_cval(A),m,args.margs);

    return ca_modp_vardnsexpdegspsterm_conddelete(A,args);
}

static void ca_modp_vardnsexpdegspsterm_args_fileOut(id aFiler,ca_modp_vardnsexpdegspsterm_args *args)
{
    ca_modp_args_fileOut(aFiler,&args->sargs);
    ca_vardnsmonom_args_fileOut(aFiler,&args->margs);
}
static void ca_modp_vardnsexpdegspsterm_args_fileIn(id aFiler,ca_modp_vardnsexpdegspsterm_args *args)
{
    ca_modp_args_fileIn(aFiler,&args->sargs);
    ca_vardnsmonom_args_fileIn(aFiler,&args->margs);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_fileOut(id aFiler,ca_modp_vardnsexpdegspsterm_t self ,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_vardnsexpdegspsterm_check(self,args));
    
    ca_modp_fileOut(aFiler,ca_modp_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
    ca_vardnsmonom_fileOut(aFiler,ca_modp_vardnsexpdegspsterm_monom_cval(self),args.margs);
    
    return self->nlink;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspsterm_fileIn_cns(id aFiler,ca_modp_vardnsexpdegspsterm_t C ,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t self = ca_modp_vardnsexpdegspsterm_alloc(args);
    
    ca_modp_fileIn(aFiler,ca_modp_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
    ca_vardnsmonom_fileIn(aFiler,ca_modp_vardnsexpdegspsterm_monom_cval(self),args.margs);
    
    return ca_modp_vardnsexpdegspsterm_cns(C,self);
}

static int ca_modp_vardnsexpdegspslist_check(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t prev;
    
    for(prev = NULL;A!=AT;prev = A,A = A->nlink) {
	if (ca_modp_vardnsexpdegspsterm_check(A,args) == 0) return 0;
	if (prev && ca_modp_vardnsexpdegspsterm_cmp(prev,A,args) <= 0) return 0;
    }
    
    return 1;
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_copy_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) { C = ca_modp_vardnsexpdegspsterm_copy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_modp_vardnsexpdegspslist_copy(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_copy_cns(C,A,AT,args),CT);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_deepcopy_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) { C = ca_modp_vardnsexpdegspsterm_deepcopy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_modp_vardnsexpdegspslist_deepcopy(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_deepcopy_cns(C,A,AT,args),CT);
}

static void ca_modp_vardnsexpdegspslist_clear(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) { ca_modp_vardnsexpdegspsterm_clear(A,args);A = A->nlink; }
}

static void ca_modp_vardnsexpdegspslist_free(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_t t;for(;A!=AT;A=t) { t = A->nlink;ca_modp_vardnsexpdegspsterm_free(A,args); }
}

static int ca_modp_vardnsexpdegspslist_count(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static unsigned ca_modp_vardnsexpdegspslist_hash(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    return ca_modp_vardnsexpdegspslist_count(A,AT);
}

static int ca_modp_vardnsexpdegspslist_iseq(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_t B, ca_modp_vardnsexpdegspsterm_t BT,ca_modp_vardnsexpdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_modp_vardnsexpdegspsterm_iseq(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_modp_vardnsexpdegspslist_isop(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_vardnsexpdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_modp_vardnsexpdegspsterm_isop(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_modp_vardnsexpdegspslist_deg(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    return (A==AT)?-1:ca_modp_vardnsexpdegspsterm_deg(A,args);
}

static int ca_modp_vardnsexpdegspslist_maxdeg(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    int max = (A==AT)?-1:ca_modp_vardnsexpdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_modp_vardnsexpdegspsterm_deg(A,args);if (max < d) max = d;
    }
    return max;
}

static int ca_modp_vardnsexpdegspslist_ord(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    return (A==AT)?-1:ca_modp_vardnsexpdegspsterm_deg(AT->plink,args);
}

static int ca_modp_vardnsexpdegspslist_mindeg(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    int min = (A==AT)?-1:ca_modp_vardnsexpdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_modp_vardnsexpdegspsterm_deg(A,args);if (min > d) min = d;
    }
    return min;
}

static void ca_modp_vardnsexpdegspslist_insert(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(ca_modp_iszero(s,args.sargs) == NO);
    
    /* adding to zero */
    if (A == AT) {
	ca_modp_vardnsexpdegspsterm_copymonomsc(A,m,s,A,args);return;
    }
    
    /* tail append */
    if (ca_modp_vardnsexpdegspsterm_cmpmonom(AT->plink,m,args) > 0) {
	ca_modp_vardnsexpdegspsterm_copymonomsc(AT->plink,m,s,AT,args);return;
    }
    
    /* insertion */
    while (A != AT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmpmonom(A,m,args);
	if (cmp == 0) {
	    ca_modp_vardnsexpdegspsterm_inaddsc(A,+1,s,+1,args);return;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_copymonomsc(A->plink,m,s,A,args);return;
	}
	A = A->nlink;
    }
 
    fprintf(stderr,"Shouldn't happen\n");
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_ooneg_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsexpdegspsterm_ooneg_cns(C,A,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsexpdegspslist_ooneg(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_ooneg_cns(C,A,AT,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inneg(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_inneg(A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oonegv_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    if (v == -1) {
	return ca_modp_vardnsexpdegspslist_ooneg_cns(C,A,AT,args);
    } else {
	return ca_modp_vardnsexpdegspslist_copy_cns(C,A,AT,args);
    }
}

static void ca_modp_vardnsexpdegspslist_oonegv(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args),CT);
}

static void ca_modp_vardnsexpdegspslist_innegv(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    if (v == -1) ca_modp_vardnsexpdegspslist_inneg(A,AT,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oodbl_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsexpdegspsterm_oodbl_cns(C,A,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsexpdegspslist_oodbl(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oodbl_cns(C,A,AT,v,args),CT);
}

static void ca_modp_vardnsexpdegspslist_indbl(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_indbl(A,v,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_ooadd_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_ooadd_cns(C,A,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_ooadd(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_ooadd_cns(C,A,AT,v,B,BT,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inadd(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inadd(A,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oomulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsexpdegspslist_oomulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inmulsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_inmulsc(A,s,v,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oomulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsexpdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsexpdegspslist_oomulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m, ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oomulmonomsc_cns(C,A,AT,m,s,v,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_inmulmonomsc(A,m,s,v,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_ooaddmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_ooaddmulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_ooaddmulsc_cns(C,A,AT,v,B,BT,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inaddmulsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inaddmulsc(A,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_ooaddmulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_ooaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_ooaddmulmonomsc_cns(C,A,AT,v,B,BT,m,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inaddmulmonomsc(A,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oomulscadd_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_oomulscadd(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT, int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oomulscadd_cns(C,A,AT,s,v,B,BT,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inmulscadd(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inmulscadd(A,s,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oomulscaddmulsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_oomulscaddmulsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oomulscaddmulsc_cns(C,A,AT,s,v,B,BT,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inmulscaddmulsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inmulscaddmulsc(A,s,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_oomulscaddmulmonomsc_cns(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsexpdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsexpdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_oomulscaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m, ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_oomulscaddmulmonomsc_cns(C,A,AT,s,v,B,BT,m,t,w,args),CT);
}

static void ca_modp_vardnsexpdegspslist_inmulscaddmulmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsexpdegspsterm_t B,ca_modp_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsexpdegspsterm_inmulscaddmulmonomsc(A,s,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsexpdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsexpdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static BOOL ca_modp_vardnsexpdegspslist_oodivsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_modp_vardnsexpdegspsterm_t nextC = ca_modp_vardnsexpdegspsterm_oodivsc_cns(C,A,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_modp_vardnsexpdegspsterm_glue(C,CT);return failed;
}

static void ca_modp_vardnsexpdegspslist_indivsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_indivsc(A,s,args);
}

static BOOL ca_modp_vardnsexpdegspslist_oodivmonomsc(ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m, ca_modp_t s,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_modp_vardnsexpdegspsterm_t nextC = ca_modp_vardnsexpdegspsterm_oodivmonomsc_cns(C,A,m,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_modp_vardnsexpdegspsterm_glue(C,CT);return failed;
}

static void ca_modp_vardnsexpdegspslist_indivmonomsc(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_indivmonomsc(A,m,s,args);
}

static void ca_modp_vardnsexpdegspslist_fileOut(id aFiler,ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT,ca_modp_vardnsexpdegspsterm_args args)
{
    int n;
    n = ca_modp_vardnsexpdegspslist_count(A,AT);
    [aFiler fileOut:&n type:'i'];
    while (A != AT) A = ca_modp_vardnsexpdegspsterm_fileOut(aFiler,A,args);
}

static ca_modp_vardnsexpdegspsterm_t ca_modp_vardnsexpdegspslist_fileIn_cns(id aFiler,ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    while (n--) C = ca_modp_vardnsexpdegspsterm_fileIn_cns(aFiler,C,args);
    return C;
}

static void ca_modp_vardnsexpdegspslist_fileIn(id aFiler,ca_modp_vardnsexpdegspsterm_t C,ca_modp_vardnsexpdegspsterm_t CT,ca_modp_vardnsexpdegspsterm_args args)
{
    ca_modp_vardnsexpdegspsterm_glue(ca_modp_vardnsexpdegspslist_fileIn_cns(aFiler,C,args),CT);
}

static void ca_modp_vardnsexpdegspspol_ix_init(ca_modp_vardnsexpdegspspol_ixt ix,id self)
{
    ca_modp_vardnsexpdegspspol_t v = [self ca_modp_vardnsexpdegspspol_value];
    ix->head = v;ix->ix = v;
}

static void ca_modp_vardnsexpdegspspol_ix_clear(ca_modp_vardnsexpdegspspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_modp_vardnsexpdegspspol_ix_count_aux_aux(ca_modp_vardnsexpdegspsterm_t A,ca_modp_vardnsexpdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static int ca_modp_vardnsexpdegspspol_ix_count_aux(ca_modp_vardnsexpdegspspol_t self)
{
    return ca_modp_vardnsexpdegspspol_ix_count_aux_aux(self->nlink,self);
}

static int ca_modp_vardnsexpdegspspol_ix_count(ca_modp_vardnsexpdegspspol_ixt ix)
{
    /* merde ! can't include entire genlist for ca_modp_vardnsexpdegspslist_count */
    return ca_modp_vardnsexpdegspspol_ix_count_aux(ix->head); 
}

static int ca_modp_vardnsexpdegspspol_ix_isempty_aux(ca_modp_vardnsexpdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static int ca_modp_vardnsexpdegspspol_ix_isempty(ca_modp_vardnsexpdegspspol_ixt ix)
{
    /* can't include entire genpol for ca_modp_vardnsexpdegspspol_isempty */
    return ca_modp_vardnsexpdegspspol_ix_isempty_aux(ix->head); 
}

static int ca_modp_vardnsexpdegspspol_ix_ismatch(ca_modp_vardnsexpdegspspol_ixt ix)
{
    return (ix->ix != NULL && ix->ix != ix->head);
}

static void ca_modp_vardnsexpdegspspol_ix_tofirst(ca_modp_vardnsexpdegspspol_ixt ix)
{
    ix->ix = ix->head;
}

static void ca_modp_vardnsexpdegspspol_ix_tolast(ca_modp_vardnsexpdegspspol_ixt ix)
{
    ix->ix = NULL;
}

static void ca_modp_vardnsexpdegspspol_ix_to(ca_modp_vardnsexpdegspspol_ixt ix,int i)
{
    fprintf(stderr,"memberAt: notImplemented\n");
}

static void ca_modp_vardnsexpdegspspol_ix_next(ca_modp_vardnsexpdegspspol_ixt ix)
{
    if (ix->ix == NULL) {
	/* remain on position just after last term */
    } else {
	ix->ix = ix->ix->nlink; /* next term or head = position after last */
	if (ix->ix == ix->head) ix->ix = NULL;
    }
}

static void ca_modp_vardnsexpdegspspol_ix_prev(ca_modp_vardnsexpdegspspol_ixt ix)
{
    if (ix->ix == ix->head) {
	/* remain on position just before first term */
    } else {
	if (ix->ix == NULL) {
	    ix->ix = ix->head->plink; /* last term or head = position before first */
	} else {
	    ix->ix = ix->ix->plink; /* previous term or head = position before first */
	}
    }
}
static int ca_modp_vardnsexpdegspspol_check(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    assert (self->data == 0xcafebabe);
    return ca_modp_vardnsexpdegspslist_check(self->nlink,self,args);
}

static ca_modp_vardnsexpdegspspol_t ca_modp_vardnsexpdegspspol_id2t(id object)
{
    return [object ca_modp_vardnsexpdegspspol_value];
}

static ca_modp_vardnsexpdegspspol_c *ca_modp_vardnsexpdegspspol_id2c(id object)
{
    return [object ca_modp_vardnsexpdegspspol_reference];
}

static id ca_modp_vardnsexpdegspspol_t2id(id proto,ca_modp_vardnsexpdegspspol_t f)
{
    return [proto ca_modp_vardnsexpdegspspol_value:f];
}

static id ca_modp_vardnsexpdegspspol_c2id(id proto,ca_modp_vardnsexpdegspspol_c *r)
{
    return [proto ca_modp_vardnsexpdegspspol_reference:r];
}

static int ca_modp_vardnsexpdegspspol_isempty(ca_modp_vardnsexpdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static void ca_modp_vardnsexpdegspspol_init(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    self->data = 0xcafebabe;self->nlink = self;self->plink = self;
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
}

static void ca_modp_vardnsexpdegspspol_copy(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_copy(c,a->nlink,a,c,args);
}

static void ca_modp_vardnsexpdegspspol_deepcopy(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_deepcopy(c,a->nlink,a,c,args);
}

static void ca_modp_vardnsexpdegspspol_move(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    if (ca_modp_vardnsexpdegspspol_isempty(a)) {
	ca_modp_vardnsexpdegspspol_init(c,args);
    } else {
	ca_modp_vardnsexpdegspspol_init(c,args);
	ca_modp_vardnsexpdegspsterm_glue(c,a->nlink);ca_modp_vardnsexpdegspsterm_glue(a->plink,c);
    }
    
    a->data = 0;assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_clear(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_clear(self->nlink,self,args);
    self->data = 0;ca_modp_vardnsexpdegspslist_free(self->nlink,self,args);
}

static int ca_modp_vardnsexpdegspspol_count(ca_modp_vardnsexpdegspspol_t self)
{
    return ca_modp_vardnsexpdegspslist_count(self->nlink,self);
}

static unsigned ca_modp_vardnsexpdegspspol_hash(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspslist_hash(a->nlink,a,args);
}

static int ca_modp_vardnsexpdegspspol_iseq(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_t b,ca_modp_vardnsexpdegspspol_args args)
{
    return (a==b)?YES:ca_modp_vardnsexpdegspslist_iseq(a->nlink,a,b->nlink,b,args);
}

static int ca_modp_vardnsexpdegspspol_isone(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    /* if the leading term is one, then it is one */
    return (ca_modp_vardnsexpdegspspol_isempty(a))?0:ca_modp_vardnsexpdegspsterm_isone(a->nlink,args);
}

static int ca_modp_vardnsexpdegspspol_isminusone(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    /* if the leading term is minus one, then it is minus one */
    return (ca_modp_vardnsexpdegspspol_isempty(a))?0:ca_modp_vardnsexpdegspsterm_isminusone(a->nlink,args);
}

static int ca_modp_vardnsexpdegspspol_deg(ca_modp_vardnsexpdegspspol_t self ,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspslist_deg(self->nlink,self,args);
}

static int ca_modp_vardnsexpdegspspol_ord(ca_modp_vardnsexpdegspspol_t self ,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspslist_ord(self->nlink,self,args);
}

static int ca_modp_vardnsexpdegspspol_maxdeg(ca_modp_vardnsexpdegspspol_t self ,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspslist_maxdeg(self->nlink,self,args);
}

static int ca_modp_vardnsexpdegspspol_mindeg(ca_modp_vardnsexpdegspspol_t self ,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspslist_mindeg(self->nlink,self,args);
}

static ca_modp_t ca_modp_vardnsexpdegspspol_scalat(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_ixt ix,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_modp_vardnsexpdegspsterm_scal(ix->ix,args);
}

static ca_vardnsmonom_t ca_modp_vardnsexpdegspspol_monomat(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_ixt ix,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_modp_vardnsexpdegspsterm_monom(ix->ix);
}

static ca_modp_t ca_modp_vardnsexpdegspspol_scalfirst(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_isempty(self) == NO);
    return ca_modp_vardnsexpdegspsterm_scal(self->nlink,args);
}

static ca_vardnsmonom_t ca_modp_vardnsexpdegspspol_monomfirst(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_isempty(self) == NO);
    return ca_modp_vardnsexpdegspsterm_monom(self->nlink);
}

static void ca_modp_vardnsexpdegspspol_remove(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_check(self,args) && ca_modp_vardnsexpdegspspol_isempty(self) == NO);
    ca_modp_vardnsexpdegspsterm_delete(self->nlink,args);
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
}

static void ca_modp_vardnsexpdegspspol_insert(ca_modp_vardnsexpdegspspol_t self,ca_vardnsmonom_t m,ca_modp_t s,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
    ca_modp_vardnsexpdegspslist_insert(self->nlink,self,m,s,args);
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
}

static int ca_modp_vardnsexpdegspspol_iszero(ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    return ca_modp_vardnsexpdegspspol_isempty(self);
}

static int ca_modp_vardnsexpdegspspol_isop(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_t b,ca_modp_vardnsexpdegspspol_args args)
{
    assert(a != b);return ca_modp_vardnsexpdegspslist_isop(a->nlink,a,b->nlink,b,args);
}

static void ca_modp_vardnsexpdegspspol_ooneg(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_ooneg(c,a->nlink,a,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inneg(ca_modp_vardnsexpdegspspol_t a,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inneg(a->nlink,a,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oodbl(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oodbl(c,a->nlink,a,v,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_indbl(ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_indbl(a->nlink,a,v,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_ooadd(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,int w,ca_modp_vardnsexpdegspspol_args args)
{
    assert(a!=b);
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_ooadd(c,a->nlink,a,v,b->nlink,b,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inadd(ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,int w,ca_modp_vardnsexpdegspspol_args args)
{
    assert(a!=b);
    ca_modp_vardnsexpdegspslist_inadd(a->nlink,a,v,b->nlink,b,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oomulsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oomulsc(c,a->nlink,a,s,+1,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inmulsc(ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inmulsc(a->nlink,a,s,+1,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_ooaddmulsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,ca_modp_t t,int w, ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_ooaddmulsc(c,a->nlink,a,v,b->nlink,b,t,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inaddmulsc(ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inaddmulsc(a->nlink,a,v,b->nlink,b,t,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oomulscadd(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,int w, ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oomulscadd(c,a->nlink,a,s,v,b->nlink,b,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inmulscadd(ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inmulscadd(a->nlink,a,s,v,b->nlink,b,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oomulscaddmulsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,ca_modp_t t,int w, ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oomulscaddmulsc(c,a->nlink,a,s,v,b->nlink,b,t,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inmulscaddmulsc(ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inmulscaddmulsc(a->nlink,a,s,v,b->nlink,b,t,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static BOOL ca_modp_vardnsexpdegspspol_oodivsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,ca_modp_vardnsexpdegspspol_args args)
{
    BOOL failed;
    ca_modp_vardnsexpdegspspol_init(c,args);
    failed = ca_modp_vardnsexpdegspslist_oodivsc(c,a->nlink,a,s,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
    if (failed) ca_modp_vardnsexpdegspspol_clear(c,args);
    return failed;
}

static void ca_modp_vardnsexpdegspspol_indivsc(ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_indivsc(a->nlink,a,s,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oomulmonomsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oomulmonomsc(c,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inmulmonomsc(ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inmulmonomsc(b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(b,args));
}

static void ca_modp_vardnsexpdegspspol_ooaddmulmonomsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w, ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_ooaddmulmonomsc(c,a->nlink,a,v,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inaddmulmonomsc(ca_modp_vardnsexpdegspspol_t a,int v,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inaddmulmonomsc(a->nlink,a,v,b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static void ca_modp_vardnsexpdegspspol_oomulscaddmulmonomsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w, ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(c,args);ca_modp_vardnsexpdegspslist_oomulscaddmulmonomsc(c,a->nlink,a,s,v,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
}

static void ca_modp_vardnsexpdegspspol_inmulscaddmulmonomsc(ca_modp_vardnsexpdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,int w,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_inmulscaddmulmonomsc(a->nlink,a,s,v,b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsexpdegspspol_check(a,args));
}

static BOOL ca_modp_vardnsexpdegspspol_oodivmonomsc(ca_modp_vardnsexpdegspspol_t c,ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t, ca_modp_vardnsexpdegspspol_args args)
{
    BOOL failed;
    ca_modp_vardnsexpdegspspol_init(c,args);
    failed = ca_modp_vardnsexpdegspslist_oodivmonomsc(c,b->nlink,b,m,t,c,args);
    assert(ca_modp_vardnsexpdegspspol_check(c,args));
    if (failed) ca_modp_vardnsexpdegspspol_clear(c,args);
    return failed;
}

static void ca_modp_vardnsexpdegspspol_indivmonomsc(ca_modp_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_modp_t t,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspslist_indivmonomsc(b->nlink,b,m,t,args);
    assert(ca_modp_vardnsexpdegspspol_check(b,args));
}

static void ca_modp_vardnsexpdegspspol_args_fileOut(id aFiler,ca_modp_vardnsexpdegspspol_args *args)
{
    ca_modp_vardnsexpdegspsterm_args_fileOut(aFiler,args);
}
static void ca_modp_vardnsexpdegspspol_args_fileIn(id aFiler,ca_modp_vardnsexpdegspspol_args *args)
{
    ca_modp_vardnsexpdegspsterm_args_fileIn(aFiler,args);
}

static void ca_modp_vardnsexpdegspspol_fileOut(id aFiler,ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
    ca_modp_vardnsexpdegspslist_fileOut(aFiler,self->nlink,self,args);
}

static void ca_modp_vardnsexpdegspspol_fileIn(id aFiler,ca_modp_vardnsexpdegspspol_t self,ca_modp_vardnsexpdegspspol_args args)
{
    ca_modp_vardnsexpdegspspol_init(self,args);ca_modp_vardnsexpdegspslist_fileIn(aFiler,self,self,args);
    assert(ca_modp_vardnsexpdegspspol_check(self,args));
}

static ca_modp_vardnsexpdegspspol_args ca_modp_vardnsexpdegspspol_getargs(id self)
{
    ca_modp_vardnsexpdegspspol_args res;id monomialZero = [self monomialZero];
    res.sargs = ca_modp_getargs([monomialZero scalar]);
    res.margs = ca_vardnsmonom_getargs([monomialZero terms]);
    return res; 
}
@implementation ca_modp_vardnsexpdegsps_polynomial : ca_vardnsexpdegsps_polynomial

- (ca_modp_vardnsexpdegspspol_t)ca_modp_vardnsexpdegspspol_value
{
    return &value;
}

- ca_modp_vardnsexpdegspspol_value:(ca_modp_vardnsexpdegspspol_t)aValue
{
    self = [self clone];
    ca_modp_vardnsexpdegspspol_copy(&value,aValue,ca_modp_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

- (ca_modp_vardnsexpdegspspol_t)ca_modp_vardnsexpdegspspol_reference
{
    return &value;
}

- ca_modp_vardnsexpdegspspol_reference:(ca_modp_vardnsexpdegspspol_c *)aReference
{
    self = [self clone];
    ca_modp_vardnsexpdegspspol_move(&value,aReference,ca_modp_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_copy(&c,&value,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- deepCopy
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_deepcopy(&c,&value,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_modp_vardnsexpdegspspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_modp_vardnsexpdegspspol_isone(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_modp_vardnsexpdegspspol_isminusone(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- (unsigned) hash
{
    return ca_modp_vardnsexpdegspspol_hash(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_modp_vardnsexpdegspspol_iseq(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_id2t(b),ca_modp_vardnsexpdegspspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_modp_vardnsexpdegspspol_iseq(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_id2t(b),ca_modp_vardnsexpdegspspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_modp_vardnsexpdegspspol_iszero(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_modp_vardnsexpdegspspol_iszero(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_modp_vardnsexpdegspspol_isop(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_id2t(b),ca_modp_vardnsexpdegspspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_modp_vardnsexpdegspspol_isop(ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_id2t(b),ca_modp_vardnsexpdegspspol_getargs(self));
}

- negate
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_ooneg(&c,ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_modp_vardnsexpdegspspol_inneg(ca_modp_vardnsexpdegspspol_id2c(self),ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_oodbl(&c,ca_modp_vardnsexpdegspspol_id2t(self),v,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_modp_vardnsexpdegspspol_indbl(ca_modp_vardnsexpdegspspol_id2c(self),v,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_oodbl(&c,ca_modp_vardnsexpdegspspol_id2t(self),+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_modp_vardnsexpdegspspol_indbl(ca_modp_vardnsexpdegspspol_id2c(self),+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_modp_vardnsexpdegspspol_ooadd(&c,ca_modp_vardnsexpdegspspol_id2t(self),v,ca_modp_vardnsexpdegspspol_id2t(b),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_modp_vardnsexpdegspspol_inadd(ca_modp_vardnsexpdegspspol_id2c(self),v,ca_modp_vardnsexpdegspspol_id2t(b),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_modp_vardnsexpdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsexpdegspspol_ooadd(&c,ca_modp_vardnsexpdegspspol_id2t(self),+1,ca_modp_vardnsexpdegspspol_id2t(b),+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_modp_vardnsexpdegspspol_inadd(ca_modp_vardnsexpdegspspol_id2c(self),+1,ca_modp_vardnsexpdegspspol_id2t(b),+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_modp_vardnsexpdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsexpdegspspol_ooadd(&c,ca_modp_vardnsexpdegspspol_id2t(self),+1,ca_modp_vardnsexpdegspspol_id2t(b),-1,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_modp_vardnsexpdegspspol_inadd(ca_modp_vardnsexpdegspspol_id2c(self),+1,ca_modp_vardnsexpdegspspol_id2t(b),-1,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_modp_vardnsexpdegspspol_deg(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- (int) lastDegree
{
    return ca_modp_vardnsexpdegspspol_ord(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- (int) maxDegree
{
    return ca_modp_vardnsexpdegspspol_maxdeg(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- (int) minDegree
{
    return ca_modp_vardnsexpdegspspol_mindeg(&value,ca_modp_vardnsexpdegspspol_getargs(self));
}

- asScalar
{
    if ([self numMembers] == 1) {
	id asScalar,lastTerm,aSequence = [self eachMember];
	asScalar = (lastTerm = [aSequence lastMember])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self numMembers] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachMember];
	asSymbol = (lastTerm = [aSequence lastMember])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    assert(scalarZero && termZero && monomialZero && symbols);
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols numMembers] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);

    ca_modp_vardnsexpdegspspol_check(&value,ca_modp_vardnsexpdegspspol_getargs(self));
    return self;
}

- _setUpScalarZero:aScalar symbols:aCltn
{
    assert([aScalar isZero] && [aCltn notEmpty]);
    assert([aScalar class_vardnsexpdegsps_polynomial] == [self class]);

    aScalar  = [aScalar cheapCopy];
    termZero = [CATermClass newScalar:aScalar symbol:nil exponent:0];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [CAMonomialClass newScalar:aScalar symbols:aCltn];

    scalarZero = aScalar;
    symbols    = aCltn;

    ca_modp_vardnsexpdegspspol_init(&value,ca_modp_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

+ newScalarZero:aScalar symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar symbols:aCltn];
}

- clear
{
    ca_modp_vardnsexpdegspspol_clear(&value,ca_modp_vardnsexpdegspspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- empty
{
    ca_modp_vardnsexpdegspspol_c c;assert([self check]);
    ca_modp_vardnsexpdegspspol_init(&c,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- (int) numMonomials
{
    assert([self check]);return ca_modp_vardnsexpdegspspol_count(&value);
}

- eachMonomial
{
    id aSequence = [ca_modp_vardnsexpdegspspol_sequence newContent:[self cheapCopy]];
    return [CASequenceClass newCarrier:aSequence];
}

- elt_ca_modp_vardnsexpdegspspol_ix:(ca_modp_vardnsexpdegspspol_ixt)ix
{
    if (ca_modp_vardnsexpdegspspol_ix_ismatch(ix)) {
	id s,m;ca_modp_t sc;ca_vardnsmonom_t mc;
	sc = ca_modp_vardnsexpdegspspol_scalat(&value,ix,ca_modp_vardnsexpdegspspol_getargs(self));
	mc = ca_modp_vardnsexpdegspspol_monomat(&value,ix,ca_modp_vardnsexpdegspspol_getargs(self));
	s  = ca_modp_t2id([monomialZero scalar],sc);
	m  = ca_vardnsmonom_t2id([monomialZero terms],mc);
	return [monomialZero scalar:s terms:m];
    } else {
	return nil;
    }
}

- removeMonomial
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id m,s;ca_modp_t sc;ca_vardnsmonom_t mc;
	[self checkRefCountOne];assert([self check]);
	
	sc = ca_modp_vardnsexpdegspspol_scalfirst(&value,ca_modp_vardnsexpdegspspol_getargs(self));
	mc = ca_modp_vardnsexpdegspspol_monomfirst(&value,ca_modp_vardnsexpdegspspol_getargs(self));
	s  = ca_modp_t2id([monomialZero scalar],sc);
	m  = ca_vardnsmonom_t2id([monomialZero terms],mc);
	lt = [monomialZero scalar:s terms:m];
	
	ca_modp_vardnsexpdegspspol_remove(&value,ca_modp_vardnsexpdegspspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertMonomial:aMonomial
{
    if ([aMonomial isZero])  {
	[aMonomial free];return self;
    } else {
	ca_modp_c c;ca_modp_t r;ca_vardnsmonom_t m;
	assert([aMonomial check]);[self checkRefCountOne];assert([self check]);
	r = ca_modp_id2ref([aMonomial scalar],&c);
	m = ca_vardnsmonom_id2t([aMonomial terms]);
	ca_modp_vardnsexpdegspspol_insert(&value,m,r,ca_modp_vardnsexpdegspspol_getargs(self));
	assert([aMonomial check]);[aMonomial free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyScalar:s
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c sc;
    ca_modp_vardnsexpdegspspol_oomulsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- multiplySelfScalar:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyScalar:s]);
    ca_modp_vardnsexpdegspspol_inmulsc(ca_modp_vardnsexpdegspspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- divideScalar:s
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c sc;
    BOOL failed = ca_modp_vardnsexpdegspspol_oodivsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsexpdegspspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- divideSelfScalar:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideScalar:s]);
    ca_modp_vardnsexpdegspspol_indivsc(ca_modp_vardnsexpdegspspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyScalar:b:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c bc;
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_ooaddmulsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),v,ca_modp_vardnsexpdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyScalar:b:(int)w
{
    ca_modp_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyScalar:b:w]);
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_inaddmulsc(ca_modp_vardnsexpdegspspol_id2c(self),v,ca_modp_vardnsexpdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c ac;
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_oomulscadd(&c,ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsexpdegspspol_id2t(B),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B:(int)w
{
    ca_modp_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_inmulscadd(ca_modp_vardnsexpdegspspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsexpdegspspol_id2t(B),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c ac,bc;
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_oomulscaddmulsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsexpdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_modp_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyScalar:b:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modp_vardnsexpdegspspol_inmulscaddmulsc(ca_modp_vardnsexpdegspspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsexpdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}
- multiplyMonomial:aMonomial
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;
    
    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    ca_modp_vardnsexpdegspspol_oomulmonomsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),m,s,+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- multiplySelfMonomial:aMonomial
{
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self multiplyMonomial:aMonomial]);

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    ca_modp_vardnsexpdegspspol_inmulmonomsc(ca_modp_vardnsexpdegspspol_id2c(self),m,s,+1,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyMonomial:aMonomial:(int)v
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    ca_modp_vardnsexpdegspspol_oomulmonomsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),m,s,v,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfMonomial:aMonomial:(int)v
{
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self _multiplyMonomial:aMonomial:v]);

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    ca_modp_vardnsexpdegspspol_inmulmonomsc(ca_modp_vardnsexpdegspspol_id2c(self),m,s,v,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- divideMonomial:aMonomial
{
    BOOL failed;
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    failed = ca_modp_vardnsexpdegspspol_oodivmonomsc(&c,ca_modp_vardnsexpdegspspol_id2t(self),m,s,ca_modp_vardnsexpdegspspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- divideSelfMonomial:aMonomial
{
    ca_modp_t s;
    ca_modp_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self divideMonomial:aMonomial]);
    
    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_modp_id2ref([aMonomial scalar],&sc);
    ca_modp_vardnsexpdegspspol_indivmonomsc(ca_modp_vardnsexpdegspspol_id2c(self),m,s,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyMonomial:b:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c bc;
    ca_modp_t bs;
    ca_vardnsmonom_t bm;

    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_modp_id2ref([b scalar],&bc);
    ca_modp_vardnsexpdegspspol_ooaddmulmonomsc(&c,ca_modp_vardnsexpdegspspol_id2c(self),v,ca_modp_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyMonomial:b:(int)w
{
    ca_modp_c bc;
    ca_modp_t bs;
    ca_vardnsmonom_t bm;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyMonomial:b:w]);
    
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_modp_id2ref([b scalar],&bc);
    ca_modp_vardnsexpdegspspol_inaddmulmonomsc(ca_modp_vardnsexpdegspspol_id2c(self),v,ca_modp_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyMonomial:b:(int)w
{
    ca_modp_vardnsexpdegspspol_c c;
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    ca_vardnsmonom_t bm;
    
    as = ca_modp_id2ref(a,&ac);
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_modp_id2ref([b scalar],&bc);
    ca_modp_vardnsexpdegspspol_oomulscaddmulmonomsc(&c,ca_modp_vardnsexpdegspspol_id2c(self),as,v,ca_modp_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_modp_vardnsexpdegspspol_getargs(self));
    return ca_modp_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyMonomial:b:(int)w
{
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    ca_vardnsmonom_t bm;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyMonomial:b:w]);
    
    as = ca_modp_id2ref(a,&ac);
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_modp_id2ref([b scalar],&bc);
    ca_modp_vardnsexpdegspspol_inmulscaddmulmonomsc(ca_modp_vardnsexpdegspspol_id2c(self),as,v,ca_modp_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_modp_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_modp_vardnsexpdegspspol_args args;
    assert([self check]);
    [super fileOutOn:aFiler];
    args = ca_modp_vardnsexpdegspspol_getargs(self);
    ca_modp_vardnsexpdegspspol_args_fileOut(aFiler,&args);
    ca_modp_vardnsexpdegspspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_modp_vardnsexpdegspspol_args args;
    [super fileInFrom:aFiler];
    ca_modp_vardnsexpdegspspol_args_fileIn(aFiler,&args);
    ca_modp_vardnsexpdegspspol_fileIn(aFiler,&value,args);
    return self;
}

@end

