#include "cakit.h"

static ca_modp_args ca_modp_getargs(id self)
{
   ca_modp_args args;args.p = [self modulus];return args;
}

static int ca_modp_size(ca_modp_args args)
{
    return 1;				/* pointer increment */
}

static int ca_modp_bsize(ca_modp_args args)
{
    return sizeof(ca_modp_c);		/* bytes */
}

static ca_modp_t ca_modp_id2t(id object)
{
    return [object ca_modp_value];	/* read-only */
}

static ca_modp_t ca_modp_id2ref(id object,ca_modp_c *ref)
{
    return [object ca_modp_value];	/* read-only, scratch space "ref" */
}

static ca_modp_c *ca_modp_id2c(id object)
{
    return [object ca_modp_reference];	/* write for inplace arithmetic */
}

static id ca_modp_t2id(id proto,ca_modp_t f)
{
    return [proto ca_modp_value:f];	/* new object containing copy of "f" */
}

static id ca_modp_c2id(id proto,ca_modp_c* c)
{
    return [proto ca_modp_reference:c];/* new object containing "c" */
}

static ca_modp_t ca_modp_fromintvalue(int i,ca_modp_args args)
{
    ca_modp_t red;
    
    if (i==0) return 0;
    if (i <0) {
	red = -i;
	red %= args.p;	/* it's dangerous to % int & unsigned */
	red = (red)?args.p-red:0;
    } else {
	red  = i;
	red %= args.p;	/* dito */
    }
    
    return red;
}

static int ca_modp_check(ca_modp_t self,ca_modp_args args)
{
    assert(self < args.p);return 1;
}

static ca_modp_t ca_modp_zero(ca_modp_args args)
{
    return 0;
}

static ca_modp_t ca_modp_one(ca_modp_args args)
{
    return 1;
}

static unsigned ca_modp_hash(ca_modp_t a,ca_modp_args args)
{
    return a;
}

static int ca_modp_iseq(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a == b);
}

static int ca_modp_isop(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a==0)?(b==0):(a+b==args.p);
}

static int ca_modp_iszero(ca_modp_t a,ca_modp_args args)
{
    return a==0;
}

static int ca_modp_isone(ca_modp_t a,ca_modp_args args)
{
    return a==1;
}

static int ca_modp_isminusone(ca_modp_t a,ca_modp_args args)
{
    return (a+1)==args.p;
}

static ca_modp_t ca_modp_neg(ca_modp_t a,ca_modp_args args)
{
    assert(ca_modp_check(a,args));return (a)?args.p-a:0;
}

static ca_modp_t ca_modp_dbl(ca_modp_t a,ca_modp_args args)
{
    ca_modp_t c = a+a;
    assert(args.p != 2 && ca_modp_check(a,args));
    if (c) return (c<args.p)?c:c-args.p; else return 0;
}

static ca_modp_t ca_modp_add(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    ca_modp_t c = a+b;
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return ((c>=args.p)?c-args.p:c); /* compiles better than c<p?c:c-p */
}

static ca_modp_t ca_modp_sub(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a>=b)?(a-b):(args.p-(b-a));
}

static ca_modp_t ca_modp_mul(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    unsigned c = a * b;		/* limitation to 16bits to be able to do this
				 * russian peasant 50% slower
                                 */
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_sqr(ca_modp_t a,ca_modp_args args)
{
    unsigned c = a * a;
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_inv(ca_modp_t x,ca_modp_args args)
{
    unsigned q,s;
    int u1,u2,us;  /* intermediate values can be negative */
    ca_modp_t a,b,inv;
    
    assert(x != 0 && x != 1 && x+1 != args.p);
    
    a = x;b = args.p;u1 = 1;u2 = 0;
    while (b) {
	s  = a%b;q = a/b;
	a  = b;b = s;
	u1 -= q*u2;			
	us  = u1;u1 = u2;u2 = us;
    }
    
    inv = ((u1<0)?u1+args.p:u1); /* keep result positive */
    
    assert(ca_modp_check(inv,args));
    assert(ca_modp_isone(ca_modp_mul(x,inv,args),args));
    return inv;
}

static ca_modp_t ca_modp_div(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(b!=0);
    return ca_modp_mul(a,ca_modp_inv(b,args),args);
}

static ca_modp_t ca_modp_pow(ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_t o;
    
    if (n==0) return ca_modp_one(args);
    if (n==1) return a;
    if (n==2) return ca_modp_sqr(a,args);
    
    o = ca_modp_one(args);
    while (n>1) {
	if (n&1) o = ca_modp_mul(o,a,args);
	a = ca_modp_sqr(a,args);
	n >>= 1;
    }
    
    return ca_modp_mul(o,a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_copy_val.fab should not be included"
#endif

static void ca_modp_clear(ca_modp_c *c,ca_modp_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_modp_t)0xcafebabe;
    #endif
}

static void ca_modp_copy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_deepcopy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_move(ca_modp_c *c,ca_modp_c *a,ca_modp_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_modp_t)0xcafebabe;
    #endif
}

#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static void ca_modp_oozero(ca_modp_c *c,ca_modp_args args)
{
    *c = ca_modp_zero(args);
}

static void ca_modp_inzero(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_zero(args);
}

static void ca_modp_inneg(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_neg(*a,args);
}

static void ca_modp_ooneg(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_neg(a,args);
}

static ca_modp_t ca_modp_dblv(ca_modp_t a,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_dbl(a,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_indbl(ca_modp_c *a,int v,ca_modp_args args)
{
    *a = ca_modp_dblv(*a,v,args);
}

static void ca_modp_oodbl(ca_modp_c *c,ca_modp_t a,int v,ca_modp_args args)
{
    *c = ca_modp_dblv(a,v,args);
}

static ca_modp_t ca_modp_addvw(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    ca_modp_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_modp_add(a,b,args);
	} else {
	    c = ca_modp_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_modp_sub(b,a,args);
	} else {
	    c = ca_modp_neg(ca_modp_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_modp_inadd(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addvw(*a,v,b,w,args);
}

static void ca_modp_ooadd(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addvw(a,v,b,w,args);
}
#if !ca_modp_isvalue
#error "ca_modp_mul_val.fab should not be included"
#endif

static void ca_modp_insqr(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_sqr(*a,args);
}

static void ca_modp_oosqr(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_sqr(a,args);
}

static void ca_modp_inmul(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_mul(*a,b,args);
}

static void ca_modp_oomul(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_mul(a,b,args);
}
#if !ca_modp_isvalue
#error "ca_modp_inv_val.fab should not be included"
#endif

static void ca_modp_ininv(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_inv(*a,args);
}

static void ca_modp_ooinv(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_inv(a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_div_val.fab should not be included"
#endif

static void ca_modp_indiv(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_div(*a,b,args);
}

static BOOL ca_modp_oodiv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_div(a,b,args);return SUCCESS;
}
#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static ca_modp_t ca_modp_mulv(ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_mul(a,b,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_inmulv(ca_modp_c *a,ca_modp_t b,int v,ca_modp_args args)
{
    *a = ca_modp_mulv(*a,b,v,args);
}

static void ca_modp_oomulv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    *c = ca_modp_mulv(a,b,v,args);
}

static ca_modp_t ca_modp_addmul(ca_modp_t a,int v,ca_modp_t b,ca_modp_t s,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_mul(b,s,args),w,args);
}

static ca_modp_t ca_modp_addsqr(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_sqr(b,args),w,args);
}

static ca_modp_t ca_modp_muladd(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,b,w,args);
}

static ca_modp_t ca_modp_muladdmul(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,ca_modp_mul(b,t,args),w,args);
}

static void ca_modp_inaddmul(ca_modp_c *a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_addmul(*a,v,b,t,w,args);
}

static void ca_modp_ooaddmul(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_addmul(a,v,b,t,w,args);
}

static void ca_modp_inaddsqr(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addsqr(*a,v,b,w,args);
}

static void ca_modp_ooaddsqr(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addsqr(a,v,b,w,args);
}

static void ca_modp_inmuladd(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_muladd(*a,s,v,b,w,args);
}

static void ca_modp_oomuladd(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_muladd(a,s,v,b,w,args);
}

static void ca_modp_inmuladdmul(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_modp_oomuladdmul(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_muladdmul(a,s,v,b,t,w,args);
}
static void ca_modp_args_fileOut(id aFiler,ca_modp_args *args) {}
static void ca_modp_args_fileIn(id aFiler,ca_modp_args *args) {}

static void ca_modp_fileOut(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileOut:c type:'S'];
}

static void ca_modp_fileIn(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileIn:c type:'S'];
}


/*
 * Computer Algebra Kit (c) 1993,97 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_unimonom_size(ca_unimonom_args args)
{
    return 1;
}

static int ca_unimonom_bsize(ca_unimonom_args args)
{
    return sizeof(ca_unimonom_t);
}

static int ca_unimonom_check(ca_unimonom_t self,ca_unimonom_args args)
{
    assert(self >= 0);return 1;
}

static void ca_unimonom_clear(ca_unimonom_c *c,ca_unimonom_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_copy(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_args args)
{
    *c = a;
}

static void ca_unimonom_move(ca_unimonom_c *c,ca_unimonom_c *a,ca_unimonom_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_args_fileOut(id aFiler,ca_unimonom_args *args) {}
static void ca_unimonom_args_fileIn(id aFiler,ca_unimonom_args *args) {}

static void ca_unimonom_fileOut(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileOut:c type:'i'];
}

static void ca_unimonom_fileIn(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileIn:c type:'i'];
}

static unsigned ca_unimonom_hash(ca_unimonom_t self,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_deg(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_isone(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self == 0;
}

static int ca_unimonom_iseq(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == b;
}

static int ca_unimonom_iseqmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == x+b;
}

static int ca_unimonom_cmp(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-b;
}

static int ca_unimonom_cmpmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-x-b;
}

static void ca_unimonom_oomul(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    *c = a + b;
}

static void ca_unimonom_inmul(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    *c += b;
}

static BOOL ca_unimonom_oodiv(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    if (a < b) {
	return FAILURE;
    } else {
	*c = a - b;
	return SUCCESS;
    }
}

static void ca_unimonom_indiv(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    if (*c < b) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
    *c -= b;
}

/*
 * scalar is accessed by the C "->" operator since having variable sized
 * scalars _and_ monomials gets difficult on machines like the DEC alpha
 * (where we'd have to manually pad to get pointer aligned structures).
 * 
 * the monomial (exponent) is however not accessed by the C "->" operator
 * but rather by functions that can take offsets as arguments
 * this is important the monomials in variable dense representation
 * (fixarray of exp)
 *
 * stes - mar31, 1997 - CHANGES: for DEC alpha coefficient needs to be aligned
 * properly, put it before the monomial and drop the idea of variable sized
 * coefficients.
 */

static ca_unimonom_c *ca_modp_vardnsrecdegspsterm_monom_cval(ca_modp_vardnsrecdegspsterm_t self)
{
    return &self->data;
}

static ca_unimonom_t ca_modp_vardnsrecdegspsterm_monom(ca_modp_vardnsrecdegspsterm_t self)
{
#if ca_unimonom_pervalue
    return self->data;
#else
    return  ca_modp_vardnsrecdegspsterm_monom_cval(self);
#endif
}

/* stes - mar31,1997
 * change for DEC alpha to exclude varsized scalars
 */

static ca_modp_c *ca_modp_vardnsrecdegspsterm_scal_cval(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
	return &self->scal;
}

static ca_modp_t ca_modp_vardnsrecdegspsterm_scal(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
#if ca_modp_pervalue
	return self->scal;
#else
    return  ca_modp_vardnsrecdegspsterm_scal_cval(self,args);
#endif
}

static int ca_modp_vardnsrecdegspsterm_check(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_check(ca_modp_vardnsrecdegspsterm_scal(self,args),args.sargs));
    assert(!ca_modp_iszero(ca_modp_vardnsrecdegspsterm_scal(self,args),args.sargs));
    assert(ca_unimonom_check(ca_modp_vardnsrecdegspsterm_monom(self),args.margs));
    return 1;
}

static int ca_modp_vardnsrecdegspsterm_bsize(ca_modp_vardnsrecdegspsterm_args args)
{
	/* this is sizeof(ca_unimonom_c) too much but why bother ? */
    return sizeof(struct ca_modp_vardnsrecdegspsterm) + ca_unimonom_bsize(args.margs);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_alloc(ca_modp_vardnsrecdegspsterm_args args)
{
    return (ca_modp_vardnsrecdegspsterm_t)OC_Malloc(ca_modp_vardnsrecdegspsterm_bsize(args));
}

static void ca_modp_vardnsrecdegspsterm_clear(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_unimonom_clear(ca_modp_vardnsrecdegspsterm_monom_cval(self),args.margs);
    ca_modp_clear(ca_modp_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
}

static void ca_modp_vardnsrecdegspsterm_free(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    /* invalidating links not always helps : buggy code might work because
     * links were cleared...        self->nlink = NULL;self->plink = NULL;
     */
    OC_Free(self);
}

static int ca_modp_vardnsrecdegspsterm_cmp(ca_modp_vardnsrecdegspsterm_t a,ca_modp_vardnsrecdegspsterm_t b,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(a,args) && ca_modp_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_cmp(ca_modp_vardnsrecdegspsterm_monom(a),ca_modp_vardnsrecdegspsterm_monom(b),args.margs);
}

static int ca_modp_vardnsrecdegspsterm_cmpmonom(ca_modp_vardnsrecdegspsterm_t a,ca_unimonom_t b,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(a,args) && ca_unimonom_check(b,args.margs));
    return ca_unimonom_cmp(ca_modp_vardnsrecdegspsterm_monom(a),b,args.margs);
}

static int ca_modp_vardnsrecdegspsterm_cmpmul(ca_modp_vardnsrecdegspsterm_t a,ca_modp_vardnsrecdegspsterm_t b,ca_unimonom_t m ,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(a,args) && ca_modp_vardnsrecdegspsterm_check(b,args));
    assert(ca_unimonom_check(m,args.margs));
    return ca_unimonom_cmpmul(ca_modp_vardnsrecdegspsterm_monom(a),ca_modp_vardnsrecdegspsterm_monom(b),m,args.margs);
}

static int ca_modp_vardnsrecdegspsterm_deg(ca_modp_vardnsrecdegspsterm_t a,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_unimonom_deg(ca_modp_vardnsrecdegspsterm_monom(a),args.margs);
}

static int ca_modp_vardnsrecdegspsterm_iszero(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_iszero(ca_modp_vardnsrecdegspsterm_scal(self,args),args.sargs);
}

static int ca_modp_vardnsrecdegspsterm_isone(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_isone(ca_modp_vardnsrecdegspsterm_scal(self,args),args.sargs) && ca_unimonom_isone(ca_modp_vardnsrecdegspsterm_monom(self),args.margs);
}

static int ca_modp_vardnsrecdegspsterm_isminusone(ca_modp_vardnsrecdegspsterm_t self,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_isminusone(ca_modp_vardnsrecdegspsterm_scal(self,args),args.sargs) && ca_unimonom_isone(ca_modp_vardnsrecdegspsterm_monom(self),args.margs);
}

static int ca_modp_vardnsrecdegspsterm_iseq(ca_modp_vardnsrecdegspsterm_t a,ca_modp_vardnsrecdegspsterm_t b,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(a,args) && ca_modp_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_iseq(ca_modp_vardnsrecdegspsterm_monom(a),ca_modp_vardnsrecdegspsterm_monom(b),args.margs) && ca_modp_iseq(ca_modp_vardnsrecdegspsterm_scal(a,args),ca_modp_vardnsrecdegspsterm_scal(b,args),args.sargs);
}

static int ca_modp_vardnsrecdegspsterm_isop(ca_modp_vardnsrecdegspsterm_t a,ca_modp_vardnsrecdegspsterm_t b,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(a,args) && ca_modp_vardnsrecdegspsterm_check(b,args));
    return ca_unimonom_iseq(ca_modp_vardnsrecdegspsterm_monom(a),ca_modp_vardnsrecdegspsterm_monom(b),args.margs) && ca_modp_isop(ca_modp_vardnsrecdegspsterm_scal(a,args),ca_modp_vardnsrecdegspsterm_scal(b,args),args.sargs);
}

static void ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A)
{
    C->nlink = A;A->plink = C;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A)
{
    ca_modp_vardnsrecdegspsterm_glue(C,A);return A;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_delete(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t C = A->nlink;
    ca_modp_vardnsrecdegspsterm_glue(A->plink,C);
    ca_modp_vardnsrecdegspsterm_clear(A,args);
    ca_modp_vardnsrecdegspsterm_free(A,args);
    return C;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_condcns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t new,ca_modp_vardnsrecdegspsterm_args args)
{
    if (ca_modp_vardnsrecdegspsterm_iszero(new,args)) {
	ca_modp_vardnsrecdegspsterm_clear(new,args);ca_modp_vardnsrecdegspsterm_free(new,args);return C;
    } else {
	assert(ca_modp_vardnsrecdegspsterm_check(new,args));return ca_modp_vardnsrecdegspsterm_cns(C,new);
    }
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_conddelete(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    if (ca_modp_vardnsrecdegspsterm_iszero(A,args)) {
	return ca_modp_vardnsrecdegspsterm_delete(A,args);
    } else {
	return A->nlink;
    }
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_copymonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    
    ca_modp_copy(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_modp_vardnsrecdegspsterm_check(new,args));return ca_modp_vardnsrecdegspsterm_cns(C,new);
}

static void ca_modp_vardnsrecdegspsterm_copymonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_copymonomsc_cns(C,m,s,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_deepcopymonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    
    ca_modp_deepcopy(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_modp_vardnsrecdegspsterm_check(new,args));return ca_modp_vardnsrecdegspsterm_cns(C,new);
}

static void ca_modp_vardnsrecdegspsterm_deepcopymonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_deepcopymonomsc_cns(C,m,s,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_copy_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_vardnsrecdegspsterm_copymonomsc_cns(C,ca_modp_vardnsrecdegspsterm_monom(A),ca_modp_vardnsrecdegspsterm_scal(A,args),args);
}

static void ca_modp_vardnsrecdegspsterm_copy(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_copy_cns(C,A,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_deepcopy_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_vardnsrecdegspsterm_deepcopymonomsc_cns(C,ca_modp_vardnsrecdegspsterm_monom(A),ca_modp_vardnsrecdegspsterm_scal(A,args),args);
}

static void ca_modp_vardnsrecdegspsterm_deepcopy(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_deepcopy_cns(C,A,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooneg_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO);

    ca_modp_ooneg(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_cns(C,new);
}

static void ca_modp_vardnsrecdegspsterm_ooneg(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_ooneg_cns(C,A,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inneg(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO);
    ca_modp_inneg(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),args.sargs);
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO);
    return A->nlink;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oonegv_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    return (v == -1)?ca_modp_vardnsrecdegspsterm_ooneg_cns(C,A,args):ca_modp_vardnsrecdegspsterm_copy_cns(C,A,args);
}

static void ca_modp_vardnsrecdegspsterm_oonegv(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oonegv_cns(C,A,v,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_innegv(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    return (v == -1)?ca_modp_vardnsrecdegspsterm_inneg(A,args):A->nlink;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oodbl_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO);

    ca_modp_oodbl(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),v,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsrecdegspsterm_oodbl(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oodbl_cns(C,A,v,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_indbl(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO);
    ca_modp_indbl(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),v,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooaddsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    ca_modp_ooadd(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),v,B,w,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inaddsc(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inadd(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),v,B,w,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooadd_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == ca_modp_vardnsrecdegspsterm_cmp(A,B,args));
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO && ca_modp_vardnsrecdegspsterm_iszero(B,args) == NO);
    return ca_modp_vardnsrecdegspsterm_ooaddsc_cns(C,A,v,ca_modp_vardnsrecdegspsterm_scal(B,args),w,args);
}

static void ca_modp_vardnsrecdegspsterm_ooadd(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_ooadd_cns(C,A,v,B,w,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inadd(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == ca_modp_vardnsrecdegspsterm_cmp(A,B,args));
    assert(ca_modp_vardnsrecdegspsterm_iszero(A,args) == NO && ca_modp_vardnsrecdegspsterm_iszero(B,args) == NO);
    return ca_modp_vardnsrecdegspsterm_inaddsc(A,v,ca_modp_vardnsrecdegspsterm_scal(B,args),w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    ca_modp_oomulv(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,v,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsrecdegspsterm_oomulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A, ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulsc(ca_modp_vardnsrecdegspsterm_t A, ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inmulv(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,v,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    ca_modp_oomulv(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,v,args.sargs);
    ca_unimonom_oomul(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),m,args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static void ca_modp_vardnsrecdegspsterm_oomulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inmulv(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,v,args.sargs);
    ca_unimonom_inmul(ca_modp_vardnsrecdegspsterm_monom_cval(A),m,args.margs);

    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooaddscmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    ca_modp_ooaddmul(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),v,B,t,w,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inaddscmulsc(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inaddmul(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),v,B,t,w,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooaddmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_ooaddscmulsc_cns(C,A,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_ooaddmulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
}

static void ca_modp_vardnsrecdegspsterm_ooaddmulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspsterm_ooaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inaddmulsc(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_inaddscmulsc(A,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsrecdegspsterm_inaddscmulsc(A,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulscaddsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    ca_modp_oomuladd(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,v,B,w,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulscaddsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inmuladd(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,v,B,w,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulscadd_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_oomulscaddsc_cns(C,A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),w,args);
}

static void ca_modp_vardnsrecdegspsterm_oomulscadd(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulscadd(ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_inmulscaddsc(A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulscaddscmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);
    ca_modp_oomuladdmul(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,v,B,t,w,args.sargs);
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulscaddscmulsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_inmuladdmul(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,v,B,t,w,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulscaddmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsrecdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static void ca_modp_vardnsrecdegspsterm_oomulscaddmulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspsterm_oomulscaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulscaddmulsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmp(A,B,args)));
    return ca_modp_vardnsrecdegspsterm_inmulscaddscmulsc(A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_inmulscaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(0 == (ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args)));
    return ca_modp_vardnsrecdegspsterm_inmulscaddscmulsc(A,s,v,ca_modp_vardnsrecdegspsterm_scal(B,args),t,w,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oodivsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    BOOL failed;
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    failed = ca_modp_oodiv(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,args.sargs);
    if (failed) {
	ca_modp_vardnsrecdegspsterm_free(new,args);return NULL;
    }
    
    ca_unimonom_copy(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_modp_vardnsrecdegspsterm_oodivsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A, ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    if ((C = ca_modp_vardnsrecdegspsterm_oodivsc_cns(C,A,s,args))) {
	ca_modp_vardnsrecdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_indivsc(ca_modp_vardnsrecdegspsterm_t A, ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_indiv(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,args.sargs);
    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_oodivmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    BOOL sfailed,mfailed;
    ca_modp_vardnsrecdegspsterm_t new = ca_modp_vardnsrecdegspsterm_alloc(args);

    sfailed = ca_modp_oodiv(ca_modp_vardnsrecdegspsterm_scal_cval(new,args),ca_modp_vardnsrecdegspsterm_scal(A,args),s,args.sargs);
    mfailed = ca_unimonom_oodiv(ca_modp_vardnsrecdegspsterm_monom_cval(new),ca_modp_vardnsrecdegspsterm_monom(A),m,args.margs);
    
    if (sfailed || mfailed) {
	/* both failures are possible in the multivariate case */
	ca_modp_vardnsrecdegspsterm_free(new,args);return NULL;
    }
    
    return ca_modp_vardnsrecdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_modp_vardnsrecdegspsterm_oodivmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    if ((C = ca_modp_vardnsrecdegspsterm_oodivmonomsc_cns(C,A,m,s,args))) {
	ca_modp_vardnsrecdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_indivmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_unimonom_t m, ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_indiv(ca_modp_vardnsrecdegspsterm_scal_cval(A,args),s,args.sargs);
    ca_unimonom_indiv(ca_modp_vardnsrecdegspsterm_monom_cval(A),m,args.margs);

    return ca_modp_vardnsrecdegspsterm_conddelete(A,args);
}

static void ca_modp_vardnsrecdegspsterm_args_fileOut(id aFiler,ca_modp_vardnsrecdegspsterm_args *args)
{
    ca_modp_args_fileOut(aFiler,&args->sargs);
    ca_unimonom_args_fileOut(aFiler,&args->margs);
}
static void ca_modp_vardnsrecdegspsterm_args_fileIn(id aFiler,ca_modp_vardnsrecdegspsterm_args *args)
{
    ca_modp_args_fileIn(aFiler,&args->sargs);
    ca_unimonom_args_fileIn(aFiler,&args->margs);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_fileOut(id aFiler,ca_modp_vardnsrecdegspsterm_t self ,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_vardnsrecdegspsterm_check(self,args));
    
    ca_modp_fileOut(aFiler,ca_modp_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
    ca_unimonom_fileOut(aFiler,ca_modp_vardnsrecdegspsterm_monom_cval(self),args.margs);
    
    return self->nlink;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspsterm_fileIn_cns(id aFiler,ca_modp_vardnsrecdegspsterm_t C ,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t self = ca_modp_vardnsrecdegspsterm_alloc(args);
    
    ca_modp_fileIn(aFiler,ca_modp_vardnsrecdegspsterm_scal_cval(self,args),args.sargs);
    ca_unimonom_fileIn(aFiler,ca_modp_vardnsrecdegspsterm_monom_cval(self),args.margs);
    
    return ca_modp_vardnsrecdegspsterm_cns(C,self);
}

static int ca_modp_vardnsrecdegspslist_check(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t prev;
    
    for(prev = NULL;A!=AT;prev = A,A = A->nlink) {
	if (ca_modp_vardnsrecdegspsterm_check(A,args) == 0) return 0;
	if (prev && ca_modp_vardnsrecdegspsterm_cmp(prev,A,args) <= 0) return 0;
    }
    
    return 1;
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_copy_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) { C = ca_modp_vardnsrecdegspsterm_copy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_modp_vardnsrecdegspslist_copy(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_copy_cns(C,A,AT,args),CT);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_deepcopy_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) { C = ca_modp_vardnsrecdegspsterm_deepcopy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_modp_vardnsrecdegspslist_deepcopy(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_deepcopy_cns(C,A,AT,args),CT);
}

static void ca_modp_vardnsrecdegspslist_clear(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) { ca_modp_vardnsrecdegspsterm_clear(A,args);A = A->nlink; }
}

static void ca_modp_vardnsrecdegspslist_free(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_t t;for(;A!=AT;A=t) { t = A->nlink;ca_modp_vardnsrecdegspsterm_free(A,args); }
}

static int ca_modp_vardnsrecdegspslist_count(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static unsigned ca_modp_vardnsrecdegspslist_hash(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    return ca_modp_vardnsrecdegspslist_count(A,AT);
}

static int ca_modp_vardnsrecdegspslist_iseq(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_t B, ca_modp_vardnsrecdegspsterm_t BT,ca_modp_vardnsrecdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_modp_vardnsrecdegspsterm_iseq(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_modp_vardnsrecdegspslist_isop(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_vardnsrecdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_modp_vardnsrecdegspsterm_isop(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_modp_vardnsrecdegspslist_deg(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    return (A==AT)?-1:ca_modp_vardnsrecdegspsterm_deg(A,args);
}

static int ca_modp_vardnsrecdegspslist_maxdeg(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    int max = (A==AT)?-1:ca_modp_vardnsrecdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_modp_vardnsrecdegspsterm_deg(A,args);if (max < d) max = d;
    }
    return max;
}

static int ca_modp_vardnsrecdegspslist_ord(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    return (A==AT)?-1:ca_modp_vardnsrecdegspsterm_deg(AT->plink,args);
}

static int ca_modp_vardnsrecdegspslist_mindeg(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    int min = (A==AT)?-1:ca_modp_vardnsrecdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_modp_vardnsrecdegspsterm_deg(A,args);if (min > d) min = d;
    }
    return min;
}

static void ca_modp_vardnsrecdegspslist_insert(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(ca_modp_iszero(s,args.sargs) == NO);
    
    /* adding to zero */
    if (A == AT) {
	ca_modp_vardnsrecdegspsterm_copymonomsc(A,m,s,A,args);return;
    }
    
    /* tail append */
    if (ca_modp_vardnsrecdegspsterm_cmpmonom(AT->plink,m,args) > 0) {
	ca_modp_vardnsrecdegspsterm_copymonomsc(AT->plink,m,s,AT,args);return;
    }
    
    /* insertion */
    while (A != AT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmpmonom(A,m,args);
	if (cmp == 0) {
	    ca_modp_vardnsrecdegspsterm_inaddsc(A,+1,s,+1,args);return;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_copymonomsc(A->plink,m,s,A,args);return;
	}
	A = A->nlink;
    }
 
    fprintf(stderr,"Shouldn't happen\n");
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_ooneg_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsrecdegspsterm_ooneg_cns(C,A,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsrecdegspslist_ooneg(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_ooneg_cns(C,A,AT,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inneg(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_inneg(A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oonegv_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    if (v == -1) {
	return ca_modp_vardnsrecdegspslist_ooneg_cns(C,A,AT,args);
    } else {
	return ca_modp_vardnsrecdegspslist_copy_cns(C,A,AT,args);
    }
}

static void ca_modp_vardnsrecdegspslist_oonegv(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args),CT);
}

static void ca_modp_vardnsrecdegspslist_innegv(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    if (v == -1) ca_modp_vardnsrecdegspslist_inneg(A,AT,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oodbl_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsrecdegspsterm_oodbl_cns(C,A,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsrecdegspslist_oodbl(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oodbl_cns(C,A,AT,v,args),CT);
}

static void ca_modp_vardnsrecdegspslist_indbl(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_indbl(A,v,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_ooadd_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_ooadd_cns(C,A,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_ooadd(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_ooadd_cns(C,A,AT,v,B,BT,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inadd(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inadd(A,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oomulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsrecdegspslist_oomulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inmulsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_inmulsc(A,s,v,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oomulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) {
	C = ca_modp_vardnsrecdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_modp_vardnsrecdegspslist_oomulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m, ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oomulmonomsc_cns(C,A,AT,m,s,v,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_inmulmonomsc(A,m,s,v,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_ooaddmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_ooaddmulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_ooaddmulsc_cns(C,A,AT,v,B,BT,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inaddmulsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inaddmulsc(A,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_ooaddmulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_ooaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_ooaddmulmonomsc_cns(C,A,AT,v,B,BT,m,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inaddmulmonomsc(A,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oomulscadd_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_oomulscadd(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT, int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oomulscadd_cns(C,A,AT,s,v,B,BT,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inmulscadd(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inmulscadd(A,s,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oomulscaddmulsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_oomulscaddmulsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oomulscaddmulsc_cns(C,A,AT,s,v,B,BT,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inmulscaddmulsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inmulscaddmulsc(A,s,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_oomulscaddmulmonomsc_cns(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_modp_vardnsrecdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_modp_vardnsrecdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_modp_vardnsrecdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_modp_vardnsrecdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_oomulscaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m, ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_oomulscaddmulmonomsc_cns(C,A,AT,s,v,B,BT,m,t,w,args),CT);
}

static void ca_modp_vardnsrecdegspslist_inmulscaddmulmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,int v,ca_modp_vardnsrecdegspsterm_t B,ca_modp_vardnsrecdegspsterm_t BT,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_modp_vardnsrecdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_modp_vardnsrecdegspsterm_inmulscaddmulmonomsc(A,s,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_modp_vardnsrecdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_modp_vardnsrecdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_modp_vardnsrecdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_modp_vardnsrecdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static BOOL ca_modp_vardnsrecdegspslist_oodivsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_modp_vardnsrecdegspsterm_t nextC = ca_modp_vardnsrecdegspsterm_oodivsc_cns(C,A,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_modp_vardnsrecdegspsterm_glue(C,CT);return failed;
}

static void ca_modp_vardnsrecdegspslist_indivsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_indivsc(A,s,args);
}

static BOOL ca_modp_vardnsrecdegspslist_oodivmonomsc(ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m, ca_modp_t s,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_modp_vardnsrecdegspsterm_t nextC = ca_modp_vardnsrecdegspsterm_oodivmonomsc_cns(C,A,m,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_modp_vardnsrecdegspsterm_glue(C,CT);return failed;
}

static void ca_modp_vardnsrecdegspslist_indivmonomsc(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspsterm_args args)
{
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_indivmonomsc(A,m,s,args);
}

static void ca_modp_vardnsrecdegspslist_fileOut(id aFiler,ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT,ca_modp_vardnsrecdegspsterm_args args)
{
    int n;
    n = ca_modp_vardnsrecdegspslist_count(A,AT);
    [aFiler fileOut:&n type:'i'];
    while (A != AT) A = ca_modp_vardnsrecdegspsterm_fileOut(aFiler,A,args);
}

static ca_modp_vardnsrecdegspsterm_t ca_modp_vardnsrecdegspslist_fileIn_cns(id aFiler,ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    while (n--) C = ca_modp_vardnsrecdegspsterm_fileIn_cns(aFiler,C,args);
    return C;
}

static void ca_modp_vardnsrecdegspslist_fileIn(id aFiler,ca_modp_vardnsrecdegspsterm_t C,ca_modp_vardnsrecdegspsterm_t CT,ca_modp_vardnsrecdegspsterm_args args)
{
    ca_modp_vardnsrecdegspsterm_glue(ca_modp_vardnsrecdegspslist_fileIn_cns(aFiler,C,args),CT);
}

static void ca_modp_vardnsrecdegspspol_ix_init(ca_modp_vardnsrecdegspspol_ixt ix,id self)
{
    ca_modp_vardnsrecdegspspol_t v = [self ca_modp_vardnsrecdegspspol_value];
    ix->head = v;ix->ix = v;
}

static void ca_modp_vardnsrecdegspspol_ix_clear(ca_modp_vardnsrecdegspspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_modp_vardnsrecdegspspol_ix_count_aux_aux(ca_modp_vardnsrecdegspsterm_t A,ca_modp_vardnsrecdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static int ca_modp_vardnsrecdegspspol_ix_count_aux(ca_modp_vardnsrecdegspspol_t self)
{
    return ca_modp_vardnsrecdegspspol_ix_count_aux_aux(self->nlink,self);
}

static int ca_modp_vardnsrecdegspspol_ix_count(ca_modp_vardnsrecdegspspol_ixt ix)
{
    /* merde ! can't include entire genlist for ca_modp_vardnsrecdegspslist_count */
    return ca_modp_vardnsrecdegspspol_ix_count_aux(ix->head); 
}

static int ca_modp_vardnsrecdegspspol_ix_isempty_aux(ca_modp_vardnsrecdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static int ca_modp_vardnsrecdegspspol_ix_isempty(ca_modp_vardnsrecdegspspol_ixt ix)
{
    /* can't include entire genpol for ca_modp_vardnsrecdegspspol_isempty */
    return ca_modp_vardnsrecdegspspol_ix_isempty_aux(ix->head); 
}

static int ca_modp_vardnsrecdegspspol_ix_ismatch(ca_modp_vardnsrecdegspspol_ixt ix)
{
    return (ix->ix != NULL && ix->ix != ix->head);
}

static void ca_modp_vardnsrecdegspspol_ix_tofirst(ca_modp_vardnsrecdegspspol_ixt ix)
{
    ix->ix = ix->head;
}

static void ca_modp_vardnsrecdegspspol_ix_tolast(ca_modp_vardnsrecdegspspol_ixt ix)
{
    ix->ix = NULL;
}

static void ca_modp_vardnsrecdegspspol_ix_to(ca_modp_vardnsrecdegspspol_ixt ix,int i)
{
    fprintf(stderr,"memberAt: notImplemented\n");
}

static void ca_modp_vardnsrecdegspspol_ix_next(ca_modp_vardnsrecdegspspol_ixt ix)
{
    if (ix->ix == NULL) {
	/* remain on position just after last term */
    } else {
	ix->ix = ix->ix->nlink; /* next term or head = position after last */
	if (ix->ix == ix->head) ix->ix = NULL;
    }
}

static void ca_modp_vardnsrecdegspspol_ix_prev(ca_modp_vardnsrecdegspspol_ixt ix)
{
    if (ix->ix == ix->head) {
	/* remain on position just before first term */
    } else {
	if (ix->ix == NULL) {
	    ix->ix = ix->head->plink; /* last term or head = position before first */
	} else {
	    ix->ix = ix->ix->plink; /* previous term or head = position before first */
	}
    }
}
static int ca_modp_vardnsrecdegspspol_check(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    assert (self->data == 0xcafebabe);
    return ca_modp_vardnsrecdegspslist_check(self->nlink,self,args);
}

static ca_modp_vardnsrecdegspspol_t ca_modp_vardnsrecdegspspol_id2t(id object)
{
    return [object ca_modp_vardnsrecdegspspol_value];
}

static ca_modp_vardnsrecdegspspol_c *ca_modp_vardnsrecdegspspol_id2c(id object)
{
    return [object ca_modp_vardnsrecdegspspol_reference];
}

static id ca_modp_vardnsrecdegspspol_t2id(id proto,ca_modp_vardnsrecdegspspol_t f)
{
    return [proto ca_modp_vardnsrecdegspspol_value:f];
}

static id ca_modp_vardnsrecdegspspol_c2id(id proto,ca_modp_vardnsrecdegspspol_c *r)
{
    return [proto ca_modp_vardnsrecdegspspol_reference:r];
}

static int ca_modp_vardnsrecdegspspol_isempty(ca_modp_vardnsrecdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static void ca_modp_vardnsrecdegspspol_init(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    self->data = 0xcafebabe;self->nlink = self;self->plink = self;
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
}

static void ca_modp_vardnsrecdegspspol_copy(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_copy(c,a->nlink,a,c,args);
}

static void ca_modp_vardnsrecdegspspol_deepcopy(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_deepcopy(c,a->nlink,a,c,args);
}

static void ca_modp_vardnsrecdegspspol_move(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    if (ca_modp_vardnsrecdegspspol_isempty(a)) {
	ca_modp_vardnsrecdegspspol_init(c,args);
    } else {
	ca_modp_vardnsrecdegspspol_init(c,args);
	ca_modp_vardnsrecdegspsterm_glue(c,a->nlink);ca_modp_vardnsrecdegspsterm_glue(a->plink,c);
    }
    
    a->data = 0;assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_clear(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_clear(self->nlink,self,args);
    self->data = 0;ca_modp_vardnsrecdegspslist_free(self->nlink,self,args);
}

static int ca_modp_vardnsrecdegspspol_count(ca_modp_vardnsrecdegspspol_t self)
{
    return ca_modp_vardnsrecdegspslist_count(self->nlink,self);
}

static unsigned ca_modp_vardnsrecdegspspol_hash(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspslist_hash(a->nlink,a,args);
}

static int ca_modp_vardnsrecdegspspol_iseq(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_t b,ca_modp_vardnsrecdegspspol_args args)
{
    return (a==b)?YES:ca_modp_vardnsrecdegspslist_iseq(a->nlink,a,b->nlink,b,args);
}

static int ca_modp_vardnsrecdegspspol_isone(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    /* if the leading term is one, then it is one */
    return (ca_modp_vardnsrecdegspspol_isempty(a))?0:ca_modp_vardnsrecdegspsterm_isone(a->nlink,args);
}

static int ca_modp_vardnsrecdegspspol_isminusone(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    /* if the leading term is minus one, then it is minus one */
    return (ca_modp_vardnsrecdegspspol_isempty(a))?0:ca_modp_vardnsrecdegspsterm_isminusone(a->nlink,args);
}

static int ca_modp_vardnsrecdegspspol_deg(ca_modp_vardnsrecdegspspol_t self ,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspslist_deg(self->nlink,self,args);
}

static int ca_modp_vardnsrecdegspspol_ord(ca_modp_vardnsrecdegspspol_t self ,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspslist_ord(self->nlink,self,args);
}

static int ca_modp_vardnsrecdegspspol_maxdeg(ca_modp_vardnsrecdegspspol_t self ,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspslist_maxdeg(self->nlink,self,args);
}

static int ca_modp_vardnsrecdegspspol_mindeg(ca_modp_vardnsrecdegspspol_t self ,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspslist_mindeg(self->nlink,self,args);
}

static ca_modp_t ca_modp_vardnsrecdegspspol_scalat(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_ixt ix,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_modp_vardnsrecdegspsterm_scal(ix->ix,args);
}

static ca_unimonom_t ca_modp_vardnsrecdegspspol_monomat(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_ixt ix,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_modp_vardnsrecdegspsterm_monom(ix->ix);
}

static ca_modp_t ca_modp_vardnsrecdegspspol_scalfirst(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_isempty(self) == NO);
    return ca_modp_vardnsrecdegspsterm_scal(self->nlink,args);
}

static ca_unimonom_t ca_modp_vardnsrecdegspspol_monomfirst(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_isempty(self) == NO);
    return ca_modp_vardnsrecdegspsterm_monom(self->nlink);
}

static void ca_modp_vardnsrecdegspspol_remove(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_check(self,args) && ca_modp_vardnsrecdegspspol_isempty(self) == NO);
    ca_modp_vardnsrecdegspsterm_delete(self->nlink,args);
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
}

static void ca_modp_vardnsrecdegspspol_insert(ca_modp_vardnsrecdegspspol_t self,ca_unimonom_t m,ca_modp_t s,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
    ca_modp_vardnsrecdegspslist_insert(self->nlink,self,m,s,args);
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
}

static int ca_modp_vardnsrecdegspspol_iszero(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspspol_isempty(self);
}

static int ca_modp_vardnsrecdegspspol_isop(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_t b,ca_modp_vardnsrecdegspspol_args args)
{
    assert(a != b);return ca_modp_vardnsrecdegspslist_isop(a->nlink,a,b->nlink,b,args);
}

static void ca_modp_vardnsrecdegspspol_ooneg(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_ooneg(c,a->nlink,a,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inneg(ca_modp_vardnsrecdegspspol_t a,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inneg(a->nlink,a,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oodbl(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oodbl(c,a->nlink,a,v,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_indbl(ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_indbl(a->nlink,a,v,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_ooadd(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,int w,ca_modp_vardnsrecdegspspol_args args)
{
    assert(a!=b);
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_ooadd(c,a->nlink,a,v,b->nlink,b,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inadd(ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,int w,ca_modp_vardnsrecdegspspol_args args)
{
    assert(a!=b);
    ca_modp_vardnsrecdegspslist_inadd(a->nlink,a,v,b->nlink,b,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oomulsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oomulsc(c,a->nlink,a,s,+1,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inmulsc(ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inmulsc(a->nlink,a,s,+1,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_ooaddmulsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,ca_modp_t t,int w, ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_ooaddmulsc(c,a->nlink,a,v,b->nlink,b,t,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inaddmulsc(ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inaddmulsc(a->nlink,a,v,b->nlink,b,t,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oomulscadd(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,int w, ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oomulscadd(c,a->nlink,a,s,v,b->nlink,b,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inmulscadd(ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inmulscadd(a->nlink,a,s,v,b->nlink,b,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oomulscaddmulsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,ca_modp_t t,int w, ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oomulscaddmulsc(c,a->nlink,a,s,v,b->nlink,b,t,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inmulscaddmulsc(ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inmulscaddmulsc(a->nlink,a,s,v,b->nlink,b,t,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static BOOL ca_modp_vardnsrecdegspspol_oodivsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,ca_modp_vardnsrecdegspspol_args args)
{
    BOOL failed;
    ca_modp_vardnsrecdegspspol_init(c,args);
    failed = ca_modp_vardnsrecdegspslist_oodivsc(c,a->nlink,a,s,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
    if (failed) ca_modp_vardnsrecdegspspol_clear(c,args);
    return failed;
}

static void ca_modp_vardnsrecdegspspol_indivsc(ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_indivsc(a->nlink,a,s,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oomulmonomsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oomulmonomsc(c,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inmulmonomsc(ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inmulmonomsc(b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(b,args));
}

static void ca_modp_vardnsrecdegspspol_ooaddmulmonomsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w, ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_ooaddmulmonomsc(c,a->nlink,a,v,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inaddmulmonomsc(ca_modp_vardnsrecdegspspol_t a,int v,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inaddmulmonomsc(a->nlink,a,v,b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static void ca_modp_vardnsrecdegspspol_oomulscaddmulmonomsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w, ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(c,args);ca_modp_vardnsrecdegspslist_oomulscaddmulmonomsc(c,a->nlink,a,s,v,b->nlink,b,m,t,w,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
}

static void ca_modp_vardnsrecdegspspol_inmulscaddmulmonomsc(ca_modp_vardnsrecdegspspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_inmulscaddmulmonomsc(a->nlink,a,s,v,b->nlink,b,m,t,w,args);
    assert(ca_modp_vardnsrecdegspspol_check(a,args));
}

static BOOL ca_modp_vardnsrecdegspspol_oodivmonomsc(ca_modp_vardnsrecdegspspol_t c,ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t, ca_modp_vardnsrecdegspspol_args args)
{
    BOOL failed;
    ca_modp_vardnsrecdegspspol_init(c,args);
    failed = ca_modp_vardnsrecdegspslist_oodivmonomsc(c,b->nlink,b,m,t,c,args);
    assert(ca_modp_vardnsrecdegspspol_check(c,args));
    if (failed) ca_modp_vardnsrecdegspspol_clear(c,args);
    return failed;
}

static void ca_modp_vardnsrecdegspspol_indivmonomsc(ca_modp_vardnsrecdegspspol_t b,ca_unimonom_t m,ca_modp_t t,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspslist_indivmonomsc(b->nlink,b,m,t,args);
    assert(ca_modp_vardnsrecdegspspol_check(b,args));
}

static void ca_modp_vardnsrecdegspspol_args_fileOut(id aFiler,ca_modp_vardnsrecdegspspol_args *args)
{
    ca_modp_vardnsrecdegspsterm_args_fileOut(aFiler,args);
}
static void ca_modp_vardnsrecdegspspol_args_fileIn(id aFiler,ca_modp_vardnsrecdegspspol_args *args)
{
    ca_modp_vardnsrecdegspsterm_args_fileIn(aFiler,args);
}

static void ca_modp_vardnsrecdegspspol_fileOut(id aFiler,ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
    ca_modp_vardnsrecdegspslist_fileOut(aFiler,self->nlink,self,args);
}

static void ca_modp_vardnsrecdegspspol_fileIn(id aFiler,ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    ca_modp_vardnsrecdegspspol_init(self,args);ca_modp_vardnsrecdegspslist_fileIn(aFiler,self,self,args);
    assert(ca_modp_vardnsrecdegspspol_check(self,args));
}

static ca_modp_vardnsrecdegspspol_args ca_modp_vardnsrecdegspspol_getargs(id self)
{
    ca_modp_vardnsrecdegspspol_args res;
    res.sargs = ca_modp_getargs([[self termZero] coefficient]);
    res.margs = 0; /* unused */
    return res; 
}

static int ca_modp_vardnsrecdegspspol_expat(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_ixt ix,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspspol_monomat(self,ix,args);
}

static int ca_modp_vardnsrecdegspspol_expfirst(ca_modp_vardnsrecdegspspol_t self,ca_modp_vardnsrecdegspspol_args args)
{
    return ca_modp_vardnsrecdegspspol_monomfirst(self,args);
}
@implementation ca_modp_vardnsrecdegsps_polynomial : ca_vardnsrecdegsps_polynomial

- _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    id aSymbol;
    assert([aCoef isZero] && [aScalar isZero]);
    assert([aCoef class_vardnsrecdegsps_polynomial] == [self class]);
    aScalar  = [aScalar cheapCopy];
    aSymbol  = [[aCltn lastMember] cheapCopy];
    termZero = [CATermClass newScalarZero:aScalar coefficient:aCoef symbol:aSymbol exponent:1];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [CAMonomialClass newScalar:aScalar symbols:aCltn];
    scalarZero   = aScalar;
    symbols      = aCltn;
    
    /* in vardns case, init takes a capacity */
    ca_modp_vardnsrecdegspspol_init(&value,ca_modp_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

+ newScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn];
}

- empty
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    /* in vardns case, init takes a capacity */
    ca_modp_vardnsrecdegspspol_init(&c,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- (int) numTerms
{
    /* seems to take getargs in degdns case */
    assert([self check]);return ca_modp_vardnsrecdegspspol_count(&value);
}

- (ca_modp_vardnsrecdegspspol_t)ca_modp_vardnsrecdegspspol_value
{
    return &value;
}

- ca_modp_vardnsrecdegspspol_value:(ca_modp_vardnsrecdegspspol_t)aValue
{
    self = [self clone];
    ca_modp_vardnsrecdegspspol_copy(&value,aValue,ca_modp_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

- (ca_modp_vardnsrecdegspspol_t)ca_modp_vardnsrecdegspspol_reference
{
    return &value;
}

- ca_modp_vardnsrecdegspspol_reference:(ca_modp_vardnsrecdegspspol_c *)aReference
{
    self = [self clone];
    ca_modp_vardnsrecdegspspol_move(&value,aReference,ca_modp_vardnsrecdegspspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    ca_modp_vardnsrecdegspspol_copy(&c,&value,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- deepCopy
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    ca_modp_vardnsrecdegspspol_deepcopy(&c,&value,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_modp_vardnsrecdegspspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_modp_vardnsrecdegspspol_isone(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_modp_vardnsrecdegspspol_isminusone(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- (unsigned) hash
{
    return ca_modp_vardnsrecdegspspol_hash(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_modp_vardnsrecdegspspol_iseq(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_id2t(b),ca_modp_vardnsrecdegspspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_modp_vardnsrecdegspspol_iseq(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_id2t(b),ca_modp_vardnsrecdegspspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_modp_vardnsrecdegspspol_iszero(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_modp_vardnsrecdegspspol_iszero(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_modp_vardnsrecdegspspol_isop(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_id2t(b),ca_modp_vardnsrecdegspspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_modp_vardnsrecdegspspol_isop(ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_id2t(b),ca_modp_vardnsrecdegspspol_getargs(self));
}

- negate
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    ca_modp_vardnsrecdegspspol_ooneg(&c,ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_modp_vardnsrecdegspspol_inneg(ca_modp_vardnsrecdegspspol_id2c(self),ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    ca_modp_vardnsrecdegspspol_oodbl(&c,ca_modp_vardnsrecdegspspol_id2t(self),v,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_modp_vardnsrecdegspspol_indbl(ca_modp_vardnsrecdegspspol_id2c(self),v,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_modp_vardnsrecdegspspol_c c;assert([self check]);
    ca_modp_vardnsrecdegspspol_oodbl(&c,ca_modp_vardnsrecdegspspol_id2t(self),+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_modp_vardnsrecdegspspol_indbl(ca_modp_vardnsrecdegspspol_id2c(self),+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_modp_vardnsrecdegspspol_ooadd(&c,ca_modp_vardnsrecdegspspol_id2t(self),v,ca_modp_vardnsrecdegspspol_id2t(b),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_modp_vardnsrecdegspspol_inadd(ca_modp_vardnsrecdegspspol_id2c(self),v,ca_modp_vardnsrecdegspspol_id2t(b),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_modp_vardnsrecdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsrecdegspspol_ooadd(&c,ca_modp_vardnsrecdegspspol_id2t(self),+1,ca_modp_vardnsrecdegspspol_id2t(b),+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_modp_vardnsrecdegspspol_inadd(ca_modp_vardnsrecdegspspol_id2c(self),+1,ca_modp_vardnsrecdegspspol_id2t(b),+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_modp_vardnsrecdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsrecdegspspol_ooadd(&c,ca_modp_vardnsrecdegspspol_id2t(self),+1,ca_modp_vardnsrecdegspspol_id2t(b),-1,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_modp_vardnsrecdegspspol_inadd(ca_modp_vardnsrecdegspspol_id2c(self),+1,ca_modp_vardnsrecdegspspol_id2t(b),-1,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_modp_vardnsrecdegspspol_deg(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- (int) lastDegree
{
    return ca_modp_vardnsrecdegspspol_ord(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- (int) maxDegree
{
    return ca_modp_vardnsrecdegspspol_maxdeg(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- (int) minDegree
{
    return ca_modp_vardnsrecdegspspol_mindeg(&value,ca_modp_vardnsrecdegspspol_getargs(self));
}

- asScalar
{
    if ([self numMembers] == 1) {
	id asScalar,lastTerm,aSequence = [self eachMember];
	asScalar = (lastTerm = [aSequence lastMember])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self numMembers] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachMember];
	asSymbol = (lastTerm = [aSequence lastMember])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols numMembers] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);
    assert([termZero isZero]&&[termZero exponent]==1&&[termZero symbol]!=nil);
    assert([[symbols lastMember] isEqual:[termZero symbol]]);

    ca_modp_vardnsrecdegspspol_check(&value,ca_modp_vardnsrecdegspspol_getargs(self));
    return self;
}

- clear
{
    ca_modp_vardnsrecdegspspol_clear(&value,ca_modp_vardnsrecdegspspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- eachTerm
{
    id aSequence = [ca_modp_vardnsrecdegspspol_sequence newContent:[self cheapCopy]];
    return [CASequenceClass newCarrier:aSequence];
}

- elt_ca_modp_vardnsrecdegspspol_ix:(ca_modp_vardnsrecdegspspol_ixt)ix
{
    if (ca_modp_vardnsrecdegspspol_ix_ismatch(ix)) {
	id c,s;int e;ca_modp_t t;
	t = ca_modp_vardnsrecdegspspol_scalat(&value,ix,ca_modp_vardnsrecdegspspol_getargs(self));
	e = ca_modp_vardnsrecdegspspol_expat(&value,ix,ca_modp_vardnsrecdegspspol_getargs(self));
	c = ca_modp_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	return [termZero coefficient:c symbol:s exponent:e];
    } else {
	return nil;
    }
}
- removeTerm
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id c,s;int e;ca_modp_t t;
	[self checkRefCountOne];assert([self check]);
	
	t = ca_modp_vardnsrecdegspspol_scalfirst(&value,ca_modp_vardnsrecdegspspol_getargs(self));
	e = ca_modp_vardnsrecdegspspol_expfirst(&value,ca_modp_vardnsrecdegspspol_getargs(self));
	c = ca_modp_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	lt = [termZero coefficient:c symbol:s exponent:e];
	
	ca_modp_vardnsrecdegspspol_remove(&value,ca_modp_vardnsrecdegspspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertTerm:aTerm
{
    if ([aTerm isZero])  {
	[aTerm free];return self;
    } else {
	ca_modp_c c;ca_modp_t r;int e;
	assert([aTerm check]);[self checkRefCountOne];assert([self check]);
	r = ca_modp_id2ref([aTerm coefficient],&c);e = [aTerm exponent];
	ca_modp_vardnsrecdegspspol_insert(&value,e,r,ca_modp_vardnsrecdegspspol_getargs(self));
	assert([aTerm check]);[aTerm free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyCoefficient:s
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c sc;
    ca_modp_vardnsrecdegspspol_oomulsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- multiplySelfCoefficient:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyCoefficient:s]);
    ca_modp_vardnsrecdegspspol_inmulsc(ca_modp_vardnsrecdegspspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- divideCoefficient:s
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c sc;
    BOOL failed = ca_modp_vardnsrecdegspspol_oodivsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegspspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- divideSelfCoefficient:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideCoefficient:s]);
    ca_modp_vardnsrecdegspspol_indivsc(ca_modp_vardnsrecdegspspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c bc;
    ca_modp_vardnsrecdegspspol_ooaddmulsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),v,ca_modp_vardnsrecdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_modp_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyCoefficient:b:w]);
    ca_modp_vardnsrecdegspspol_inaddmulsc(ca_modp_vardnsrecdegspspol_id2c(self),v,ca_modp_vardnsrecdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c ac;
    ca_modp_vardnsrecdegspspol_oomulscadd(&c,ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegspspol_id2t(B),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B:(int)w
{
    ca_modp_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B:w]);
    ca_modp_vardnsrecdegspspol_inmulscadd(ca_modp_vardnsrecdegspspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegspspol_id2t(B),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c ac,bc;
    ca_modp_vardnsrecdegspspol_oomulscaddmulsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_modp_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyCoefficient:b:w]);
    ca_modp_vardnsrecdegspspol_inmulscaddmulsc(ca_modp_vardnsrecdegspspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegspspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}
- multiplyTerm:aTerm
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegspspol_oomulmonomsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),[aTerm exponent],s,+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- multiplySelfTerm:aTerm
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyTerm:aTerm]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegspspol_inmulmonomsc(ca_modp_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,+1,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyTerm:aTerm:(int)v
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegspspol_oomulmonomsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),[aTerm exponent],s,v,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfTerm:aTerm:(int)v
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self _multiplyTerm:aTerm:v]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegspspol_inmulmonomsc(ca_modp_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,v,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- divideTerm:aTerm
{
    int e;
    ca_modp_vardnsrecdegspspol_c c;
    BOOL failed;
    ca_modp_t s;
    ca_modp_c sc;
    
    e = [aTerm exponent];
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    failed=ca_modp_vardnsrecdegspspol_oodivmonomsc(&c,ca_modp_vardnsrecdegspspol_id2t(self),e,s,ca_modp_vardnsrecdegspspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- divideSelfTerm:aTerm
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideTerm:aTerm]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegspspol_indivmonomsc(ca_modp_vardnsrecdegspspol_id2c(self),[aTerm exponent],s,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyTerm:b:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c bc;
    ca_modp_t bs;
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegspspol_ooaddmulmonomsc(&c,ca_modp_vardnsrecdegspspol_id2c(self),v,ca_modp_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyTerm:b:(int)w
{
    ca_modp_c bc;
    ca_modp_t bs;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyTerm:b:w]);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegspspol_inaddmulmonomsc(ca_modp_vardnsrecdegspspol_id2c(self),v,ca_modp_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_modp_vardnsrecdegspspol_c c;
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    as = ca_modp_id2ref(a,&ac);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegspspol_oomulscaddmulmonomsc(&c,ca_modp_vardnsrecdegspspol_id2c(self),as,v,ca_modp_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegspspol_getargs(self));
    return ca_modp_vardnsrecdegspspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyTerm:b:w]);
    as = ca_modp_id2ref(a,&ac);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegspspol_inmulscaddmulmonomsc(ca_modp_vardnsrecdegspspol_id2c(self),as,v,ca_modp_vardnsrecdegspspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegspspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_modp_vardnsrecdegspspol_args args;
    [super fileOutOn:aFiler];
    args = ca_modp_vardnsrecdegspspol_getargs(self); /* can't do this in fileIn case */
    ca_modp_vardnsrecdegspspol_args_fileOut(aFiler,&args);
    ca_modp_vardnsrecdegspspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_modp_vardnsrecdegspspol_args args;
    [super fileInFrom:aFiler];
    ca_modp_vardnsrecdegspspol_args_fileIn(aFiler,&args);
    ca_modp_vardnsrecdegspspol_fileIn(aFiler,&value,args);
    return self;
}

@end

