#include "cakit.h"

static ca_modp_args ca_modp_getargs(id self)
{
   ca_modp_args args;args.p = [self modulus];return args;
}

static int ca_modp_size(ca_modp_args args)
{
    return 1;				/* pointer increment */
}

static int ca_modp_bsize(ca_modp_args args)
{
    return sizeof(ca_modp_c);		/* bytes */
}

static ca_modp_t ca_modp_id2t(id object)
{
    return [object ca_modp_value];	/* read-only */
}

static ca_modp_t ca_modp_id2ref(id object,ca_modp_c *ref)
{
    return [object ca_modp_value];	/* read-only, scratch space "ref" */
}

static ca_modp_c *ca_modp_id2c(id object)
{
    return [object ca_modp_reference];	/* write for inplace arithmetic */
}

static id ca_modp_t2id(id proto,ca_modp_t f)
{
    return [proto ca_modp_value:f];	/* new object containing copy of "f" */
}

static id ca_modp_c2id(id proto,ca_modp_c* c)
{
    return [proto ca_modp_reference:c];/* new object containing "c" */
}

static ca_modp_t ca_modp_fromintvalue(int i,ca_modp_args args)
{
    ca_modp_t red;
    
    if (i==0) return 0;
    if (i <0) {
	red = -i;
	red %= args.p;	/* it's dangerous to % int & unsigned */
	red = (red)?args.p-red:0;
    } else {
	red  = i;
	red %= args.p;	/* dito */
    }
    
    return red;
}

static int ca_modp_check(ca_modp_t self,ca_modp_args args)
{
    assert(self < args.p);return 1;
}

static ca_modp_t ca_modp_zero(ca_modp_args args)
{
    return 0;
}

static ca_modp_t ca_modp_one(ca_modp_args args)
{
    return 1;
}

static unsigned ca_modp_hash(ca_modp_t a,ca_modp_args args)
{
    return a;
}

static int ca_modp_iseq(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a == b);
}

static int ca_modp_isop(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a==0)?(b==0):(a+b==args.p);
}

static int ca_modp_iszero(ca_modp_t a,ca_modp_args args)
{
    return a==0;
}

static int ca_modp_isone(ca_modp_t a,ca_modp_args args)
{
    return a==1;
}

static int ca_modp_isminusone(ca_modp_t a,ca_modp_args args)
{
    return (a+1)==args.p;
}

static ca_modp_t ca_modp_neg(ca_modp_t a,ca_modp_args args)
{
    assert(ca_modp_check(a,args));return (a)?args.p-a:0;
}

static ca_modp_t ca_modp_dbl(ca_modp_t a,ca_modp_args args)
{
    ca_modp_t c = a+a;
    assert(args.p != 2 && ca_modp_check(a,args));
    if (c) return (c<args.p)?c:c-args.p; else return 0;
}

static ca_modp_t ca_modp_add(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    ca_modp_t c = a+b;
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return ((c>=args.p)?c-args.p:c); /* compiles better than c<p?c:c-p */
}

static ca_modp_t ca_modp_sub(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a>=b)?(a-b):(args.p-(b-a));
}

static ca_modp_t ca_modp_mul(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    unsigned c = a * b;		/* limitation to 16bits to be able to do this
				 * russian peasant 50% slower
                                 */
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_sqr(ca_modp_t a,ca_modp_args args)
{
    unsigned c = a * a;
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_inv(ca_modp_t x,ca_modp_args args)
{
    unsigned q,s;
    int u1,u2,us;  /* intermediate values can be negative */
    ca_modp_t a,b,inv;
    
    assert(x != 0 && x != 1 && x+1 != args.p);
    
    a = x;b = args.p;u1 = 1;u2 = 0;
    while (b) {
	s  = a%b;q = a/b;
	a  = b;b = s;
	u1 -= q*u2;			
	us  = u1;u1 = u2;u2 = us;
    }
    
    inv = ((u1<0)?u1+args.p:u1); /* keep result positive */
    
    assert(ca_modp_check(inv,args));
    assert(ca_modp_isone(ca_modp_mul(x,inv,args),args));
    return inv;
}

static ca_modp_t ca_modp_div(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(b!=0);
    return ca_modp_mul(a,ca_modp_inv(b,args),args);
}

static ca_modp_t ca_modp_pow(ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_t o;
    
    if (n==0) return ca_modp_one(args);
    if (n==1) return a;
    if (n==2) return ca_modp_sqr(a,args);
    
    o = ca_modp_one(args);
    while (n>1) {
	if (n&1) o = ca_modp_mul(o,a,args);
	a = ca_modp_sqr(a,args);
	n >>= 1;
    }
    
    return ca_modp_mul(o,a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_copy_val.fab should not be included"
#endif

static void ca_modp_clear(ca_modp_c *c,ca_modp_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_modp_t)0xcafebabe;
    #endif
}

static void ca_modp_copy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_deepcopy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_move(ca_modp_c *c,ca_modp_c *a,ca_modp_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_modp_t)0xcafebabe;
    #endif
}

#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static void ca_modp_oozero(ca_modp_c *c,ca_modp_args args)
{
    *c = ca_modp_zero(args);
}

static void ca_modp_inzero(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_zero(args);
}

static void ca_modp_inneg(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_neg(*a,args);
}

static void ca_modp_ooneg(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_neg(a,args);
}

static ca_modp_t ca_modp_dblv(ca_modp_t a,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_dbl(a,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_indbl(ca_modp_c *a,int v,ca_modp_args args)
{
    *a = ca_modp_dblv(*a,v,args);
}

static void ca_modp_oodbl(ca_modp_c *c,ca_modp_t a,int v,ca_modp_args args)
{
    *c = ca_modp_dblv(a,v,args);
}

static ca_modp_t ca_modp_addvw(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    ca_modp_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_modp_add(a,b,args);
	} else {
	    c = ca_modp_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_modp_sub(b,a,args);
	} else {
	    c = ca_modp_neg(ca_modp_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_modp_inadd(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addvw(*a,v,b,w,args);
}

static void ca_modp_ooadd(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addvw(a,v,b,w,args);
}
#if !ca_modp_isvalue
#error "ca_modp_mul_val.fab should not be included"
#endif

static void ca_modp_insqr(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_sqr(*a,args);
}

static void ca_modp_oosqr(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_sqr(a,args);
}

static void ca_modp_inmul(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_mul(*a,b,args);
}

static void ca_modp_oomul(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_mul(a,b,args);
}
#if !ca_modp_isvalue
#error "ca_modp_inv_val.fab should not be included"
#endif

static void ca_modp_ininv(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_inv(*a,args);
}

static void ca_modp_ooinv(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_inv(a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_div_val.fab should not be included"
#endif

static void ca_modp_indiv(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_div(*a,b,args);
}

static BOOL ca_modp_oodiv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_div(a,b,args);return SUCCESS;
}
#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static ca_modp_t ca_modp_mulv(ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_mul(a,b,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_inmulv(ca_modp_c *a,ca_modp_t b,int v,ca_modp_args args)
{
    *a = ca_modp_mulv(*a,b,v,args);
}

static void ca_modp_oomulv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    *c = ca_modp_mulv(a,b,v,args);
}

static ca_modp_t ca_modp_addmul(ca_modp_t a,int v,ca_modp_t b,ca_modp_t s,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_mul(b,s,args),w,args);
}

static ca_modp_t ca_modp_addsqr(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_sqr(b,args),w,args);
}

static ca_modp_t ca_modp_muladd(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,b,w,args);
}

static ca_modp_t ca_modp_muladdmul(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,ca_modp_mul(b,t,args),w,args);
}

static void ca_modp_inaddmul(ca_modp_c *a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_addmul(*a,v,b,t,w,args);
}

static void ca_modp_ooaddmul(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_addmul(a,v,b,t,w,args);
}

static void ca_modp_inaddsqr(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addsqr(*a,v,b,w,args);
}

static void ca_modp_ooaddsqr(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addsqr(a,v,b,w,args);
}

static void ca_modp_inmuladd(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_muladd(*a,s,v,b,w,args);
}

static void ca_modp_oomuladd(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_muladd(a,s,v,b,w,args);
}

static void ca_modp_inmuladdmul(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_modp_oomuladdmul(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_muladdmul(a,s,v,b,t,w,args);
}
static void ca_modp_args_fileOut(id aFiler,ca_modp_args *args) {}
static void ca_modp_args_fileIn(id aFiler,ca_modp_args *args) {}

static void ca_modp_fileOut(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileOut:c type:'S'];
}

static void ca_modp_fileIn(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileIn:c type:'S'];
}


/*
 * Computer Algebra Kit (c) 1993,95 by Comp.Alg.Objects.  All Rights Reserved.
 */

#if ca_modp_pervalue
static ca_modp_t ca_modpseq_tval(ca_modpseq_t self)
{
    return *self;
}
static ca_modp_c *ca_modpseq_cval(ca_modpseq_t self)
{
    return self;
}
#else
static ca_modp_t ca_modpseq_tval(ca_modpseq_t self)
{
    return self;
}
static ca_modp_c *ca_modpseq_cval(ca_modpseq_t self)
{
    return self;
}
#endif

static int ca_modpseq_check(ca_modpseq_t self,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_check(ca_modpseq_tval(self),args);
	self += ca_modp_size(args);
    }
    return 1;
}

static ca_modpseq_t ca_modpseq_alloc(int n,ca_modp_args args)
{
    return (ca_modpseq_t)OC_Malloc(n * ca_modp_bsize(args));
}

static ca_modpseq_t ca_modpseq_realloc(ca_modpseq_t self,int n,ca_modp_args args)
{
    return (ca_modpseq_t)OC_Realloc(self,n * ca_modp_bsize(args));
}

static ca_modpseq_t ca_modpseq_free(ca_modpseq_t self,ca_modp_args args)
{
    OC_Free(self);return NULL;
}

static void ca_modpseq_oozero(ca_modpseq_t c,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_oozero(ca_modpseq_cval(c),args);
	c += ca_modp_size(args);
    }
}

static void ca_modpseq_copy(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_copy(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_deepcopy(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_deepcopy(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

#if ca_modp_isvalue
static void ca_modpseq_clear(ca_modpseq_t self,int n,ca_modp_args args)
{
    /* nothing to clear */
}
#else
static void ca_modpseq_clear(ca_modpseq_t self,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_clear(self,args);
	self+=ca_modp_size(args);
    }
}
#endif /* ca_modp_isvalue */

static int ca_modpseq_iseq(ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_modp_iseq(ca_modpseq_tval(a),ca_modpseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_modp_size(args);b+=ca_modp_size(args);
	}
    }
    return 1;
}

static unsigned ca_modpseq_hash(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    return (n)?ca_modp_hash(ca_modpseq_tval(a),args):0;
}

static void ca_modpseq_place(ca_modpseq_t self,ca_modp_t a,int i,ca_modp_args args)
{
    ca_modp_c *c = ca_modpseq_cval(self + i * ca_modp_size(args));
    ca_modp_clear(c,args);ca_modp_copy(c,a,args);
}

static int ca_modpseq_insertlast(ca_modpseq_t self,ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_c *c;
    assert(n >= 0);
    c = ca_modpseq_cval(self + n * ca_modp_size(args));
    ca_modp_copy(c,a,args);return ++n;
}

static int ca_modpseq_insertfirst(ca_modpseq_t self,ca_modp_t a,int n,ca_modp_args args)
{
    int j;
    assert(n > 0);
    self += n * ca_modp_size(args);
    j=n;while(j--) {
	ca_modpseq_t t;
	t = self - ca_modp_size(args);
	ca_modp_move(ca_modpseq_cval(self),ca_modpseq_cval(t),args);
	self = t;
    }
    ca_modp_copy(ca_modpseq_cval(self),a,args);
    return ++n;
}

static int ca_modpseq_insertat(ca_modpseq_t self,ca_modp_t a,int i,int n,ca_modp_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_modp_size(args);
    return i + ca_modpseq_insertfirst(self,a,n-i,args);
}

static int ca_modpseq_freelast(ca_modpseq_t self,int n,ca_modp_args args)
{
    ca_modp_c *c;assert(0 < n);
    --n;c = ca_modpseq_cval(self + n * ca_modp_size(args));
    ca_modp_clear(c,args);return n;
}

static int ca_modpseq_freefirst(ca_modpseq_t self,int n,ca_modp_args args)
{
    int j;
    assert(0 < n);
    ca_modp_clear(ca_modpseq_cval(self),args);
    --n;j=n;while(j--) {
	ca_modpseq_t t;
	t = self+ca_modp_size(args);
	ca_modp_move(ca_modpseq_cval(self),ca_modpseq_cval(t),args);
	self = t;
    }
    return n;
}

static int ca_modpseq_freeat(ca_modpseq_t self,int i,int n,ca_modp_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_modp_size(args);
    return i + ca_modpseq_freefirst(self,n-i,args);
}

static ca_modp_t ca_modpseq_at(ca_modpseq_t self,int i,ca_modp_args args)
{
    ca_modp_t res;assert(0 <= i);
    res = ca_modpseq_tval(self + i * ca_modp_size(args));
    assert(ca_modp_check(res,args));
    return res;
}

static ca_modp_t ca_modpseq_last(ca_modpseq_t self,int n,ca_modp_args args)
{
    return ca_modpseq_at(self,n-1,args);
}

static ca_modp_c *ca_modpseq_cvalat(ca_modpseq_t self,int i,ca_modp_args args)
{
    assert(0 <= i);return ca_modpseq_cval(self + i * ca_modp_size(args));
}

static int ca_modpseq_iszero(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	if (ca_modp_iszero(ca_modpseq_tval(a),args)) {
	    a+=ca_modp_size(args);
	} else {
	    return 0;
	}
    }
    return 1;
}

static int ca_modpseq_isop(ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_modp_isop(ca_modpseq_tval(a),ca_modpseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_modp_size(args);b+=ca_modp_size(args);
	}
    }
    return 1;
}

static void ca_modpseq_ooneg(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_ooneg(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inneg(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_inneg(ca_modpseq_cval(a),args);
	a += ca_modp_size(args);
    }
}

static void ca_modpseq_oonegv(ca_modpseq_t c,ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(v*v == 1);if (v == -1) ca_modpseq_ooneg(c,a,n,args);
    else ca_modpseq_copy(c,a,n,args);
}

static void ca_modpseq_innegv(ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(v*v == 1);if (v == -1) ca_modpseq_inneg(a,n,args);
}

static void ca_modpseq_oodbl(ca_modpseq_t c,ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_oodbl(ca_modpseq_cval(c),ca_modpseq_tval(a),v,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_indbl(ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_indbl(ca_modpseq_cval(a),v,args);
	a += ca_modp_size(args);
    }
}

static void ca_modpseq_ooadd(ca_modpseq_t c,ca_modpseq_t a,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_ooadd(ca_modpseq_cval(c),ca_modpseq_tval(a),v,ca_modpseq_tval(b),w,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inadd(ca_modpseq_t a,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inadd(ca_modpseq_cval(a),v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_oomul(ca_modpseq_cval(c),ca_modpseq_tval(a),s,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulsc(ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_inmul(ca_modpseq_cval(a),s,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscv(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_oomulv(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscv(ca_modpseq_t a,ca_modp_t s,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_inmulv(ca_modpseq_cval(a),s,v,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_ooaddmulsc(ca_modpseq_t c,ca_modpseq_t a,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_ooaddmul(ca_modpseq_cval(c),ca_modpseq_tval(a),v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscadd(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_oomuladd(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscaddmulsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_oomuladdmul(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_inaddmulsc(ca_modpseq_t a,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inaddmul(ca_modpseq_cval(a),v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscadd(ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inmuladd(ca_modpseq_cval(a),s,v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscaddmulsc(ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inmuladdmul(ca_modpseq_cval(a),s,v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static BOOL ca_modpseq_oodivsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    int i;
    BOOL failed;
    ca_modpseq_t orgc = c;
    assert(n >= 0);
   
    for(i=0;i<n;i++) {
    
	failed = ca_modp_oodiv(ca_modpseq_cval(c),ca_modpseq_tval(a),s,args);
	
	if (failed) {
	    ca_modpseq_clear(orgc,i - 1,args);
	    return FAILURE;
	} else {
	    c+=ca_modp_size(args);a+=ca_modp_size(args);
	}
    }
    
    return SUCCESS;
}

static void ca_modpseq_indivsc(ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_indiv(ca_modpseq_cval(a),s,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_oodotsqr(ca_modp_c *c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    ca_modp_oozero(c,args);
    while (n--) {
	ca_modp_inaddsqr(c,+1,ca_modpseq_tval(a),+1,args);a += ca_modp_size(args);
    }
}

static void ca_modpseq_oodotmul(ca_modp_c *c,ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0);
    ca_modp_oozero(c,args);
    while (n--) {
	ca_modp_inaddmul(c,+1,ca_modpseq_tval(a),ca_modpseq_tval(b),+1,args);
	a += ca_modp_size(args);b += ca_modp_size(args);
    }
}

static void ca_modpseq_args_fileOut(id aFiler,ca_modp_args *args)
{
    ca_modp_args_fileOut(aFiler,args);
}
static void ca_modpseq_args_fileIn(id aFiler,ca_modp_args *args)
{
    ca_modp_args_fileIn(aFiler,args);
}

static void ca_modpseq_fileOut(id aFiler,ca_modpseq_t self,int n,ca_modp_args args)
{
    assert(n>=0);
    while (n--) {
	ca_modp_fileOut(aFiler,ca_modpseq_cval(self),args);
	self += ca_modp_size(args);
    }
}

static void ca_modpseq_fileIn(id aFiler,ca_modpseq_t self,int n,ca_modp_args args)
{
    assert(n>=0);
    while (n--) {
	ca_modp_fileIn(aFiler,ca_modpseq_cval(self),args);
	self += ca_modp_size(args);
    }
}

static ca_modpvec_args ca_modpvec_getargs(id self)
{
    ca_modpvec_args args;
    args.sargs = ca_modp_getargs([self scalarZero]);
    return args;
}

static int ca_modpvec_check(ca_modpvec_t self,ca_modpvec_args args)
{
    assert(0 <= self->n && self->n <= self->c);
    ca_modpseq_check(self->ptr,self->n,args.sargs);return 1;
}

static ca_modpvec_t ca_modpvec_id2t(id object)
{
    return [object ca_modpvec_value];
}

static ca_modpvec_c *ca_modpvec_id2c(id object)
{
    return [object ca_modpvec_reference];
}

static id ca_modpvec_t2id(id proto,ca_modpvec_t f)
{
    return [proto ca_modpvec_value:f];
}

static id ca_modpvec_c2id(id proto,ca_modpvec_c *r)
{
    return [proto ca_modpvec_reference:r];
}

static void ca_modpvec_init(ca_modpvec_t self,int n,ca_modpvec_args args)
{
    self->n   = n;
    self->c   = n;
    self->ptr = ca_modpseq_alloc(n,args.sargs);
}

static void ca_modpvec_initcount(ca_modpvec_t c,int n,ca_modpvec_args args)
{
    ca_modpvec_init(c,n,args);
    ca_modpseq_oozero(c->ptr,n,args.sargs);
}

static void ca_modpvec_initcapacity(ca_modpvec_t self,int n,ca_modpvec_args args)
{
    self->n   = 0;
    self->c   = n;
    self->ptr = ca_modpseq_alloc(n,args.sargs);
}

static void ca_modpvec_copy(ca_modpvec_t c,ca_modpvec_t a,ca_modpvec_args args)
{
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_copy(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_modpvec_deepcopy(ca_modpvec_t c,ca_modpvec_t a,ca_modpvec_args args)
{
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_deepcopy(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_modpvec_move(ca_modpvec_t c,ca_modpvec_t a,ca_modpvec_args args)
{
    c->n   = a->n;
    c->c   = a->c;
    c->ptr = a->ptr;a->ptr = NULL;
}

static void ca_modpvec_destroy(ca_modpvec_t self,ca_modpvec_args args)
{
    /*
     * clearing memory allocated by ca_modpvec_init after a failed operation
     * can't do a ca_modpseq_clear in that case as the handler doesnt know
     * what elements have been allocated
     */
     
    self->ptr = ca_modpseq_free(self->ptr,args.sargs);
}

static void ca_modpvec_clear(ca_modpvec_t self,ca_modpvec_args args)
{
    ca_modpseq_clear(self->ptr,self->n,args.sargs);
    ca_modpvec_destroy(self,args);
}

static int ca_modpvec_count(ca_modpvec_t self)
{
    return self->n;
}

static unsigned ca_modpvec_hash(ca_modpvec_t a,ca_modpvec_args args)
{
    return ca_modpseq_hash(a->ptr,a->n,args.sargs);
}

static int ca_modpvec_iseq(ca_modpvec_t a,ca_modpvec_t b,ca_modpvec_args args)
{
    assert(a != b);
    return (a->n == b->n && ca_modpseq_iseq(a->ptr,b->ptr,a->n,args.sargs));
}

static void ca_modpvec_expand(ca_modpvec_t self,ca_modpvec_args args)
{
    assert(self->n == self->c);
    self->c   = self->n + self->c + 1;
    self->ptr = ca_modpseq_realloc(self->ptr,self->c,args.sargs);
    assert(self->n < self->c);
}

static void ca_modpvec_insertfirst(ca_modpvec_t self,ca_modp_t a,ca_modpvec_args args)
{
    assert(self->n <= self->c);
    if (self->n == self->c) ca_modpvec_expand(self,args);
    self->n = ca_modpseq_insertlast(self->ptr,a,self->n,args.sargs);
}

static void ca_modpvec_insertat(ca_modpvec_t self,ca_modp_t a,int i,ca_modpvec_args args)
{
    assert(0 < i && i <= self->n && self->n <= self->c);
    if (self->n == self->c) ca_modpvec_expand(self,args);
    
    if (i == self->n) {
    self->n = ca_modpseq_insertfirst(self->ptr,a,self->n,args.sargs);
    } else {
    self->n = ca_modpseq_insertat(self->ptr,a,self->n-i-1,self->n,args.sargs);
    }
}

static void ca_modpvec_place(ca_modpvec_t self,ca_modp_t a,int i,ca_modpvec_args args)
{
    assert(0 <= i && i < self->n);
    ca_modpseq_place(self->ptr,a,self->n - i - 1,args.sargs);
}

static void ca_modpvec_freefirst(ca_modpvec_t self,ca_modpvec_args args)
{
    assert(self->n > 0);
    self->n = ca_modpseq_freelast(self->ptr,self->n,args.sargs);
}

static void ca_modpvec_freeat(ca_modpvec_t self,int i,ca_modpvec_args args)
{
    assert(0 < i && i < self->n);
    self->n = ca_modpseq_freeat(self->ptr,self->n-i-1,self->n,args.sargs);
}


static ca_modp_t ca_modpvec_at(ca_modpvec_t self,int ix,ca_modpvec_args args)
{
    assert(0 <= ix && ix < self->n);
    return ca_modpseq_at(self->ptr,self->n - ix - 1,args.sargs);
}

static ca_modp_t ca_modpvec_first(ca_modpvec_t self,ca_modpvec_args args)
{
    assert(self->n > 0);
    return ca_modpseq_last(self->ptr,self->n,args.sargs);
}

static ca_modp_t ca_modpvec_elt(ca_modpvec_t self,ca_modpvec_ixt ix,ca_modpvec_args args)
{
    assert(ix->n == self->n);
    return ca_modpvec_at(self,ix->i,args);
}

static int ca_modpvec_iszero(ca_modpvec_t self,ca_modpvec_args args)
{
    return ca_modpseq_iszero(self->ptr,self->n,args.sargs);
}

static int ca_modpvec_isop(ca_modpvec_t a,ca_modpvec_t b,ca_modpvec_args args)
{
    assert(a != b);
    return (a->n == b->n && ca_modpseq_isop(a->ptr,b->ptr,a->n,args.sargs));
}

static void ca_modpvec_ooneg(ca_modpvec_t c,ca_modpvec_t a,ca_modpvec_args args)
{
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_ooneg(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_modpvec_inneg(ca_modpvec_t a,ca_modpvec_args args)
{
    ca_modpseq_inneg(a->ptr,a->n,args.sargs);
}

static void ca_modpvec_oodbl(ca_modpvec_t c,ca_modpvec_t a,int v,ca_modpvec_args args)
{
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_oodbl(c->ptr,a->ptr,v,a->n,args.sargs);
}

static void ca_modpvec_indbl(ca_modpvec_t a,int v,ca_modpvec_args args)
{
    ca_modpseq_indbl(a->ptr,v,a->n,args.sargs);
}

static void ca_modpvec_checkcount(int a,int b)
{
    if (a != b) { fprintf(stderr,"Vector : Dimensions Not Equal\n");abort(); }
}

static void ca_modpvec_ooadd(ca_modpvec_t c,ca_modpvec_t a,int v,ca_modpvec_t b,int w,ca_modpvec_args args)
{
    assert(a!=b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_ooadd(c->ptr,a->ptr,v,b->ptr,w,a->n,args.sargs);
}

static void ca_modpvec_inadd(ca_modpvec_t a,int v,ca_modpvec_t b,int w,ca_modpvec_args args)
{
    assert(a!=b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpseq_inadd(a->ptr,v,b->ptr,w,a->n,args.sargs);
}

static void ca_modpvec_oomulsc(ca_modpvec_t c,ca_modpvec_t a,ca_modp_t s,ca_modpvec_args args)
{
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_oomulsc(c->ptr,a->ptr,s,a->n,args.sargs);
}

static void ca_modpvec_inmulsc(ca_modpvec_t a,ca_modp_t s,ca_modpvec_args args)
{
    ca_modpseq_inmulsc(a->ptr,s,a->n,args.sargs);
}

static void ca_modpvec_ooaddmulsc(ca_modpvec_t c,ca_modpvec_t a,int v,ca_modpvec_t b,ca_modp_t s,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_ooaddmulsc(c->ptr,a->ptr,v,b->ptr,s,w,a->n,args.sargs);
}

static void ca_modpvec_inaddmulsc(ca_modpvec_t a,int v,ca_modpvec_t b,ca_modp_t s,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpseq_inaddmulsc(a->ptr,v,b->ptr,s,w,a->n,args.sargs);
}

static void ca_modpvec_oomulscadd(ca_modpvec_t c,ca_modpvec_t a,ca_modp_t s,int v,ca_modpvec_t b,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_oomulscadd(c->ptr,a->ptr,s,v,b->ptr,w,a->n,args.sargs);
}

static void ca_modpvec_inmulscadd(ca_modpvec_t a,ca_modp_t s,int v,ca_modpvec_t b,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpseq_inmulscadd(a->ptr,s,v,b->ptr,w,a->n,args.sargs);
}

static void ca_modpvec_oomulscaddmulsc(ca_modpvec_t c,ca_modpvec_t a,ca_modp_t s,int v,ca_modpvec_t b,ca_modp_t t,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpvec_init(c,a->n,args);
    ca_modpseq_oomulscaddmulsc(c->ptr,a->ptr,s,v,b->ptr,t,w,a->n,args.sargs);
}

static void ca_modpvec_inmulscaddmulsc(ca_modpvec_t a,ca_modp_t s,int v,ca_modpvec_t b,ca_modp_t t,int w,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpseq_inmulscaddmulsc(a->ptr,s,v,b->ptr,t,w,a->n,args.sargs);
}

static BOOL ca_modpvec_oodivsc(ca_modpvec_t c,ca_modpvec_t a,ca_modp_t s,ca_modpvec_args args)
{
    BOOL failed;
    
    ca_modpvec_init(c,a->n,args);
    failed = ca_modpseq_oodivsc(c->ptr,a->ptr,s,a->n,args.sargs);
    if (failed) ca_modpvec_destroy(c,args);
    
    return failed;
}

static void ca_modpvec_indivsc(ca_modpvec_t a,ca_modp_t s,ca_modpvec_args args)
{
    ca_modpseq_indivsc(a->ptr,s,a->n,args.sargs);
}

static void ca_modpvec_oodotmul(ca_modp_c *c,ca_modpvec_t a,ca_modpvec_t b,ca_modpvec_args args)
{
    assert(a != b);ca_modpvec_checkcount(a->n,b->n);
    ca_modpseq_oodotmul(c,a->ptr,b->ptr,a->n,args.sargs);
}

static void ca_modpvec_oodotsqr(ca_modp_c *c,ca_modpvec_t a,ca_modpvec_args args)
{
    ca_modpseq_oodotsqr(c,a->ptr,a->n,args.sargs);
}

static void ca_modpvec_args_fileOut(id aFiler,ca_modpvec_args *args)
{
    ca_modpseq_args_fileOut(aFiler,&args->sargs);
}
static void ca_modpvec_args_fileIn(id aFiler,ca_modpvec_args *args)
{
    ca_modpseq_args_fileIn(aFiler,&args->sargs);
}

static void ca_modpvec_fileOut(id aFiler,ca_modpvec_t self,ca_modpvec_args args)
{
    [aFiler fileOut:&self->n type:'i'];
    ca_modpseq_fileOut(aFiler,self->ptr,self->n,args.sargs);
}

static void ca_modpvec_fileIn(id aFiler,ca_modpvec_t self,ca_modpvec_args args)
{
    int n;
    [aFiler fileIn:&n type:'i']; 
    ca_modpvec_init(self,n,args);
    ca_modpseq_fileIn(aFiler,self->ptr,self->n,args.sargs);
}

static void ca_modpvec_ix_init(ca_modpvec_ixt ix,id self)
{
    ca_modpvec_t v = [self ca_modpvec_value];
    ix->n = v->n;
}

static void ca_modpvec_ix_copy(ca_modpvec_ixt c,ca_modpvec_ixt a)
{
    c->i = a->i;c->n = a->n;
}

static void ca_modpvec_ix_clear(ca_modpvec_ixt ix)
{
    /* nothing to clear */
}

static int ca_modpvec_ix_count(ca_modpvec_ixt ix)
{
    return ix->n;
}

static int ca_modpvec_ix_isempty(ca_modpvec_ixt ix)
{
    return ix->n == 0;
}

static int ca_modpvec_ix_ismatch(ca_modpvec_ixt ix)
{
    assert(-1 <= ix->i && ix->i <= ix->n);
    return 0 <= ix->i && ix->i < ix->n;
}

static void ca_modpvec_ix_tofirst(ca_modpvec_ixt ix)
{
    ix->i = -1;	/* next -> first */
}

static void ca_modpvec_ix_tolast(ca_modpvec_ixt ix)
{
    ix->i = ix->n; /* prev -> last */
}

static void ca_modpvec_ix_to(ca_modpvec_ixt ix,int i)
{
    /* the actual position is n-i-1; conversion done in elting routine... */
    if (i < 0) ix->i = -1;else if (i > ix->n) ix->i = ix->n;
    else ix->i = i-1; /* next -> at i */
}

static void ca_modpvec_ix_next(ca_modpvec_ixt ix)
{
    ix->i = (ix->i < ix->n)?ix->i + 1:ix->n;
}

static void ca_modpvec_ix_prev(ca_modpvec_ixt ix)
{
    ix->i = (0 <= ix->i)?ix->i - 1:-1;
}
@implementation ca_modp_vector : ca_vector

- check
{
    [super check];
    [scalarZero check];ca_modpvec_check(&value,ca_modpvec_getargs(self));
    return self;
}


- _setUpScalarZero:aScalarZero numScalars:(int)numScalars
{
    scalarZero = aScalarZero;
    ca_modpvec_initcount(&value,numScalars,ca_modpvec_getargs(self));
    assert([self check]);return self;
}

+ newScalarZero:aScalarZero numScalars:(int)numScalars
{
    return [[super new] _setUpScalarZero:aScalarZero numScalars:numScalars];
}

- copy
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_copy(&c,&value,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- deepCopy
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_deepcopy(&c,&value,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- clear
{
    scalarZero = [scalarZero free];
    ca_modpvec_clear(&value,ca_modpvec_getargs(self));
    return [super clear];
}

- (ca_modpvec_t)ca_modpvec_value
{
    return &value;
}

- ca_modpvec_value:(ca_modpvec_t)aValue
{
    self = [self clone];
    ca_modpvec_copy(&value,aValue,ca_modpvec_getargs(self));
    assert([self check]);return self;
}

- (ca_modpvec_t)ca_modpvec_reference
{
    return &value;
}

- ca_modpvec_reference:(ca_modpvec_c *)aReference
{
    self = [self clone];
    ca_modpvec_move(&value,aReference,ca_modpvec_getargs(self));
    assert([self check]);return self;
}

- clone
{
    self = [super clone];
    scalarZero = [scalarZero cheapCopy];
    return self;
}

- capacity:(int)aCapacity
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_initcapacity(&c,aCapacity,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- numScalars:(int)numScalars
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_initcount(&c,numScalars,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}


- scalarZero
{
    return scalarZero;
}

- (int) numScalars
{
    assert([self check]);return ca_modpvec_count(&value);
}


- insertScalar:aScalar
{
    ca_modp_c c;
    assert([aScalar check]);[self checkRefCountOne];assert([self check]);
    ca_modpvec_insertfirst(&value,ca_modp_id2ref(aScalar,&c),ca_modpvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}

- insertScalar:aScalar at:(int)i
{
    ca_modp_c c;
    assert(aScalar);[self checkRefCountOne];assert([self check]);
    ca_modpvec_insertat(&value,ca_modp_id2ref(aScalar,&c),i,ca_modpvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}


- removeScalar
{
    if ([self numScalars]) {
	id res;
	ca_modp_t t = ca_modpvec_first(&value,ca_modpvec_getargs(self));
	assert([self check]);[self checkRefCountOne];
	res = ca_modp_t2id(scalarZero,t);
	ca_modpvec_freefirst(&value,ca_modpvec_getargs(self));
	assert([self check]);assert([res check]);[self invalidate];return res;
    } else {
	return nil;
    }
}

- removeScalarAt:(int)i
{
    id res;
    assert([self check]);[self checkRefCountOne];
    res = ca_modp_t2id(scalarZero,ca_modpvec_at(&value,i,ca_modpvec_getargs(self)));
    ca_modpvec_freeat(&value,i,ca_modpvec_getargs(self));
    assert([self check]);assert([res check]);[self invalidate];return res;
}


- placeScalar:aScalar at:(int)i
{
    ca_modp_c c;
    assert([aScalar check]);assert([self check]);[self checkRefCountOne];
    ca_modpvec_place(&value,ca_modp_id2ref(aScalar,&c),i,ca_modpvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}

- replaceScalarAt:(int)i with:aScalar
{
    id res;
    ca_modp_c c;
    ca_modp_t t = ca_modpvec_at(&value,i,ca_modpvec_getargs(self));
    assert([self check]);[self checkRefCountOne];
    res = ca_modp_t2id(scalarZero,t);
    ca_modpvec_place(&value,ca_modp_id2ref(aScalar,&c),i,ca_modpvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);assert([res check]);[self invalidate];return res;
}


- eachScalar
{
    id aSequence = [ca_modpvec_sequence newContent:[self cheapCopy]];
    return [CASequenceClass newCarrier:aSequence];
}

- elt_ca_modpvec_ix:(ca_modpvec_ixt)ix
{
    if (ca_modpvec_ix_ismatch(ix)) {
	ca_modp_t t = ca_modpvec_elt(&value,ix,ca_modpvec_getargs(self));
	return ca_modp_t2id(scalarZero,t);
    } else {
	return nil;
    }
}


- dotMultiply:b
{
    ca_modp_c c;	/* carrier for scalar */
    ca_modpvec_oodotmul(&c,&value,[b ca_modpvec_value],ca_modpvec_getargs(self));
    /* now promote the carrier to object */
    /* all of this used to be very difficult in 0.6 */
    return ca_modp_c2id(scalarZero,&c);
}

- dotSquare
{
    ca_modp_c c;	/* carrier for scalar */
    ca_modpvec_oodotsqr(&c,&value,ca_modpvec_getargs(self));
    /* now promote the carrier to object */
    /* all of this used to be very difficult in 0.6 */
    return ca_modp_c2id(scalarZero,&c);
}


- fileOutOn:aFiler
{
    ca_modpvec_args args;
    assert([self check]);
    [super fileOutOn:aFiler];
    args = ca_modpvec_getargs(self);
    ca_modpvec_args_fileOut(aFiler,&args);
    ca_modpvec_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_modpvec_args args;
    [super fileInFrom:aFiler];
    ca_modpvec_args_fileIn(aFiler,&args);
    ca_modpvec_fileIn(aFiler,&value,args);
    return self;
}

- (unsigned) hash
{
    return ca_modpvec_hash(ca_modpvec_id2t(self),ca_modpvec_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_modpvec_iseq(ca_modpvec_id2t(self),ca_modpvec_id2t(b),ca_modpvec_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_modpvec_iseq(ca_modpvec_id2t(self),ca_modpvec_id2t(b),ca_modpvec_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_modpvec_iszero(ca_modpvec_id2t(self),ca_modpvec_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_modpvec_iszero(ca_modpvec_id2t(self),ca_modpvec_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_modpvec_isop(ca_modpvec_id2t(self),ca_modpvec_id2t(b),ca_modpvec_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_modpvec_isop(ca_modpvec_id2t(self),ca_modpvec_id2t(b),ca_modpvec_getargs(self));
}

- negate
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_ooneg(&c,ca_modpvec_id2t(self),ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_modpvec_inneg(ca_modpvec_id2c(self),ca_modpvec_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_oodbl(&c,ca_modpvec_id2t(self),v,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_modpvec_indbl(ca_modpvec_id2c(self),v,ca_modpvec_getargs(self));
    return [self invalidate];
}

- double
{
    ca_modpvec_c c;assert([self check]);
    ca_modpvec_oodbl(&c,ca_modpvec_id2t(self),+1,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_modpvec_indbl(ca_modpvec_id2c(self),+1,ca_modpvec_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_modpvec_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_modpvec_ooadd(&c,ca_modpvec_id2t(self),v,ca_modpvec_id2t(b),w,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_modpvec_inadd(ca_modpvec_id2c(self),v,ca_modpvec_id2t(b),w,ca_modpvec_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_modpvec_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modpvec_ooadd(&c,ca_modpvec_id2t(self),+1,ca_modpvec_id2t(b),+1,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_modpvec_inadd(ca_modpvec_id2c(self),+1,ca_modpvec_id2t(b),+1,ca_modpvec_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_modpvec_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modpvec_ooadd(&c,ca_modpvec_id2t(self),+1,ca_modpvec_id2t(b),-1,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_modpvec_inadd(ca_modpvec_id2c(self),+1,ca_modpvec_id2t(b),-1,ca_modpvec_getargs(self));
    return [self invalidate];
}

- multiplyScalar:s
{
    ca_modpvec_c c;
    ca_modp_c sc;
    ca_modpvec_oomulsc(&c,ca_modpvec_id2t(self),ca_modp_id2ref(s,&sc),ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- multiplySelfScalar:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyScalar:s]);
    ca_modpvec_inmulsc(ca_modpvec_id2c(self),ca_modp_id2ref(s,&sc),ca_modpvec_getargs(self));
    return [self invalidate];
}

- divideScalar:s
{
    ca_modpvec_c c;
    ca_modp_c sc;
    BOOL failed = ca_modpvec_oodivsc(&c,ca_modpvec_id2t(self),ca_modp_id2ref(s,&sc),ca_modpvec_getargs(self));
    return (failed)?nil:ca_modpvec_c2id(self,&c);
}

- divideSelfScalar:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideScalar:s]);
    ca_modpvec_indivsc(ca_modpvec_id2c(self),ca_modp_id2ref(s,&sc),ca_modpvec_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyScalar:b:(int)w
{
    ca_modpvec_c c;
    ca_modp_c bc;
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modpvec_ooaddmulsc(&c,ca_modpvec_id2t(self),v,ca_modpvec_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyScalar:b:(int)w
{
    ca_modp_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyScalar:b:w]);
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modpvec_inaddmulsc(ca_modpvec_id2c(self),v,ca_modpvec_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modpvec_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B:(int)w
{
    ca_modpvec_c c;
    ca_modp_c ac;
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_modpvec_oomulscadd(&c,ca_modpvec_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modpvec_id2t(B),w,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B:(int)w
{
    ca_modp_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_modpvec_inmulscadd(ca_modpvec_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modpvec_id2t(B),w,ca_modpvec_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_modpvec_c c;
    ca_modp_c ac,bc;
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modpvec_oomulscaddmulsc(&c,ca_modpvec_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modpvec_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modpvec_getargs(self));
    return ca_modpvec_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_modp_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyScalar:b:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_modpvec_inmulscaddmulsc(ca_modpvec_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modpvec_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modpvec_getargs(self));
    return [self invalidate];
}
@end

