
/*
 * Computer Algebra Kit (c) 1993,95 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"

@implementation ca_obj_varspsrecdegsps_polynomial
- check
{
    int n;
    id last;
    
    [super check];
    [scalarZero check];[termZero check];[monomialZero check];
    assert([scalarZero isEqual:[termZero coefficient]]);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([monomialZero isVariableSparse]);
    
    last = nil;n = [terms numMembers];
    while (n--) {
	id term = [terms memberAt:n];assert([term notZero]);
	if ([term isCoefficient]) {
	    assert([[term coefficient] sameClass:scalarZero]);
	}
	if (last) assert([term compareExponents:last] < 0);
	[term check];last = term;
    }

    return self;
}

+ new
{
    return [self notImplemented:_cmd];
}

- _setUpScalarZero:aScalar
{
    assert([aScalar isZero]);
    assert([aScalar class_varspsrecdegsps_polynomial] == [self class]);
  
    aScalar      = [aScalar cheapCopy];
    termZero     = [CATermClass newScalar:aScalar symbol:nil exponent:0];
    aScalar      = [aScalar cheapCopy];
    monomialZero = [CAMonomialClass newScalar:aScalar];
    scalarZero   = aScalar;
    terms        = [CACollectionClass new];
  
    assert([self check]);return self;
}

+ newScalarZero:aScalar
{
    return [[super new] _setUpScalarZero:aScalar];
}

- copy
{
    int i,n;id new;

    /* don't use [terms copy] -> sends cheapCopy */
    new = [terms empty];
    for(n = [terms numMembers],i=0;i<n;i++) {
	id term = [terms memberAt:i];[new insert:[term copy]];
    }

    self = [self clone];terms = new;return self;
}

- deepCopy
{
    int i,n;id new;

    /* don't use [terms deepCopy] -> sends copy */
    new = [terms empty];
    for(n = [terms numMembers],i=0;i<n;i++) {
	id term = [terms memberAt:i];[new insert:[term deepCopy]];
    }

    self = [self clone];terms = new;return self;
}

- clear
{
    scalarZero = [scalarZero free];
    termZero = [termZero free];
    monomialZero = [monomialZero free];
    terms = [terms free];
    return [super clear];
}

- _terms { return terms; } /* super temporal of course */

- scalarZero { return scalarZero; }
- termZero   { return termZero; }
- monomialZero { return monomialZero; }

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    terms = nil;return self;
}

- _terms:cltn
{
    self = [self clone];terms = cltn;return self;
}

- empty
{
    return [self _terms:[CACollectionClass new]];
}

- newScalarZero:aScalar
{
   return [self notImplemented:_cmd];
}

- newScalarZero:aScalar symbols:aCltn
{
   return [self notImplemented:_cmd];
}

- emptyScalarZero:aZero
{
    id classCarrier = [aZero class_varspsrecdegsps_polynomial];
    return [classCarrier newScalarZero:aZero];
}

- emptyVariableDense:aCltn
{
    id classCarrier = [scalarZero class_vardnsrecdegsps_polynomial];
    return [classCarrier newScalarZero:[scalarZero cheapCopy] symbols:aCltn];
}

- emptyVariableSparse
{
    return [self empty];
}

- emptyExpanded
{
    id classCarrier = [scalarZero class_varspsexpdegsps_polynomial];
    return [classCarrier newScalarZero:[scalarZero cheapCopy]];
}

- emptyRecursive
{
    return [self empty];
}

- emptyDegreeSparse
{
    return [self empty];
}

- emptyDegreeDense
{
    id classCarrier = [scalarZero class_varspsrecdegdns_polynomial];
    return [classCarrier newScalarZero:[scalarZero cheapCopy]];
}

- (BOOL) isEmpty
{
    return [terms numMembers] == 0;
}

- (BOOL) isZero
{
    return [terms numMembers] == 0;
}

- (BOOL) isOne
{
    return [terms numMembers] == 1 && [[terms lastMember] isOne];
}

- (BOOL) isMinusOne
{
    return [terms numMembers] == 1 && [[terms lastMember] isMinusOne];
}

- (BOOL) sameClass:b
{
    return [super sameClass:b] && [scalarZero isEqual:[b scalarZero]];
}

- (BOOL) isEqual:b
{
    assert([self sameClass:b]);return [terms isEqual:[b _terms]];
}

- asScalar
{
    return ([terms numMembers] == 1)?[[terms lastMember] asScalar]:nil;
}

- asSymbol
{
    return ([terms numMembers] == 1)?[[terms lastMember] asSymbol]:nil;
}

- (int) numTerms
{
    return [terms numMembers];
}

- eachTerm
{
    return [terms eachMemberReversed];
}

- removeTerm
{
    id lt;
    assert([self checkRefCountOne] == self && [self check] == self);
    lt = [terms removeMember];[self invalidate];
    return lt;
}

- insertTerm:aTerm
{
    id c;
    assert([self checkRefCountOne] == self && [self check] && [aTerm check]);
    
    if ([aTerm isZero])  { [aTerm free];return self; }
    if ([aTerm symbol] || [[aTerm coefficient] sameClass:scalarZero])  {
	int n = [terms numMembers];
	while (n--) {
	    int cmp;
	    id term = [terms memberAt:n];
	    cmp = [term compareExponents:aTerm];
	    if (cmp == 0) {
		id sum = [term add:aTerm];[aTerm free];
		if ([sum isZero]) {
		    [[terms removeMemberAt:n] free];[sum free];
		    assert([self check]);return [self invalidate];
		} else  {
		    [[terms replaceMemberAt:n with:sum] free];
		    assert([self check]);return [self invalidate];
		}
	    } else {
		if (cmp > 0) {
		    continue;
		} else {
		    [terms insert:aTerm at:n+1];
		    assert([self check]);return [self invalidate];
		}
	    }
	}
	[terms insert:aTerm at:0];
	assert([self check]);return [self invalidate];
    } else {
	/* "insert" of a lifted term */
	assert([aTerm isCoefficient]&&[[aTerm coefficient] isVariableSparse]);
	self = [self addSelf:[aTerm coefficient]];
	[aTerm free];return self;
    }
}

@end

