
/*
 * Computer Algebra Kit (c) 1993,95 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"

@implementation ca_varsps_monomial
- check
{
    int n;
    id last;
    
    [super check];assert(termOne != nil && terms != nil);
    [termOne check];[terms check];
    
    last = nil;n = [terms numMembers];
    while(n--) {
	id term = [terms memberAt:n];
	assert([[term coefficient] isOne] && [term exponent] > 0);
	if (last) assert([last compareSymbols:term] < 0);
	last = term;
    }

    return self;
}

- _setUpScalarOne:aOne
{
    assert([aOne isOne]);
    termOne = [CATermClass newScalar:aOne symbol:nil exponent:0];
    terms   = [CACollectionClass new];
    assert([self check]);return self;
}

+ newScalarOne:aOne
{
    return [[super new] _setUpScalarOne:aOne];
}

- copy
{
    return [self _terms:[terms copy]];
}

- deepCopy
{
    return [self _terms:[terms deepCopy]];
}

- clear
{
    termOne = [termOne free];terms = [terms free];return [super clear];
}

- clone
{
    self = [super clone];termOne = [termOne cheapCopy];
    terms = nil;return self;
}

- _terms
{
    return terms;
}

- _terms:cltn
{
    self = [self clone];terms = cltn;assert([self check]);return self;
}
    
- empty
{
    return [self _terms:[CACollectionClass new]];
}

- termOne { return termOne; }
- (BOOL)isVariableSparse { return YES; }
- (BOOL)isVariableDense  { return NO;  }

- (BOOL) isEmpty
{
    return [terms numMembers] == 0;
}

- (BOOL) isOne
{
    return [terms numMembers] == 0;
}

- (int) compare:b
{
    int cmp,n,m;
    id bterms = [b _terms];

    n = [terms numMembers];
    m = [bterms numMembers];
    
    while (n && m) {
	id term  = [terms memberAt:--n];
	id bterm = [bterms memberAt:--m];
	if ((cmp = [term compareExponents:bterm])) return cmp;
    }
    
    if (n) return +1;
    if (m) return -1;
    return 0;
}

- (BOOL) isEqual:b
{
    id bTerms = [b _terms];
    return (terms==bTerms)?YES:[terms isEqual:bTerms];
}

- (unsigned) hash
{
    return [terms hash];
}

- asSymbol
{
    return ([terms numMembers] == 1)?[[terms lastMember] asSymbol]:nil;
}

- symbol:aSymbol
{
    id term = [termOne symbol:aSymbol];
    assert([termOne symbol] == nil && term); /* never fails since symbol is nil */
    return [[self empty] insertTerm:term];
}

- (int) numTerms
{
    return [terms numMembers];
}

- eachTerm
{
    return [terms eachMemberReversed];
}

- removeTerm
{
    id lt;
    assert([self checkRefCountOne] == self && [self check] == self);
    lt = [terms removeMember];[self invalidate];
    return lt;
}

- insertTerm:aTerm
{
    int n;
    
    [self checkRefCountOne];
    assert([self check] && [[aTerm coefficient] isOne]);
    if ([aTerm exponent] == 0) { [aTerm free];return self; }
    
    n = [terms numMembers];
    while (n--) {
	int cmp;
	id term = [terms memberAt:n];
	cmp = [term compareSymbols:aTerm];
	if (cmp == 0) {
	    id product = [term multiply:aTerm];[aTerm free];
	    if ([product exponent] < 0) { [product free];return nil; }
	    if ([product exponent] == 0) {
		[[terms removeMemberAt:n] free];[product free];
		assert([self check]);return [self invalidate];
	    } else  {
		[[terms replaceMemberAt:n with:product] free];
		assert([self check]);return [self invalidate];
	    }
	} else {
	    if (cmp < 0) {
		continue;
	    } else {
		if ([aTerm exponent] < 0) {
		    return nil;
		} else {
		    [terms insert:aTerm at:n+1];
		    assert([self check]);return [self invalidate];
		}
	    }
	}
    }
    
    if ([aTerm exponent] < 0) {
	return nil;
    } else {
	[terms insert:aTerm at:0];
	assert([self check]);return [self invalidate];
    }
}

@end

