
/*
 * Portable Object Compiler (c) 1996, 97 by Stes & Lerman.  All Rights Reserved.
 */

#ifdef __PORTABLE_OBJC__

#ifndef __OBJECT_INCLUDED__
#define __OBJECT_INCLUDED__
#include "Object.h" /* Stepstone Object.h assumes #import */
#endif

#include "Block.h" /* idTrue */

@implementation Object
/*****************************************************************************
 *
 * Class Management
 *
 ****************************************************************************/

+ initialize
{ 
	return self; 
}

+ subclasses
{
    id c = [[Object idOfSTR:"OrdCltn"] new];
    addSubclassesTo(c,[self name]);
    return c;
}
+ poseAs:superClass
{
    objcrt_poseAs((struct objcrt_shared *)self,(struct objcrt_shared *)superClass);
    return self;
}

+ addMethodsTo:superClass
{
    objcrt_addMethodsTo((struct objcrt_shared *)self,(struct objcrt_shared *)superClass);
    return self;
}

+ subclass:(STR)name
{
    return (id) objcrt_subclass(name,(struct objcrt_shared *)self,0,0);
}

+ inheritsFrom:aClass
{
    BOOL x;
    x=objcrt_inheritsFrom((struct objcrt_shared *)self,(STR)[aClass name]);
    return (x)?[Block idTrue]:[Block idFalse];
}

- initialize
{
	return self;
}

/*****************************************************************************
 *
 * Creating, Copying and Freeing
 *
 ****************************************************************************/


+ new
{
	id newObject = (*OC_ALLOC_VEC)(self, 0);return newObject;
}

- new
{
	return [self notImplemented:_cmd];
}

- copy
{
	id newObject = (*OC_COPY_VEC)(self, 0);return newObject;
}

- deepCopy
{
	id newObject = (*OC_COPY_VEC)(self, 0);return newObject;
}

- free
{
	return (OC_DEALLOC_VEC)?(*OC_DEALLOC_VEC)(self):nil; 
}

+ free
{ 
	return nil; 
}

/*****************************************************************************
 *
 * Identity
 *
 ****************************************************************************/


- self
{
    return self;
}

- yourself
{
    return self;
}

- class
{
    return [isa class];
}

- superclass
{
    return [isa superclass];
}

+ class
{
    return self;
}

+ superclass
{
       return (id) ((struct objcrt_shared *)self)->clsSuper; 
}

- add:anObject
{
    return [self notImplemented:_cmd]; /* mmh */
} 

- (STR) name
{
	return [isa name];
}

+ (STR) name
{
	return ((struct objcrt_shared *)self)->clsName; 
}


- findClass:(STR)aClassName
{ 
	return (*OC_CVT_TO_ID_VEC)(aClassName);
}

- (SEL) selOfSTR:(STR)aSelName
{

	SEL aSel = (*OC_CVT_TO_SEL_VEC)(aSelName);

	if (aSel) {
	    return aSel;
	} else {
	    STR msg = "'%s' not in hashtable.";
	    [errorBlock(OC_SELNOTFOUND,msg,aSelName) raise]; 
	    return NULL;
	}
}

- idOfSTR:(STR)aClassName
{
	id aClass = [self findClass:aClassName];
	if (aClass) {
	    return aClass;
	} else {
	    STR msg = "'%s' not linked in application.";
	    return [errorBlock(OC_CLASSNOTFOUND,msg,aClassName) raise]; 
	}
}

/*****************************************************************************
 *
 * Comparing
 *
 ****************************************************************************/


- (unsigned) hash
{
	/* MIPSpro 64bit : cast a size_t to unsigned and avoid a warning
	* it is not incorrect for |hash| to ignore the upper 32bit
	*/

	return (unsigned) (self - nil);
}

- (BOOL) isEqual:anObject
{
	return (BOOL) (anObject == self); 
}

- idEqual:anObject
{
	return ([self isEqual:anObject])?[Block idTrue]:[Block idFalse];
}

+ (BOOL) isEqual:anObject
{
	return (BOOL) (self == anObject);
}

- (BOOL) isSame:anObject
{ 
	return (BOOL) (self == anObject);
}

- (BOOL) notEqual:anObject
{
	return [self isEqual:anObject] == NO;
}

- (BOOL) notSame:anObject
{ 
	return (BOOL) ![self isSame:anObject]; 
}

- (int) compare:anObject
{
	[self notImplemented:_cmd];return 0;
}

/*****************************************************************************
 *
 * Class and Methods
 *
 ****************************************************************************/


- (BOOL) respondsTo:(SEL)aSelector
{
	unsigned n; 
	struct objcrt_slt *smt; 
	struct objcrt_shared * cls;
	
	cls = (struct objcrt_shared *) isa; 

	do {
		for (smt = cls->clsDispTable, n = cls->clsSizDict; n--; smt++) {
			/* selectors can be compared with == */
			if (smt->_cmd == aSelector) return YES;
		}
	} while ((cls = cls->clsSuper));

	return NO;
}

- (BOOL) isMemberOf:aClass
{
	return (BOOL) (isa == aClass);
}

- (BOOL) isKindOf:aClass
{
	struct objcrt_shared *cls = (struct objcrt_shared *) isa;
	struct objcrt_shared *aCls = (struct objcrt_shared *) aClass;

	for(;cls;cls = cls->clsSuper) if (cls == aCls) return YES;

	return NO;
}

/*****************************************************************************
 *
 * Blocks 
 *
 ****************************************************************************/


- (BOOL) isTrue
{
    return self != idFalse;
}

- (BOOL) isFalse
{
    return self == idFalse;
}

- ifTrue:aBlock
{
    return ([self isTrue])?[aBlock value]:nil;
}
- ifTrue:trueBlock ifFalse:falseBlock
{
    return ([self isTrue])?[trueBlock value]:[falseBlock value];
}

- ifFalse:aBlock
{
    return ([self isFalse])?[aBlock value]:nil;
}

- ifFalse:falseBlock ifTrue:trueBlock
{
    return ([self isFalse])?[falseBlock value]:[trueBlock value];
}

/*****************************************************************************
 *
 * Error Handling
 *
 ****************************************************************************/


- subclassResponsibility:(SEL)aSelector
{
	int tag = OC_SUBCLASSRESPONSIBILITY;
	STR msg = "%s should have implemented '%s'.";
	return [errorBlock(tag,msg,[self name],aSelector) raise];
}

- subclassResponsibility
{
	return [self subclassResponsibility:"method"];
}

- shouldNotImplement:(SEL)aSelector from:superClass
{
    	int tag = OC_SHOULDNOTIMPLEMENT;
	STR msg = "%s should NOT have implemented '%s'.";
	return [errorBlock(tag,msg,[self name],aSelector) raise];
}

- notImplemented:(SEL)aSelector
{
    	int tag = OC_NOTIMPLEMENTED;
	STR msg = "%s does not implement '%s'.";
	return [errorBlock(tag,msg,[self name],aSelector) raise];
}

- doesNotRecognize:(SEL)aSelector
{
    	int tag = OC_DOESNOTRECOGNIZE;
	STR msg = "%s does not recognize selector '%s'.";
	return [errorBlock(tag,msg,[self name],aSelector) raise];
}

- doesNotUnderstand:(SEL)aSelector
{
    return [self doesNotRecognize: aSelector];
}

- error:(STR)format,...
{
	OC_VA_LIST ap;

	/* use OC macros here for porting to SunOS4 */

	OC_VA_START(ap,format);
	(*OC_ERROR_VEC)(self,format, ap); 
	OC_VA_END(ap);

	return self;
}

/*****************************************************************************
 *
 * Method Implemenation Lookup
 *
 ****************************************************************************/


- (IMP) methodFor:(SEL)aSelector
{
	return _imp(self,aSelector);
}

+ (IMP) instanceMethodFor:(SEL)aSelector
{
	return _impSuper(self,aSelector);
}

/*****************************************************************************
 *
 * Method Performing
 *
 ****************************************************************************/


- perform:(SEL)aSelector
{
   /* the usual all C cast */
   return (*(id(*)(id,SEL))_imp(self, aSelector))(self, aSelector);
}

- perform:(SEL)aSelector with:anObject
{
   /* the usual all C cast */
   return (*(id(*)(id,SEL,id))_imp(self, aSelector))(self,aSelector,anObject);
}

- perform:(SEL)aSelector with:anObject with:otherObject
{
   /* the usual all C cast */
   return (*(id(*)(id,SEL,id,id))_imp(self, aSelector))(self,aSelector,anObject,otherObject);
}

- perform:(SEL)aSelector with:anObject with:otherObject with:thirdObj
{
   /* the usual all C cast */
   return (*(id(*)(id,SEL,id,id,id))_imp(self, aSelector))(self,aSelector,anObject,otherObject,thirdObj);
}
/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- print
{
    return [self printToFile:stdout];
}

+ print
{
    printf([self name]);
    return self;
}


- printLine
{
    [self print];
    printf("\n");
    return self;
}

- show
{
	return (*_showOn)(self, 0); 
}

- printToFile:(FILE *)aFile
{
     return self;
}


#ifdef __STPSTN__
- printOn:(IOD)anIOD
{
    return [self printToFile:anIOD];
}
#else
- printOn:(FILE *)anIOD
{
    return [self printToFile:anIOD];
}
#endif /* __STPSTN__ */

/*****************************************************************************
 *
 * Version
 *
 ****************************************************************************/

+ (STR) objcrtRevision
{
    return __objcrt_revision__;
}

/*****************************************************************************
 *
 * AsciiFiler Methods for Object Class
 *
 ****************************************************************************/

+ readFrom:(STR)aFileName
{ 
	return (*_readFrom)(aFileName); 
}
- (BOOL) storeOn:(STR)aFileName
{ 
	return (*_storeOn)(aFileName, self); 
}
/*****************************************************************************
 *
 * AsciiFiler Methods for Object Class
 *
 ****************************************************************************/

- fileOutOn:aFiler
{
	[aFiler fileOut:& isa type:'#'];
	return [self fileOutIdsFor:aFiler];
}
+ fileInFrom:aFiler
{
	id newObject = (*OC_ALLOC_VEC)(self, 0);
	return [newObject fileInFrom:aFiler];
}

- fileInFrom:aFiler
{
	return [self fileInIdsFrom:aFiler];
}
- awake
{
	return self; 
}

- awakeFrom:aFiler
{
	return [ self awake ];
}


@end

#endif /* __PORTABLE_OBJC__ */


