
/*
 * Portable Object Compiler (c) 1996,97 by Stes & Lerman.  All Rights Reserved.
 */

#ifdef __PORTABLE_OBJC__

#ifndef __objcrt__
#define __objcrt__

#include "config.h"

#include <stdio.h>  /* FILE */
#include <stddef.h> /* size_t */

#ifndef OBJCRTEXPORT
#define OBJCRTEXPORT /* empty */
#endif

#if OBJCRT_PTHREADS
#include <pthread.h>  /* POSIX 1003.1c thread-safe messenger */
#endif
#if OBJCRT_CTHREADS
#include <cthreads.h> /* Mach C thread-safe messenger */
#endif
#if OBJCRT_BOEHM
#include <gc.h>       /* objcInit() vectors to use Hans-J. Boehm gc */
#endif

/* The "id" type is assumed to be defined by the compiler
 * (in the case of objc or objcc, by emitting a typedef)
 * This also allows for cross compiles with gcc in ObjC mode
 * because a definition here would redefine its builtin "id".
 */

/* The traditional Objective C types (see Brad Cox' book)
 */

typedef char *SEL;	/* Uniqued String for Selector */
typedef char *STR;	/* C, NULL-terminated, String */
typedef char BOOL;	/* Boolean */
typedef FILE *IOD;      /* I/O Descriptor */
typedef id SHR;         /* type of class, for us, it's id */

#ifdef __cplusplus
typedef id (*IMP)(...);	/* Method pointer */
#else
typedef id (*IMP)();	/* Method pointer */
#endif

/* The traditional Objective C defines 
 */

#ifndef YES
#define YES (BOOL)1 /* Boolean TRUE */
#endif

#ifndef NO
#define NO (BOOL)0 /* Boolean FALSE */
#endif

#ifndef nil
#define nil (id)0 /* id of Nil instance */
#endif

/* varargs for OC_ERROR_VEC error reporting
 */
 
#if OBJCRT_USE_STDARG
#include <stdarg.h>
#define OC_VA_LIST va_list
#define OC_VA_START(ap,larg) va_start(ap,larg)
#define OC_VA_ARG(ap,type) va_arg(ap,type)
#define OC_VA_END(ap) va_end(ap)
#else
#include <varargs.h>
#define OC_VA_LIST va_list
#define OC_VA_START(ap,larg) va_start(ap)
#define OC_VA_ARG(ap,type) va_arg(ap,type)
#define OC_VA_END(ap) va_end(ap)
#endif /* OBJCRT_USE_STDARG */

/* Types for Objective C classes.
 *
 * These SHOULD MATCH what the compiler is emitting (for struct _PRIVATE
 * and struct _SHARED).
 * 
 * The compiler emits the types, as to be able to generate code that can
 * work with the Stepstone runtime.
 *
 * NOTE: The Stepstone compiler emits struct _PRIVATE and struct _SHARED,
 * only for a file containing a class definition.
 *
 * All of this is sort of ugly, and error prone (since the types need to
 * be kept in sync), but needed for cross-compiles (different runtimes
 * & compilers)...
 */

struct objcrt_private {			
	struct objcrt_shared *isa;
	unsigned short attr;
	unsigned short objID;
};

struct objcrt_shared {
	struct objcrt_shared *isa;
	struct objcrt_shared *clsSuper;
	char *clsName;
	char *clsTypes;	
	short clsSizInstance;
	short clsSizDict;
	struct objcrt_slt *clsDispTable;
	long clsStatus;
	struct objcrt_modDescriptor *clsMod;
	unsigned clsVersion;
	struct objcrt_shared *clsCats; /* obsolete, will go away */
	id *clsGlbl; /* only if CLS_POSING */
};

/* Masks for clsStatus */

#define CLS_FACTORY 0x1L
#define CLS_META 0x2L
#define CLS_INITIALIZED 0x4L
#define CLS_POSING 0x8L /* unused, but present in, Stepstn. */
#define CLS_MAPPED 0x10L
/* POC extensions */
#define CLS_CATS 0x20L /* obsoleted, will go away */
#define CLS_CAT 0x40L /* new category implementation 10/97 */

/* CLS_POSING isn't actually used it seems for Stepstone
 * (although they define it).  Let's use this flag to check whether
 * the class has a clsGlbl field (which our implementation uses)
 */

#define CLS_HASPOSING(aCls) ((aCls)->clsStatus & CLS_POSING)

/* Dispatch table entry.
 * Used in struct _SHARED as _SLT...
 * Must match what the compiler (objc or objcc) is emitting
 */

struct objcrt_slt {
	SEL _cmd;
	IMP _imp;
};

/* Use Descriptor for automatic initialization (as opposed to postLink).
 * Should match struct useDescriptor emitted by compiler.
 */

struct objcrt_useDescriptor {
	int processed;
	struct objcrt_useDescriptor *next;
	struct objcrt_useDescriptor ***uses;
	struct objcrt_modDescriptor *(*bind)();
};

/*
 * HashTable Types
 */

typedef struct hashedSelector {
	struct hashedSelector *next;
	STR key;
} HASH, *PHASH;

/*
 * Module Types.
 * These types should be compatible with what the compiler generates.
 */

typedef struct objcrt_modDescriptor MOD, *PMOD;
typedef struct objcrt_methodDescriptor	METH, *PMETH; /* for static binding */

struct objcrt_modDescriptor {
	STR	modName;
	STR	modVersion;
	long	modStatus;
	SEL	modMinSel;
	SEL	modMaxSel;
	struct objcrt_shared **modClsLst; /* single or array of class globals */
	short	modSelRef;
	SEL	*modSelTbl;
	PMETH	modMapTbl; /* not used */
};

struct objcrt_modEntry { 
	PMOD (*modLink)();
	PMOD modInfo;
};

struct objcrt_methodDescriptor { /* for static binding - not used */
	id	*cls;
	SEL	*sel;
	IMP	imp;
};

/* Masks for modStatus
 * MOD_MORETHANONE is a stes extension : compiler flags when more
 * than one class per module; in this case modClsLst is '0' terminated.
 */

#define MOD_EARLYBIND	0x1L /* unused, old Stepstone */
#define MOD_MAPPED		0x2L /* marked after selectors mapped */
#define MOD_MORETHANONE 0x4L

/* C Messenger interface - try to be compatible */

/* msg tracing */
extern BOOL msgFlag;
extern void OBJCRTEXPORT setMsgFlag(BOOL);
extern FILE *msgIOD;
extern void OBJCRTEXPORT setMsgIOD(FILE*);

extern BOOL allocFlag;   /* allocation tracing */
extern BOOL dbgFlag;     /* verbose dbg() calls */
extern BOOL noCacheFlag; /* disable messenger caching */
extern BOOL noNilRcvr;   /* do not allow msg. to nil */

/* all in the name of backw. compat. */
extern FILE *dbgIOD;
#ifndef NAMEOF
#define NAMEOF(myObject) ([myObject name])
#endif

SEL OBJCRTEXPORT selUid(STR);
STR OBJCRTEXPORT selName(SEL);
void OBJCRTEXPORT dbg(char *fmt, ...);
void OBJCRTEXPORT addObjcModules(void *modPtr);

/* the following def does not completely correspond to the objcc all
 * C messenger, since it's typing classes to SHR instead of "id".
 * To avoid a clash, allow to not prototype them (minor problem
 * as the compiler emitted sources will prototype them anyhow)
 */

#if OBJCRT_PROTOTYPE_MESSENGER
IMP OBJCRTEXPORT _imp(id,SEL);
IMP OBJCRTEXPORT _impSuper(id,SEL);

#endif

/* can't make this public unless making OBJCRT_USE_PREFIXED_NAMES public */
#if OBJCRT_USE_PREFIXED_NAMES

#define OC_ALLOC_VEC		oc_alloc
#define OC_DEALLOC_VEC		oc_dealloc
#define OC_COPY_VEC		oc_copy
#define OC_ERROR_VEC		oc_error
#define OC_CVT_TO_ID_VEC	oc_cvtToId
#define OC_CVT_TO_SEL_VEC	oc_cvtToSel
#define OC_OBJCINIT_FNC		oc_objcInit

#else

#define OC_ALLOC_VEC		_alloc
#define OC_DEALLOC_VEC		_dealloc
#define OC_COPY_VEC		_copy
#define OC_ERROR_VEC		_error
#define OC_CVT_TO_ID_VEC	_cvtToId
#define OC_CVT_TO_SEL_VEC	_cvtToSel
#define OC_OBJCINIT_FNC		_objcInit

#endif /* OBJCRT_USE_PREFIXED_NAMES */

/* the "extern" def for vectors is important for SGI cc */

extern id (*OC_ALLOC_VEC)(id, unsigned int); /*allocate a new object*/
extern id (*OC_DEALLOC_VEC)(id); /* deallocate an object */
extern id (*OC_COPY_VEC)(id,unsigned int); /* shallow copy an object */
extern id (*OC_ERROR_VEC)(id,STR,OC_VA_LIST); /* error handler */

extern id (*OC_CVT_TO_ID_VEC)(STR); /* convert string name to class id */
extern SEL (*OC_CVT_TO_SEL_VEC)(STR); /* convert string to selector */

#define OC_STOREON_VEC		_storeOn
#define OC_READFROM_VEC		_readFrom

extern id (*_fileIn)(FILE*);
extern BOOL (*_fileOut)(FILE*,id);
void OBJCRTEXPORT setFileIn(id (*f)(FILE *));
void OBJCRTEXPORT setFileOut(BOOL (*f)(FILE *,id));
extern BOOL (*OC_STOREON_VEC)(STR,id);  
extern id (*OC_READFROM_VEC)(STR);

extern id (*_showOn)(id,unsigned); 

extern void *(*objcrt_malloc)(size_t);
extern void *(*objcrt_malloc_atomic)(size_t);
extern void *(*objcrt_realloc)(void*,size_t);
extern void (*objcrt_free)(void*);

void OBJCRTEXPORT *OC_Malloc(size_t);
/* allocate memory from the system */
void OBJCRTEXPORT *OC_MallocAtomic(size_t); 
/* same as OC_Malloc, but contains no pointers */
void OBJCRTEXPORT *OC_Calloc(size_t); 
/* same as OC_Malloc, but all zeroes */
void OBJCRTEXPORT *OC_Realloc(void *,size_t); 
/* reallocate memory from the system */
void OBJCRTEXPORT *OC_Free(void *data); 
/* deallocate OC_Malloc'ed memory and return NULL */

void addSubclassesTo(id c,STR name);
BOOL objcrt_inheritsFrom(struct objcrt_shared *c,STR name);

struct objcrt_shared  *objcrt_subclass(STR name,struct objcrt_shared *superClass,int ivars,int cvars);

#if OBJCRT_PTHREADS || OBJCRT_CTHREADS
extern void (*objcrt_cache_lock)(void);
extern void (*objcrt_cache_unlock)(void);
#endif

void OBJCRTEXPORT objcrt_addMethodsTo(struct objcrt_shared *src,struct objcrt_shared *dst);

void objcrt_poseAs(struct objcrt_shared *posing,struct objcrt_shared *target);

#endif /* __objcrt__ */

#endif /* __PORTABLE_OBJC__ */


