
/*
 * Portable Object Compiler (c) 1996,97 by Stes & Lerman.  All Rights Reserved.
 */

#ifdef __PORTABLE_OBJC__

#ifndef __OBJECT_INCLUDED__
#define __OBJECT_INCLUDED__
#include "Object.h" /* Stepstone Object.h assumes #import */
#endif

#include "Block.h"    /* blockRaise */
#include <assert.h>
#include <stdlib.h>
#include <string.h>   /* memset */

static BOOL ISFACTORY(struct objcrt_shared *cls)
{
	return (cls)->clsStatus & CLS_FACTORY;
}

static BOOL ISMETA(struct objcrt_shared *cls)
{
	return (cls)->clsStatus & CLS_META;
}

static BOOL ISCAT(struct objcrt_shared *cls)
{
	return (cls)->clsStatus & CLS_CAT;
}

static struct objcrt_shared * GETMETA(struct objcrt_shared *cls)
{
	return ISMETA(cls) ? cls : cls->isa;
}

static BOOL ISINITIALIZED(struct objcrt_shared *cls)
{
	return GETMETA(cls)->clsStatus & CLS_INITIALIZED;
}

static void MARKINITIALIZED(struct objcrt_shared *cls)
{
	GETMETA(cls)->clsStatus |= CLS_INITIALIZED;
}

static BOOL ISMAPPED(struct objcrt_shared *aCls)
{
	return (aCls)->clsStatus & CLS_MAPPED;
}

static void MARKMAPPED(struct objcrt_shared *aCls)
{
	(aCls)->clsStatus |= CLS_MAPPED;
}

static id XXXGetClass(id anObject)
{
   return (id) ((struct objcrt_private *) anObject) -> isa;
}

static void XXXSetClass(id anObject,id aClass)
{
   ((struct objcrt_private *)anObject)->isa = (struct objcrt_shared *)aClass;
}

static short XXXGetInstanceSize(id aClass)
{
   return ((struct objcrt_shared *) aClass) -> clsSizInstance;
}


#define SIZEHASHTABLE 73 /* default initial size */

static PHASH *hashList;
static int nHashLists;

#if 0
static BOOL MOD_ISMAPPED(struct objcrt_modDescriptor *aMod)
{
	return (aMod)->modStatus & MOD_MAPPED;
}
#endif

static void MOD_MARKMAPPED(struct objcrt_modDescriptor *aMod)
{
	(aMod)->modStatus |= MOD_MAPPED;
}

static BOOL MOD_ISMORETHANONE(struct objcrt_modDescriptor *aMod)
{
	return (aMod)->modStatus & MOD_MORETHANONE; /* TRUE for 1.4.x */
}

/* _objcModules is defined as NULL for auto-initialization
 * it's defined as the table of BIND functions that need to be called
 * in the postlink version
 */

/* defining _objcModules as "extern" (as it should) will break
 * compiler-recompiles using < 1.1.6 bootstrap compilers (because these
 * compilers don't emit an _objcModules definition).
 */

/* 11/97 _objcModules not defined at all any more for Windows NT DLL's !
 * (NOSHARED case)
 */

#ifndef OBJCRT_NOSHARED
#if OBJCRT_SCOPE_OBJCMODULES_EXTERN
extern struct objcrt_modEntry *_objcModules;
#else
struct objcrt_modEntry *_objcModules;
#endif
#endif

/* array for when loading modules (using dlopen() or similar)
 * and then calling addObjcModules() on that _objcModules array
 */

#define MAX_TABLE_MODULES 32
static int table_modules_count = 0;
static struct objcrt_modEntry *table_modules[MAX_TABLE_MODULES]; 

/*****************************************************************************
 *
 * Memory management functions
 *
 ****************************************************************************/

/* function wrappers around malloc & co in case they are implemented
 * as macros, or in case they have special calling conventions */

static void *dflt_malloc(size_t s) { return malloc(s); }
static void *dflt_realloc(void *p,size_t s) { return realloc(p,s); }
static void dflt_free(void *p) { free(p); }

/* access malloc via function pointer : the -gc option reassigns the vector */

void *(*objcrt_malloc)(size_t) = dflt_malloc;
void *(*objcrt_malloc_atomic)(size_t) = dflt_malloc; /* no difference */
void *(*objcrt_realloc)(void*,size_t) = dflt_realloc;
void (*objcrt_free)(void*) = dflt_free;

static id OC_Malloc_exception;
static id OC_Realloc_exception;

void OBJCRTEXPORT *OC_Malloc(size_t nBytes)
{
    void *data;
    if (nBytes > 32 * 1024) dbg("OC_Malloc call for %i bytes\n",nBytes);

    /* DEC alpha malloc() returns NULL for 0 bytes */
    data = (*objcrt_malloc)((nBytes)?nBytes:sizeof(void*));

    /* try until the handler aborts or frees memory */
    while (!data) {
        [OC_Malloc_exception raise]; 
	data = (*objcrt_malloc)((nBytes)?nBytes:sizeof(void*));
    }

    return data;
}

void OBJCRTEXPORT *OC_MallocAtomic(size_t nBytes)
{
    void *data;
    if (nBytes > 32 * 1024) dbg("OC_Malloc call for %i bytes\n",nBytes);

    /* DEC alpha malloc() returns NULL for 0 bytes */
    data = (*objcrt_malloc_atomic)((nBytes)?nBytes:sizeof(void*));

    /* try until the handler aborts or frees memory */
    while (!data) {
        [OC_Malloc_exception raise]; 
	data = (*objcrt_malloc_atomic)((nBytes)?nBytes:sizeof(void*));
    }

    return data;
}

void OBJCRTEXPORT *OC_Calloc(size_t nBytes)
{
    char *p = (char *)OC_Malloc(nBytes);
#if OBJCRT_USE_MEMSET
    memset(p,0,nBytes);
#else
    p += nBytes;while (nBytes--) *--p = 0;
#endif
    return (void *)p;
}

void OBJCRTEXPORT *OC_Realloc(void *data, size_t nBytes)
{
    if (nBytes > 32 * 1024) dbg("OC_Realloc call for %i bytes\n",nBytes);
    if (!data) [OC_Realloc_exception raise];

    data = (*objcrt_realloc)(data,(nBytes)?nBytes:sizeof(void*));

    /* try until the handler aborts or frees memory */
    while (!data) {
        [OC_Realloc_exception raise];
        data = (*objcrt_realloc)(data,(nBytes)?nBytes:sizeof(void*));
    }

    return data;
}

void OBJCRTEXPORT *OC_Free(void *data)
{
    if (objcrt_free) (*objcrt_free)(data);
    return NULL;
}

/*****************************************************************************
 *
 * Object Allocator function vectors
 *
 ****************************************************************************/

/*
 * We generally type classes as "id", unlike Stepstone objc (SHR)
 * or NeXT objc (Class).  Also, we don't support indexed variables.
 */

static id objcrt_alloc(id aClass,unsigned int nBytes)
{
	id anObject;
	unsigned aSize;

	if (nBytes) [errorBlock(OC_NOIXVARS,
	    "indexed variables are not supported") raise];

	if (!aClass) [errorBlock(OC_MEMERROR,"alloc: nil class") raise];

	aSize = XXXGetInstanceSize(aClass);

	/* by default, +new zeroes ivars */
	anObject = (id)OC_Calloc(aSize);

	XXXSetClass(anObject,aClass);
	return anObject;
}

/*
 * realloc is anyhow just for ixvars, which we don't do
 * _realloc clashes with IRIX 6.2 version.
 *
 * In addition objcrt_realloc is now used for the gc stuff !
 */

#if 0
static id objcrt_realloc(id anObject,unsigned int nBytes)
{
	return nil;
}
#endif

static id objcrt_copy(id anObject,unsigned int nBytes)
{
	char *p,*q;
	id newObject;
	unsigned aSize;
	id aClass = XXXGetClass(anObject);

	if (nBytes) [errorBlock(OC_NOIXVARS,
	    "indexed variables are not supported") raise];

	aSize = XXXGetInstanceSize(aClass);

	/* by default, -copy bytecopies ivars */
	newObject = (id)OC_Malloc(aSize);
	p = (char *)newObject;
	q = (char *)anObject;
#if OBJCRT_USE_MEMSET
	memcpy(p,q,aSize);
#else
	while (aSize--) *p++ = *q++;
#endif

	assert(XXXGetClass(newObject) == aClass);
	return newObject;
}

static id objcrt_dealloc(id anObject)
{
	XXXSetClass(anObject,nil);

	/* it might be an idea to zero all ivars
	* but then again, this might also hide bugs
	*/

	OC_Free(anObject);
	return nil;
}

id (*OC_ALLOC_VEC)(id, unsigned int) = objcrt_alloc;
id (*OC_DEALLOC_VEC)(id) = objcrt_dealloc;
id (*OC_COPY_VEC)(id, unsigned int) = objcrt_copy;

/* clashes with IRIX 6.2 and isn't really used */
#if 0
id (*OC_REALLOC_VEC)(id, unsigned int) = objcrt_realloc;
#endif

/*****************************************************************************
 *
 * Error Handler
 *
 * Especially useful to be declared as a 'function vector' so that the handler
 * can be replaced at runtime by something else.
 *
 * DEC Alpha OSF/1 does not allow for a '0' argument for va_list
 *
 ****************************************************************************/

static id objcrt_error(id self, STR fmt, OC_VA_LIST ap) 
{
	/* Fixed bug here that would send a |name| to self.
	* Recursion for freedHandler (by sending again msg to freed object)
	*/

	fflush(stderr);
	fprintf(stderr,"objcrt error: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr,"\n");
	fflush(stderr);
	
	abort();return nil;
}

id (*OC_ERROR_VEC)(id self, STR fmt, OC_VA_LIST ap) = objcrt_error;

/*****************************************************************************
 *
 * Default Filer Functions
 *
 * They don't do anything, they just tell the programmer to use AsciiFiler.
 *
 ****************************************************************************/

static void noFiler(void)
{
	id e=errorBlock(OC_NOFILER,"No Filer class linked into application.");
	[e raise];
}

static id objcrt_readFrom(STR aFileName)
{
	FILE *f;
	id r = nil;
	if ((f = fopen(aFileName, "r"))) { r = (*_fileIn)(f);fclose(f); }
	return r;
}

static BOOL objcrt_storeOn(STR aFileName,id anObject)
{
	FILE *f;
	BOOL r = NO;
	if ((f = fopen(aFileName, "w"))) { r = (*_fileOut)(f,anObject);fclose(f); }
	return r;
}

static id objcrt_filein(FILE* f) { noFiler();return nil; }
static BOOL objcrt_fileout(FILE* f,id anObject) {noFiler();return NO;}

id (*OC_READFROM_VEC)(STR) = objcrt_readFrom;
BOOL (*OC_STOREON_VEC)(STR,id) = objcrt_storeOn;
id (*_fileIn)(FILE*) = objcrt_filein;
BOOL (*_fileOut)(FILE*,id) = objcrt_fileout;
void OBJCRTEXPORT setFileIn(id (*f)(FILE *)) { _fileIn = f; }
void OBJCRTEXPORT setFileOut(BOOL (*f)(FILE *,id)) { _fileOut = f; }

/* function scoped extern since it can be useful from within debugger
 * in particular if it's not easy to send the |show| message from the debugger
 */

id OBJCRTEXPORT __showOn(id self,unsigned level /* unused */)
{
	(*_fileOut)(stderr,self);return self;
}

id (*_showOn)(id,unsigned) = __showOn;

/*****************************************************************************
 *
 * Hash Table Maintenance
 *
 * Selectors are uniqued strings.
 *
 ****************************************************************************/

static int strCmp(char *s1,char *s2)
{
    int r;
    int c1, c2;

    while (1) {
	c1 = *s1++;
	c2 = *s2++;
	if (c1  == '\0') return (c2 == 0)?0:-1;
	if (c2  == '\0') return 1;
	if ((r = c1 - c2)) return r;
    }

    return 0;
}

static unsigned strHash(char *s)
{
    unsigned hash = 0;
    
    while (1) {
	if (*s == '\0') break;else hash ^= *s++;
	if (*s == '\0') break;else hash ^= (*s++ << 8);
	if (*s == '\0') break;else hash ^= (*s++ << 16);
	if (*s == '\0') break;else hash ^= (*s++ << 24);
    }
    
    return hash;
}

static void hashInit()
{
	int i;
	nHashLists = SIZEHASHTABLE;

	hashList = (PHASH *)OC_Malloc(nHashLists * sizeof(PHASH)); 

	for (i = 0; i < nHashLists; i++) hashList[i] = 0;
}

static PHASH hashNew(STR key, PHASH link)
{
	PHASH obj = (PHASH)OC_Malloc(sizeof(HASH)); 
	obj->next = link;
	obj->key = key;
	return obj;
}

static PHASH search(STR key, int *slot, PHASH *prev)
{
	PHASH target;

	*slot = strHash(key) % nHashLists;
	*prev = 0;
	target = hashList[*slot];

	while (target && (strCmp(key, target->key) != 0)) {
		*prev = target;
		target = target->next;
	}

	return target;
}

static PHASH hashEnter(STR key, int slot)
{
	assert(key != NULL);
	if (slot < 0) slot = strHash(key) % nHashLists;
	hashList[slot] = hashNew(key, hashList[slot]);
	return hashList[slot];
}

static PHASH hashLookup(STR key, int *slot)
{
	PHASH prev;return search(key, slot, &prev);
}

#if 0
static PHASH hashLookupByUid(SEL uid)
{
	int n = 0;
	PHASH target = NULL;

	while (n < nHashLists) {
		target = hashList[n];
		while (target) {
			if (target->key != uid) {
				target = target->next;
			} else {
				return target;
			}
		}
		n++;
	}

	return target;
}

static PHASH hashRemove(STR key)
{
	int slot;
	PHASH prev, target;

	if (target = search(key, &slot, &prev)) {
		if (prev) {
			prev->next = target->next;
		} else {
			hashList[slot] = target->next;
		}
	}

	return target;
}
#endif

static void mapSelRefs(PMOD info, SEL *sel)
{
	PHASH ent;
	int i, slot;

	for (i = 0; i < info->modSelRef; i++) {

		if (!(ent = hashLookup(info->modSelTbl[i], &slot))) {
			ent = hashEnter(info->modSelTbl[i], slot);
		}

		/* 'unique' the string (replace by shared one) */

		if (sel != 0 && *sel == info->modSelTbl[i])
			*sel = ent->key;

		info->modSelTbl[i] = ent->key; 
	}

	/* mark module as mapped */
	MOD_MARKMAPPED(info);
}

static void mapUnmappedClass(struct objcrt_shared *cls)
{
	PHASH ent;
	int i, slot;

	for (i = 0; i < cls->clsSizDict; i++) {
	    if (!(ent = hashLookup(cls->clsDispTable[i]._cmd, &slot))) {
		    ent = hashEnter(cls->clsDispTable[i]._cmd, slot);
	    }
	
	    /* unique string */
	    cls->clsDispTable[i]._cmd = ent->key; 
	}
    
	MARKMAPPED(cls);
}

static void mapClass(struct objcrt_shared *cls)
{
	if (!ISMAPPED(cls)) mapUnmappedClass(cls);

	/* also do the meta */
	while (ISFACTORY(cls)) {
	    cls = cls->isa;
	    if (!ISMAPPED(cls)) mapUnmappedClass(cls);
	}
}

/*****************************************************************************
 *
 * Runtime Initialization
 *
 ****************************************************************************/

static BOOL _objcinitFlag; /* YES after initialization */

/*
 *  modlist is the handle of a linked list of modules with _BIND
 *  entries found so far.  Since the order of entries is not
 *  important, it can be maintained efficiently with only a single
 *  pointer. 
 */

static struct objcrt_useDescriptor *modlist = 0;
static int bindcnt = 0; /* _BIND entries found so far */

#ifndef OBJCRT_NOSHARED
extern struct objcrt_useDescriptor *OCU_main; /* entry point for the program */
#endif

/*
 *  Recursively process each of the module use descriptors
 */

static void processModule(struct objcrt_useDescriptor *desc)
{
    struct objcrt_useDescriptor ***nxt;

    /*  Mark this one as processed to break any cycles */
    desc->processed = 1;

    /* process each of the pointers in turn */
    for(nxt = desc->uses; *nxt ; nxt++)
    {
	/*
	 *  If this pointer is not zero and the thing it points to has
	 *  not been processed, process it.  The code would look simpler
	 *  if we examined the pointer and the processed flag at the 
	 *  beginning of this function, but that would involve extra
	 *  function calls which just test and return.
	 */

	if(**nxt && !((**nxt)->processed)) processModule(**nxt);
    }

    /*
     *  If this module has a bind entry point, add it to the linked
     *  list of modules with bind entries and increment the count of
     *  modules with bind entries that have been processed.
     */

    if (desc->bind)
    {
        desc->next = modlist;
        modlist = desc;
	bindcnt++;
    }
}

static struct objcrt_modEntry *findAllModules(struct objcrt_useDescriptor *desc)
{
    unsigned aSize;
    struct objcrt_modEntry *tmp;
    struct objcrt_useDescriptor *md;
    struct objcrt_modEntry *theModules;

    if(desc && !(desc->processed)) processModule(desc);

    aSize = (bindcnt+1)*sizeof(struct objcrt_modEntry);
    theModules = (struct objcrt_modEntry *)OC_Malloc(aSize);

    /*
     *  initialize it with the entries we have just found
     */

    for(md = modlist, tmp = theModules; md; md = md->next, tmp++)
    {
	tmp->modLink = md->bind;
        tmp->modInfo = 0;
    }

    /* makeSelsUnique() depends on modLink of last element null'ed */
    tmp->modLink = 0;
    tmp->modInfo = 0;

    return theModules;
}

static void makeSelsUnique(struct objcrt_modEntry *modPtr)
{
    static int needHashInit = 1;

    /* we can get here from objcInit() or from addObjcModules()
     * for some shared library that gets loaded _before_ the main()
     */

    if (needHashInit) { hashInit(); needHashInit = 0; }

    for (; modPtr && modPtr->modLink; modPtr++) {
	    PMOD info = modPtr->modInfo = (*modPtr->modLink)();

	    if (info) {
		    struct objcrt_shared **cls;

		    if (info->modSelTbl) mapSelRefs(info, 0);

		    cls = modPtr->modInfo->modClsLst;
		    
		    if (MOD_ISMORETHANONE(modPtr->modInfo)) {
			    while (*cls) mapClass(*cls++);
		    } else {
			    if (cls) mapClass(*cls);
		    }
	    }
    }
}

static void doInitialize(struct objcrt_shared *aCls)
{
	struct objcrt_shared *super;

	if (ISINITIALIZED(aCls)) return;

	/* force initialization of superclasses first */
	/* if we're a category, this will also force initialization of class */
	/* (which is just the superclass) */

	super = aCls->clsSuper;
	if (super && !ISINITIALIZED(super)) doInitialize(super);

	/* It's possible my superclass has sent a message back to me
	 * from its 'initialize' method, which has caused me to get
	 * initialized! Check again.
	 */

	if (ISINITIALIZED(aCls)) return;
	MARKINITIALIZED(aCls);

	if (ISCAT(aCls)) objcrt_addMethodsTo(aCls,super);
	[ (id)aCls initialize ];
}

static void initializeModules(struct objcrt_modEntry *modPtr)
{
	for (; modPtr->modInfo ; modPtr++)
	{
		struct objcrt_shared *aCls;
		struct objcrt_shared **cls = modPtr->modInfo->modClsLst;
		
		if (MOD_ISMORETHANONE(modPtr->modInfo)) {
			while((aCls = *cls++)) if(!ISINITIALIZED(aCls)) doInitialize(aCls);
		} else {
			if (cls) if (!ISINITIALIZED(aCls = *cls)) doInitialize(aCls);
		}
	}
}

/*
 * This function can be called to add modules to the runtime.
 * (dynamically loaded modules for instance)
 */

static void initObjcModules(struct objcrt_modEntry *modPtr)
{
    makeSelsUnique(modPtr);
    initializeModules(modPtr);
}

void OBJCRTEXPORT addObjcModules(void *p)
{
    struct objcrt_modEntry *modPtr = (struct objcrt_modEntry *)p;
    if (!modPtr) {fprintf(stderr,"addObjcModules with NULL argument");exit(1);}

    /* check whether we're being called from a shlib that gets loaded
     * _before_ objcInit() has been executed; in this case we want to
     * delay adding the classes, since most likely superclasses (such
     * as Object) are not yet loaded.
     *
     * We simply remember the modules, and main()'s _objcInit() will
     * take care of those.
     */

    table_modules[table_modules_count++] = modPtr;
    if (!_objcinitFlag) return;

    if (table_modules_count + 1 == MAX_TABLE_MODULES) {
	[errorBlock(0,"too many calls to addObjcModules()") raise];
    }

    initObjcModules(modPtr);
}

/*
 * The init function should not be prototyped in a header, but should match
 * the init call emitted by compiler.
 *
 * Some systems have an objcInit() in their system libs, so use some other
 * name in these cases.
 *
 * Traditionally, _objcInit() was getting its modules from a global
 * _objcModules.  In the NOSHARED case, we use the _objcInitNoShared() call
 * instead.
 */

int OBJCRTEXPORT _objcInitNoShared(
    struct objcrt_modEntry *_objcModules,struct objcrt_useDescriptor *OCU_main)
{
        STR s;
	int i;

	if (_objcinitFlag) {
		return 1;
	} else {

	/* this is not a constant on cygwin32 */
	dbgIOD = stderr;
	if ((s = (getenv("OBJCRTDBG")))) {
	    dbgFlag = YES;
	    if (strcmp(s,"stderr") == 0) {
		dbgIOD = stderr;
	    } else {
		dbgIOD = fopen(s,"w");
		setbuf(dbgIOD,NULL);
	    }
	}

	msgIOD = stderr;
	if ((s = (getenv("OBJCRTMSG")))) {
	    msgFlag = YES;
	    if (strcmp(s,"stderr") == 0) {
		msgIOD = stderr;
	    } else {
		msgIOD = fopen(s,"w");
		setbuf(msgIOD,NULL);
	    }
	}

	/* trick here, is that in the Boehm case, objcrt_free has been set
	* to NULL in the main() of the program, so OC_Free is a NOP
	* and we make -free also a NOP
	*/
	
	if (!objcrt_free) OC_DEALLOC_VEC = 0;
	
	/* Do auto-initialize if _objcModules is zero.  Otherwise,
	* assume that it is the list of all bind functions to be called
	*/

	if (!_objcModules) _objcModules=findAllModules(OCU_main);
	addObjcModules(_objcModules);

	/* do initialize of modules that were already registered
	 * via shlibs that got loaded before _objcInit()
	 * first call all BIND functions (including OBJCBIND_objcrt!)
	 * then start sending +initialize messages.
	 */

	for(i=0;i<table_modules_count;i++) makeSelsUnique(table_modules[i]);
	for(i=0;i<table_modules_count;i++) initializeModules(table_modules[i]);

	/* initialize the Malloc exception blocks cause we can't allocate
	* them when we run out of memory !
	* This must be done after initializing the runtime
	* (because errorBlock() sends an Objective-C message).
	*/

        OC_Malloc_exception = errorBlock(OC_MEMERROR,
         	"OC_Malloc: out of memory");
        OC_Realloc_exception = errorBlock(OC_MEMERROR,
         	"OC_Realloc: out of memory");
    
    	/* finished _objcInit(). it's now safe for addObjcModules() to
	 * call initObjcModules()
	 */

	    _objcinitFlag = YES;

        /* Stepstone objcc returns maxSelector, probably not used */
	    return 0;
	}
}

/* traditional entry point _objcInit, or prefixed entry point oc_objcInit
 * when we have to live together with another runtime
 */

#ifndef OBJCRT_NOSHARED
int OBJCRTEXPORT OC_OBJCINIT_FNC(int debug, BOOL traceInit)
{
    return _objcInitNoShared(_objcModules,OCU_main);
}
#endif

/*****************************************************************************
 *
 * Conversion String, Selector, Class
 *
 ****************************************************************************/

STR OBJCRTEXPORT selName(SEL sel)
{
	return (STR)(sel); /* trivial for our runtime */
}

SEL OBJCRTEXPORT selUid(STR sel)
{
	int slot;
	PHASH retVal;
	if ((retVal = hashLookup(sel,&slot))) return (SEL) retVal->key;
	return NULL;
}

static SEL cvtToSel(STR aString) 
{ 
	SEL retval;
	if(aString == NULL) return NULL;
	retval = selUid(aString);
	return (retval == aString) ? (SEL)0 : retval;
}

static id cvtToId(STR aClassName) 
{ 
        int i;
	struct objcrt_modEntry * modPtr;

	if (aClassName == NULL) return nil;

        for (i=0;i<table_modules_count;i++) {
	    for(modPtr=table_modules[i];modPtr && modPtr->modLink;modPtr++) {
		struct objcrt_shared **cls,*aCls;

		cls = modPtr->modInfo->modClsLst;

		if (MOD_ISMORETHANONE(modPtr->modInfo)) {
	           while ((aCls = *cls++)) {
			if (strCmp(aCls->clsName,aClassName)==0) return (id)aCls;
			}
		} else {
			if (cls) {
			aCls = *cls;
			if (strCmp(aCls->clsName,aClassName)==0) return (id)aCls;
			}
		}
	    }
	}
	
	return nil;
}

/* for cplusplus the (STR) cannot be () */
id (*OC_CVT_TO_ID_VEC)(STR)= cvtToId;
SEL (*OC_CVT_TO_SEL_VEC)(STR)= cvtToSel;

/*****************************************************************************
 *
 * Finding subclasses.
 *
 ****************************************************************************/

static BOOL isSubclassOf(struct objcrt_shared *aCls,STR aClassname)
{
    struct objcrt_shared *clsSuper = aCls->clsSuper;
    return (clsSuper == NULL)?0:strCmp(clsSuper->clsName,aClassname) == 0;
}

BOOL objcrt_inheritsFrom(struct objcrt_shared *aCls,STR name)
{
    struct objcrt_shared *clsSuper;

    while ((clsSuper=aCls->clsSuper)) {
	if (strCmp(clsSuper->clsName,name) == 0) return YES;
    }

    return NO;
}

void  addSubclassesTo(id c,STR aClassname)
{ 
    int i;
    struct objcrt_modEntry * modPtr;

    for (i=0;i<table_modules_count;i++) {
	for(modPtr=table_modules[i];modPtr && modPtr->modLink;modPtr++) {
	    struct objcrt_shared **cls,*aCls;

	    cls = modPtr->modInfo->modClsLst;

	    if (MOD_ISMORETHANONE(modPtr->modInfo)) {
	       while ((aCls = *cls++)) {
		   if (isSubclassOf(aCls,aClassname)) [c add:(id)aCls];
	       }	
	    } else {
		    if (cls) {
		    aCls = *cls;
		    if (isSubclassOf(aCls,aClassname)) [c add:(id)aCls];
	       }	
	    }
	}
    }
}

/*****************************************************************************
 *
 * Allocation of Classes (Dynamic Subclassing)
 *
 ****************************************************************************/

static struct objcrt_shared *newShared(int aSize)
{
    return (struct objcrt_shared *)OC_Malloc(aSize);
}

static struct objcrt_shared *newMeta(STR name,struct objcrt_shared *superClass)
{
    struct objcrt_shared *n = newShared(sizeof(struct objcrt_shared));

    n->isa = superClass->isa;
    n->clsSuper = superClass;
    n->clsName = name; /* strdup ? */
    n->clsTypes = superClass->clsTypes;
    n->clsSizInstance = superClass->clsSizInstance;
    n->clsSizDict = 0;
    n->clsDispTable = NULL;
    n->clsStatus = 0;
    n->clsMod = NULL; /* unused anyhow */
    n->clsVersion = superClass->clsVersion;
    n->clsGlbl = NULL;

    return n;
}

static struct objcrt_shared *newClass(STR name,struct objcrt_shared *superClass)
{
    struct objcrt_shared *meta, *n;

    meta = newMeta(name,superClass->isa);
    n    = newShared(meta->clsSizInstance); /* includes cvars */

    n->isa = meta;
    n->clsSuper = superClass;
    n->clsName = name; /* strdup ? */
    n->clsTypes = superClass->clsTypes;
    n->clsSizInstance = superClass->clsSizInstance;
    n->clsSizDict = 0;
    n->clsDispTable = NULL;
    n->clsStatus = 0;
    n->clsMod = NULL; /* unused anyhow */
    n->clsVersion = superClass->clsVersion;
    n->clsGlbl = NULL; /* could set this to clsLst */

    return n;
}

static PMOD newModDesc(struct objcrt_shared **clsLst)
{
    PMOD mod = (PMOD)OC_Malloc(sizeof(MOD));

    mod->modName = "dynamic";
    mod->modName = "1.7.3";
    /* not MOD_MORETHANONE (clsLst is not NULL terminated) */
    mod->modStatus = 0;
    mod->modMinSel = NULL;
    mod->modMaxSel = NULL;
    mod->modClsLst = clsLst;
    mod->modSelRef = 0;
    mod->modSelTbl = NULL;
    mod->modMapTbl = NULL;

    return mod;
}

static PMOD dynMod; /* ugh */
static PMOD dynBIND(void) { return dynMod; }

static void addModEntry(struct objcrt_shared *aCls)
{
    struct objcrt_shared **clsLst;
    struct objcrt_modEntry *entry;
    entry = (struct objcrt_modEntry *)OC_Malloc(sizeof(struct objcrt_modEntry));

    clsLst  = OC_Malloc(sizeof(struct objcrt_shared *));
    *clsLst = aCls;
    dynMod  = newModDesc(clsLst);

    entry->modLink = dynBIND;
    entry->modInfo = NULL;

    addObjcModules(entry); /* dynBIND will use dynMod */
}

struct objcrt_shared  *objcrt_subclass(STR name,struct objcrt_shared *superClass,int ivars,int cvars)
{
    struct objcrt_shared *n = newClass(name,superClass);
    addModEntry(n);
    return n;
}

/*****************************************************************************
 *
 * Message Errors & Diagnostics
 *
 ****************************************************************************/

#if 0
static void _prnStack(FILE *f) 
{
    fprintf(f, "Use your debugger to see a stack backtrace\n");
    fflush(f);
    abort();
}
#endif

/*
 * This function is used when the selTransTbl table (for selector lookup)
 * is not initialized. (Checked for every selector, not enabled by default).
 */

char OBJCRTEXPORT *objcrt_bindError(char *m)
{
	fprintf(stderr,"objcrt initialization error (_OBJCBIND_%s)",m);
	return NULL;
}

static void _prnFrame(FILE *f, id obj, SEL sel, BOOL isSuper) 
{
        struct objcrt_private *self = (struct objcrt_private *)obj;

	if (self) {
	    fprintf(f,(isSuper ? "[%s %s] super\n" :
                                 "[%s %s]\n"),self->isa->clsName, sel);
	} else {
	    fprintf(f,(isSuper ? "[%s] super\n" : "[%s]\n"), sel);
	}

	putc('\n', f);
}

static id msgError (id rcv, char *fmt, ...) 
{ 
	OC_VA_LIST vp;
	OC_VA_START(vp,fmt);
	(*OC_ERROR_VEC)(rcv, fmt, vp);
	OC_VA_END(vp);
	return rcv;
}

static id _errHandler(id self, SEL sel) 
{
	if (sel == @selector(doesNotRecognize:)) { /* possible ! */
		char *msg = "Object does not recognize -doesNotRecognize:";
		return msgError(self,msg);
	} else {
		return [self doesNotRecognize:sel];
	}
}

static id _freedHandler(id self, SEL sel) 
{
	return msgError(self,"message %s sent to freed object=%lx", sel, self);
}

/* stes - feb 12, 97 - scoped extern for inlineCache messenger */
/* stes - sep 4, 97 - implemented noNilRcvr option */

BOOL noNilRcvr = NO;

id OBJCRTEXPORT _nilHandler(id self, SEL sel) 
{
	return (noNilRcvr)?msgError(self,"message %s sent to nil",sel):nil;
}

/*****************************************************************************
 *
 * C Messenger
 *
 * The name "C" messenger comes from the fact that this messenger is
 * implemented in "pure C".  It does not depend on small, auxiliary routines
 * for "building stackframes", usually written in assembler.
 *
 ****************************************************************************/

#define	CACHESIZE 1031 /* a prime number */

static id  clsCache[CACHESIZE];
static SEL selCache[CACHESIZE];
static IMP impCache[CACHESIZE];

static void flushCache(void)
{
    int i = 0;
    for(i=0;i<CACHESIZE;i++) clsCache[i] = nil;
}

FILE *msgIOD;
BOOL msgFlag = NO;
void OBJCRTEXPORT setMsgFlag(BOOL f) { msgFlag = f; }
void OBJCRTEXPORT setMsgIOD(FILE* f) {msgIOD = f; }

BOOL noCacheFlag = NO;
BOOL dbgFlag = NO;
BOOL allocFlag = NO;
FILE *dbgIOD;

/* optional thread-safe messenger - works via function vectors */
/* so that runtime doesn't need to be recompiled to work with  */
/* various threads packages                                    */

void (*objcrt_cache_lock)(void);
void (*objcrt_cache_unlock)(void);
#define CACHE_LOCK if (objcrt_cache_lock) (*objcrt_cache_lock)();
#define CACHE_UNLOCK if (objcrt_cache_unlock) (*objcrt_cache_unlock)();

/* computing index of IMP pointer in method look-up cache      */

#define CACHE_INDEX(aSel,aShr) \
	(int)((((unsigned long)aSel) ^ ((unsigned long)aShr)) % CACHESIZE)


static IMP _getImp(struct objcrt_shared *cls, SEL sel,int index)
{
	struct objcrt_shared *wCls = cls /* working class */;

	do {
		int n; 
		struct objcrt_slt *smt = wCls->clsDispTable;

		for (n = 0; n < wCls->clsSizDict; n++, smt++)
		{
			/* selectors can be compared with '==' */
			if ( sel == smt->_cmd)
			    {
			    if (noCacheFlag) {
				    return smt->_imp;
			    } else {
				    IMP imp = smt->_imp;
				    CACHE_LOCK
				    selCache[index] = sel;
				    clsCache[index] = (id)cls;
				    impCache[index] = imp;
				    CACHE_UNLOCK
				    return imp;
			    }
			}
		}
	} while ((wCls = wCls->clsSuper));

	return (IMP)_errHandler;
}

/* AIX cc doesnt like & address operator in front of address specs
 * so don't use &_nilHandler.
 */

IMP OBJCRTEXPORT _imp(id aRecvr,SEL aSel)
{
        int index;
	BOOL inSuper = NO;
	struct objcrt_shared *shr;

	if (msgFlag) _prnFrame(msgIOD, aRecvr, aSel, inSuper);
	if (!aRecvr) return (IMP)_nilHandler;
	shr = ((struct objcrt_private *)aRecvr)->isa;
	if (!shr) return (IMP)_freedHandler;

	/* try cache hit before extra function call _getImp() */
	index = CACHE_INDEX(aSel,shr);

	if (!noCacheFlag) {
		CACHE_LOCK
		if (clsCache[index] == (id)shr && selCache[index] == aSel) {
			IMP imp = impCache[index]; CACHE_UNLOCK return imp;
		}
		CACHE_UNLOCK
	}

	return _getImp(shr, aSel, index);
}

IMP OBJCRTEXPORT _impSuper(id aClass,SEL aSel)
{
        int index;
	BOOL inSuper = YES;

	if (msgFlag) _prnFrame(msgIOD, aClass, aSel, inSuper);
	if (!aClass) return (IMP)_nilHandler;
	
	/* try cache hit before extra function call _getImp() */
	index = CACHE_INDEX(aSel,aClass);

	if (!noCacheFlag) {
		CACHE_LOCK
		if (clsCache[index] == aClass && selCache[index] == aSel) {
			IMP imp = impCache[index]; CACHE_UNLOCK return imp;
		}
		CACHE_UNLOCK
	}

	return _getImp((struct objcrt_shared *)aClass, aSel, index);
}

/*
 * This function is for compatibility with Stepstone.
 * Used in Producer.
 */

void OBJCRTEXPORT dbg(char *fmt, ...) 
{
        if (dbgFlag) { 
		OC_VA_LIST arglist;
		OC_VA_START(arglist,fmt);
                vfprintf (dbgIOD, fmt, arglist);
                OC_VA_END(arglist);
                fflush(dbgIOD);
        }
}

/*****************************************************************************
 *
 * Adding Methods (as for Categories)
 *
 ****************************************************************************/


static struct objcrt_slt *newDispTable(int n)
{
    return (struct objcrt_slt *)OC_Malloc(n * sizeof(struct objcrt_slt));
}

static void copyDispTable(struct objcrt_slt *dst,struct objcrt_slt *src,int n)
{
    while (n--) { dst->_imp = src->_imp; dst->_cmd = src->_cmd; dst++;src++; }
}

static void freeDispTable(struct objcrt_slt *self)
{
    /* in any case, don't free the statically (compiler) allocated ones */
}

static void addIMethodsTo(struct objcrt_shared *src,struct objcrt_shared *dst)
{
    struct objcrt_slt *n = newDispTable(dst->clsSizDict + src->clsSizDict);
    copyDispTable(n,dst->clsDispTable,dst->clsSizDict);
    copyDispTable(n+dst->clsSizDict,src->clsDispTable,src->clsSizDict);
    dst->clsSizDict += src->clsSizDict;
    freeDispTable(dst->clsDispTable);
    dst->clsDispTable = n;
}

void OBJCRTEXPORT objcrt_addMethodsTo(struct objcrt_shared *src,struct objcrt_shared *dst)
{
    STR srcName,dstName;

    /* can happen from inside +initialize */
    if (src == dst) return;

    /* check direct subclass; otherwise, if A has subclass B, and B
     * has subclass C, and [C poseAs:A] then B will inherit from both
     * C and A
     */

    srcName = src->clsName;
    dstName = dst->clsName;

    if (src->clsSuper != dst) {
	[(id)src error:"addMethodsTo: %s not direct subclass of %s.",
	    srcName,dstName];
    }

    if (src->clsSizInstance != dst->clsSizInstance) {
	[(id)src error:"addMethodsTo: %s adds instance variables to %s.",
	    srcName,dstName];
    }

    addIMethodsTo(src,dst);
    addIMethodsTo(src->isa,dst->isa);

    /* flush message caches */
    flushCache();
}

/*****************************************************************************
 *
 * Class Posing
 *
 ****************************************************************************/

void objcrt_poseAs(struct objcrt_shared *posing,struct objcrt_shared *target)
{
    int i;
    struct objcrt_modEntry * modPtr;
    STR newName,posingName,targetName;

    /* can happen from inside +initialize */
    if (posing == target) return;

    /* check direct subclass; otherwise, if A has subclass B, and B
     * has subclass C, and [C poseAs:A] then B will inherit from both
     * C and A
     */

    posingName = posing->clsName;
    targetName = target->clsName;

    if (!CLS_HASPOSING(target)) {
	[(id)target error:"poseAs: %s needs to be recompiled",targetName];
    }

    if (posing->clsSuper != target) {
	[(id)posing error:"poseAs: %s not direct subclass of %s.",
	    posingName,targetName];
    }

    if (posing->clsSizInstance != target->clsSizInstance) {
	[(id)posing error:"poseAs: %s adds instance variables to %s.",
	    posingName,targetName];
    }

    /* first change names.  this means that findClass: will return
     * the posing class in the future; the old class can still be
     * obtained as "%<classname>".
     */

    newName = (char *)OC_Malloc(strlen(targetName) + 2 + 1);
    strcpy(newName,"_%");strcpy(newName+2,targetName);

    posing->clsName = targetName;
    posing->isa->clsName = target->isa->clsName;
    target->clsName = newName + 1;
    target->isa->clsName = newName;

    /* now "patch" the hierarchy.  look for subclasses of 'target'
     * and (if != posing) make their clsSuper point to posing.
     */

    for (i=0;i<table_modules_count;i++) {
	for(modPtr=table_modules[i];modPtr && modPtr->modLink;modPtr++) {
	    struct objcrt_shared **cls,*aCls;

	    cls = modPtr->modInfo->modClsLst;

	    if (MOD_ISMORETHANONE(modPtr->modInfo)) {
	       while ((aCls = *cls++)) {
		    if (aCls == posing) continue;
		    if (aCls->clsSuper == target) {
			aCls->clsSuper = posing;
			aCls->isa->clsSuper = posing->isa;
		    }
	       }
	    } else {
		    if (cls) {
		    aCls = *cls;
		    if (aCls == posing) continue;
		    if (aCls->clsSuper == target) {
			aCls->clsSuper = posing;
			aCls->isa->clsSuper = posing->isa;
		    }
		}
	    }
	}
    }

    /* finally, assign the class external */
    assert(CLS_HASPOSING(target));
    if (target->clsGlbl) *(target->clsGlbl) = (id)posing;

    /* flush message caches */
    flushCache();
}

#endif /* __PORTABLE_OBJC__ */


