
/*
 * Portable Object Compiler (c) 1998 by Stes & Lerman.  All Rights Reserved.
 */


#include "objpak.h"

#define DEFAULT_CAPACITY (16)

@implementation IdArray

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/

static void idary_init(idary_t self,id v,int n,int c)
{
    id *p = (id *)OC_MallocAtomic(c * sizeof(id));
    assert(0 <= n && n <= c);
    
    self->count    = n;
    self->capacity = c;
    self->ptr      = p;

    while (n--) *p++ = v;
}

+ new
{
    id newObj = [super new];
    idary_init([newObj idaryValue],nil,0,DEFAULT_CAPACITY);
    return newObj;
}

+ new:(unsigned)n
{
    id newObj = [super new];
    idary_init([newObj idaryValue],nil,n,n);
    return newObj;
}

+ with:(unsigned)nArgs,...
{
    /* use OC macros for porting to SunOS4 */
    unsigned i;
    OC_VA_LIST vp;
    id newObject = [self new:nArgs];

    /* #if 0 this piece of code if problems with stdarg
     * typically this means the driver is not configured with
     * the right -builtintype
     * or builtinfunction flags (because the macros might expand to these)
     * 
     * alternative solution: check the .P output (-retain) and 
     * do a setenv OBJCOPT -builtinfunction __builtin_foo
     * (and please let me know)
     */

#if 1
    OC_VA_START(vp,nArgs);
    for(i=0;i<nArgs;i++) {
		id anObject = OC_VA_ARG(vp,id);
		[newObject at:i put:anObject];
	}
    OC_VA_END(vp);
#endif

    return newObject;
}

static void id_ncpy(id *p,id *q,int n)
{
    while (n--) *p++ = *q++;
}

static void idary_copy(idary_t dst,idary_t src)
{
    int n = src->count;
    
    dst->count    = n;
    dst->capacity = n + 1;
    dst->ptr      = (id *)OC_MallocAtomic(dst->capacity * sizeof(id));

    id_ncpy(dst->ptr,src->ptr,n);
}

- copy
{
    id aCopy = [super copy];
    idary_copy([aCopy idaryValue],[self idaryValue]);
    return aCopy;
}

- deepCopy
{
    return [self copy];
}

static void idary_clear(idary_t self)
{
    OC_Free(self->ptr);self->ptr = NULL;
}

- free
{
    idary_clear([self idaryValue]);return [super free];
}

/*****************************************************************************
 *
 * Interrogation
 *
 ****************************************************************************/

- (idary_t) idaryValue
{
    return &value;
}

- (unsigned) size
{
    return [self idaryValue]->count;
}

static id idary_at(idary_t self,int i)
{
    if (0 <= i && i < self->count) {
	return (self->ptr)[i];
    } else {
	[errorBlock(OC_BOUNDSERROR,"index out of bounds") raise];
	return nil;
    }
}

- at:(unsigned)anOffset
{
    return idary_at([self idaryValue],anOffset);
}

static id idputid(id *self,id c)
{
    id r = *self;*self = c;return r;
}

static id idary_putat(idary_t self,int i,id c)
{
    if (0 <= i && i < self->count) {
	return idputid(self->ptr + i,c);
    } else {
	[errorBlock(OC_BOUNDSERROR,"index out of bounds") raise];
	return nil;
    }
}

- at:(unsigned)anOffset put:anObject
{
    return idary_putat([self idaryValue],anOffset,anObject);
}

/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- printToFile:(FILE *)aFile
{
    unsigned i,n = [self size];

    if (n) {
	id o = [self at:0];
	if (o) [o printToFile:aFile]; else fprintf(aFile,"nil");
    }

    for(i=0;i<n;i++) {
	id o = [self at:i];
	fprintf(aFile,",");
	if (o) [o printToFile:aFile]; else fprintf(aFile,"nil");
    }

    return self;
}

/*****************************************************************************
 *
 * Archiving
 *
 ****************************************************************************/

#ifdef __PORTABLE_OBJC__
static void id_fileout(id aFiler,id *p,int n)
{
    while (n--) [aFiler fileOut:p++ type:'@'];
}

static void id_filein(id aFiler,id *p,int n)
{
    while (n--) [aFiler fileIn:p++ type:'@'];
}

static void idary_fileout(id aFiler,idary_t self)
{
    [aFiler fileOut:&self->count type:'i'];
    [aFiler fileOut:&self->capacity type:'i'];
    id_fileout(aFiler,self->ptr,self->count);
}

static void idary_filein(id aFiler,idary_t self)
{
    [aFiler fileIn:&self->count type:'i'];
    [aFiler fileIn:&self->capacity type:'i'];
    self->ptr = (id *)OC_MallocAtomic(self->capacity * sizeof(id));
    id_filein(aFiler,self->ptr,self->count);
}

- fileOutOn:aFiler
{
    [super fileOutOn:aFiler];idary_fileout(aFiler,&value);return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler];idary_filein(aFiler,&value);return self;
}

#endif /* __PORTABLE_OBJC__ */

@end

