
/*
 * Portable Object Compiler (c) 1997 by David Stes.  All Rights Reserved.
 */

#include "objpak.h"

@implementation Rectangle

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


+ new
{
    return [self origin:[Point new] corner:[Point new]];
}

- origin:p corner:q
{
    origin = p;
    corner = q;
    return self;
}


+ origin:p corner:q
{
    return [[super new] origin:p corner:q];
}

- copy
{
    return [self origin:[origin copy] corner:[corner copy]];
}

- deepCopy
{
    return [self origin:[origin deepCopy] corner:[corner deepCopy]];
}

- free
{
    corner = [corner free];origin = [origin free];return [super free];
}

/*****************************************************************************
 *
 * Interrogation
 *
 ****************************************************************************/


- (int) left
{
    return [origin x];
}

- (int) right
{
    return [corner x];
}

- (int) top
{
    return [corner y];
}

- (int) bottom
{
    return [origin y];
}

- (int) width
{
    return [corner x] - [origin x];
}

- (int) height
{
    return [corner y] - [origin y];
}

- origin
{
    return origin;
}

- corner
{
    return corner;
}

/*****************************************************************************
 *
 * Comparing
 *
 ****************************************************************************/


- (BOOL) contains:aPoint
{
    int xLoc = [aPoint x];
    int yLoc = [aPoint y];
    return ([origin x] <= xLoc && xLoc <= [corner x]) 
          && ([origin y] <= yLoc && yLoc <= [corner y]);
}

- (unsigned) hash
{
    return [origin hash] ^ [corner hash];
}

- (BOOL) isEqual:aRectangle
{
    /* Stepstone isKindOf: takes id but isa is SHR */
    return [aRectangle isKindOf:(id)isa] &&
    		[origin isEqual:[aRectangle origin]] &&
		     [corner isEqual:[aRectangle corner]];
}

/*****************************************************************************
 *
 * Arithmetic
 *
 ****************************************************************************/


- moveBy:aPoint
{
    [origin moveBy:aPoint];
    [corner moveBy:aPoint];
    return self;
}

/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- printToFile:(FILE *)aFile
{
    fprintf(aFile,"(");
    [origin printToFile:aFile];
    fprintf(aFile,",");
    [corner printToFile:aFile];
    fprintf(aFile,")");
    return self;
}

@end

