
/*
 * Portable Object Compiler (c) 1997,98 by Stes & Lerman.  All Rights Reserved.
 */

#include "objpak.h"

@implementation Set

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/

static void objptr_zero(id *p,int c)
{
    while (c--) *p++ = nil;
}

#ifndef NDEBUG
static int objptr_count(id *p,int c)
{
    int n = 0;
    while (c--) if (*p++) n++;
    return n;
}
#endif

static void objset_init(objset_t self,int n,int c)
{
    assert(0 <= n && n <= c);
    self->count    = n;
    self->capacity = c;
    self->ptr      = (id *)OC_Malloc(c * sizeof(id));
}

static void objset_initzero(objset_t self,int c)
{
    objset_init(self,0,16);
    objptr_zero(self->ptr,c);
}

+ new
{
    return [self new:16];
}

+ new:(unsigned)n
{
    id newObj = [super new];
    objset_initzero([newObj objsetValue],n);
    return newObj;
}

/*****************************************************************************
 *
 * Creating with contents.
 *
 ****************************************************************************/

+ with:(int)nArgs,...
{
    id newObject;

    /* use OC macros for porting to SunOS4 */
    OC_VA_LIST vp;

    newObject = [self new];

    /* #if 0 this piece of code if problems with stdarg
     * typically this means the driver is not configured with
     * the right -builtintype
     * or builtinfunction flags (because the macros might expand to these)
     * 
     * alternative solution: check the .P output (-retain) and 
     * do a setenv OBJCOPT -builtinfunction __builtin_foo
     * (and please let me know)
     */

#if 1
    OC_VA_START(vp,nArgs);
    while (nArgs-- > 0) {
		id anObject = OC_VA_ARG(vp,id);
		[newObject add:anObject];
	}
    OC_VA_END(vp);
#endif

    return newObject;
}

+ with:firstObject with:nextObject
{
    return [[[self new] add:firstObject] add:nextObject];
}

+ add:firstObject
{
    return [[self new] add:firstObject];
}

static void objptr_copy(id *p,id *q,int c)
{
    while (c--) *p++ = *q++;
}

static void objset_copy(objset_t dst,objset_t src)
{
    objset_init(dst,src->count,src->capacity);
    objptr_copy(dst->ptr,src->ptr,src->capacity);
}

- copy
{
    id aCopy = [super copy];
    objset_copy([aCopy objsetValue],[self objsetValue]);
    return aCopy;
}

static void objptr_deepcopy(id *p,id *q,int c)
{
    while (c--) { id obj = *q++;*p++ = (obj)?[obj deepCopy]:nil; }
}

static void objset_deepcopy(objset_t dst,objset_t src)
{
    objset_init(dst,src->count,src->capacity);
    objptr_deepcopy(dst->ptr,src->ptr,src->capacity);
}

- deepCopy
{
    /* not all Object implementations have |deepCopy| */
    /* so we send |copy| to super instead             */

    id aCopy = [super copy];
    objset_deepcopy([aCopy objsetValue],[self objsetValue]);
    return aCopy;
}

static void objset_empty(objset_t self)
{
    self->count = 0;objptr_zero(self->ptr,self->capacity);
}

- emptyYourself
{
    objset_empty([self objsetValue]);return self;
}

- addYourself
{
    /* trivial for sets, since there are no duplicates */
    return self;
}

static void objptr_clear(id *p,int c)
{
    while (c--) { id obj = *p;*p++ = (obj)?[obj free]:nil; }
}

static void objset_freecontents(objset_t self)
{
    self->count = 0;objptr_clear(self->ptr,self->capacity);
    assert(objptr_count(self->ptr,self->capacity) == 0);
}

- freeContents
{
    objset_freecontents([self objsetValue]);return self;
}

static void objset_clear(objset_t self)
{
    self->count = 0;self->capacity = 0;OC_Free(self->ptr);self->ptr = NULL;
}

- free
{
    objset_clear([self objsetValue]);return [super free];
}

/*****************************************************************************
 *
 * Interrogation
 *
 ****************************************************************************/

- (objset_t) objsetValue
{
    return &value;
}

static int objset_count(objset_t self)
{
    assert(objptr_count(self->ptr,self->capacity) == self->count);
    return self->count;
}

- (unsigned) size
{
    return (unsigned)objset_count([self objsetValue]);
}

- (BOOL) isEmpty
{
    return objset_count([self objsetValue]) == 0;
}

- idEmpty
{
    return ([self isEmpty])?[Block idTrue]:[Block idFalse];
}

- eachElement
{
    id aCarrier = [SetSequence over:self];
    return [Sequence over:aCarrier];
}

/*****************************************************************************
 *
 * Comparing
 *
 ****************************************************************************/

- (unsigned) hash
{
    [self notImplemented:_cmd]; return 0;
}

- (BOOL) isEqual:aSet
{
    [self notImplemented:_cmd]; return NO;
}

/*****************************************************************************
 *
 * Adding
 *
 ****************************************************************************/

static id *objptr_find(id *p,id obj,int n)
{
    id *begin,*now,*end;

    begin = p;
    now   = p + (([obj hash]) % n);
    end   = p + n;
    
    for(;n--;now++) {
	if (now >= end) now = begin;
	if (*now == nil || [*now isEqual:obj]) return now;
    }
    
    fprintf(stderr,"objset_find: table full shouldn't happen");return NULL;
}

static id *objset_find(objset_t self,id obj)
{
    assert(obj != nil);return objptr_find(self->ptr,obj,self->capacity);
}

- add:anObject
{
    [self addNTest:anObject];
    return self;
}

static BOOL objset_needsexpand(objset_t self)
{
    return 2 * self->count > self->capacity;
}

/* don't use "new" as variable name (for C++) */
static void objptr_rehash(id *newp,int newc,id *old,int oldc)
{
    while(oldc--) {
	id obj = *old++;
	id *newend = newp + newc;
	if (obj) {
	    id *pos = newp + ([obj hash] % ((unsigned)newc));
	    while (*pos) { pos++;if (pos == newend) pos = newp; }
	    *pos = obj;
	}
    }
}

static void objset_rehash(objset_t self)
{
    int c;
    id *newp,*old;
    
    c = self->capacity;old = self->ptr;newp = (id *)OC_Malloc(c * sizeof(id));
    
    assert(objptr_count(old,c) == self->count);
    objptr_zero(newp,c);objptr_rehash(newp,c,old,c);
    assert(objptr_count(newp,c) == self->count);
    
    OC_Free(old); self->ptr = newp;
}

static void objset_expand(objset_t self)
{
    id *newp,*old;
    int newc,oldc;
    
    oldc = self->capacity;old = self->ptr;
    newc = 1 + 2 * oldc; newp = (id *)OC_Malloc(newc * sizeof(id));
    
    assert(objptr_count(old,oldc) == self->count);
    objptr_zero(newp,newc);objptr_rehash(newp,newc,old,oldc);
    assert(objptr_count(newp,newc) == self->count);
    
    OC_Free(old); self->ptr = newp; self->capacity = newc;
}

static id objset_add(objset_t self,id obj)
{
    id *p;
    
    if (objset_needsexpand(self)) objset_expand(self);
    
    if (*(p = objset_find(self,obj))) {
	return nil;
    } else {
	self->count++;return *p = obj;
    }
}

- addNTest:anObject
{
    if (anObject) {
	return objset_add([self objsetValue],anObject); 
    } else {
	return nil;
    }
}

static id objset_filter(objset_t self,id obj)
{
    id *p;
    
    if (objset_needsexpand(self)) objset_expand(self);
    
    if (*(p = objset_find(self,obj))) {
	return *p;
    } else {
	self->count++;*p = obj;return nil;
    }
}

- filter:anObject
{
    if (anObject) {
	id t = objset_filter([self objsetValue],anObject); 
	if (t) { [anObject free];return t; } else { return anObject; }
    } else {
	return nil;
    }
}

- add:anObject ifDuplicate:aBlock
{
    if (anObject) {
	id t = objset_filter([self objsetValue],anObject); 
	if (t) { [aBlock value];return t; } else { return anObject; }
    } else {
	return nil;
    }
}

static id objset_replace(objset_t self,id obj)
{
    id *p;
    
    if (objset_needsexpand(self)) objset_expand(self);
    
    if (*(p = objset_find(self,obj))) {
	id tmp = *p;*p = obj;return tmp;
    } else {
	self->count++;*p = obj;return nil;
    }
}

- replace:anObject
{
    if (anObject) {
	return objset_replace([self objsetValue],anObject); 
    } else {
	return nil;
    }
}

/*****************************************************************************
 *
 * Removing
 *
 ****************************************************************************/

static id objset_remove(objset_t self,id obj)
{
    id *p;
    
    if (*(p = objset_find(self,obj))) {
	id tmp = *p;
	*p = nil;self->count--;
	objset_rehash(self);return tmp;
    } else {
	return nil;
    }
}

- remove:oldObject
{
    if (oldObject) {
	return objset_remove([self objsetValue],oldObject); 
    } else {
	return nil;
    }
}

- remove:oldObject ifAbsent:exceptionBlock
{
    if (oldObject) {
	return objset_remove([self objsetValue],oldObject); 
    } else {
	return nil;
    }
}

/*****************************************************************************
 *
 * Testing Contents
 *
 ****************************************************************************/

- (BOOL) includesAllOf:aCol
{
    if (self == aCol) {
	return YES;
    } else {
	id e,seq = [aCol eachElement];
	while ((e=[seq next])) if (![self includes:e]) {
	    [seq free];return NO;
	}
	[seq free];return YES;
    }
}

- (BOOL) includesAnyOf:aCol
{
    if (self == aCol) {
	return YES;
    } else {
	id e,seq = [aCol eachElement];
	while ((e=[seq next])) if ([self includes:e]) {
	    [seq free];return YES;
	}
	[seq free];return NO;
    }
}

/*****************************************************************************
 *
 * Adding and Removing Contents
 *
 ****************************************************************************/

- addContentsTo:aCol
{
    if (self == aCol) {
	[self addYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [self eachElement];
	while ((anElement = [elements next])) [aCol add:anElement];
	elements = [elements free];
    }
	
    return aCol;
}

- addContentsOf:aCol
{
    if (self == aCol) {
	[self addYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [aCol eachElement];
	while ((anElement = [elements next])) [self add:anElement];
	elements = [elements free];
    }
	
    return self;
}
- addAll:aCol
{
    return [self addContentsOf:aCol];
}

- removeContentsOf:aCol
{
    if (self == aCol) {
	return [self emptyYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [aCol eachElement];
	while ((anElement = [elements next])) [self remove:anElement];
	elements = [elements free];
	
	return self;
    }
}

- removeContentsFrom:aCol
{
    if (self == aCol) {
	[self emptyYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [self eachElement];
	while ((anElement = [elements next])) [aCol remove:anElement];
	elements = [elements free];
    }
	
    return self;
}

- removeAll:aCol
{
    return [self removeContentsOf:aCol];
}

/*****************************************************************************
 *
 * Converting to different type of collection.
 *
 ****************************************************************************/

- asSet
{
    /* Stepstone isKindOf takes a id but class returns SHR */
    if ([self isKindOf:(id)[Set class]]) {
	return self;
    } else {
	return [[Set new] addContentsOf:self];
    }
}

- asOrdCltn
{
    /* Stepstone isKindOf takes a id but class returns SHR */
    if ([self isKindOf:(id)[OrdCltn class]]) {
	return self;
    } else {
	return [[OrdCltn new] addContentsOf:self];
    }
}

/*****************************************************************************
 *
 * Block methods
 *
 ****************************************************************************/

- do:aBlock
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) [aBlock value:anElement];
    elements = [elements free];
    
    return self;
}

- detect:aBlock
{
    id elements;
    id anElement;

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if ([[aBlock value:anElement] isTrue]) {
	    elements = [elements free];
	    return anElement; 
	}
    }

    elements = [elements free];
    return nil;
}

- detect:aBlock ifNone:noneBlock
{
    id elements;
    id anElement;

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if ([[aBlock value:anElement] isTrue]) {
	    elements = [elements free];return anElement; 
	}
    }

    elements = [elements free];
    return [noneBlock value];
}

- select:testBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if ([[testBlock value:anElement] isTrue]) [newObject add:anElement];
    }

    elements = [elements free];
    return newObject;
}

- reject:testBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if ([[testBlock value:anElement] isFalse]) [newObject add:anElement];
    }

    elements = [elements free];
    return newObject;
}

- collect:transformBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	id anImage = [ transformBlock value:anElement ];
	if (anImage) [ newObject add:anImage ]; 
    }

    elements = [elements free];
    return newObject;
}

- (unsigned) count:aBlock
{
    id elements;
    id anElement;
    unsigned c = 0;

    elements = [self eachElement];
    while ((anElement = [elements next])) {
	if ([aBlock value:anElement]) c++;
    }
    elements = [elements free];

    return c;
}

/*****************************************************************************
 *
 * Making elements perform actions.
 *
 ****************************************************************************/

- elementsPerform:(SEL)aSelector
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector with:anObject];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject with:otherObject
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector with:anObject with:otherObject];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject with:otherObject with:thirdObj
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
     [anElement perform:aSelector with:anObject with:otherObject with:thirdObj];
    }
    elements = [elements free];
    
    return self;
}
/*****************************************************************************
 *
 * Locating
 *
 ****************************************************************************/

- find:anObject
{
    return (anObject)?*objset_find([self objsetValue],anObject):nil;
}

- (BOOL) contains:anObject
{
    return (BOOL)([self find:anObject] ? YES : NO); 
}

- (BOOL) includes:anObject
{
    return [self contains:anObject];
}

- (unsigned) occurrencesOf:anObject
{
    return (unsigned)([self find:anObject] ? 1 : 0); 
}

/*****************************************************************************
 *
 * Combining
 *
 ****************************************************************************/

- intersection:aSet
{
    if (self == aSet) {
	return [self copy];
    } else {
	id anElement,elements;
	id intersection = [isa new];
	
	elements = [self eachElement];
	while ((anElement = [elements next])) {
	    if ([aSet find:anElement]) [intersection add:anElement];
	}
	elements = [elements free];
	
	return intersection;
    }
}

- union:aSet
{
    if (self == aSet) {
	return [self copy];
    } else {
	return [[self copy] addContentsOf:aSet];
    }
}

- difference:aSet
{
    if (self == aSet) {
	return [isa new];
    } else {
	id difference = [self copy];
	[aSet removeContentsFrom:difference];
	return difference;
    }
}

/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- printToFile:(FILE *)aFile
{
    [[[self eachElement] printToFile:aFile] free];return self;
}

/*****************************************************************************
 *
 * Archiving
 *
 ****************************************************************************/

#ifdef __PORTABLE_OBJC__
static void objptr_fileout(id aFiler,id *a,int n)
{
    while (n--) { id obj = *a++; if (obj) [aFiler fileOut:&obj type:'@']; }
}

static void objptr_filein(id aFiler,id *a,int n)
{
	/* The idea here is that the set is not usable until a "rehash"
	* is performed, in the awakeFrom: method.
	* The hash-value of the objects themselves cannot, by the way, be
	* computed before the |awakeFrom:| message is being sent.
	*/
	
    while (n--) [aFiler fileIn:a++ type:'@'];
}

static void objset_fileout(id aFiler,objset_t self)
{
    int n = self->count;
    [aFiler fileOut:&n type:'i'];
    objptr_fileout(aFiler,self->ptr,self->capacity);
}

static void objset_filein(id aFiler, objset_t self)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    objset_init(self,n,n);
    objptr_filein(aFiler,self->ptr,n);
}

- fileOutOn:aFiler
{
    [super fileOutOn:aFiler]; objset_fileout(aFiler,&value); return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler]; objset_filein(aFiler,&value); return self;
}

- awakeFrom:aFiler
{
	/* double size of capacity and rehash */
	objset_expand(&value); return self;
}
#endif /* __PORTABLE_OBJC__ */

@end

