
/*
 * Portable Object Compiler (c) 1997 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"

id currentBlock; /* inside outputBlockDefs compound */

@implementation BlockExpression
- setUpBlock
{
	static int blockCount = 1;
	count = blockCount++;
	return self;
}

+ new
{
	return [[super new] setUpBlock];
}

- free
{
    return nil;
}

- blockFunctionName
{
	return [String sprintf:"blockFun%i",count];
}

- blockDataName
{
    return [String sprintf:"blockVar%i",count];
}

- blockDataTypeName
{
    return [String sprintf:"struct blockData%i",count];
}

- scope
{
    return scope;
}

- setArguments:aNode
{
	arguments = aNode;return self;
}

- setDecls:dNode body:bNode
{
	id lToken = [Token str:"{"];
	id rToken = [Token str:"}"];
	
	/* wrap statement in compound: handle inlinecaches and tempvars */
	compoundStatement = 
	[CompoundStatement lbrace:lToken decls:dNode body:bNode rbrace:rToken];

	return self;
}

- setStatements:aNode
{
    return [self setDecls:nil body:aNode];
}

- setExpression:aNode
{
    id rNode = [Identifier str:"return"];
    id sNode = [Token str:";"];

    /* wrap return expression in compound */
    aNode = [[[[Node new] add:rNode] add:aNode] add:sNode];

    return [self setStatements:aNode];
}

- (int) numArgsOfBlock
{
    if (arguments) {
        int n = [[arguments subnodes] size];
	assert(n % 2 == 1); /* comma list: arg,arg,arg */
	return (n + 1)/2;
    } else {
	return 0;
    }
}

- addTempVarsTo:aStmt
{
        
	return self;
}

- addBlocksTo:aStmt
{
        [aStmt addBlock:self];
	/* do not descend, ie. don't get subblocks */
	return self;
}

- addInlineCachesTo:aStmt
{
	return self;
}

- addScope:aScope
{
    if (!scope) scope = [Collection new];
    [scope addIfAbsent:aScope];  /* no dups in -isSame sense */
    return self;
}

- outputBlockDataTypeOn:aStream
{
    int n = [scope size];

    /* some compilers warn about empty structs */

    if (n) {
	int i;
	STR type = [[self blockDataTypeName] str];

	[aStream printStr:"%s {\n",type];
	for(i=0;i<n;i++) {
	    STR t,d;id s = [scope at:i];
	    t = [[s scopeDataTypeName] str];
	    d = [[s scopeDataName] str];
	    [aStream printStr:"%s *%s;\n",t,d];
	}
	[aStream printStr:"};\n"];
	[aStream printNewline];
    }

    return self;
}

- packBlockDataOn:aStream
{
    /* wrap a blockData around the scope pointers */
    /* this is output-ing a comma expression */

    int n = [scope size];

    if (n) {
	int i;
	STR t = [[self blockDataTypeName] str];
	STR d = [[self blockDataName] str];

	[aStream printStr:"%s=(%s *)OC_Malloc(sizeof(%s)),\n",d,t,t];

	for(i=0;i<n;i++) {
	    STR ds,ts;
	    id s = [scope at:i];
	    ds = [[s scopeDataName] str];
	    ts = [[s scopeDataTypeName] str];
	    [aStream printStr:"%s->%s=%s,",d,ds,ds];
	}
    }

    return self;
}

- unpackBlockDataOn:aStream
{
    /* sort of opposite of packBlockData */

    int n = [scope size];

    if (n) {
	int i;
	for(i=0;i<n;i++) {
	    STR ds,ts;
	    id s = [scope at:i];
	    ts = [[s scopeDataTypeName] str];
	    ds = [[s scopeDataName] str];
	    [aStream printStr:"%s *%s= data->%s;\n",ts,ds,ds];
	}
    }

    return self;
}

- processBlocks
{
    /* now copy what's needed of the enclosing scope */
    /* this depends on the pushScope/popScope mechanism of compounds */

    [compoundStatement addReferencedScopeToBlock:self];	

    /* descend for nested blocks */
    [Identifier pushScope:compoundStatement];
    [super processBlocks];
    [Identifier popScope];

    return self;
}

- pushScope
{
    int i,n = [scope size];
    for(i=0;i<n;i++) [Identifier pushScope:[scope at:i]];
    return self;
}

- popScope
{
    int i,n = [scope size];
    for(i=0;i<n;i++) [Identifier popScope];
    return self;
}

- outputBlockDefsOn:aStream
{
	STR tName;
        STR fnName = [[self blockFunctionName] str];

	/* first descend into body for nested blocks */
	[compoundStatement outputBlockDefsOn:aStream];

	[self outputBlockDataTypeOn:aStream];
	[aStream printNewline];

	/* tName is "void" if there is no data */
	if ([scope size]) {
	    tName = [[self blockDataTypeName] str];
	} else {
	    tName = "void";
	}

	[aStream printStr:"static id %s(id thisBlock,%s *data",fnName,tName];

	if (arguments) {
	    	[aStream printStr:","];
		[arguments outputWithoutTagsOn:aStream];
	}

	[aStream printStr:")\n"];
	[aStream printStr:"{\n"];

	[self unpackBlockDataOn:aStream];


	[self pushScope];

	[aStream printStr:"if (1)\n"];

	[LineTag forceLineTag];
	[compoundStatement translateOn:aStream];

	[self popScope];

	/* return by default - if not yet returned - the block itself */
	[aStream printNewline];
	[aStream printStr:"return thisBlock;\n"];
	[aStream printStr:"}\n"];
	return self;
}

- translateOn:aStream
{
	if ([options blocks]) {
		int n = [self numArgsOfBlock];
		STR data,f = [[self blockFunctionName] str];
		
		/* start comma expression */
		[aStream printStr:"("];

		if ([scope size]) {
		    data = [[self blockDataName] str];
		    [self packBlockDataOn:aStream];
		} else {
		    data = "(void*)0";
		}

		[aStream printStr:"newBlock(%i,(IMP)%s,%s))",n,f,data];
		[LineTag forceLineTag];

		/* this is for automatic initialization */
		[Module usesEntry:[String str:"newBlock"]];
	} else {
		fatalError("blocks are not allowed with -noBlocks");
		[aStream printStr:"(id)0"];
	}
	
	return self;
}

@end

