
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"
#include "assert.h"

@implementation CompoundStatement
- setUpLbrace:lToken decls:aDecls body:aBody rbrace:rToken
{
    	static int theCount = 0;

	/* for generating unique types (blocks) */
    	count = theCount++;

	/* using tokens for the braces, for nice lineTags */

	lBrace = lToken;[self add:lToken];

        decls  = aDecls;
	if (aDecls) {
	    [self add:aDecls];

	    /* for local vars */
	    [self setDeclarators:aDecls];

	    /* descend to look for msg. expressions in initializers */
	    [aDecls addTempVarsTo:self];

	    /* descend to look for message expressions for inlineCaches */
	    if ([options inlineCache]) [aDecls addInlineCachesTo:self];

	    /* descend to look fo blocks inside this compound */
	    [aDecls addBlocksTo:self];
	}

	body   = aBody;

	if (aBody) {
	    [self add:aBody];

	    /* descend to look for message expressions that need tempVars */
	    [aBody addTempVarsTo:self];

	    /* descend to look for message expressions for inlineCaches */
	    if ([options inlineCache]) [aBody addInlineCachesTo:self];

	    /* descend to look fo blocks inside this compound */
	    [aBody addBlocksTo:self];
	}
	

    /* it is allowed to set rToken nil and include the rToken in the body */
	rBrace = rToken;if (rToken) [self add:rToken];

	return self;
}

+ lbrace:lToken body:aBody rbrace:rToken
{
	id newObj = [super new];
	[newObj setUpLbrace:lToken decls:nil body:aBody rbrace:rToken];
	return newObj;
}

+ lbrace:lToken decls:aDecls body:aBody rbrace:rToken
{
	id newObj = [super new];
	[newObj setUpLbrace:lToken decls:aDecls body:aBody rbrace:rToken];
	return newObj;
}


- setDeclarators:declarators
{
    localVariables = [declarators variables];
    localTypes     = [declarators types];
    return self;
}

- blockVariables
{
    return blockVariables;
}

- addLocalVars:vars types:types
{
    if (!localVariables) localVariables = [Collection new];
    if (!localTypes) localTypes = [Collection new];

    /* this is to add self, _cmd to this compound */
    [localVariables addContentsOf:vars];
    [localTypes addContentsOf:types];

    return self;
}

- pushLocals
{
    if (localVariables) {
       [Identifier pushLocalVars:localVariables types:localTypes];
    }

    return self;
}

- popLocals
{
    if (localVariables) {
	[Identifier popLocals:localVariables];
    }

    return self;
}

- referenceBlockVar:anIdentifier
{
    id matching;
    if ((matching = [localVariables findMatching:anIdentifier])) {
	unsigned i = [localVariables offsetOf:matching];

	if (!blockVariables) blockVariables = [Collection new];
	if (!blockTypes) blockTypes = [Collection new];

	[blockTypes add:[localTypes removeAt:i]];
	[blockVariables add:[localVariables removeAt:i]];
	assert([blockTypes size] == [blockVariables size]);

	return self; /* found identifier in this scope */
    } else {
	if ([blockVariables findMatching:anIdentifier]) {
	    return self; /* found identifier in this scope */
	}
    }

    return nil; /* continue search */
}


- addTempVar:aString
{
	if (!tmpVars) tmpVars = [Collection new];
	[tmpVars add:aString];
	return self;
}

- addBlock:aBlock
{
	if (!blocks) blocks = [Collection new];
	[blocks add:aBlock];
	return self;
}

- addInlineCache:aString
{
	if (!inlineCaches) inlineCaches = [Collection new];
	[inlineCaches add:aString];
	return self;
}

- addTempVarsTo:aStmt
{
	return self;
}

- addInlineCachesTo:aStmt
{
	return self;
}

/* stes - may/2/97 - limit length of line output here */

#define NUM_VARS_PER_LINE 5

- printTmpVars:aStream from:(int)i to:(int)j
{
	[aStream printStr:"id "];
	if (i<j) [aStream printObject:[tmpVars at:i]];
	for(i++;i<j;i++) {
		[aStream printStr:","];
		[aStream printObject:[tmpVars at:i]];
	}
	[aStream printStr:";\n"];
	return self;
}

- printTempVars:aStream
{
	int i,j;
	int n = [tmpVars size];

	assert(n > 0);

	for(i=0;i<n;i = j) {
		j = i + NUM_VARS_PER_LINE;
		j = (j > n)?n:j;
		[self printTmpVars:aStream from:i to:j];
	}
	
	[LineTag forceLineTag];
	return self;
}


- printInlineCaches:aStream from:(int)i to:(int)j
{
	[aStream printStr:"static struct objcrt_inlineCache "];
	if (i<j) [aStream printObject:[inlineCaches at:i]];
	for(i++;i<j;i++) {
		[aStream printStr:","];
		[aStream printObject:[inlineCaches at:i]];
	}
	[aStream printStr:";\n"];
	return self;
}

- printInlineCaches:aStream
{
	int i,j;
	int n = [inlineCaches size];

	assert(n > 0);

	for(i=0;i<n;i = j) {
		j = i + NUM_VARS_PER_LINE;
		j = (j > n)?n:j;
		[self printInlineCaches:aStream from:i to:j];
	}
	
	[LineTag forceLineTag];
	return self;
}

- printBlockData:aStream
{
    int i,n = [blocks size]; 

    for(i=0;i<n;i++) {
	id aBlock = [blocks at:i];
        if ([[aBlock scope] size]) {
	   STR type,var;
	   type   = [[aBlock blockDataTypeName] str];
           var    = [[aBlock blockDataName] str];
	   [aStream printStr:"%s *%s;\n",type,var];
        }
    }

    return self;
}


- scopeDataTypeName
{
    return [String sprintf:"struct scopeData%i",count];
}

- scopeDataName
{
    return [String sprintf:"scopeVar%i",count];
}

- outputScopeDataTypeOn:aStream
{
    int n = [blockVariables size];
    assert(n == [blockTypes size]);

    [aStream printNewline];

    /* some compilers warn about empty structs */

    if (n) {
	int i;
	STR type = [[self scopeDataTypeName] str];

	[aStream printStr:"%s {\n",type];
	for(i=0;i<n;i++) {
	    [[blockTypes at:i] outputWithoutTagsOn:aStream];
	    [aStream printStr:" "];
	    [[blockVariables at:i] outputWithoutTagsOn:aStream];
	    [aStream printStr:";\n"];
	}
	[aStream printStr:"};\n"];
    }

    return self;
}

- createScopeOn:aStream
{
    /* no newline here */

    int n = [blockVariables size];
    assert(n == [blockTypes size]);

    if (n) {
	STR t = [[self scopeDataTypeName] str];
	STR d = [[self scopeDataName] str];

	[aStream printStr:"%s *%s = (%s *)OC_Malloc(sizeof(%s));",t,d,t,t];
    }

    [LineTag forceLineTag];
    return self;
}

- copyScopeOn:aStream
{
    int i,n = [blockVariables size];
    STR d = [[self scopeDataName] str];

    assert(n == [blockTypes size]);

    for(i=0;i<n;i++) {
	    [aStream printStr:"%s->",d];
	    [[blockVariables at:i] outputWithoutTagsOn:aStream];
	    [aStream printStr:" = "];
	    [[blockVariables at:i] outputWithoutTagsOn:aStream];
	    [aStream printStr:";"];
    }

    if (n) [LineTag forceLineTag];
    return self;
}


- processBlocks
{
    if (blocks) {
	[Identifier pushScope:self];
	[super processBlocks];
	[Identifier popScope];
    }

    return self;
}

- outputBlockDefsOn:aStream
{
    [self outputScopeDataTypeOn:aStream];
    return [super outputBlockDefsOn:aStream];
}

- translateOn:aStream
{
	[lBrace translateOn:aStream];

	[aStream printNewline];

	/* generate temporary variables for message expressions */
	/* in this compound statement */
	if (tmpVars) [self printTempVars:aStream];

	/* generate static caches for the inline cache messenger */
	if (inlineCaches) [self printInlineCaches:aStream];

	/* declare block data if any blocks in this compound */
	if (blocks) [self printBlockData:aStream];

	if (isMain) {
	/* generate a call to the runtime initializer
         * note that the assignment to a tmp var "_ocT" is necessary
	 * because the initCall can be emitted in front of a block of variable
	 * declarators
	 */

	/* since this line has no equivalent in .m file, fake linetag
	 * make it seem as if it's the line of the main() function itself
	 */
	        [outertag outputOn:aStream];
		[aStream printStr:"int _ocT = objc_main_init();\n"];
		[LineTag forceLineTag];
	}

        [self pushLocals];[Identifier pushScope:self];

	/* do this before decls so that id aBlock = [ ... ]; works */
	if ([blockVariables size]) {
	        if (lBrace) [[lBrace lineTag] outputOn:aStream];
	        /* no newline here, otherwise have to emit a linetag */
		[self createScopeOn:aStream];
	}

	/* translate decl so that initializer (e.g. id x = [...];)
	 * works;  left-hand side is not translated 
	 */
	if (decls) {[decls translateOn:aStream];}

	/* this needs to be done after decls
	 * the right way to do this could perhaps be to eliminate 
	 * declarations of variables that are accessed via scopePtr
	 */
	if ([blockVariables size]) {
	        /* no newline here, otherwise have to emit a linetag */
		[self copyScopeOn:aStream];
	}

	/* make sure begin of statements have correct linetag */
	[aStream printNewline];
	[LineTag forceLineTag];

	if (body) {[body translateOn:aStream];}
	[self popLocals];[Identifier popScope];

	[rBrace translateOn:aStream];

	return self;
}

- mainTranslateOn:aStream linetag:aTag
{
	isMain = YES;
	outertag = aTag;
	return [self translateOn:aStream];
}

@end

