
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"

@implementation FunctionDefinition
- setUpSpecifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt
{
	specifiers = aSpec;if (aSpec) [self add:aSpec]; /* can be nil */

	declarators = aDecl;[self add:aDecl];
	extDeclarators = extDecl;if (extDecl) [self add:extDecl];
	statements = aStmt;[self add:aStmt];

	nameIdent = [declarators findIdentifier];
	name = [nameIdent string];
	if (!name) fatalError("function name not found");

	if ((aSpec!=nil) && [aSpec containsStatic]) {
		[Module staticEntry:name];		
	} else {
		[Module definesEntry:name];
	}

	return self;
}

+ specifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt
{
	id newObj = [super new];
	[newObj setUpSpecifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt];
	return newObj;
}

- mainPrologueOn:aStream
{
	STR fmt;
	id initCall = [options initCall];
	BOOL noShared = ([options sharedData] == NO);
	
	/* stes - may18/97 moved the xtern call to this place out of
	* function body for cplus
	*/
	
	if (initCall) {
		STR f = [initCall str];
		if (noShared) {

		[aStream printStr:"extern struct useDescriptor *OCU_main;\n"];
		[aStream printStr:"extern struct modEntry *_objcModules;\n"];

		if ([options cplus]) {
		   fmt = "extern \"C\" int %s(void *mods,void *desc);\n";
		} else {
		   fmt = "extern int %s(void *mods,void *desc);\n";
		}

		} else {

		if ([options cplus]) {
		   fmt = "extern \"C\" int %s(int debug,int traceInit);\n";
		} else {
		   fmt = "extern int %s(int debug,int traceInit);\n";
		}

		}
		[aStream printStr:fmt,f];
	} else {
		fatalError("no initialization call defined");
	}

	if ([options filerCode]) {
		[aStream printStr:([options cplus])?"extern \"C\"":"extern"];
		[aStream printStr:" void *_OBJCBIND_ascfiler(void);\n"];

		/* this is only for the noShared, noPostlink case */
		/* so that autoinit will visit the AsciiFiler module */
		[Module usesEntry:[String str:"AsciiFiler"]];
	}

	/* stes - 6/97 for thread-safe runtime without recompiling runtime */
	/* can't use MUTEX_INITIALIZER or macros in .i files */
	
	if ([options cthreads]) {
	fmt = "static struct mutex cacheLock = {0,0};\n";
	[aStream printStr:fmt];
	fmt = "static void lockCacheLock(void) {if (!mutex_try_lock(&cacheLock)) mutex_wait_lock(&cacheLock);}\n";
	[aStream printStr:fmt];
	fmt = "static void unlockCacheLock(void) {(&cacheLock)->lock = 0;}\n";
	[aStream printStr:fmt];
	}
	
	if ([options pthreads]) {
	fmt = "static pthread_mutex_t cacheLock;\n";
	[aStream printStr:fmt];
	fmt = "static void lockCacheLock(void) {pthread_mutex_lock(&cacheLock);}\n";
	[aStream printStr:fmt];
	fmt = "static void unlockCacheLock(void) {pthread_mutex_unlock(&cacheLock);}\n";
	[aStream printStr:fmt];
	}

	/* I'm never going to understand C++. Sigh.  Seems
	 * if I define this static then my program doesnt link
	 */

	if (![options cplus]) [aStream printStr:"static "];
	[aStream printStr:"int objc_main_init(void)\n"];
	[aStream printStr:"{\n"];


	if ([options pthreads]) {
		[aStream printStr:"pthread_mutex_init(&cacheLock,0);\n"];
	}
	
	/* set function vectors */
	if ([options cthreads] || [options pthreads]) {
		[aStream printStr:"objcrt_cache_lock = lockCacheLock;\n"];
		[aStream printStr:"objcrt_cache_unlock = unlockCacheLock;\n"];
	}
	
	if ([options boehm]) {
		[aStream printStr:"objcrt_malloc = GC_malloc;\n"];
		[aStream printStr:"objcrt_malloc_atomic = GC_malloc_atomic;\n"];
		[aStream printStr:"objcrt_realloc = GC_realloc;\n"];
		/* GC_free is more expensive */
		/* registering '0' makes OC_Free() and -free a NOP */
		[aStream printStr:"objcrt_free = 0;\n"];
	}
	
	if ([options filerCode]) {
	        /* this is to pull in the AsciiFiler class */
		[aStream printStr:"_OBJCBIND_ascfiler();\n"];
	}
	
	if (![options cache]) {[aStream printStr:"noCacheFlag=1;\n"];}
	if (![options nilRcvr]) {[aStream printStr:"noNilRcvr=1;\n"];}

	/* objcInit() call */
	if (noShared) {
	[aStream printStr:"%s(_objcModules,OCU_main);\n",[initCall str]];
	} else {
	[aStream printStr:"%s(0,0);\n",[initCall str]];
	}
	/* for auto-init */
	[Module usesEntry:initCall];

	[aStream printStr:"return 0;\n}\n\n"];
	return self;
}

/* override this method to generate a call to the initFunction */
/* of the Objective C runtime                                  */

- translateOn:aStream
{
	if ([name isEqual:[options mainFunction]]) {
	
		[self mainPrologueOn:aStream];
		[LineTag forceLineTag];

		/* this is the entry point for objcInit() to intialize */
		/* all classes at startup */
		if (specifiers) [specifiers translateOn:aStream];
		[declarators translateOn:aStream];
		if (extDeclarators) [extDeclarators translateOn:aStream];
		[statements mainTranslateOn:aStream linetag:[nameIdent lineTag]];
		
		return self;
	} else {
		return [super translateOn:aStream];
	}
}

@end

