
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"
#include "assert.h"

id currentMethodImplementation;

@implementation MethodImplementation

- free
{
    return [super free];
}

- (BOOL) isFactory
{
	id declarator = [subnodes at:0];
	return [declarator isFactory];
}

/* 
 * outputBlockDefsOn: is overridden so that the block can reference
 * instance variables
 *
 */

- processBlocks
{
    id declarator = [subnodes at:0];
    id compound = [subnodes at:1];
    id aClass = [ClassDefinition current];

    if (!aClass) {
	    fatalError("method implementation outside class definition");
    }


    /* allow method arguments + self,_cmd to be referenced in Block */
    [declarator addLocalsTo:compound];

    /* allow iVar / cVar to be referenced */
    currentMethodImplementation=self;
    [aClass makeSetOfIVarsCurrent];

    [super processBlocks];

    [aClass clearCurrentSetOfIVars];
    currentMethodImplementation=nil;

    return self;
}

- outputBlockDefsOn:aStream
{
	id aClass = [ClassDefinition current];


        [aClass willTranslateMethodImplementationOn:aStream];

	/* remember which method implementation for "super" context */
	/* limit the scope of ivars to method implementations */

	currentMethodImplementation=self;
	[aClass makeSetOfIVarsCurrent];

	[super outputBlockDefsOn:aStream];

	[aClass clearCurrentSetOfIVars];
	currentMethodImplementation=nil;

	return self;
}
    
- translateOn:aStream
{
	id declarator = [subnodes at:0];
	id compoundStmt = [subnodes at:1];
	id aClass = [ClassDefinition current];
	if (!aClass) {
		fatalError("method implementation outside class definition");
	}

	[aClass willTranslateMethodImplementationOn:aStream];

	assert([declarator isKindOf:[MethodDeclarator class]]);

	[aStream printNewline];
	[aStream printNewline];
	[LineTag forceLineTag];

	[declarator translateOn:aStream];

	/* remember which method implementation for "super" context */
	/* limit the scope of ivars to method implementations */
	currentMethodImplementation=self;
	[aClass makeSetOfIVarsCurrent];

	[compoundStmt translateOn:aStream];

	[aClass clearCurrentSetOfIVars];
	currentMethodImplementation=nil;

	return self;
}

@end

