
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"

@implementation OutputStream
/*****************************************************************************
 *
 * Utility function
 *
 ****************************************************************************/

/*
 * Opening and closing a file.  We're not dealing with wrapping around
 * files (include).  This is because we're processing cpp output.
 */

FILE* openFile(STR name,STR modfs)
{
	FILE *f;

	if ((f = fopen(name,modfs)) == NULL) {
		fprintf(stderr,"objc1: Unable to open %s\n",name);
		exit(1);
	}

	return f;
}

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


- setUpFile:(FILE *)f
{
	file = f;
	return self;
}

+ new
{
	return [self newFile:stdout];
}

+ str:(STR)s
{
	return [self newFile:openFile(s,"w")];
}

+ newFile:(FILE *)f
{
	/* stes - feb 23,1997
         * the default is to use block buffered output
         * but for debugging unbuffered output is very useful
         */

	if ([options isBuffered]==NO) setbuf(f,NULL);

	return [[super new] setUpFile:f];
}

- close
{
	fclose(file);return self;
}

/*****************************************************************************
 *
 * Output methods
 *
 ****************************************************************************/

- printNewline
{
	fprintf(file,"\n");
	return self;
}

- printChar:(char)aChar
{
	fprintf(file,"%c",aChar);
	return self;
}

- printToken:(STR)token
{
	/* watch out for % in tokenName */
	fprintf(file,"%s",token);
	return self;
}

- printSelector:aString
{
	if ([options doSelTransTbl]) {
		BOOL check = [options checkBindCalled];
	    int offset = [Module selTransTblOffsetFor:aString];

		if (check) [self printStr:"(selTransTbl)?"];
		
		[self printStr:"selTransTbl[%i",offset];
		if ([options comments]) [self printStr:" /* %s */",[aString str]];
		[self printStr:"]"];
		
		if (check) {
			STR modName = [[Module bindName] str];
			[self printStr:":objcrt_bindError(\"%s\")",modName];
		}
	} else {
	    [self printStr:"\"%s\"",[aString str]];
	}

	return self;
}

- printStr:(STR)format,...
{
	OC_VA_LIST ap;
	/* use OC macros for port to SunOS4 */

	OC_VA_START(ap,format);
	vfprintf(file,format,ap);
	OC_VA_END(ap);

	return self;
}

- printObject:anObject
{
	[anObject printToFile:file];
	return self;
}

@end

